/*
 * Copyright (c) 2017
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
grammar EPQL;

@header {
    package de.umr.ep.queryengine.ql.generated;
}

// Parser entry
start : (statement) (SEMICOLON statement)* (SEMICOLON)? ;

// Generic statement
statement 
  : outerQuery
  | unionQuery
  | exceptQuery
  | createStreamStatement
  | dropStreamStatement
  | insertEventStatement
  ;

// SELECT clause
selectClause : SELECT (dist=DISTINCT)? (asterisk='*' | projections=projectionList);

// FROM clause
fromClause
  : FROM source=innerQuery (',' innerQuery)*
  ;

// JOIN clause
joinClause
  : source1=innerQuery JOIN source2=innerQuery ON '(' bExp ')'
  | source1=innerQuery JOIN source2=innerQuery USING '(' jIdent=identifier ')'
  ;

boundIdentifier: (identifier '.')? identifier;

// Identifiers such as names and any non-reserved keywords
identifier : ID;

// Attribute ( [prefix.]name [AS rename] )
attribute : (prefix=identifier '.')? (name=identifier (AS rename=identifier)? | asterisk='*');

// Attribute order by entry
orderByEntry : name=boundIdentifier (ASC|DESC)?;

// List of order by entries
orderByList: orderByEntry (',' orderByList)?;

// WHERE clause
whereClause: WHERE bExp;

// An attribute in the GROUP BY clause
groupByAttribute: (prefix=identifier '.')? name=identifier;

// GROUP BY clause
groupByClause : GROUP BY groupByAttribute (',' groupByAttribute)*;

// HAVING clause
havingClause: HAVING havingExpression;

// LIMIT clause
limitClause : LIMIT num=NUMBER;

// ORDER BY clause
orderByClause : ORDER BY orderByList;

// WITHIN clause
withinClause: WITHIN '(' timeExpression ',' timeExpression ')';

// WINDOW clause
windowClause: WINDOW '('
    ( COUNT count=NUMBER EVENTS ( JUMP countJump=NUMBER EVENTS )? |
    TIME ( time=timeNumber ) ( JUMP timeJump=timeNumber )? ) ')';

// TimeExpression
timeExpression
  : NOW # NowTime
  | timeNumber # NumberTime
  | left=timeExpression op=(ADD|SUBTRACT) right=timeExpression # ExpressionTime
  ;

timeNumber
  : num=NUMBER ( ms=( MILLISECOND | MILLISECONDS ) | sec=( SECOND | SECONDS ) | min=( MINUTE | MINUTES ) | hour=( HOUR | HOURS ) | day=( DAY | DAYS ) | week=( WEEK | WEEKS ) )
  ;

// stream clause :- identifier [WINDOW clause]
stream
  : name=identifier ((AS)? alias=identifier)?
  ;

// Aggregate identifier
aggregateIdentifier: COUNT | SUM | MIN | MAX | AVG | ST_EXTENT;

// Aggregates
aggregate
  : method=aggregateIdentifier '(' (input=boundIdentifier|asterisk='*') ')'
  ;

// An outer query
outerQuery
  : selectClause fromClause (whereClause)? (groupByClause (havingClause)?)? (withinClause)? (windowClause)? (orderByClause)? (limitClause)?
  | selectClause fromClause matchRecognizeClause
  ;

unionQuery
  : outerQueryPart UNION outerQueryPart
  ;

exceptQuery
  : outerQueryPart EXCEPT (all=ALL)? outerQueryPart
  ;

outerQueryPart
  : outerQuery
  ;

// Inner query, i.e. a data source
innerQuery
  : subQuery (AS)? name=identifier #SubQueryRule
  | '(' join=joinClause ')' #Join
  | stream #StreamRule
  ;

// A simple subquery
subQuery
  : '(' statement ')' #SubqueryStatement
  | '(' join=joinClause ')' #SubqueryJoin
  ;

// An attribute definition within a projection
projectionAttribute
  : agg=aggregate (AS rename=identifier)? #AggAtt
  | name=boundIdentifier (AS rename=identifier)? #PlainAtt
  | havingArithmeticExpression AS name=identifier #ArithAtt
  | spatialFunction AS name=identifier #SpatialAtt
  ;

// A list of projected attributes
projectionList: projectionAttribute (',' projectionAttribute)*;

// Atomic expression
atom
  : atomAttribute
  |  prevAtomAttribute
  | atomVal
  ;

atomAttribute : (prefix=identifier '.')? name=identifier;
prevAtomAttribute: PREV '(' prefix=identifier '.' name=identifier ')' ;
atomVal: atomValue;

atomValue
  : (STRING_SINGLE_QUOTED | STRING) #AtomString
  | NUMBER # AtomNum
  ;

// A WKT geometry constant
geometryAtom
  : atomGeometry
  | atomAttribute;

atomGeometry: GEOMETRY_DEFINITION;

// Defines a list of expressions
aExpList
  : aExp (',' aExp)*
  ;

// Artihmetic expression
aExp
  : left=aExp op=(ADD|SUBTRACT|MULTIPLY|DIVIDE|MODULO) right=aExp # ArithOpExp
//  | '-' aExp # ArithNegExp
  | atom # ArithAtomExp
  | '(' aExp ')' # ArithParExp
  | arithmeticFunction # arithmeticFunc
  ;

arithmeticFunction
    : funcName=identifier '(' (atom)? (',' atom)* ')'
    ;

// Interval
// syntax: {([} a1,...,an {)]}  for interval lists
//         {([} a1 : an {)]}          for interval ranges

interval
  : '(' aExpList ')' # IntervalList
  | open=( '(' | '[' ) a=aExp ':' b=aExp close=( ')' | ']' ) # IntervalRange
  ;

// Scalar subquery
aSubQueryExp
 : res = subQuery
 ;

scalarArithArg
 : aExp #ArithExpRule
 | aSubQueryExp #ArithSubQueryRule
 ;

scalarHavingArithArg
 : havingArithmeticExpression #HavingArithExpRule
 | aSubQueryExp #HavingArithSubQueryRule
 ;

// Boolean expressions
bExp
// Compound expressions
  : TRUE #TrueBoolExp
  | FALSE #FalseBoolExp
  | left=bExp op=(AND|OR) right=bExp # BoolOpExp
  | NOT right=bExp #NegatedBoolOp
// Relational expressions
  | left=scalarArithArg op=(EQUALS | NOT_EQUALS | GREATER | LESS | GREATER_EQUALS | LESS_EQUALS | LIKE) right=scalarArithArg # BoolRelExp
  | left=geometryAtom op=(ST_CONTAINS | ST_COVEREDBY | ST_COVERS | ST_CROSSES | ST_DISJOINT | ST_EQUALS | ST_EQUALSEXACT | ST_INTERSECTS | ST_OVERLAPS | ST_TOUCHES | ST_WITHIN) right=geometryAtom # BoolSpatialExp
  | left=geometryAtom op=(ST_ISEMPTY | ST_ISRECTANGLE | ST_ISSIMPLE | ST_ISVALID) # UnaryBoolSpatialExp
// Higher order boolean expressions
  | left=aExp negate=NOT? IN interval # BoolInExp
  | left=aExp negate=NOT? BETWEEN right1=aExp AND right2=aExp # BoolBetweenExp
  | EXISTS subQuery #BoolExistsExp
// Rule closure
  | '(' bExp ')' # BoolParExp
  ;

// Arithmetic expression in HAVING clause, allowing aggregates
havingArithmeticExpression
  : left=havingArithmeticExpression op=(ADD|SUBTRACT|MULTIPLY|DIVIDE|MODULO) right=havingArithmeticExpression # HavingArithOpExp
//  | '-' havingArithmeticExpression # HavingArithNegExp
  | atom # HavingArithAtomExp
  | agg=aggregateIdentifier '(' (prefix=identifier '.')? attr=identifier ')' # HavingaggregateExpression
  | '(' havingArithmeticExpression ')' # HavingArithParExp
  ;

// Boolean expression in HAVING clause, allowing aggregates
havingExpression
  : TRUE #HavingTrueBoolExp
  | FALSE #HavingFalseBoolExp
  | left=havingExpression op=(AND|OR) right=havingExpression # HavingBoolOpExp
  | NOT right=havingExpression #HavingNegatedBoolOp
// Relational expressions
 | left=scalarHavingArithArg op=(EQUALS | NOT_EQUALS | GREATER | LESS | GREATER_EQUALS | LESS_EQUALS | LIKE) right=scalarHavingArithArg # HavingBoolRelExp
// Higher order boolean expressions
 | left=havingArithmeticExpression negate=NOT? IN interval # HavingBoolInExp
 | left=havingArithmeticExpression negate=NOT? BETWEEN right1=havingArithmeticExpression AND right2=havingArithmeticExpression # HavingBoolBetweenExp
// Rule closure
 | '(' havingExpression ')' # HavingBoolParExp
 ;

spatialFunction
  : op=ST_ENVELOPE '(' source=geometryAtom ')'
  ;

 // ===============================================================================
 // =============================== Pattern Matching ==============================
 // ===============================================================================

matchRecognizeClause
    : MATCH_RECOGNIZE '(' measureClause patternClause defineClause ')'
    ;

patternClause
    : PATTERN patternDefinition
    ;

patternDefinition
    :  (patternList| subpattern='(' patternDefinition ')' (asterisk='*'|plus='+')?)+
    ;

patternList
    : patternSymbol (patternSymbol)*
    ;

patternSymbol
    : name=identifier (asterisk='*'|plus='+')?
    ;

defineClause
    : DEFINE symbolDefinition (',' symbolDefinition)*
    ;

symbolDefinition
    : symbol=identifier AS expression=bExp
    ;

measureClause
    : MEASURES measureAttributes = measureAttributeList
    ;

measureAttributeList
    : measureAttribute (',' measureAttribute)*
    ;

measureAttribute
    : LAST '(' symbol=identifier '.' attr=identifier ')' AS rename=identifier #lastMeasureAttribute
    | FIRST '(' symbol=identifier '.' attr=identifier ')'  AS rename=identifier #firstMeasureAttribute
    | symbol=identifier '.' attr=identifier AS rename=identifier #plainMeasureAttribute
    ;

 // ===============================================================================
 // =========================== Data Definition language ==========================
 // ===============================================================================

dropStreamStatement
  : DROP STREAM name=identifier
  ;

// Create stream statement
createStreamStatement
  : CREATE STREAM name=identifier '(' attributeDefinitionList ')'
  ;

attributeDefinitionList
  : attributeDefinition (',' attributeDefinitionList)?
  ;

attributeDefinition
  : identifier dataTypeDefinition (NOT NULL)?
  ;

dataTypeDefinition
  : INTEGER #IntegerDataType
  | BYTE #ByteDataType
  | SHORT #ShortDataType
  | LONG # LongDataType
  | FLOAT #FloatDataType
  | DOUBLE #DoubleDataType
  | STRING_TYPE '(' num=NUMBER ')' #StringDataType
  ;

// Insert statement
insertEventStatement
  : INSERT INTO name=identifier VALUES '(' attributeList ')'
  ;

attributeList
  : atomValue (',' attributeList)?
  ;

/**
 * LEXER RULES
 */

// Quoted strings, atomics and whitespaces
fragment ESCAPED_DOUBLE_QUOTE : '\\"';
fragment ESCAPED_SINGLE_QUOTE : '\\\'';
fragment DIGIT: [0-9];

// Quoted strings, atomics and whitespaces
STRING: '"' ( ESCAPED_DOUBLE_QUOTE | ~('\n'|'\r') )*? '"';
STRING_SINGLE_QUOTED: '\'' ( ESCAPED_SINGLE_QUOTE | ~('\n'|'\r') )*? '\'';
NUMBER: ('-')? (DIGIT+ | DIGIT* '.' DIGIT+) ('f'|'F')?;
WS: [ \t\r\n]+ -> skip; // Discard whitespaces
SEMICOLON : ';';
COMMA: ',';

// Binary spatial operators
ST_CONTAINS: S T '_' C O N T A I N S;
ST_COVEREDBY: S T '_' C O V E R E D B Y;
ST_COVERS: S T '_' C O V E R S;
ST_CROSSES: S T '_' C R O S S E S;
ST_DISJOINT: S T '_' D I S J O I N T;
ST_EQUALS: S T '_' E Q U A L S;
ST_EQUALSEXACT: S T '_' E Q U A L S E X A C T;
ST_INTERSECTS: S T '_' I N T E R S E C T S;
ST_OVERLAPS: S T '_' O V E R L A P S;
ST_TOUCHES: S T '_' T O U C H E S;
ST_WITHIN: S T '_' W I T H I N;

// Unary spatial operators
ST_ISEMPTY: S T '_' I S E M P T Y;
ST_ISRECTANGLE: S T '_' I  S R E C T A N G L E;
ST_ISSIMPLE: S T '_' I S S I M P L E;
ST_ISVALID: S T '_' I S V A L I D;

ST_ENVELOPE: S T '_' E N V E L O P E;

// Spatial keywords
fragment POINT: 'POINT' | 'point';
fragment LINESTRING: 'LINESTRING' | 'linestring';
fragment LINE: 'LINE' | 'line';
fragment POLYGON: 'POLYGON' | 'polygon';
fragment EMPTY_GEOMETRY: 'EMPTY' | 'empty';

fragment GEOMETRY_TYPES: POINT | LINESTRING | LINE | POLYGON;
fragment WKT_STRING: '('+ [0-9 ,-.]* ')'+;
GEOMETRY_DEFINITION : GEOMETRY_TYPES WS* (WKT_STRING | EMPTY_GEOMETRY);

// Artihmetic operators
ADD : '+';
SUBTRACT : '-';
MULTIPLY : '*';
DIVIDE : '/';
MODULO : '%'; 

// Relational operators
EQUALS : '=';
NOT_EQUALS : '!=' | '<>';
GREATER : '>';
LESS : '<';
GREATER_EQUALS : '>=' | '!<';
LESS_EQUALS : '<=' | '!>';

// Literal tokens for case-insensitive keywords
fragment A : 'A' | 'a';
fragment B : 'B' | 'b';
fragment C : 'C' | 'c';
fragment D : 'D' | 'd';
fragment E : 'E' | 'e';
fragment F : 'F' | 'f';
fragment G : 'G' | 'g';
fragment H : 'H' | 'h';
fragment I : 'I' | 'i';
fragment J : 'J' | 'j';
fragment K : 'K' | 'k';
fragment L : 'L' | 'l';
fragment M : 'M' | 'm';
fragment N : 'N' | 'n';
fragment O : 'O' | 'o';
fragment P : 'P' | 'p';
fragment Q : 'Q' | 'q';
fragment R : 'R' | 'r';
fragment S : 'S' | 's';
fragment T : 'T' | 't';
fragment U : 'U' | 'u';
fragment V : 'V' | 'v';
fragment W : 'W' | 'w';
fragment X : 'X' | 'x';
fragment Y : 'Y' | 'y';
fragment Z : 'Z' | 'z';

// Logical operators
TRUE: T R U E;
FALSE: F A L S E;
ALL : A L L;
AND : A N D;
ANY : A N Y;
BETWEEN : B E T W E E N;
EXISTS : E X I S T S;
IN : I N;
LIKE : L I K E;
NOT : N O T;
OR : O R;
IS_NULL : I S ' ' N U L L;
UNIQUE : 'UNIQUE';

// QL keywords
SELECT : S E L E C T;
DISTINCT: D I S T I N C T;
FROM: F R O M;
WHERE: W H E R E;
AS: A S;
GROUP: G R O U P;
BY: B Y;
HAVING: H A V I N G;
LIMIT: L I M I T;
JOIN: J O I N;
ON: O N;
USING: U S I N G;
ORDER: O R D E R;
UNION: U N I O N;
EXCEPT: E X C E P T;
WITHIN: W I T H I N;
WINDOW: W I N D O W;
JUMP: J U M P;
PATTERN: P A T T E R N;
DEFINE: D E F I N E;
NOW : N O W;

// DDL keywords
DROP: D R O P;
CREATE: C R E A T E;
STREAM: S T R E A M;
NULL: N U L L;
INSERT: I N S E R T;
INTO: I N T O;
VALUES: V A L U E S;

// Ordering keywords
ASC: A S C;
DESC: D E S C;

// Aggregation keywords
COUNT : C O U N T;
SUM: S U M;
MIN: M I N;
MAX: M A X;
AVG: A V G;
ST_EXTENT: S T '_' E X T E N T;

// WindowExpression keywords
MILLISECONDS : M I L L I S E C O N D S;
MILLISECOND : M I L L I S E C O N D;
MINUTES : M I N U T E S;
MINUTE : M I N U T E;
SECONDS : S E C O N D S;
SECOND : S E C O N D;
HOURS : H O U R S;
HOUR : H O U R;
DAYS : D A Y S;
DAY : D A Y;
WEEKS : W E E K S;
WEEK : W E E K;

// JEPC datatypes
BYTE : B Y T E;
SHORT : S H O R T;
INTEGER : I N T E G E R;
LONG : L O N G;
FLOAT : F L O A T;
DOUBLE : D O U B L E;
STRING_TYPE : S T R I N G;
GEOMETRY : G E O M E T R Y;

// Pattern keywords
LAST: L A S T;
FIRST: F I R S T;
MEASURES: M E A S U R E S;
MATCH_RECOGNIZE: 'MATCH_RECOGNIZE';
PREV: P R E V;

TIME : T I M E;
EVENTS : E V E N T S;

ID:[a-zA-Z]+[a-zA-Z0-9_]*; // Non-Keyword identifiers

DATE:[0-9][0-9]'.'[0-9][0-9]'.'[0-9][0-9][0-9][0-9]; // Date format
