/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.common;

/**
 * Models an aggregation entry. Each entry
 * consists of an attribute and an aggregate
 * type, e.g. <code>sum</code> or <code>max</code>.
 *
 * @author Marc Seidemann
 */
public class AggregationAttribute {

    /**
     * Represents all available aggregate types.
     */
    public static enum Aggregate {
        SUM,
        COUNT,
        MIN,
        MAX,
        AVG,
        SPATIAL_EXTENT
    }

    /**
     * The name of the aggregated attribute.
     */
    protected String attribute;
    /**
     * The type of aggregation.
     */
    protected Aggregate aggregate;

    /**
     * Creates a new aggregation entry.
     *
     * @param attribute the name of the aggregated attribute
     * @param aggregate the aggregate type
     */
    public AggregationAttribute(String attribute, Aggregate aggregate) {
        this.attribute = attribute;
        this.aggregate = aggregate;
    }

    /**
     * Returns the name of the resulting aggregate attribute column.
     *
     * @return the name of the aggregated attribute
     */
    public String getColumnName() {
        return aggregate.toString().toLowerCase()+"("+attribute+")";
    }

    /**
     * Returns the data type of the aggregation attribute.
     * @return the type of the attribute
     */
    public Attribute.DataType getAttributeType() {
        if (aggregate == Aggregate.SPATIAL_EXTENT)
            return Attribute.DataType.GEOMETRY;
        else
            return Attribute.DataType.DOUBLE;
    }

    /**
     * Returns the name of the attribute that should be aggregated.
     *
     * @return the name of the attribute that should be aggregated.
     */
    public String getAttributeName() {
        return attribute;
    }

    /**
     * Sets the name of the attribute that should be aggregated.
     *
     * @param name the name of the attribute that should be aggregated.
     */
    public void setAttributeName(String name) {
        this.attribute = name;
    }

    /**
     * Returns the type of aggregation,
     * e.g. <code>sum</code> or <code>max</code>.
     *
     * @return the type of aggregation
     */
    public Aggregate getType() {
        return aggregate;
    }

    @Override
    public boolean equals(Object obj) {
        if (obj instanceof AggregationAttribute) {
            AggregationAttribute att = (AggregationAttribute)obj;
            return att.aggregate == aggregate && att.attribute.equals(attribute);
        } else
            return false;
    }
}
