/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.common;

/**
 * Represents a generic number. This number can be compared
 * to other types of numbers and also can be applied to other
 * types of number. Operations on two comparable numbers
 * result in an integer number iff both operands are also
 * integer numbers itself. Otherwise, the result will be a
 * floating point number.
 *
 * @param <N> the type of the number
 *
 * @author Marc Seidemann
 */
public class ComparableNumber<N extends Number> extends Number implements Comparable<ComparableNumber>{

    /**
     * The contained number.
     */
    private N number;

    /**
     * Creates a new IntervalNumber for the given java number
     * @param number the number
     */
    public ComparableNumber(N number) {
        this.number = number;
    }

    /**
     * Compares this number with the given one.
     *
     * @param n the number to be compared with
     * @return the result of the comparison
     */
    public int compareTo(ComparableNumber n) {
        if (isNumeric() && n.isNumeric())
            return Long.compare(number.longValue(), n.getNumber().longValue());
        else
            return Double.compare(number.doubleValue(), n.getNumber().doubleValue());
    }

    /**
     * Returns if this number represents a floating point number.
     *
     * @return <tt>True</tt> if this number if a floating point
     * number, <tt>false</tt> otherwise
     */
    private boolean isFloatingPoint() {
        return number instanceof Float || number instanceof Double;
    }

    /**
     * Returns if this number represents a numeric, i.e. an integer number (byte,short,integer,long).
     *
     * @return <tt>True</tt> if this number if an integer number
     * (byte,short,integer,long), <tt>false</tt> otherwise
     */
    private boolean isNumeric() {
        return number instanceof Byte || number instanceof Short || number instanceof Integer || number instanceof Long;
    }

    /**
     * Adds the given number to this number and returns the result.
     * If both numbers are integer numbers, the result is an integer
     * number again. Otherwise, the result is a floating point number.
     *
     * @param n the number to be added
     * @return this+n
     */
    public ComparableNumber add(ComparableNumber n) {
        if (isNumeric() && n.isNumeric())
            return new ComparableNumber(number.longValue()+n.getNumber().longValue());
        else
            return new ComparableNumber(number.doubleValue()+n.getNumber().doubleValue());
    }

    /**
     * Subtracts the given number from this number and returns the result.
     * If both numbers are integer numbers, the result is an integer
     * number again. Otherwise, the result is a floating point number.
     *
     * @param n the number to be subtracted
     * @return this-n
     */
    public ComparableNumber subtract(ComparableNumber n) {
        if (isNumeric() && n.isNumeric())
            return new ComparableNumber(number.longValue()-n.getNumber().longValue());
        else
            return new ComparableNumber(number.doubleValue()-n.getNumber().doubleValue());
    }

    /**
     * Multiplies the given number with this number and returns the result.
     * If both numbers are integer numbers, the result is an integer
     * number again. Otherwise, the result is a floating point number.
     *
     * @param n the number to be multiplied with
     * @return this*n
     */
    public ComparableNumber mult(ComparableNumber n) {
        if (isNumeric() && n.isNumeric())
            return new ComparableNumber(number.longValue()*n.getNumber().longValue());
        else
            return new ComparableNumber(number.doubleValue()*n.getNumber().doubleValue());
    }

    /**
     * Divides this number by the given number and returns the result.
     * If both numbers are integer numbers, the result is an integer
     * number again. Otherwise, the result is a floating point number.
     *
     * @param n the number to be divided by
     * @return this/n
     */
    public ComparableNumber divide(ComparableNumber n) {
        if (isNumeric() && n.isNumeric())
            return new ComparableNumber(number.longValue()/n.getNumber().longValue());
        else
            return new ComparableNumber(number.doubleValue()/n.getNumber().doubleValue());
    }

    /**
     * Calculates the modulo of this number and the given one and returns the result.
     * If both numbers are integer numbers, the result is an integer
     * number again. Otherwise, the result is a floating point number.
     *
     * @param n the number to calculate the modulo with
     * @return this%n
     */
    public ComparableNumber mod(ComparableNumber n) {
        if (isNumeric() && n.isNumeric())
            return new ComparableNumber(number.longValue()%n.getNumber().longValue());
        else
            return new ComparableNumber(number.doubleValue()%n.getNumber().doubleValue());
    }

    /**
     * Returns the underlying number.
     *
     * @return the underlying number
     */
    public N getNumber() {
        return number;
    }


    @Override
    public String toString() {
        return number.toString();
    }

    @Override
    public int intValue() {
        return number.intValue();
    }

    @Override
    public long longValue() {
        return number.longValue();
    }

    @Override
    public float floatValue() {
        return number.floatValue();
    }

    @Override
    public double doubleValue() {
        return number.doubleValue();
    }
}
