package de.umr.ep.common;

import java.util.ArrayList;
import java.util.List;

/**
 * Utility operations for schemas.
 *
 * @author Marc Seidemann
 */
public class SchemaUtils {

    /**
     * Defines a qualifier id.
     */
    public static final String QUALIFIER_ID = "q_id";
    
    public static final String ATTR_TSTART = "tstart";
    public static final String ATTR_TEND   = "tend";

    // ==========================================================================

    /**
     * Returns the join schema of this schema and the given one.
     *
     * @param s2 the schema to join with
     * @return the joined schema
     */
    public static List<Attribute> join(List<Attribute> s1, List<Attribute> s2) {
        List<Attribute> atts = new ArrayList<>();
        List<Attribute> timeAttributes = new ArrayList<>(2);
        for(int i = 0; i < s1.size(); i++) {
            if (isTimeAttribute(s1.get(i)))
                timeAttributes.add(new Attribute(s1.get(i)));
            else
                atts.add(new Attribute(s1.get(i)));
        }
        for(int i = 0; i < s2.size(); i++) {
            if (!atts.contains(s2.get(i)) && !isTimeAttribute(s2.get(i)))
                atts.add(new Attribute(s2.get(i)));
        }
        for (Attribute att : timeAttributes)
            att.removeProperty(QUALIFIER_ID);
        atts.addAll(timeAttributes);
        return atts;
    }

    // ==========================================================================

    public static List<Attribute> copy(List<Attribute> schema) {
        List<Attribute> result = new ArrayList<>();
        for (int i = 0; i < schema.size(); i++) {
            result.add(new Attribute(schema.get(i)));
        }
        return result;
    }

    // ==========================================================================

    /**
     * Returns if the given attribute is a time attribute.
     *
     * @param attribute the attribute
     * @return <tt>True</tt> if it is a time attribute, <tt>false</tt> otherwise
     */
    public static boolean isTimeAttribute(Attribute attribute) {
        return isTimeAttribute(attribute.getName());
    }
    
    /**
     * Returns if the given attribute is a time attribute.
     *
     * @param attribute the attribute
     * @return <tt>True</tt> if it is a time attribute, <tt>false</tt> otherwise
     */
    public static boolean isTimeAttribute(String attributeName) {
        return (ATTR_TSTART.equalsIgnoreCase(attributeName) || ATTR_TEND.equalsIgnoreCase(attributeName));
    }

    // ==========================================================================

    /**
     * Sets an identifier for this schema.
     *
     * @param identifier the new identifier
     */
    public static void setIdentifier(List<Attribute> schema, String identifier) {
        for(Attribute att : schema)
            if (!isTimeAttribute(att))
                att.setProperty(QUALIFIER_ID, identifier);
    }

    // ==========================================================================

    /**
     * Looks up an attribute with the given name
     * @param schema the schema to look up in
     * @param name the atttribute's name
     * @return the attribute or null, if there is no attribute with the given name
     */
    public static Attribute getAttribute(List<Attribute> schema, String name) {
        for(Attribute att : schema)
            if (att.getName().equals(name))
                return att;
        return null;
    }

    // ==========================================================================

    /**
     * Returns if the given schema contains an attribute with the given identifier.
     * @param schema the schema to look up in
     * @param name the atttribute's name
     * @return the attribute or null, if there is no attribute with the given name
     */
    public static boolean containsAttribute(List<Attribute> schema, String name) {
        for(Attribute att : schema)
            if (att.getName().equals(name))
                return true;
        return false;
    }

    // ==========================================================================

    public static List<String> projectNames(List<Attribute> schema) {
        List<String> result = new ArrayList<>(schema.size());
        for (int i = 0; i < schema.size(); i++) {
            result.add(schema.get(i).getName());
        }
        return result;
    }

    // ==========================================================================

    /**
     * Returns the fully qualified attribute name.
     *
     * @param attribute the attribute
     * @return the qualified attribute name
     */
    public static String getQualifiedAttributeName(Attribute attribute) {
        String qualifier = attribute.getProperty(QUALIFIER_ID);
        if (qualifier != null)
            return qualifier+"."+attribute.getName();
        else
            return attribute.getName();
    }

    public static String getIdentifier(List<Attribute> attributes) {
        if (attributes.size() > 0 && attributes.get(0).hasProperty(QUALIFIER_ID))
            return attributes.get(0).getProperty(QUALIFIER_ID);
        else
            return null;
    }
}
