package de.umr.ep.event;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import de.umr.ep.common.Attribute;


public class EventSchema implements Iterable<Attribute>, Serializable {

    private static final byte VERSION_1 = 1;
    
    public static final byte CURRENT_VERSION = VERSION_1;
    
	/**
     * 
     */
    private static final long serialVersionUID = 1L;
    
    private final List<Attribute> attributes = new ArrayList<>();

	public EventSchema(Attribute... attributes) throws SchemaException {
		// validate
		if (attributes.length < 1)
			throw new IllegalArgumentException("A schema requires at least one attribute");

		Set<String> names = new HashSet<>();
		for (Attribute a : attributes)
			if (!names.add(a.getName()))
				throw new SchemaException("The attribute-names must be unique inside an EventSchema: " + a.getName());

		// Copy the attributes -- for safety, even if they are immutable right
		// now
		for (Attribute a : attributes)
			this.attributes.add(new Attribute(a));
	}
	
	public EventSchema union(EventSchema other) throws SchemaException {
		return union(other,null,null);
	}

	public EventSchema union(EventSchema other, String prefixl, String prefixr) throws SchemaException {
		List<Attribute> result = new ArrayList<>();

		for (Attribute a : attributes) {
			String name = prefixl != null ? prefixl + "." + a.getName() : a.getName();
			result.add(new Attribute(name, a.getType()));
		}
		for (Attribute a : other.attributes) {
			String name = prefixr != null ? prefixr + "." + a.getName() : a.getName();
			result.add(new Attribute(name, a.getType()));
		}
		Attribute[] as = result.toArray(new Attribute[result.size()]);
		return new EventSchema(as);
	}

	public int getNumAttributes() {
		return attributes.size();
	}

	public int getAttributeIndex(String name) throws SchemaException {
		int idx = 0;
		String uName = name.toUpperCase();
		for (Attribute a : attributes) {
			if (a.getName().equals(uName))
				return idx;
			else
				idx++;
		}
		throw new SchemaException("No attribute named: " + name + " found in schema.");
	}
	
	public Attribute getAttribute( int index ) {
	    return attributes.get(index);
	}

	public Attribute byName(String name) throws SchemaException {
	    String uName = name.toUpperCase();
		for (Attribute a : attributes) {
			if (a.getName().equals(uName))
				return a;
		}
		throw new SchemaException("No attribute named: " + name + " found in schema.");
	}

	@Override
	public Iterator<Attribute> iterator() {
		return attributes.iterator();
	}

	@Override
	public int hashCode() {
		return ((attributes == null) ? 0 : attributes.hashCode());
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		EventSchema other = (EventSchema) obj;
		if (attributes == null) {
			if (other.attributes != null)
				return false;
		} else if (!attributes.equals(other.attributes))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "EventSchema: " + attributes;
	}
	
	public void write( DataOutput out ) throws IOException {
	    // Write version
	    out.writeByte(CURRENT_VERSION);
	    // Write number of attributes
	    out.writeByte(attributes.size());
	    for ( Attribute attr : attributes )
	        attr.write(out);
	}
	
	public static EventSchema read( DataInput in ) throws IOException {
	    // Read version
	    byte version = in.readByte();

	    // This is for version 1 -- apply a switch on version increase
	    byte numAttribs = in.readByte();
	    
	    List<Attribute> attribs = new ArrayList<>(numAttribs);
	    
	    for ( int i = 0; i < numAttribs; i++ ) {
	        Attribute attrib = new Attribute();
	        attrib.read(in);
	        attribs.add(attrib);
	    }
	    try {
            return new EventSchema(attribs.toArray(new Attribute[attribs.size()]));
        } catch (SchemaException e) {
            throw new IOException("Error while deserializing event schema",e);
        }
	}
	
}
