/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.operators;



import de.umr.ep.common.AggregationAttribute;
import de.umr.ep.common.Attribute;
import de.umr.ep.event.TimeRepresentation;
import de.umr.ep.queryengine.LogicalOperator;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * Implementation of a logical aggregation operator node.
 *
 * @author Marc Seidemann
 */
public class Aggregation extends LogicalOperator {

    /**
     * The input operator.
     */
    protected LogicalOperator input;
    /**
     * Stores all aggregated attributes and the corresponding aggregation type
     */
    protected List<AggregationAttribute> aggregates;
    /**
     * Stores the number of timestamp values at the end of an event.
     */
    private int skip;

    /**
     * Creates a new logical aggregation operator.
     *
     * @param input the input operator
     * @param aggregates the list of aggregates to be calculated
     */
    public Aggregation(LogicalOperator input, AggregationAttribute... aggregates) {
        this(input, Arrays.asList(aggregates));
    }

    /**
     * Creates a new logical aggregation operator.
     *
     * @param input the input operator
     * @param aggregates the list of aggregates to be calculated
     */
    public Aggregation(LogicalOperator input, List<AggregationAttribute> aggregates) {
        this.input = input;
        this.aggregates = aggregates;
        setInput(input);
    }

    /**
     * Returns the input node.
     *
     * @return the input node
     */
    public LogicalOperator getInput() {
        return input;
    }

    /**
     * Sets the input node.
     *
     * @param node the new input node
     */
    public void setInput(LogicalOperator node) {
        this.input = node;
        if (input != null) {
            TimeRepresentation rep = TimeRepresentation.fromSchema(input.getSchema());
            skip = (rep==TimeRepresentation.INTERVAL)?2: (rep==TimeRepresentation.POINT)?1:0;
        }
    }

    /**
     * Returns the aggregated attributes and the corresponding aggregation.
     *
     * @return the aggregation entries
     */
    public List<AggregationAttribute> getAggregates() {
        return aggregates;
    }

    @Override
    public List<Attribute> getSchema() {
        List<Attribute> iSchema = input.getSchema();

        List<Attribute> result = new ArrayList<>(iSchema);
//        Collections.copy(result, iSchema);
        for (int i = 0; i < aggregates.size(); i++) {
            result.add(result.size() - skip, new Attribute(aggregates.get(i).getColumnName(), aggregates.get(i).getAttributeType()));
        }
        return result;
    }

    @Override
    public void setInput(LogicalOperator input, int i) {
        setInput(input);
    }

    @Override
    public int getInputCount() {
        return 1;
    }

    @Override
    public LogicalOperator getInput(int i) {
        return getInput();
    }

    @Override
    public String toString() {
        String res = "";
        for(AggregationAttribute entry : aggregates)
            res += entry.getColumnName() + ", ";
        if (aggregates.size() > 0)
            res = res.substring(0, res.length()-2);
        return '\u03B1' + "["+res+"]";
    }
}
