/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.operators;

import de.umr.ep.common.Attribute;
import de.umr.ep.queryengine.LogicalOperator;

import java.util.List;

/**
 * Implementation of a logical EXCEPT operator node.
 */
public class Except extends LogicalOperator {

    /**
     * The left input that should be removed from.
     */
    protected LogicalOperator left;
    /**
     * The right input that should be removed from the left.
     */
    protected LogicalOperator right;
    /**
     * Indicates whether all occurrences of a value from right
     * should be removed in left.
     */
    protected boolean all;

    /**
     * Creates a new except operator for the given input.
     *
     * @param left the left input operator
     * @param right the right input operator
     * @param all indicates whether all occurrences should be removed
     */
    public Except(LogicalOperator left, LogicalOperator right, boolean all) {
        this.left = left;
        this.right = right;
        this.all = all;
    }

    @Override
    public List<Attribute> getSchema() {
        return left.getSchema();
    }

    @Override
    public void setInput(LogicalOperator input, int i) {
        if (i == 0)
            left = input;
        else
            right = input;
    }

    @Override
    public int getInputCount() {
        return 2;
    }

    @Override
    public LogicalOperator getInput(int i) {
        if (i == 0)
            return left;
        else
            return right;
    }

    /**
     * Returns the left input operator.
     *
     * @return the left input operator
     */
    public LogicalOperator getLeft() {
        return left;
    }

    /**
     * Returns the right input operator.
     *
     * @return the right input operator
     */
    public LogicalOperator getRight() {
        return right;
    }

    /**
     * Returns whether all occurrences of a value from right
     * should be removed in left.
     *
     * @return <tt>True</tt> if all occurrences should be removed,
     *         <tt>false</tt> otherwise
     */
    public boolean isAll() {
        return all;
    }

    @Override
    public String toString() {
        return "-";
    }
}
