/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.operators;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.SchemaProvider;
import de.umr.ep.common.SchemaUtils;
import de.umr.ep.queryengine.LogicalOperator;

import java.util.List;

/**
 * Implementation of a logical from operator node.
 *
 * @author Marc Seidemann
 */
public class From extends LogicalOperator {

    /**
     * The schema provider.
     */
    protected SchemaProvider schemaProvider;
    /**
     * The name of the stream.
     */
    protected String name;
    /**
     * Stores the stream`s schema.
     */
    protected List<Attribute> schema;
    /**
     * Stores the stream`s alias.
     */
    protected String alias;

    /**
     * Creates a new logical from operator node.
     *
     * @param schemaProvider the schema provider
     * @param name the name of the stream
     */
    public From(SchemaProvider schemaProvider, String name) {
        this.schemaProvider = schemaProvider;
        this.name = name;
        this.schema = SchemaUtils.copy(schemaProvider.getSchema(name));
    }

    /**
     * Creates a new logical from operator node.
     *
     * @param schemaProvider the schema provider
     * @param name the name of the stream
     */
    public From(SchemaProvider schemaProvider, String name, String alias) {
        this(schemaProvider, name);
        this.alias = alias;
        SchemaUtils.setIdentifier(this.schema, alias);
    }

    /**
     * Returns the event schemaProvider this logical from refers to.
     *
     * @return the event schemaProvider
     */
    public SchemaProvider getSchemaProvider() {
        return schemaProvider;
    }

    /**
     * Returns the stream`s name of this from operation.
     *
     * @return the name of the stream
     */
    public String getName() {
        return name;
    }

    /**
     * Returns the streams alias. If no alias is
     * defined, this method returns the stream`s name.
     *
     * @return the alias of this stream
     */
    public String getAlias() {
        return alias;
    }

    @Override
    public List<Attribute> getSchema() {
        return schema;
    }

    @Override
    public void setInput(LogicalOperator input, int i) {
        throw new UnsupportedOperationException("From operator has no input");
    }

    @Override
    public String toString() {
        return name;
    }

    @Override
    public int getInputCount() {
        return 0;
    }

    @Override
    public LogicalOperator getInput(int i) {
        return null;
    }
}
