/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.operators;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.SchemaUtils;
import de.umr.ep.queryengine.LogicalOperator;
import de.umr.ep.queryengine.predicates.BooleanExpression;

import java.util.List;

/**
 * Implementation of a logical JOIN operator node.
 *
 * @author Marc Seidemann
 */
public class Join extends LogicalOperator {

    /**
     * The left join input.
     */
    protected LogicalOperator left;
    /**
     * The right join input.
     */
    protected LogicalOperator right;
    /**
     * The boolean join predicate.
     */
    private BooleanExpression predicate;

    /**
     * Creates a new logical join operator.
     *
     * @param left the left join input
     * @param right the right join input
     * @param predicate the join predicate
     */
    public Join(LogicalOperator left, LogicalOperator right, BooleanExpression predicate) {
        this.left = left;
        this.right = right;
        this.predicate = predicate;
    }

    /**
     * Returns the left input operator.
     *
     * @return the left input operator
     */
    public LogicalOperator getLeft() {
        return left;
    }

    /**
     * Returns the right input operator.
     *
     * @return the right input operator
     */
    public LogicalOperator getRight() {
        return right;
    }

    /**
     * Returns the join predicate.
     *
     * @return the join predicate
     */
    public BooleanExpression getPredicate() {
        return predicate;
    }

    /**
     * Sets the join predicate.
     *
     * @param predicate the new join predicate
     */
    public void setPredicate(BooleanExpression predicate) {
        this.predicate = predicate;
    }

    /**
     * Sets the left input operator.
     *
     * @param node the new left input operator
     */
    public void setLeft(LogicalOperator node) {
        this.left = node;
    }

    /**
     * Sets the right input operator.
     *
     * @param node the new right input operator
     */
    public void setRight(LogicalOperator node) {
        this.right = node;
    }

    /**
     * Creates a new logical join node.
     *
     * @param left the left join input
     * @param right the right join input
     */
    public Join(LogicalOperator left, LogicalOperator right) {
        this.left = left;
        this.right = right;
    }

    @Override
    public List<Attribute> getSchema() {
        return SchemaUtils.join(left.getSchema(),(right.getSchema()));
    }

    @Override
    public void setInput(LogicalOperator input, int i) {
        if (i == 0)
            left = input;
        else
            right = input;
    }

    @Override
    public int getInputCount() {
        return 2;
    }

    @Override
    public LogicalOperator getInput(int i) {
        if (i == 0)
            return getLeft();
        else
            return getRight();
    }

    @Override
    public String toString() {
        return "\u22C8";
    }
}
