/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.operators;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.Attribute.DataType;
import de.umr.ep.event.TimeRepresentation;
import de.umr.ep.common.SchemaUtils;
import de.umr.ep.queryengine.LogicalOperator;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.omg.CORBA.NO_IMPLEMENT;

/**
 * Implementation of a logical projection operator node.
 *
 * @author Marc Seidemann
 */
public class Projection extends LogicalOperator {

    /**
     * The projection input.
     */
    protected LogicalOperator input;
    /**
     * The list of projected attribute names.
     */
    private List<String> attributes;

    /**
     * Creates a new logical projection operator.
     *
     * @param input the projection input
     * @param attributes the projected attribute names
     */
    public Projection(LogicalOperator input, List<String> attributes) {
        this.input = input;
        this.attributes = new ArrayList<>();
        attributes.stream().filter( a -> !SchemaUtils.isTimeAttribute(a) ).forEach(this.attributes::add);
    }

    /**
     * Creates a new logical projection operator.
     *
     * @param input the projection input
     * @param attributes the projected attribute names
     */
    public Projection(LogicalOperator input, String ... attributes) {
        this(input,Arrays.asList(attributes));
    }

    /**
     * Returns the input node.
     *
     * @return the input node
     */
    public LogicalOperator getInput() {
        return input;
    }

    /**
     * Sets the input node.
     *
     * @param node the new input node
     */
    public void setInput(LogicalOperator node) {
        this.input = node;
    }

    /**
     * Returns a list of attributes which should be projected.
     *
     * @return the list of projected attributes.
     */
    public List<String> getAttributes() {
        return attributes;
    }

    @Override
    public List<Attribute> getSchema() {
        List<Attribute> inputSchema = input.getSchema();
        TimeRepresentation tr = TimeRepresentation.fromSchema(inputSchema);
        List<Attribute> resultSchema = new ArrayList<>(attributes.size());
        for(int i = 0; i < attributes.size(); i++) {
            try {
                resultSchema.add(SchemaUtils.getAttribute(inputSchema,attributes.get(i)));
            } catch (Exception ex) {
                throw new RuntimeException("Error parsing SQL query: attribute '"+attributes.get(i)+"' does not exist in this context.");
            }
        }
        // TODO: Move to schema methods
        if ( tr != TimeRepresentation.NONE )
            resultSchema.add( new Attribute(SchemaUtils.ATTR_TSTART, DataType.LONG ));
        if ( tr == TimeRepresentation.INTERVAL )
            resultSchema.add( new Attribute(SchemaUtils.ATTR_TEND, DataType.LONG ));
        return resultSchema;
    }

    @Override
    public void setInput(LogicalOperator input, int i) {
        setInput(input);
    }

    @Override
    public String toString() {
        return '\u03C0' + Arrays.toString(attributes.toArray());
    }

    @Override
    public int getInputCount() {
        return 1;
    }

    @Override
    public LogicalOperator getInput(int i) {
        return getInput();
    }
}
