/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.operators;

import de.umr.ep.common.Attribute;
import de.umr.ep.queryengine.Constants;
import de.umr.ep.queryengine.LogicalOperator;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Implementation of a logical renaming operator node.
 *
 * @author Marc Seidemann
 */
public class Renaming extends LogicalOperator {

    /**
     * Representation of a mapping entry.
     * This class is for convenience to
     * simplify the constructor call and
     * is translated into a map entry.
     */
    public static class RenameItem {
        String name;
        String newName;

        public RenameItem(String name, String newName) {
            this.name = name;
            this.newName = newName;
        }
    }

    /**
     * The input operand.
     */
    protected LogicalOperator input;

    /**
     * Stores the rename mapping: old => new.
     */
    protected Map<String, String> renaming;

    /**
     * Creates a new logical renaming operator.
     *
     * @param input the input operator
     * @param renaming a list of renaming items
     */
    public Renaming(LogicalOperator input, RenameItem ... renaming) {
        this.input = input;
        this.renaming = new HashMap<>();
        for(RenameItem item : renaming)
            this.renaming.put(item.name, item.newName);
    }

    /**
     * Creates a new logical renaming operator.
     *
     * @param input the operator input
     * @param renaming the renaming map
     */
    public Renaming(LogicalOperator input, Map<String, String> renaming) {
        this.input = input;
        this.renaming = renaming;
    }

    /**
     * Returns the input operator.
     *
     * @return the input operator
     */
    public LogicalOperator getInput() {
        return input;
    }

    /**
     * Sets the input node.
     *
     * @param node the new input node
     */
    public void setInput(LogicalOperator node) {
        this.input = node;
    }

    /**
     * Returns a map containing the rename mapping.
     *
     * @return a map containing the renaming
     */
    public Map<String, String> getRenaming() {
        return renaming;
    }

    @Override
    public List<Attribute> getSchema() {
        List<Attribute> inputSchema = input.getSchema();
        Attribute[] output = new Attribute[inputSchema.size()];
        for (int i = 0; i < output.length; i++) {
            if (renaming.containsKey(getQualifiedAttributeName(inputSchema.get(i)))) {
                // Copy attribute and rename the copy
                output[i] = new Attribute(inputSchema.get(i));
                String newName = renaming.get(getQualifiedAttributeName(inputSchema.get(i)));
                String[] substitution = newName.split("\\.");
                if (substitution.length > 2)
                    throw new RuntimeException("Invalid renaming identifier: "+newName);
                if (substitution.length == 1)
                    output[i].setName(newName);
                else if (substitution.length == 2) {
                    output[i].setName(substitution[1]);
                    output[i].setProperty(Constants.QUALIFIER_ID, substitution[0]);
                }
            } else {
                output[i] = inputSchema.get(i);
            }
        }
        return Arrays.asList(output);
    }

    private String getQualifiedAttributeName(Attribute attribute) {
        return attribute.getName();
    }

    @Override
    public void setInput(LogicalOperator input, int i) {
        setInput(input);
    }

    @Override
    public String toString() {
        String rn = "";
        for(String s : renaming.keySet())
            rn += renaming.get(s)+"<-"+s+", ";
        if (renaming.size() > 0)
            rn = rn.substring(0, rn.length()-2);
        return '\u03C1' + "[" + rn + "]";
    }

    @Override
    public int getInputCount() {
        return 1;
    }

    @Override
    public LogicalOperator getInput(int i) {
        return getInput();
    }
}
