/*
 * Copyright (c) 2012, 2013
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.umr.ep.persistence;

import de.umr.ep.event.TimeRepresentation;

/**
 * Representation of events in the form of pairs (p,t).
 * 
 * @author Bastian Hoßbach
 */
public class Event implements Comparable<Event> {

	/**
	 * The timestamp of the event.
	 */
	private long timestamp;
	
	/**
	 * The payload of the event.
	 */
	private Object[] payload;


	/**
	 * Creates a new pair representation of the given event.
	 * 
	 * @param event the event
	 */
	public Event(Object[] event) {
        this(event, TimeRepresentation.INTERVAL);
	}

    /**
     * Creates a new Event with the given representation
     *
     * @param event the event
     * @param representation the representation of the event
     */
    public Event(Object[] event, TimeRepresentation representation) {
        this.payload = event;
        if (representation == TimeRepresentation.INTERVAL)
            timestamp = (long) payload[payload.length-2];
        else if(representation == TimeRepresentation.POINT)
            timestamp = (long) payload[payload.length-1];
        else
            timestamp = System.currentTimeMillis();
    }
	
	/**
	 * Creates a new pair representation of the given event.
	 * 
	 * @param payload the payload of the event
	 * @param timestamp the timestamp of the event
	 */
	public Event(Object[] payload, long timestamp) {
		this.payload   = payload;
		this.timestamp = timestamp;
	}

	
	/**
	 * Returns the timestamp of the event.
	 * 
	 * @return the timestamp of the event
	 */
	public long getTimestamp() {
		return timestamp;
	}


	/**
	 * Returns the pyload of the event.
	 * 
	 * @return the pyload of the event.
	 */
	public Object[] getPayload() {
		return payload;
	}


	@Override
	public String toString() {
//		return "<" + timestamp + ", " + payload + ">";
        String output = "<" + timestamp + ", " + "[";
        for (int i = 0; i < payload.length; i++) {
            output += ""+payload[i];
            if (i < payload.length-1)
                output += ",";
        }
        output += "] >";
        return output;
	}
	
	
	/*
	 * (non-Javadoc)
	 * 
	 * Events are ordered globally by time and a comparison between events
	 * is reduced to a comparison of their timestamps.
	 * 
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	@Override
	public int compareTo(Event arg0) {
		return (int)(this.getTimestamp() - arg0.getTimestamp());
	}

}