package de.umr.ep.persistence;

import java.util.Iterator;
import java.util.Set;

import de.umr.ep.common.Attribute;

public interface EventStore {
	
	/**
	 * Opens the event store. This method should be called before using the event store.
	 */
	void open();
	
	
	/**
	 * Closes the event store.
	 */
	void close();


    /**
     * Removes all data from the event store!
     */
    void clear();

	/**
	 * Registers a new event stream.
	 * 
	 * @param streamName the name of the event stream to register
	 * @param schema the schema of the event stream to register
	 */
	void registerStream(String streamName, Attribute[] schema);


	/**
     * Returns the schema for the given event stream.
     *
     * @param streamName the name of the event stream
     * @return the schema of the event stream, if a stream
     * with the given name has been registered, null otherwise
     */
    Attribute[] getSchema(String streamName);

	
	/**
	 * Removes a stream and all stored events from the event store completely.
	 * 
	 * @param streamName the name of the sevent stream to remove
	 */
	void removeStream(String streamName);


	/**
	 * Returns all registered event streams.
	 * 
	 * @return the names of all registered event streams.
	 */
	Set<String> getRegisteredStreams();
	
	
	/**
	 * Pushes a new event into the event store.
	 * 
	 * @param streamName the corresponding event stream of the new event
	 * @param event the new event that is pushed into the event store
	 */
	void pushEvent(String streamName, Object[] event);
	
	
	/**
	 * Pushes a new event into the event store.
	 * 
	 * @param streamName the corresponding event stream of the new event
	 * @param payload the new event that is pushed into the event store
	 * @param timestamp the timestamp of the new event
	 */
	void pushEvent(String streamName, Object[] payload, long timestamp);


    /**
     * Returns the complete history of the given stream, beginning at the last
     * inserted event and iterating backwards through the history.
     *
     * @return the complete history in reverse order
     */
    Iterator<Event> getBackwardsHistory(String streamName);


	/**
	 * Returns the complete history of an event stream beginning from a user-defined instant of time
	 * ordered by time.
	 * 
	 * @param streamName the event stream
	 * @param startTime the start instant of time
	 * @return the complete history of an event stream beginning from <code>startTime</code>
	 */
	Iterator<Event> getFreshestHistory(String streamName, long startTime);
	
	
	/**
	 * Returns the complete history of an event stream ordered by time(relation scan).
	 * 
	 * @param streamName the event stream
	 * @return the complete history of an event stream
	 */
	Iterator<Event> getHistory(String streamName);
	
	
	/**
	 * Returns all historical events of an event stream for a given point in time ordered by time (temporal point query).
	 * 
	 * @param streamName the event stream
	 * @param timestamp the point in time
	 * @return all historical events of an event stream for the given point in time
	 */
	Iterator<Event> getHistoryPoint(String streamName, long timestamp);
	
	
	/**
	 * Returns all historical events of an event stream for a given range of time ordered by time (temporal range query).
	 * 
	 * @param streamName the event stream
	 * @param startTime the beginning of the range
	 * @param endTime the end of the range
	 * @return all historical events of an event stream for the given range of time
	 */
	Iterator<Event> getHistoryRange(String streamName, long startTime, long endTime);

	/**
	 * Sets the capacity of the event store.
	 * 
	 * @param capacity the capacity of the event store
	 * @param capacityUnit the unit that specifies how to interpret the capacity
	 */
	void setEventStoreCapacity(long capacity);


    /**
     * Returns the capacity of the event store.
     *
     * @return the maximum capacity of the event store.
     */
    long getEventStoreCapacity();

    /**
     *
     * Suspends a stream and frees all handlers on it. This method should be called,
     * if the given stream will not be used for a considerable time. Before the given
     * stream can be written to again, it has to be resumed.
     *
     * @param streamName The name of the stream to be suspended
     */
    void suspendStream(String streamName);

    /**
     * Returns the timestamp of the last inserted event. If no event
     * has been inserted yet, then Long.MIN_VALUE is returned.
     *
     * @param streamName the name of the event stream
     * @return the last timestamp
     */
    long getLastTimestamp(String streamName);

    /**
     * Resumes a stream after being suspended. This method should be called to
     * restore the state of the stream and all required file handlers.
     *
     * @param streamName The name of the stream to be resumed
     */
    void resumeStream(String streamName);

}