/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.queryengine;

import de.umr.ep.common.SchemaProvider;
import de.umr.ep.operators.ddl.DDLOperator;
import de.umr.ep.queryengine.ql.generated.EPQLLexer;
import de.umr.ep.queryengine.ql.generated.EPQLParser;
import de.umr.ep.queryengine.ql.grammar.EPQLListener;
import org.antlr.v4.runtime.*;
import org.antlr.v4.runtime.tree.ParseTreeWalker;

import java.text.ParseException;
import java.util.List;

/**
 * Created by seidemann on 08.09.14.
 */
public class EPQueryCompiler {

    /**
     * Compiles the given query and returns a logical plan.
     *
     * @param query a valid EP-QL query
     * @param schemaProvider the schema provider providing information about the stream source schemata
     * @return a logical query plan
     */
    public static LogicalOperator compileQuery(String query, SchemaProvider schemaProvider) throws ParseException {

        // Create a new lexer instance
        EPQLLexer lexer = new EPQLLexer(new ANTLRInputStream(query));

        // Get a list of matched tokens
        CommonTokenStream tokens = new CommonTokenStream(lexer);

        // Pass the tokens to the parser
        EPQLParser parser = new EPQLParser(tokens);

        lexer.removeErrorListeners();
        parser.removeErrorListeners();
        DescriptiveErrorListener errorListener = new DescriptiveErrorListener();
        lexer.addErrorListener(errorListener);
        parser.addErrorListener(errorListener);

        // Specify out entry point
        EPQLParser.StartContext startContext = parser.start();

        // Walk it and attach our listener
        ParseTreeWalker walker = new ParseTreeWalker();
        EPQLListener listener = new EPQLListener(schemaProvider);
        walker.walk(listener, startContext);

        if (errorListener.exception != null)
            throw errorListener.exception;

        return listener.getNode();
    }

    /**
     * Translates the given DDL command(s).
     *
     * @param command the command to be executed, i.e. a DDL command
     * @param schemaProvider the schema provider providing information about the stream source schemata
     * @return the resulting DDL operators
     */
    public static List<DDLOperator> execute(String command, SchemaProvider schemaProvider) throws ParseException {
        // Create a new lexer instance
        EPQLLexer lexer = new EPQLLexer(new ANTLRInputStream(command));

        // Get a list of matched tokens
        CommonTokenStream tokens = new CommonTokenStream(lexer);

        // Pass the tokens to the parser
        EPQLParser parser = new EPQLParser(tokens);

        lexer.removeErrorListeners();
        parser.removeErrorListeners();
        DescriptiveErrorListener errorListener = new DescriptiveErrorListener();
        lexer.addErrorListener(errorListener);
        parser.addErrorListener(errorListener);

        // Specify out entry point
        EPQLParser.StartContext startContext = parser.start();

        // Walk it and attach our listener
        ParseTreeWalker walker = new ParseTreeWalker();
        EPQLListener listener = new EPQLListener(schemaProvider);
        walker.walk(listener, startContext);

        if (errorListener.exception != null)
            throw errorListener.exception;

        return listener.getDDLNodes();
    }

    // =================================================================================================================
    // ============================================== Error handling ===================================================
    // =================================================================================================================

    /**
     * Error handler
     */
    protected static class DescriptiveErrorListener extends BaseErrorListener {

        public ParseException exception;

        @Override
        public void syntaxError(Recognizer<?, ?> recognizer, Object offendingSymbol,
                                int line, int charPositionInLine,
                                String msg, RecognitionException e) {

            exception = new ParseException("Error parsing SQL query: line "+line+":"+charPositionInLine+" "+msg, charPositionInLine);
        }
    }
}
