/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.queryengine.optimization;

import de.umr.ep.queryengine.predicates.*;
import de.umr.ep.queryengine.predicates.arithmetic.ArithmeticOperation;
import de.umr.ep.queryengine.predicates.arithmetic.ArithmeticOperationTerm;
import de.umr.ep.queryengine.predicates.arithmetic.ArithmeticTerm;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

/**
 * Optimizer for boolean expressions.
 * This optimizer allows to define transformation rules as well as their application order.
 *
 * @author Marc Seidemann
 */
public class BooleanExpressionOptimizer {

    /**
     * The transformation rules to be applied (in appliance order).
     */
    private List<TransformationRule> transformationRules;

    // ===========================================================================================

    /**
     * Creates a new boolean expression optimizer.
     */
    public BooleanExpressionOptimizer() {
        this.transformationRules = new ArrayList<>();
    }

    // ===========================================================================================

    /**
     * Adds a new transformation rule.
     *
     * @param rule the rule to be added
     */
    public void addTransformationRule(TransformationRule rule) {
        this.transformationRules.add(rule);
    }

    /**
     * Returns the list of transformation rules.
     *
     * @return the transformation rules
     */
    public List<TransformationRule> getTransformationRules() {
        return transformationRules;
    }

    // ===========================================================================================

    /**
     * Applies all transformations on the given boolean expression.
     *
     * @param expression the expression to be optimized
     * @return the resulting boolean expression
     */
    public BooleanExpression optimize(BooleanExpression expression) {
        Stack<Object> stack = new Stack<>();
        transform(expression, stack);
        return (BooleanExpression) transform(stack, expression);
    }

    //====================================================================================

    /**
     * Transforms a boolean expression.
     *
     * @param expression the boolean expression to be transformed.
     * @param path the path to the root expression
     */
    private void transform(BooleanExpression expression, Stack<Object> path) {
        path.push(expression);
        if (expression instanceof NestedExpression || expression instanceof Not)
            transform(expression.getInput(0), path);
        else if (expression.getArity() == 2) {
            Stack<Object> childStack = new Stack<>();
            transform(expression.getInput(0), childStack);
            // Now, child stack only contains elements from the left child
            expression.setInput(0,(BooleanExpression) transform(childStack, expression));
            childStack = new Stack<>();
            transform(expression.getInput(1), childStack);
            expression.setInput(1, (BooleanExpression) transform(childStack, expression));
        } else if (expression instanceof Comparison) {
            Comparison be = (Comparison) expression;
            transform(be.getLeft(), path);
            transform(be.getRight(), path);
        } else if (expression instanceof Exists) {
            //  do not handle here
        }
    }

    /**
     * Transforms a term.
     *
     * @param expression the term to be transformed.
     * @param path the path to the root expression
     */
    private void transform(Term expression, Stack<Object> path) {
        path.push(expression);
        if (expression instanceof ArithmeticTerm) {
            transform(((ArithmeticTerm) expression).getOperation(), path);
        }
    }

    /**
     * Transforms an arithmetic expression.
     *
     * @param expression the arithmetic expression to be transformed.
     * @param path the path to the root expression
     */
    private void transform(ArithmeticOperation expression, Stack<Object> path) {
        path.push(expression);
        if (expression instanceof ArithmeticOperationTerm) {
            transform(((ArithmeticOperationTerm) expression).getTerm(), path);
        } else {
            for (int i = 0; i < expression.getArity(); i++) {
                transform(expression.getInput(i), path);
            }
        }
    }

    /**
     * Applies the transformations at the given expression stack.
     *
     * @param stack the stack containing all sub-expressions of a branch
     * @param parent the parent boolean expression
     * @return the resulting expression
     */
    private Object transform(Stack<Object> stack, BooleanExpression parent) {
        Object o = null;
        while(!stack.isEmpty()) {
            o = stack.pop();
            if (o != null) {
                for (TransformationRule rule : transformationRules) {
                    if (rule.getTransformationClass().isAssignableFrom(o.getClass())) {
                        Object c = rule.transform(o, stack, parent);
                        if (c != null)
                            stack.push(c);
                    }
                }
            }
        }
        return o;
    }

    //====================================================================================

    /**
     * A transformation rule.
     *
     * @param <C> the type this rule should be applied on
     */
    public static abstract class TransformationRule<C> {

        /**
         * The class this rule should be applied at.
         */
        private Class<C> transformationClass;

        /**
         * Creates a new transformation rule for the given class.
         * The rule will only be applied on objects of the given call or its sub-classes.
         *
         * @param transformationClass the class this rule should be applied at
         */
        public TransformationRule(Class<C> transformationClass) {
            this.transformationClass = transformationClass;
        }

        /**
         * Transforms the given expression C into an object. The result of the expression
         * will be pushed onto the stack and processed further, iff the result is not null.
         * This allows to apply transformation rules iteratively.
         *
         * @param c the object to be transformed
         * @param path the path to the object since the last branch
         * @param parent the parent boolean expression
         * @return the transformed object
         */
        public abstract Object transform(C c, Stack<Object> path,  BooleanExpression parent);

        /**
         * Returns the transformation class this rule refers to.
         *
         * @return the class of the rules to be transformed
         */
        public Class<C> getTransformationClass() {
            return transformationClass;
        }
    }
}
