/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.queryengine.optimization;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.SchemaUtils;
import de.umr.ep.operators.*;
import de.umr.ep.queryengine.LogicalOperator;
import de.umr.ep.queryengine.predicates.*;
import de.umr.ep.queryengine.predicates.arithmetic.ArithmeticOperationTerm;
import de.umr.ep.queryengine.predicates.arithmetic.ArithmeticTerm;

import java.util.*;

/**
 * Logical query optimizer for selection push-down
 * and join predicate push-down.
 *
 * @author Marc Seidemann
 */
public class SelectionPushDownOptimizer {

    public LogicalOperator optimize(LogicalOperator input) {
        LogicalOperator root = input;
        // Iterate plan until a SELECT clause occurs as input!
        // => push down selection and link result of the pushDown method as new input
        Stack<LogicalOperator> stack = new Stack<>();
        if (input instanceof Selection)
            root = pushDown((Selection)root);
        stack.push(root);
        while(!stack.isEmpty()) {
            LogicalOperator node = stack.pop();

            if (node instanceof Projection) {
                Projection p = (Projection)node;
                if (p.getInput() instanceof Selection) {
                    p.setInput( pushDown((Selection)p.getInput()) );
                }
                stack.push(p.getInput());
            } else if (node instanceof Aggregation) {
                Aggregation a = (Aggregation)node;
                if (a.getInput() instanceof Selection) {
                    a.setInput( pushDown((Selection)a.getInput()) );
                }
                stack.push(a.getInput());
            } else if (node instanceof Renaming) {
                Renaming r = (Renaming)node;
                if (r.getInput() instanceof Selection) {
                    r.setInput( pushDown((Selection)r.getInput()) );
                }
                stack.push(r.getInput());
            } else if (node instanceof Join) {
                Join j = (Join)node;
                pushDown(j);
                if (j.getLeft() instanceof Selection)
                    j.setLeft( pushDown((Selection)j.getLeft()) );
                if (j.getRight() instanceof Selection)
                    j.setRight( pushDown((Selection)j.getRight()) );
                stack.push(j.getRight());
                stack.push(j.getLeft());
            }
        }
        return root;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////// Selection ////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Pushes down a selection.
     *
     * @param s the selection
     * @return the final operator node this selection is pushed down to.
     */
    private LogicalOperator pushDown(Selection s) {
        LogicalOperator result = s;
        if (s.getInput() instanceof Selection) {
            Selection s2 = (Selection)s.getInput();
            s.setExpression(new And(s.getExpression(), s2.getExpression()));
            s.setInput(s2.getInput());
            return pushDown(s);
        } else if (s.getInput() instanceof Renaming) {
            Renaming r = (Renaming)s.getInput();
            Map<String,String> renaming = r.getRenaming();
            for(String renameString : renaming.keySet()) {
                // Revert the renaming in the selection expression
                renameVariable(s.getExpression(), renaming.get(renameString), renameString);
            }
            s.setInput(r.getInput());
            r.setInput(pushDown(s));
            return r;
        } else if (s.getInput() instanceof Projection) {
            Projection p = (Projection)s.getInput();
            s.setInput(p.getInput());
            p.setInput(pushDown(s));
            return p;
        } else if (s.getInput() instanceof Aggregation) {
            Aggregation a = (Aggregation)s.getInput();
            List<String> selectionVariables = determinePredicateVariables(s.getExpression());
            List<String> aggregateAttributes = SchemaUtils.projectNames(a.getSchema());
            boolean schemaValidityCheck = true;
            for(String var: selectionVariables)
                schemaValidityCheck &= !aggregateAttributes.contains(var);

            if (schemaValidityCheck) {
                s.setInput(a.getInput());
                a.setInput(pushDown(s));
                return a;
            }
        } else if (s.getInput() instanceof Join) {
            Join j = (Join)s.getInput();
            j.setPredicate(new And(j.getPredicate(), s.getExpression()));
            return j;
        }
        return result;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////// Join //////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Pushes down the join predicate to the join sources.
     *
     * @param j the join
     */
    private void pushDown(Join j) {
        PredicatePartitionResult partitionResult = partitionJoinPredicate(j.getPredicate(), j.getLeft().getSchema(), j.getRight().getSchema());

        if (partitionResult.p1 != null)
            j.setLeft( new Selection(j.getLeft(), partitionResult.p1) );
        if (partitionResult.p2 != null)
            j.setRight( new Selection(j.getRight(), partitionResult.p2) );
        if (partitionResult.p3 == null)
            j.setPredicate(new True());
        else
            j.setPredicate(partitionResult.p3);
    }

    /**
     * Partitions the given boolean expression in such a way that its
     * result consists of three different expressions:
     * 1. all expressions that can be validated on the first schema (s1) only
     * 2. all expressions that can be validated on the second schema  only
     * 3. all expressions, that have to evaluated on both schemata
     *
     * @param b the boolean expression to be partitioned
     * @param s1 the first schema
     * @param s2 the second schema
     * @return the result of the partitioned join predicate
     */
    private PredicatePartitionResult partitionJoinPredicate(BooleanExpression b, List<Attribute> s1, List<Attribute> s2) {
        PredicateNode root = PredicateNode.createNode(b, s1, s2);
        root.mark();
        PredicatePartitionResult result = new PredicatePartitionResult();

        Stack<PredicateNode> tree = new Stack<PredicateNode>();
        tree.push(root);

        while(!tree.isEmpty()) {
            PredicateNode n = tree.pop();

            if (n.getType() == PartitionType.P1) {
                result.add(n.getExpression(), PartitionType.P1);
            } else if (n.getType() == PartitionType.P2) {
                result.add(n.getExpression(), PartitionType.P2);
            } else if (n.getType() == PartitionType.P3) {

                if (n instanceof BinaryPredicateNode) { // Binary node
                    BinaryPredicateNode dbn = (BinaryPredicateNode)n;
                    if (dbn.expression instanceof Or) {
                        result.add(dbn.expression, PartitionType.P3);
                    } else {
                        // And => can be splitted!!
                        tree.push(dbn.rightNode);
                        tree.push(dbn.leftNode);
                    }
                } else {
                    UnaryPredicateNode mn = (UnaryPredicateNode)n;
                    if (mn.expression instanceof Not)
                        tree.push(mn.childNode);
                    else
                        result.add(mn.expression, PartitionType.P3);
                }

            }
        }
        return result;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////// Helper /////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Determines all variables in the given boolean expression.
     *
     * @param exp the boolean expression.
     * @return all variable names contained in the expression
     */
    private List<String> determinePredicateVariables(BooleanExpression exp) {
        ArrayList<String> result = new ArrayList<>();
        determinePredicateVariables(exp, result);
        return result;
    }

    /**
     * Determines all variables used in the given boolean expression.
     *
     * @param exp the boolean expression
     * @param l a list to collect all variable names
     */
    private void determinePredicateVariables(BooleanExpression exp, List<String> l) {
        for(int i = 0; i < exp.getArity(); i++)
            determinePredicateVariables(exp.getInput(i), l);
        if (exp instanceof Comparison) {
            determinePredicateVariables(((Comparison)exp).getLeft(), l);
            determinePredicateVariables(((Comparison)exp).getRight(), l);
        }
    }

    /**
     * Renames a variable in the given boolean expression.
     *
     * @param exp the boolean expression
     * @param var the (old) name of the variable to be renamed
     * @param newVar the new name of the variable
     */
    private void renameVariable(BooleanExpression exp, String var, String newVar) {
        for (int i = 0; i < exp.getArity(); i++) {
            renameVariable(exp.getInput(i), var, newVar);
        }
        if (exp instanceof Comparison) {
            renameVariable(((Comparison) exp).getLeft(), var, newVar);
            renameVariable(((Comparison) exp).getRight(), var, newVar);
        }
    }

    /**
     * Renames a variable in the given term.
     *
     * @param term the term
     * @param var the (old) name of the variable to be renamed
     * @param newVar the new name of the variable
     */
    private void renameVariable(Term term, String var, String newVar) {
        if (term instanceof Variable) {
            Variable v = (Variable)term;
            if (v.name.equals(var))
                v.name = newVar;
        }
    }

    /**
     * Determines the variable from the given term.
     *
     * @param term the term
     * @param l a list to collect all variable names
     */
    private void determinePredicateVariables(Term term, List<String> l) {
        if (term instanceof Variable)
            l.add(((Variable)term).name);
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////// Helper Classes /////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * The result of a predicate partitioning.
     * The expression p1 refers to the given schema (in case of a join: the left schema),
     * the expression p2 refers to a not defined schema
     * (in case of a join: the right schema)
     * and expression p3 refers to not only to p1
     */
    private static class PredicatePartitionResult {
        /**
         * Expression referring only to the given schema.
         */
        BooleanExpression p1;
        /**
         * Expression referring not at all to the given schema.
         */
        BooleanExpression p2;
        /**
         * Expression referring partially to the given schema.
         */
        BooleanExpression p3;

        /**
         * Adds the given expression to the partition result.
         *
         * @param b the expression
         * @param type defines to which expression the given one
         *             should be added to
         */
        public void add(BooleanExpression b, PartitionType type) {
            switch (type) {
                case P1:
                    if (p1 == null)
                        p1 = b;
                    else
                        p1 = new And(p1,b);
                    break;
                case P2:
                    if (p2 == null)
                        p2 = b;
                    else
                        p2 = new And(p2,b);
                    break;
                case P3:
                    if (p3 == null)
                        p3 = b;
                    else
                        p3 = new And(p3,b);
                    break;
            }
        }
    }

    /**
     * Defines the type of a partition.
     */
    private enum PartitionType {
        P1,P2,P3,CONSTANT
    }

    /**
     * Models an abstract node in the predicate tree.
     */
    private static abstract class PredicateNode {

        protected PartitionType type;
        protected List<Attribute> s1;
        protected List<Attribute> s2;
        public abstract PartitionType mark();

        /**
         * Creates a new PredicateNode for the given schema.
         *
         * @param s1 the schema to partition by
         * @param s2 the second schema
         */
        public PredicateNode(List<Attribute> s1, List<Attribute> s2) {
            this.s1 = s1;
            this.s2 = s2;
        }

        /**
         * Returns this node`s partition type.
         *
         * @return the type of the partition
         */
        public PartitionType getType() {
            return type;
        }

        /**
         * Creates a new Predicate node for the given boolean expression.
         *
         * @param be the boolean expression
         * @param s1 the schema for the first partition
         * @param s2 the second schema
         * @return the predicate node corresponding to the given expression
         */
        public static PredicateNode createNode(BooleanExpression be, List<Attribute> s1, List<Attribute> s2) {
            if (be instanceof BooleanExpression && be.getArity() == 2)
                return new BinaryPredicateNode(be,s1,s2);
            else
                return new UnaryPredicateNode(be,s1,s2);
        }

        /**
         * Returns the boolean expression that corresponds to this node.
         *
         * @return the boolean (sub-)expression corresponding to this node
         */
        public abstract BooleanExpression getExpression();

        /**
         * Determines the partition type from to inputs.
         *
         * @param t1 the first partition type
         * @param t2 the second partition type
         * @return the resulting partition type
         */
        protected PartitionType merge(PartitionType t1, PartitionType t2) {
            if (t1 == t2)
                return t1;
            else if (t1 == PartitionType.CONSTANT)
                return t2;
            else if (t2 == PartitionType.CONSTANT)
                return t1;
            else
                return PartitionType.P3;
        }
    }

    /**
     * Models an unary node in the predicate tree.
     */
    private static class UnaryPredicateNode extends PredicateNode {

        protected PredicateNode childNode;
        protected BooleanExpression expression;

        public UnaryPredicateNode(BooleanExpression exp, List<Attribute> s1, List<Attribute> s2) {
            super(s1,s2);
            this.expression = exp;
        }

        @Override
        public PartitionType mark() {

            if (expression instanceof Comparison) {
                Comparison c = (Comparison)expression;
                type = merge( mark(c.getLeft(),true), mark(c.getRight(),false) );
            } else if (expression instanceof Not) {
                Not n = (Not)expression;
                childNode = createNode(n.getInput(0), s1, s2);
                type = childNode.mark();
            }
            return type;
        }

        /**
         * Returns the partition type of a term.
         *
         * @param t the term which partition type should
         *          be determined
         * @return the partition type of t
         */
        protected PartitionType mark(Term t, boolean left) {
            Variable v;
            if (t instanceof Constant)
                return PartitionType.CONSTANT;
            if (t instanceof ArithmeticTerm) {
                ArithmeticTerm at = (ArithmeticTerm)t;
                if (at.getOperation() instanceof ArithmeticOperationTerm) {
                    ArithmeticOperationTerm aot = (ArithmeticOperationTerm) at.getOperation();
                    return mark(aot.getTerm(), left);
                } else
                    return PartitionType.P3; // Cannot be partitioned
            } else
                v = (Variable)t;

            if (left) {
                if (SchemaUtils.containsAttribute(s1, v.name))
                    return PartitionType.P1;
                else
                    return PartitionType.P2;
            } else {
                if (SchemaUtils.containsAttribute(s2, v.name))
                    return PartitionType.P2;
                else
                    return PartitionType.P1;
            }
        }

        @Override
        public BooleanExpression getExpression() {
            return expression;
        }
    }

    /**
     * Models a binary node in the predicate tree.
     */
    private static class BinaryPredicateNode extends PredicateNode {

        protected BooleanExpression expression;
        protected PredicateNode leftNode;
        protected PredicateNode rightNode;

        public BinaryPredicateNode(BooleanExpression exp, List<Attribute> s1, List<Attribute> s2) {
            super(s1,s2);
            this.expression = exp;
        }

        @Override
        public BooleanExpression getExpression() {
            return expression;
        }

        @Override
        public PartitionType mark() {
            leftNode = createNode(expression.getInput(0), s1, s2);
            rightNode = createNode(expression.getInput(1), s1, s2);
            type = merge(leftNode.mark(), rightNode.mark());
            return type;
        }
    }
}
