/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.queryengine.predicates;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.SchemaUtils;
import de.umr.ep.operators.Selection;
import de.umr.ep.operators.Subquery;
import de.umr.ep.queryengine.LogicalOperator;
import de.umr.ep.queryengine.predicates.arithmetic.ArithmeticOperation;
import de.umr.ep.queryengine.predicates.arithmetic.ArithmeticOperationTerm;
import de.umr.ep.queryengine.predicates.arithmetic.ArithmeticTerm;
import de.umr.ep.queryengine.predicates.arithmetic.NestedOperation;

import java.util.ArrayList;
import java.util.List;

/**
 * Models a scalar subquery. A scalar subquery may only return
 * a single scalar value.
 *
 * @author Marc Seidemann
 */
public class ScalarSubquery extends Term {

    /**
     * The subquery that calculates the value
     * for this scalar subquery.
     */
    protected Subquery subquery;

    /**
     * Creates a new scalar subquery for the given subquery.
     *
     * @param subquery the subquery delivering scalar values
     */
    public ScalarSubquery(Subquery subquery) {
        this.subquery = subquery;
    }

    public Subquery getSubquery() {
        return subquery;
    }

    @Override
    public String toString() {
        return subquery.toString();
    }


    /**
     * Returns if this subquery is correlated.
     *
     * @param schema the schema to check for correlation
     * @return
     */
    public boolean isCorrelated(List<Attribute> schema) {
        List<String> correlatedAttributes = getCorrelatedAttributes(schema);
        return correlatedAttributes.size()>0;
    }

    public LogicalOperator getCorrelationTarget(List<Attribute> schema) {
        return checkCorrelation(subquery.getInput(), new ArrayList<String>(), schema);
    }

    /**
     * Returns the list of correlated attributes.
     *
     * @return the list of correlated attributes
     */
    public List<String> getCorrelatedAttributes(List<Attribute> schema) {
        List<String> correlatedAttributes = new ArrayList<>();
        checkCorrelation(this.getSubquery(), correlatedAttributes, schema);
        return correlatedAttributes;
    }

    /**
     * Gathers correlated attributes in the given list
     * and returns the found selection operator using
     * this correlation.
     * If this list is empty, the subquery is not correlated.
     * Otherwise, the subquery is correlated. In that case,
     * each correlated attribute has to be checked to be
     * contained in the event`s schema and the subquery`s
     * events have to be extended with the correlated attribute values.
     *
     * @param node the operator node to check
     * @param att a list to schemaProvider correlated attributes in
     * @param schema the input event`s schema
     * @return a list with the correlated attribute names
     */
    public LogicalOperator checkCorrelation(LogicalOperator node, List<String> att, List<Attribute> schema) {
        if (node instanceof Selection) {
            Selection s = (Selection)node;
            BooleanExpression be = s.getExpression();
            checkCorrelation(be, att, s.getSchema());
            if (att.size() > 0)
                return s;
        } else {
            LogicalOperator s = null;
            for(int i = 0; i < node.getInputCount(); i++) {
                s = checkCorrelation(node.getInput(i), att, node.getInput(i).getSchema());
                if (s != null)
                    return s;
            }
        }
        return null;
    }

    /**
     * Gathers correlated attributes in the given list.
     * If this list is empty, the subquery is not correlated.
     * Otherwise, the subquery is correlated. In that case,
     * each correlated attribute has to be checked to be
     * contained in the event`s schema and the subquery`s
     * events have to be extended with the correlated attribute values.
     *
     * @param be the boolean expression to check
     * @param att a list to schemaProvider correlated attributes in
     * @param s the input event`s schema
     */
    private void checkCorrelation(BooleanExpression be, List<String> att, List<Attribute> s) {
        if (be instanceof Comparison) {
            Comparison c = (Comparison)be;
            checkCorrelation(c.getLeft(), att, s);
            checkCorrelation(c.getRight(), att, s);
        } else if (be.getArity() == 2) {
            checkCorrelation(be.getInput(0), att, s);
            checkCorrelation(be.getInput(1), att, s);
        } else if (be instanceof NestedExpression) {
            checkCorrelation( be.getInput(0), att, s);
        } else if (be instanceof Not) {
            checkCorrelation( be.getInput(0), att, s);
        }
    }

    /**
     * Gathers correlated attributes in the given list.
     * If this list is empty, the subquery is not correlated.
     * Otherwise, the subquery is correlated. In that case,
     * each correlated attribute has to be checked to be
     * contained in the event`s schema and the subquery`s
     * events have to be extended with the correlated attribute values.
     *
     * @param t the term to check
     * @param att a list to schemaProvider correlated attributes in
     * @param s the input event`s schema
     */
    private void checkCorrelation(Term t, List<String> att, List<Attribute> s) {
        if (t instanceof Variable) {
            Variable v = (Variable)t;
            if (!SchemaUtils.containsAttribute(s, v.name) && !att.contains(v.name))
                att.add(v.name);
        } else if (t instanceof ArithmeticTerm) {
            ArithmeticTerm at = (ArithmeticTerm) t;
            checkCorrelation(at.getOperation(), att, s);
        }
    }

    /**
     * Gathers correlated attributes in the given list.
     * If this list is empty, the subquery is not correlated.
     * Otherwise, the subquery is correlated. In that case,
     * each correlated attribute has to be checked to be
     * contained in the event`s schema and the subquery`s
     * events have to be extended with the correlated attribute values.
     *
     * @param op the arithmetic operation to check
     * @param att a list to schemaProvider correlated attributes in
     * @param s the input event`s schema
     */
    private void checkCorrelation(ArithmeticOperation op, List<String> att, List<Attribute> s) {
        if (op instanceof ArithmeticOperationTerm) {
            ArithmeticOperationTerm aot = (ArithmeticOperationTerm)op;
            checkCorrelation(aot.getTerm(), att, s);
        } else if (op.getArity() == 2) {
            checkCorrelation(op.getInput(0), att, s);
            checkCorrelation(op.getInput(1), att, s);
        } else if (op instanceof NestedOperation) {
            checkCorrelation(op.getInput(0), att, s);
        }
    }
}
