/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.queryengine.predicates.pattern;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.SchemaUtils;
import de.umr.ep.event.TimeRepresentation;
import de.umr.ep.queryengine.LogicalOperator;

import java.util.ArrayList;
import java.util.List;

/**
 * Implementation of a pattern matching node.
 *
 * @author Marc Seidemann
 *
 */
public class PatternMatching extends LogicalOperator {

    /**
     * The input to match.
     */
    private LogicalOperator input;

    /**
     * The pattern to be matched.
     */
    private PatternDefinition pattern;

    /**
     * The list of measured attributes.
     */
    private List<MeasureAttribute> attributes;

    /**
     * Creates a new pattern matcher.
     *
     * @param input the input operator
     * @param pattern the pattern to match
     * @param attributes the output schema definition
     */
    public PatternMatching(LogicalOperator input, PatternDefinition pattern, List<MeasureAttribute> attributes) {
        this.input = input;
        this.pattern = pattern;
        this.attributes = attributes;
        ensureTimestamps();
    }

    /**
     * Adds timestamp attributes to the schema (could be relaxed in further version).
     */
    private void ensureTimestamps() {
        if (TimeRepresentation.fromSchema(input.getSchema()) == TimeRepresentation.INTERVAL) {
            if (!attributes.get(attributes.size()-1).getAttributeName().equals("tend")) {
                attributes.add(new MeasureAttribute(attributes.get(attributes.size()-1).getSymbolName(), "tend", "tend"));
            }
            if (!attributes.get(attributes.size()-2).getAttributeName().equals("tstart")) {
                attributes.add(attributes.size()-1, new MeasureAttribute(attributes.get(attributes.size()-1).getSymbolName(), "tstart", "tstart"));
            }
        } else if (TimeRepresentation.fromSchema(input.getSchema()) == TimeRepresentation.POINT) {
            if (!attributes.get(attributes.size()-1).getAttributeName().equals("tstart")) {
                attributes.add(new MeasureAttribute(attributes.get(attributes.size()-1).getSymbolName(), "tstart", "tstart"));
            }
        }
    }

    /**
     * Returns the pattern
     *
     * @return the pattern
     */
    public PatternDefinition getPattern() {
        return pattern;
    }

    /**
     * Returns the pattern matching input.
     *
     * @return the input operator
     */
    public LogicalOperator getInput() {
        return input;
    }

    /**
     * Returns the measured attributes.
     *
     * @return the list of attributes to be measured
     */
    public List<MeasureAttribute> getAttributes() {
        return attributes;
    }

    @Override
    public List<Attribute> getSchema() {
        // Infers the schema from the defined attributes
        List<Attribute> s = input.getSchema();
        List<Attribute> result = new ArrayList<>(attributes.size());
        for(int i = 0; i < attributes.size(); i++) {
            Attribute att = new Attribute(SchemaUtils.getAttribute(s, attributes.get(i).getAttributeName()));
            att.setName(attributes.get(i).getRenaming());
            result.add(att);
        }
        return result;
    }

    @Override
    public void setInput(LogicalOperator input, int i) {
        this.input = input;
    }

    @Override
    public int getInputCount() {
        return 1;
    }

    @Override
    public LogicalOperator getInput(int i) {
        return input;
    }
}
