/*
 * Copyright (c) 2015
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.queryengine.predicates.pattern;

import de.umr.ep.queryengine.predicates.BooleanExpression;
import de.umr.ep.queryengine.predicates.True;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Representation of an symbol, i.e. the atomic unit of
 * a pattern expression.
 *
 * @author Marc Seidemann
 */
public class Symbol extends Pattern {

    /** The condition that defines the given symbol */
    private BooleanExpression condition;

    /** The symbol to be represented */
    private char symbol;

    /**
     * Defines variable bindings
     */
    private List<PatternVariable> bindings;

    /**
     * Creates a new symbol without variable binding, accepting all inputs.
     *
     * @param symbol the symbol representing this pattern
     */
    public Symbol(char symbol) {
        this(symbol, new True());
    }

    /**
     * Creates a new symbol without variable binding, accepting only such
     * states evaluating the given condition to true
     *
     * @param symbol the symbol representing this pattern
     * @param condition the condition defining the symbol
     */
    public Symbol(char symbol, BooleanExpression condition) {
        this(symbol, condition, new PatternVariable[0]);
    }

    /**
     * Creates a new symbol with the given variable binding, accepting only such
     * states evaluating the given condition to true
     *
     * @param symbol the symbol representing this pattern
     * @param condition the condition defining the symbol
     * @param bindings the variable bindings
     */
    public Symbol(char symbol, BooleanExpression condition, PatternVariable... bindings) {
        this.symbol = symbol;
        this.condition = condition;
        this.bindings = new ArrayList<>();
        this.bindings.addAll(Arrays.asList(bindings));
    }

    /**
     * Returns all variable bindings.
     *
     * @return the variable bindings
     */
    public PatternVariable[] getBindings() {
        PatternVariable[] out = new PatternVariable[bindings.size()];
        return bindings.toArray(out);
    }

    /**
     * Adds a new variable binding.
     *
     * @param binding the variable binding
     */
    public void addBinding(PatternVariable binding) {
        bindings.add(binding);
    }

    /**
     * Returns the symbol representation.
     *
     * @return the symbol representation
     */
    public char getSymbol() {
        return symbol;
    }

    /**
     * Returns the symbolic representation of this pattern.
     *
     * @return the symbolic representation of this pattern
     */
    public String getPattern() {
        return symbol+"";
    }

    /**
     * Returns the condition for this symbol.
     *
     * @return the condition defining this symbol
     */
    public BooleanExpression getCondition() {
        return condition;
    }

    /**
     * Sets the condition for this symbol.
     *
     * @param condition the condition for this symbol
     */
    public void setCondition(BooleanExpression condition) {
        this.condition = condition;
    }

    @Override
    public List<Symbol> getSymbolDefinition() {
        List<Symbol> s = new ArrayList<>();
        s.add(this);
        return s;
    }

    @Override
    public Pattern copy() {
        PatternVariable[] bind = new PatternVariable[bindings.size()];
        bind = bindings.toArray(bind);
        return new Symbol(symbol, condition, bind);
    }

    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof Symbol))
            return false;
        Symbol s = (Symbol)obj;
        return this.symbol == s.symbol;
    }

    @Override
    public String toString() {
        String result = symbol + " : ["+condition.toString()+"]";
        if (bindings.size() > 0) {
            result += " ; ";
            for (int i = 0; i < bindings.size(); i++) {
                result += bindings.get(i);
                if (i < bindings.size()-1)
                    result += ", ";
            }
        }
        return result;
    }
}
