/*
 * Copyright (c) 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.ep.queryengine.ql.grammar;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import de.umr.ep.queryengine.ql.generated.EPQLBaseListener;
import org.antlr.v4.runtime.misc.NotNull;

import de.umr.ep.common.AggregationAttribute;
import de.umr.ep.common.Attribute;
import de.umr.ep.common.CalculationAttribute;
import de.umr.ep.common.ComparableNumber;
import de.umr.ep.common.SchemaProvider;
import de.umr.ep.common.SchemaUtils;
import de.umr.ep.common.SortAttribute;
import de.umr.ep.conversions.Conversion;
import de.umr.ep.operators.Aggregation;
import de.umr.ep.operators.Calculation;
import de.umr.ep.operators.CrossProduct;
import de.umr.ep.operators.Distinct;
import de.umr.ep.operators.Except;
import de.umr.ep.operators.From;
import de.umr.ep.operators.Grouping;
import de.umr.ep.operators.Join;
import de.umr.ep.operators.Limit;
import de.umr.ep.operators.Projection;
import de.umr.ep.operators.Renaming;
import de.umr.ep.operators.Selection;
import de.umr.ep.operators.Sort;
import de.umr.ep.operators.Subquery;
import de.umr.ep.operators.Union;
import de.umr.ep.operators.WindowGrouping;
import de.umr.ep.operators.Within;
import de.umr.ep.operators.ddl.Create;
import de.umr.ep.operators.ddl.DDLOperator;
import de.umr.ep.operators.ddl.Drop;
import de.umr.ep.operators.ddl.Insert;
import de.umr.ep.operators.windows.JumpingWindow;
import de.umr.ep.queryengine.LogicalOperator;
import de.umr.ep.queryengine.predicates.And;
import de.umr.ep.queryengine.predicates.BooleanExpression;
import de.umr.ep.queryengine.predicates.Constant;
import de.umr.ep.queryengine.predicates.Equal;
import de.umr.ep.queryengine.predicates.Exists;
import de.umr.ep.queryengine.predicates.False;
import de.umr.ep.queryengine.predicates.Greater;
import de.umr.ep.queryengine.predicates.GreaterEqual;
import de.umr.ep.queryengine.predicates.Like;
import de.umr.ep.queryengine.predicates.Lower;
import de.umr.ep.queryengine.predicates.LowerEqual;
import de.umr.ep.queryengine.predicates.NestedExpression;
import de.umr.ep.queryengine.predicates.Not;
import de.umr.ep.queryengine.predicates.Or;
import de.umr.ep.queryengine.predicates.ScalarSubquery;
import de.umr.ep.queryengine.predicates.Term;
import de.umr.ep.queryengine.predicates.True;
import de.umr.ep.queryengine.predicates.Unequal;
import de.umr.ep.queryengine.predicates.Variable;
import de.umr.ep.queryengine.predicates.arithmetic.Add;
import de.umr.ep.queryengine.predicates.arithmetic.ArithmeticOperation;
import de.umr.ep.queryengine.predicates.arithmetic.ArithmeticOperationTerm;
import de.umr.ep.queryengine.predicates.arithmetic.ArithmeticTerm;
import de.umr.ep.queryengine.predicates.arithmetic.Divide;
import de.umr.ep.queryengine.predicates.arithmetic.Modulo;
import de.umr.ep.queryengine.predicates.arithmetic.Multiply;
import de.umr.ep.queryengine.predicates.arithmetic.NestedOperation;
import de.umr.ep.queryengine.predicates.arithmetic.Subtract;
import de.umr.ep.queryengine.predicates.pattern.Binding;
import de.umr.ep.queryengine.predicates.pattern.KleenePlus;
import de.umr.ep.queryengine.predicates.pattern.KleeneStar;
import de.umr.ep.queryengine.predicates.pattern.MeasureAttribute;
import de.umr.ep.queryengine.predicates.pattern.Pattern;
import de.umr.ep.queryengine.predicates.pattern.PatternDefinition;
import de.umr.ep.queryengine.predicates.pattern.PatternMatching;
import de.umr.ep.queryengine.predicates.pattern.PatternVariable;
import de.umr.ep.queryengine.predicates.pattern.Symbol;
import de.umr.ep.queryengine.predicates.spatial.GeometryConstant;
import de.umr.ep.queryengine.predicates.spatial.SpatialComparison;
import de.umr.ep.queryengine.predicates.spatial.SpatialEnvelope;
import de.umr.ep.queryengine.predicates.spatial.SpatialIsEmpty;
import de.umr.ep.queryengine.predicates.spatial.SpatialIsRectangle;
import de.umr.ep.queryengine.predicates.spatial.SpatialIsSimple;
import de.umr.ep.queryengine.predicates.spatial.SpatialIsValid;
import de.umr.ep.queryengine.predicates.spatial.SpatialOperation;
import de.umr.ep.queryengine.predicates.spatial.SpatialOperationTerm;
import de.umr.ep.queryengine.ql.generated.EPQLParser;
import de.umr.ep.queryengine.ql.generated.EPQLParser.FromClauseContext;

/**
 * Implementation of query language token listener.
 *
 * @author Marc Seidemann
 */
public class EPQLListener extends EPQLBaseListener {

    /**
     * Counter for pattern matching binding variable names
     */
    private static int idCounter = 0;

    /**
     * Represents a parsing context, i.e. the status for a (sub-)query.
     */
    private class Context {
        LogicalOperator root;
        Stack<LogicalOperator> outerQueries = new Stack<>();
        Stack<LogicalOperator> nodes = new Stack<>();
        Stack<BooleanExpression> booleanOp = new Stack<>();
        Stack<ArithmeticOperation> numbers = new Stack<>();
        List<String> attributes = new ArrayList<>();
        List<SortAttribute> sortAttributes = new ArrayList<>();
        List<AggregationAttribute> aggregationAttributes = new ArrayList<>();
        List<CalculationAttribute<?>> calculationAttributes = new ArrayList<>();
        Stack<SpatialOperation> spatialOperations = new Stack<>();
        List<Attribute> creationAttributes = new ArrayList<>();
        List<String> groupByAttributes = new ArrayList<>();
        Map<String, String> renaming = new HashMap<>();
        Stack<Object> op = new Stack<>();
        // Pattern matching
        Map<String, Symbol> symbols = new HashMap<>();
        Stack <Pattern> patterns = new Stack<>();
        List<MeasureAttribute> measureAttributes = new ArrayList<>();
        int currentSymbolIndex;

        List<PatternVariable> symbolBindings = new ArrayList<>();

    }

    // ===================================================================================

    /**
     * The DDL operations, if valid.
     */
    private List<DDLOperator> ddlOperators;
    /**
     * The root of the query.
     */
    private LogicalOperator root;
    /**
     * The currently active context.
     */
    private Context c = new Context();
    /**
     * A stack containing all contexts.
     */
    private Stack<Context> globalStack = new Stack<>();
    /**
     * The event schemaProvider the query is executed on.
     */
    private SchemaProvider schemaProvider;

    // ===================================================================================

    /**
     * Creates a new listener for query tokens.
     *
     * @param schemaProvider the schema provider providing information about the stream source schemata
     */
    public EPQLListener(SchemaProvider schemaProvider) {
        this.schemaProvider = schemaProvider;
        ddlOperators = new LinkedList<>();
    }

    // ===================================================================================
    // ==================================== SELECT =======================================
    // ===================================================================================

    @Override
    public void exitSelectClause(@NotNull EPQLParser.SelectClauseContext ctx) {
        if (ctx.asterisk == null) {
            LogicalOperator input;

            if (ctx.dist != null) {
                input = new Distinct(null, c.attributes.toArray(new String[]{}));
                c.root = input;
                c.nodes.push(input);
                SortAttribute[] sortAttributes = new SortAttribute[c.attributes.size()];
                for(int i = 0; i < sortAttributes.length; i++)
                    sortAttributes[i] = new SortAttribute(c.attributes.get(i));
                input = new Sort(null, sortAttributes);
                c.nodes.peek().setInput(input, 0);
                c.nodes.push(input);
                input = new Projection(null, c.attributes.toArray(new String[]{}));
                c.nodes.peek().setInput(input, 0);
                c.nodes.push(input);
            } else {
                input = new Projection(null, c.attributes.toArray(new String[]{}));
                c.root = input;
                c.nodes.push(input);
            }
            root = c.root;
            if (c.calculationAttributes.size() > 0) {
                input = new Calculation(null, c.calculationAttributes);
                c.nodes.peek().setInput(input, 0);
                c.nodes.push(input);
            }
            if (c.renaming.size() > 0) {
                input = new Renaming(null, c.renaming);
                c.nodes.peek().setInput(input, 0);
                c.nodes.push(input);
            }
            if (c.aggregationAttributes.size() > 0) {
                input = new Aggregation(null, c.aggregationAttributes);
                c.nodes.peek().setInput(input,0);
                c.nodes.push(input);
                c.aggregationAttributes = new ArrayList<>();
            }
        }
        c.attributes.clear();
        c.renaming = new HashMap<>();
        c.calculationAttributes = new ArrayList<>();
        super.exitSelectClause(ctx);
    }

    @Override
    public void exitAggAtt(@NotNull EPQLParser.AggAttContext ctx) {
        String name = (ctx.agg.asterisk != null)?ctx.agg.asterisk.getText():ctx.agg.input.getText();

        AggregationAttribute agg = null;
        if (ctx.agg.method.AVG() != null) {
            agg = new AggregationAttribute(name, AggregationAttribute.Aggregate.AVG);
        } else if (ctx.agg.method.SUM() != null) {
            agg = new AggregationAttribute(name, AggregationAttribute.Aggregate.SUM);
        } else if (ctx.agg.method.COUNT() != null) {
            agg = new AggregationAttribute(name, AggregationAttribute.Aggregate.COUNT);
        } else if (ctx.agg.method.MIN() != null) {
            agg = new AggregationAttribute(name, AggregationAttribute.Aggregate.MIN);
        } else if (ctx.agg.method.MAX() != null) {
            agg = new AggregationAttribute(name, AggregationAttribute.Aggregate.MAX);
        } else if (ctx.agg.method.ST_EXTENT() != null)
            agg = new AggregationAttribute(name, AggregationAttribute.Aggregate.SPATIAL_EXTENT);

        if (ctx.rename != null) {
            String rename = ctx.rename.getText();
            c.renaming.put(agg.getColumnName(), rename);
            c.attributes.add(rename);
        } else
            c.attributes.add(agg.getColumnName());
        c.aggregationAttributes.add(agg);
        super.exitAggAtt(ctx);
    }

    @Override
    public void exitPlainAtt(@NotNull EPQLParser.PlainAttContext ctx) {
        String name = ctx.name.getText();
        if (ctx.rename != null) {
            String rename = ctx.rename.getText();
            c.renaming.put(name, rename);
            name = rename;
        }
        c.attributes.add(name);
        super.exitPlainAtt(ctx);
    }

    // ===================================================================================
    // ===================================== FROM ========================================
    // ===================================================================================
    
    @Override
    public void enterFromClause(FromClauseContext ctx) {
        super.enterFromClause(ctx);
        // Save current context
        globalStack.push(c);
        c = new Context();
    }
    

    @Override
    public void exitFromClause(@NotNull EPQLParser.FromClauseContext ctx) {
        LogicalOperator source = c.nodes.pop();
        while ( !c.nodes.isEmpty() ) {
            source = new CrossProduct(source, c.nodes.pop() );
        }
        c = globalStack.pop();
        
        if (!c.nodes.isEmpty())
            c.nodes.peek().setInput(source, 0);
        else {
            root = source;
            c.root = source;
        }
        c.nodes.push(source);
        super.exitFromClause(ctx);
    }

   

    @Override
    public void exitStream(@NotNull EPQLParser.StreamContext ctx) {
        if (ctx.alias != null)
            c.nodes.push(new From(schemaProvider, ctx.name.getText(), ctx.alias.getText()));
        else
            c.nodes.push(new From(schemaProvider, ctx.name.getText()));
        super.exitStream(ctx);
    }

    // ===================================================================================
    // ====================================== JOIN =======================================
    // ===================================================================================

    @Override
    public void exitJoin(@NotNull EPQLParser.JoinContext ctx) {
        BooleanExpression predicate;

        LogicalOperator right = c.nodes.pop();
        LogicalOperator left = c.nodes.pop();
        if (ctx.join.USING() != null) {
            String rightIdent = SchemaUtils.getIdentifier(right.getSchema());
            String leftIdent = SchemaUtils.getIdentifier(left.getSchema());
            String leftVar = ((leftIdent != null)?leftIdent+"."+ctx.join.jIdent.getText():ctx.join.jIdent.getText());
            String rightVar = ((rightIdent != null)?rightIdent+"."+ctx.join.jIdent.getText():ctx.join.jIdent.getText());
            predicate = new Equal(new Variable(leftVar), new Variable(rightVar));
        }
        else
            predicate = c.booleanOp.pop();
        c.nodes.push( new Join(left, right, predicate));
        super.exitJoin(ctx);
    }

    @Override
    public void exitSubqueryJoin(@NotNull EPQLParser.SubqueryJoinContext ctx) {
        // JOIN in subquery clause
        BooleanExpression predicate;
        LogicalOperator right = c.nodes.pop();
        LogicalOperator left = c.nodes.pop();
        if (ctx.join.USING() != null) {
            String rightIdent = SchemaUtils.getIdentifier(right.getSchema());
            String leftIdent = SchemaUtils.getIdentifier(left.getSchema());
            String leftVar = ((leftIdent != null)?leftIdent+"."+ctx.join.jIdent.getText():ctx.join.jIdent.getText());
            String rightVar = ((rightIdent != null)?rightIdent+"."+ctx.join.jIdent.getText():ctx.join.jIdent.getText());
            predicate = new Equal(new Variable(leftVar), new Variable(rightVar));
        }
        else
            predicate = c.booleanOp.pop();
        if (c.nodes.isEmpty()) {
            c.root = c.nodes.push( new Join(left, right, predicate));
            root = c.root;
        }
        else
            c.nodes.push( new Join(left, right, predicate));
        super.exitSubqueryJoin(ctx);
    }

    // ===================================================================================
    // =================================== Subqueries ====================================
    // ===================================================================================

    @Override
    public void exitASubQueryExp(@NotNull EPQLParser.ASubQueryExpContext ctx) {
        globalStack.peek().nodes.push(new Subquery(root, null));
        c = globalStack.pop();
        root = c.root;
        super.exitASubQueryExp(ctx);
    }

    @Override
    public void enterArithSubQueryRule(@NotNull EPQLParser.ArithSubQueryRuleContext ctx) {
        super.enterArithSubQueryRule(ctx);
//        c.nodes.push(root);
        globalStack.push(c);
        c = new Context();
        
    }

    @Override
    public void exitArithSubQueryRule(@NotNull EPQLParser.ArithSubQueryRuleContext ctx) {
        Context old = c;
        c = globalStack.pop();
        c.numbers.push(new ArithmeticOperationTerm(new ScalarSubquery((Subquery) old.nodes.pop())));
//        root = c.nodes.pop();
        root = c.root;
        super.exitArithSubQueryRule(ctx);
    }

    @Override
    public void enterHavingArithSubQueryRule(@NotNull EPQLParser.HavingArithSubQueryRuleContext ctx) {
        super.enterHavingArithSubQueryRule(ctx);
//        c.nodes.push(root);
        globalStack.push(c);
        c = new Context();
    }

    @Override
    public void exitHavingArithSubQueryRule(@NotNull EPQLParser.HavingArithSubQueryRuleContext ctx) {
        Context old = c;
        c = globalStack.pop();
        c.op.push(new ScalarSubquery((Subquery) old.nodes.pop()));
//        root = c.nodes.pop();
        root = c.root;
        super.exitHavingArithSubQueryRule(ctx);
    }

    @Override
    public void enterSubQueryRule(@NotNull EPQLParser.SubQueryRuleContext ctx) {
        super.enterSubQueryRule(ctx);
        globalStack.push(c);
        c = new Context();
    }

    @Override
    public void enterASubQueryExp(@NotNull EPQLParser.ASubQueryExpContext ctx) {
        super.enterASubQueryExp(ctx);
        globalStack.push(c);
        c = new Context();
    }

    @Override
    public void exitSubQueryRule(@NotNull EPQLParser.SubQueryRuleContext ctx) {
        globalStack.peek().nodes.push(new Subquery(root, ctx.name.getText()));
        c = globalStack.pop();
        root = c.root;
        super.exitSubQueryRule(ctx);
    }

    // ===================================================================================
    // ===================================== WHERE =======================================
    // ===================================================================================

    @Override
    public void exitWhereClause(@NotNull EPQLParser.WhereClauseContext ctx) {
        LogicalOperator from = c.nodes.pop();
        LogicalOperator selection = new Selection(from, c.booleanOp.pop());
        if (!c.nodes.isEmpty())
            c.nodes.peek().setInput(selection, 0);
        else {
            c.root = selection;
            root = c.root;
        }
        c.nodes.push(selection);
        c.nodes.push(from);
        super.exitWhereClause(ctx);
    }

    @Override
    public void exitBoolOpExp(@NotNull EPQLParser.BoolOpExpContext ctx) {
        BooleanExpression right = c.booleanOp.pop();
        BooleanExpression left = c.booleanOp.pop();

        if (ctx.AND() != null)
            c.booleanOp.push(new And(left, right));
        else
            c.booleanOp.push(new Or(left, right));
        super.exitBoolOpExp(ctx);
    }

    @Override
    public void exitTrueBoolExp(@NotNull EPQLParser.TrueBoolExpContext ctx) {
        c.booleanOp.push(new True());
        super.exitTrueBoolExp(ctx);
    }

    @Override
    public void exitFalseBoolExp(@NotNull EPQLParser.FalseBoolExpContext ctx) {
        c.booleanOp.push(new False());
        super.exitFalseBoolExp(ctx);
    }

    @Override
    public void exitNegatedBoolOp(@NotNull EPQLParser.NegatedBoolOpContext ctx) {
        BooleanExpression right = c.booleanOp.pop();
        c.booleanOp.push(new Not(right));
        super.exitNegatedBoolOp(ctx);
    }

    @Override
    public void exitHavingBoolOpExp(@NotNull EPQLParser.HavingBoolOpExpContext ctx) {
        BooleanExpression right = c.booleanOp.pop();
        BooleanExpression left = c.booleanOp.pop();

        if (ctx.AND() != null)
            c.booleanOp.push(new And(left, right));
        else
            c.booleanOp.push(new Or(left, right));
        super.exitHavingBoolOpExp(ctx);
    }

    @Override
    public void exitHavingNegatedBoolOp(@NotNull EPQLParser.HavingNegatedBoolOpContext ctx) {
        BooleanExpression right = c.booleanOp.pop();
        c.booleanOp.push(new Not(right));
        super.exitHavingNegatedBoolOp(ctx);
    }

    @Override
    public void exitBoolParExp(@NotNull EPQLParser.BoolParExpContext ctx) {
        c.booleanOp.push(new NestedExpression(c.booleanOp.pop()));
        super.exitBoolParExp(ctx);
    }

    @Override
    public void exitBoolRelExp(@NotNull EPQLParser.BoolRelExpContext ctx) {
        Term right = new ArithmeticTerm(c.numbers.pop());
        Term left = new ArithmeticTerm(c.numbers.pop());

        // Unpack unnecessary arithmetic
        if (((ArithmeticTerm)right).getOperation() instanceof ArithmeticOperationTerm)
            right = ((ArithmeticOperationTerm)((ArithmeticTerm)right).getOperation()).getTerm();
        if (((ArithmeticTerm)left).getOperation() instanceof ArithmeticOperationTerm)
            left = ((ArithmeticOperationTerm)((ArithmeticTerm)left).getOperation()).getTerm();

        if(ctx.EQUALS() != null)
            c.booleanOp.push( new Equal(left, right));
        else if (ctx.NOT_EQUALS() != null)
            c.booleanOp.push(new Unequal(left, right));
        else if (ctx.GREATER_EQUALS() != null)
            c.booleanOp.push(new GreaterEqual(left, right));
        else if (ctx.GREATER() != null)
            c.booleanOp.push(new Greater(left, right));
        else if (ctx.LESS_EQUALS() != null)
            c.booleanOp.push(new LowerEqual(left, right));
        else if (ctx.LESS() != null)
            c.booleanOp.push(new Lower(left, right));
        else if (ctx.LIKE() != null)
            c.booleanOp.push(new Like(left, right));
        super.exitBoolRelExp(ctx);
    }

    @Override
    public void exitBoolSpatialExp(@NotNull EPQLParser.BoolSpatialExpContext ctx) {
        Term right = (Term) c.op.pop();
        Term left = (Term) c.op.pop();

        if (ctx.ST_EQUALS() != null)
            c.booleanOp.push(new SpatialComparison(SpatialComparison.SpatialPredicate.EQUALS, left, right));
        else if (ctx.ST_CONTAINS() != null)
            c.booleanOp.push(new SpatialComparison(SpatialComparison.SpatialPredicate.CONTAINS,left, right));
        else if (ctx.ST_COVEREDBY() != null)
            c.booleanOp.push(new SpatialComparison(SpatialComparison.SpatialPredicate.COVERED_BY,left, right));
        else if (ctx.ST_COVERS() != null)
            c.booleanOp.push(new SpatialComparison(SpatialComparison.SpatialPredicate.COVERS, left, right));
        else if (ctx.ST_CROSSES() != null)
            c.booleanOp.push(new SpatialComparison(SpatialComparison.SpatialPredicate.CROSSES, left, right));
        else if (ctx.ST_DISJOINT() != null)
            c.booleanOp.push(new SpatialComparison(SpatialComparison.SpatialPredicate.DISJOINT, left, right));
        else if (ctx.ST_EQUALSEXACT() != null)
            c.booleanOp.push(new SpatialComparison(SpatialComparison.SpatialPredicate.EQUALS_EXACT, left, right));
        else if (ctx.ST_INTERSECTS() != null)
            c.booleanOp.push(new SpatialComparison(SpatialComparison.SpatialPredicate.INTERSECTS, left, right));
        else if (ctx.ST_OVERLAPS() != null)
            c.booleanOp.push(new SpatialComparison(SpatialComparison.SpatialPredicate.OVERLAPS, left, right));
        else if (ctx.ST_TOUCHES() != null)
            c.booleanOp.push(new SpatialComparison(SpatialComparison.SpatialPredicate.TOUCHES, left, right));
        else if (ctx.ST_WITHIN() != null)
            c.booleanOp.push(new SpatialComparison(SpatialComparison.SpatialPredicate.WITHIN, left, right));
        super.exitBoolSpatialExp(ctx);

    }

    @Override
    public void exitUnaryBoolSpatialExp(@NotNull EPQLParser.UnaryBoolSpatialExpContext ctx) {
        Term left = (Term) c.op.pop();

        if (ctx.ST_ISEMPTY() != null)
            c.booleanOp.push(new SpatialIsEmpty(left));
        else if (ctx.ST_ISRECTANGLE() != null)
            c.booleanOp.push(new SpatialIsRectangle(left));
        else if (ctx.ST_ISSIMPLE() != null)
            c.booleanOp.push(new SpatialIsSimple(left));
        else if (ctx.ST_ISVALID() != null)
            c.booleanOp.push(new SpatialIsValid(left));
        super.exitUnaryBoolSpatialExp(ctx);
    }

    @Override
    public void enterBoolExistsExp(@NotNull EPQLParser.BoolExistsExpContext ctx) {
        super.enterBoolExistsExp(ctx);
        globalStack.push(c);
        c = new Context();
    }

    @Override
    public void exitBoolExistsExp(@NotNull EPQLParser.BoolExistsExpContext ctx) {
//        Context old = c;
        c = globalStack.pop();
        c.booleanOp.push(new Exists(new ScalarSubquery(new Subquery(root, null))));
//        root = c.nodes.peek();
        root = c.root;
        super.exitBoolExistsExp(ctx);
    }

    @Override
    public void exitAtomAttribute(@NotNull EPQLParser.AtomAttributeContext ctx) {
        if (ctx.prefix != null) {
            String varName = ctx.prefix.getText() + "." + ctx.name.getText();
            if (c.symbols.keySet().contains(ctx.prefix.getText())) {
                // Add variable binding
                String bindingName = generateSymbolId();
                c.op.push(new Binding(bindingName));
                c.symbols.get(ctx.prefix.getText()).addBinding(new PatternVariable(bindingName, ctx.name.getText()));
            } else
                c.op.push(new Variable(varName));
        }
        else
            c.op.push(new Variable(ctx.name.getText()));
        super.exitAtomAttribute(ctx);
    }

    @Override
    public void exitPrevAtomAttribute(@NotNull EPQLParser.PrevAtomAttributeContext ctx) {
        if (c.symbols.size() > 0) {
            String bindingName = generateSymbolId();
            c.op.push(new Binding(bindingName));
            c.symbolBindings.add(new PatternVariable(bindingName, ctx.name.getText()));
        }
        super.exitPrevAtomAttribute(ctx);
    }

    @Override
    public void exitAtomNum(@NotNull EPQLParser.AtomNumContext ctx) {
        c.op.push(new Constant(parseNumber(ctx.getText())));
        super.exitAtomNum(ctx);
    }

    @Override
    public void exitAtomString(@NotNull EPQLParser.AtomStringContext ctx) {
        if (ctx.STRING() != null)
            c.op.push(new Constant(ctx.STRING().getText().replace("\"","")));
        else
            c.op.push(new Constant(ctx.STRING_SINGLE_QUOTED().getText().replace("'","")));
        super.exitAtomString(ctx);
    }

    @Override
    public void exitAtomGeometry(@NotNull EPQLParser.AtomGeometryContext ctx) {
        c.op.push(new GeometryConstant(ctx.getText().toString().replace("\"","")));
        super.exitAtomGeometry(ctx);
    }

    @Override
    public void exitSpatialFunction(@NotNull EPQLParser.SpatialFunctionContext ctx) {
        if (ctx.ST_ENVELOPE() != null) {
            if (c.spatialOperations.isEmpty())
                c.spatialOperations.push(new SpatialEnvelope(new SpatialOperationTerm((Term) c.op.pop())));
            else
                c.spatialOperations.push(new SpatialEnvelope(c.spatialOperations.pop()));
        }
        super.exitSpatialFunction(ctx);
    }

    @SuppressWarnings({ "unchecked", "rawtypes" })
    @Override
    public void exitSpatialAtt(@NotNull EPQLParser.SpatialAttContext ctx) {
        String name = ctx.name.getText();
        c.calculationAttributes.add(new CalculationAttribute(name, new Conversion(c.spatialOperations.pop(), Attribute.DataType.GEOMETRY, Attribute.DataType.STRING)));
        c.attributes.add(name);
        super.exitSpatialAtt(ctx);
    }

    // ===================================================================================
    // ==================================== GROUP BY =====================================
    // ===================================================================================

    @Override
    public void exitGroupByAttribute(@NotNull EPQLParser.GroupByAttributeContext ctx) {
        if (ctx.prefix != null)
            c.groupByAttributes.add(ctx.prefix.getText()+"."+ctx.name.getText());
        else
            c.groupByAttributes.add(ctx.name.getText());
        super.exitGroupByAttribute(ctx);
    }

    @Override
    public void exitGroupByClause(@NotNull EPQLParser.GroupByClauseContext ctx) {
        Stack<LogicalOperator> src = new Stack<>();
        src.push(c.nodes.pop());
        while ( !(c.nodes.peek() instanceof Aggregation) && !(c.nodes.peek() instanceof Projection)) {
            src.push(c.nodes.pop());
        }
        LogicalOperator op = new Grouping(src.peek(), c.groupByAttributes);
        c.nodes.peek().setInput(op,0);
        c.nodes.push(op);
        while(!src.isEmpty())
            c.nodes.push(src.pop());
        super.exitGroupByClause(ctx);
    }

    // ===================================================================================
    // ===================================== HAVING ======================================
    // ===================================================================================

    @Override
    public void exitHavingBoolParExp(@NotNull EPQLParser.HavingBoolParExpContext ctx) {
        c.booleanOp.push(new NestedExpression(c.booleanOp.pop()));
        super.exitHavingBoolParExp(ctx);
    }

    @Override
    public void exitHavingaggregateExpression(@NotNull EPQLParser.HavingaggregateExpressionContext ctx) {

        String name = ctx.prefix != null?ctx.prefix.getText()+"."+ctx.attr.getText():ctx.attr.getText();

        AggregationAttribute agg = null;
        if (ctx.agg.AVG() != null) {
            agg = new AggregationAttribute(name, AggregationAttribute.Aggregate.AVG);
        } else if (ctx.agg.SUM() != null) {
            agg = new AggregationAttribute(name, AggregationAttribute.Aggregate.SUM);
        } else if (ctx.agg.COUNT() != null) {
            agg = new AggregationAttribute(name, AggregationAttribute.Aggregate.COUNT);
        } else if (ctx.agg.MIN() != null) {
            agg = new AggregationAttribute(name, AggregationAttribute.Aggregate.MIN);
        } else if (ctx.agg.MAX() != null) {
            agg = new AggregationAttribute(name, AggregationAttribute.Aggregate.MAX);
        }

        c.aggregationAttributes.add(agg);

        c.op.push(new Variable(agg.getColumnName()));
        c.numbers.push(new ArithmeticOperationTerm(new Variable(agg.getColumnName())));
        super.exitHavingaggregateExpression(ctx);
    }

    @Override
    public void exitHavingBoolRelExp(@NotNull EPQLParser.HavingBoolRelExpContext ctx) {
        Term right = (Term) c.op.pop();
        Term left = (Term) c.op.pop();

        if(ctx.EQUALS() != null)
            c.booleanOp.push( new Equal(left, right));
        else if (ctx.NOT_EQUALS() != null)
            c.booleanOp.push(new Unequal(left, right));
        else if (ctx.GREATER_EQUALS() != null)
            c.booleanOp.push(new GreaterEqual(left, right));
        else if (ctx.GREATER() != null)
            c.booleanOp.push(new Greater(left, right));
        else if (ctx.LESS_EQUALS() != null)
            c.booleanOp.push(new LowerEqual(left, right));
        else if (ctx.LESS() != null)
            c.booleanOp.push(new Lower(left, right));
        else if (ctx.LIKE() != null)
            c.booleanOp.push(new Like(left, right));
        super.exitHavingBoolRelExp(ctx);
    }

    @Override
    public void exitHavingClause(@NotNull EPQLParser.HavingClauseContext ctx) {
        // Push having up to grouping
        Stack<LogicalOperator> src = new Stack<>();
        src.push(c.nodes.pop());
        while ( !(c.nodes.peek() instanceof Aggregation)) {
            src.push(c.nodes.pop());
        }
        src.push(c.nodes.pop());

        Aggregation agg = (Aggregation) src.peek();
        List<Attribute> s = agg.getSchema();
        List<String> aggProj = new ArrayList<>();
        for(int i = 0; i < s.size(); i++)
            aggProj.add(SchemaUtils.getQualifiedAttributeName(s.get(i)));

        LogicalOperator op = new Selection(agg, c.booleanOp.pop());
        Projection p = new Projection(op, aggProj);
        if (!c.nodes.isEmpty())
            c.nodes.peek().setInput(p,0);
        else {
            root = p;
            c.root = p;
        }
        c.nodes.push(p);
        c.nodes.push(op);
        while( !(src.peek() instanceof Aggregation))
            c.nodes.push(src.pop());
        // Add aggregates

        agg.getAggregates().addAll(c.aggregationAttributes);
        c.aggregationAttributes = new ArrayList<>();
        // Add operators back to stack
        while(!src.isEmpty())
            c.nodes.push(src.pop());
        super.exitHavingClause(ctx);
    }

    // ===================================================================================
    // ===================================== WITHIN ======================================
    // ===================================================================================

    @Override
    public void exitWithinClause(@NotNull EPQLParser.WithinClauseContext ctx) {
        ArithmeticOperation end   = c.numbers.pop();
        ArithmeticOperation start = c.numbers.pop();
        
        LogicalOperator source = c.nodes.pop();
        Within within = new Within(source, start, end);

        if (!c.nodes.isEmpty())
            c.nodes.peek().setInput(within, 0);
        else {
            root = within;
            c.root = within;
        }
        c.nodes.push(within);
        c.nodes.push(source);
        super.exitWithinClause(ctx);
    }

    // ===================================================================================
    // ==================================== ORDER BY =====================================
    // ===================================================================================

    @Override
    public void exitOrderByEntry(@NotNull EPQLParser.OrderByEntryContext ctx) {
        SortAttribute.Mode mode = SortAttribute.Mode.ASCENDING;
        if (ctx.DESC() != null)
            mode = SortAttribute.Mode.DESCENDING;
        c.sortAttributes.add(new SortAttribute(ctx.name.getText(), mode));
        super.exitOrderByEntry(ctx);
    }

    @Override
    public void exitOrderByClause(@NotNull EPQLParser.OrderByClauseContext ctx) {
        c.root = new Sort(root, c.sortAttributes.toArray(new SortAttribute[]{}));
        root = c.root;
        super.exitOrderByClause(ctx);
    }

    // ===================================================================================
    // ===================================== LIMIT =======================================
    // ===================================================================================

    @Override
    public void exitLimitClause(@NotNull EPQLParser.LimitClauseContext ctx) {
        c.root = new Limit(root, Long.parseLong(ctx.num.getText()));
        root = c.root;
        super.exitLimitClause(ctx);
    }

    // ===================================================================================
    // ==================================== WINDOW =======================================
    // ===================================================================================

    @Override
    public void exitWindowClause(@NotNull EPQLParser.WindowClauseContext ctx) {

        Stack<LogicalOperator> src = new Stack<>();
        src.push(c.nodes.pop());
        while ( !c.nodes.isEmpty() && !(c.nodes.peek() instanceof Aggregation) && !(c.nodes.peek() instanceof Projection)) {
            src.push(c.nodes.pop());
        }
        LogicalOperator op = new WindowGrouping(src.peek(),  new JumpingWindow(c.numbers.pop()));
        if ( !c.nodes.isEmpty() ) {
            c.nodes.peek().setInput(op,0);
        }
        else {
            root = op;
            c.root = op;
        }
        c.nodes.push(op);
        while(!src.isEmpty())
            c.nodes.push(src.pop());
        super.exitWindowClause(ctx);
    }

    @Override
    public void exitExpressionTime(@NotNull EPQLParser.ExpressionTimeContext ctx) {
        ArithmeticOperation right = c.numbers.pop();
        ArithmeticOperation left = c.numbers.pop();
        if (ctx.ADD() != null) {
            c. numbers.push(new Add(left, right));
        } else if (ctx.SUBTRACT() != null)
            c. numbers.push(new Subtract(left, right));
        super.exitExpressionTime(ctx);
    }

    @Override
    public void exitNowTime(@NotNull EPQLParser.NowTimeContext ctx) {
        c.numbers.push(new ArithmeticOperationTerm(new Constant(System.currentTimeMillis())));
        super.exitNowTime(ctx);
    }

    @Override
    public void exitTimeNumber(@NotNull EPQLParser.TimeNumberContext ctx) {
        Long num = ((Number)parseNumber(ctx.num.getText())).longValue();
        long result = num;

        if (ctx.sec != null)
            result *= 1000L;
        else if (ctx.min != null)
            result *= 1000L * 60L;
        else if (ctx.hour != null)
            result *= 1000L * 60L * 60L;
        else if (ctx.day != null)
            result *= 1000L * 60L * 60L * 24L;
        else if (ctx.week != null)
            result *= 1000L * 60L * 60L * 24L * 7L;
        c.numbers.push(new ArithmeticOperationTerm(new Constant(result)));
        super.exitTimeNumber(ctx);
    }

    // ===================================================================================
    // ============================= Arithmetic Expressions ==============================
    // ===================================================================================

    @Override
    public void exitArithAtomExp(@NotNull EPQLParser.ArithAtomExpContext ctx) {
        Term t = (Term) c.op.peek();
        c.numbers.push(new ArithmeticOperationTerm(t));
        super.exitArithAtomExp(ctx);
    }

    @Override
    public void exitHavingArithAtomExp(@NotNull EPQLParser.HavingArithAtomExpContext ctx) {
        Term t = (Term) c.op.peek();
        c.numbers.push(new ArithmeticOperationTerm(t));
        super.exitHavingArithAtomExp(ctx);
    }

    @Override
    public void exitArithOpExp(@NotNull EPQLParser.ArithOpExpContext ctx) {
        ArithmeticOperation right = c.numbers.pop();
        ArithmeticOperation left = c.numbers.pop();
        // Operator precedence has to be considered!
        // => "*", "/" and "%" bind stronger than "+ and "-"
        if (ctx.DIVIDE() != null || ctx.MULTIPLY() != null || ctx.MODULO() != null) {
            if (left instanceof Add || left instanceof Subtract) {
                ArithmeticOperation op;
                if (ctx.MULTIPLY() != null)
                    op = new Multiply(left.getInput(1), right);
                else if (ctx.DIVIDE() != null)
                    op = new Divide(left.getInput(1), right);
                else
                    op = new Modulo(left.getInput(1), right);
                left.setInput(1, op);
                c.numbers.push(left);
            } else {
                if (ctx.MULTIPLY() != null)
                    c.numbers.push(new Multiply(left, right));
                else if (ctx.DIVIDE() != null)
                    c.numbers.push(new Divide(left, right));
                else if (ctx.MODULO() != null)
                    c.numbers.push(new Modulo(left, right));
            }
        } else {
            if (ctx.ADD() != null)
                c.numbers.push(new Add(left, right));
            else if (ctx.SUBTRACT() != null)
                c.numbers.push(new Subtract(left, right));
        }
        super.exitArithOpExp(ctx);
    }

    @Override
    public void exitHavingArithOpExp(@NotNull EPQLParser.HavingArithOpExpContext ctx) {
        ArithmeticOperation right = c.numbers.pop();
        ArithmeticOperation left = c.numbers.pop();
        // Operator precedence has to be considered!
        // => "*", "/" and "%" bind stronger than "+ and "-"
        if (ctx.DIVIDE() != null || ctx.MULTIPLY() != null || ctx.MODULO() != null) {
            if (left instanceof Add || left instanceof Subtract) {
                ArithmeticOperation op;
                if (ctx.MULTIPLY() != null)
                    op = new Multiply(left.getInput(1), right);
                else if (ctx.DIVIDE() != null)
                    op = new Divide(left.getInput(1), right);
                else
                    op = new Modulo(left.getInput(1), right);
                left.setInput(1, op);
                c.numbers.push(left);
            } else {
                if (ctx.MULTIPLY() != null)
                    c.numbers.push(new Multiply(left, right));
                else if (ctx.DIVIDE() != null)
                    c.numbers.push(new Divide(left, right));
                else if (ctx.MODULO() != null)
                    c.numbers.push(new Modulo(left, right));
            }
        } else {
            if (ctx.ADD() != null)
                c.numbers.push(new Add(left, right));
            else if (ctx.SUBTRACT() != null)
                c.numbers.push(new Subtract(left, right));
        }
        super.exitHavingArithOpExp(ctx);
    }

    @Override
    public void exitArithParExp(@NotNull EPQLParser.ArithParExpContext ctx) {
        c.numbers.push(new NestedOperation(c.numbers.pop()));
        super.exitArithParExp(ctx);
    }

    @Override
    public void exitHavingArithParExp(@NotNull EPQLParser.HavingArithParExpContext ctx) {
        c.numbers.push(new NestedOperation(c.numbers.pop()));
        super.exitHavingArithParExp(ctx);
    }

    @SuppressWarnings({ "unchecked", "rawtypes" })
    @Override
    public void exitArithAtt(@NotNull EPQLParser.ArithAttContext ctx) {
        String name = ctx.name.getText();
        c.calculationAttributes.add(new CalculationAttribute(name, c.numbers.pop()));
        c.attributes.add(name);
        super.exitArithAtt(ctx);
    }

    @Override
    public void exitArithmeticFunc(@NotNull EPQLParser.ArithmeticFuncContext ctx) {
//        String name = ctx.arithmeticFunction().funcName.getText();
        //TODO: parse argument list
        super.exitArithmeticFunc(ctx);
    }


    // ===================================================================================
    // ===================================== UNION =======================================
    // ===================================================================================

    @Override
    public void exitOuterQueryPart(@NotNull EPQLParser.OuterQueryPartContext ctx) {
        c.outerQueries.push(root);
        c.nodes.clear();
        super.exitOuterQueryPart(ctx);
    }

    @Override
    public void exitUnionQuery(@NotNull EPQLParser.UnionQueryContext ctx) {
        LogicalOperator right = c.outerQueries.pop();
        LogicalOperator left = c.outerQueries.pop();
        c.root = new Union(left, right);
        root = c.root;
        c.outerQueries.push(root);
        super.exitUnionQuery(ctx);
    }

    // ===================================================================================
    // ==================================== EXCEPT =======================================
    // ===================================================================================

    @Override
    public void exitExceptQuery(@NotNull EPQLParser.ExceptQueryContext ctx) {
        LogicalOperator right = c.outerQueries.pop();
        LogicalOperator left = c.outerQueries.pop();
        // Both inputs have to be sorted
        root = new Except( sort(left), sort(right), ctx.all!=null);
        c.outerQueries.push(root);
        super.exitExceptQuery(ctx);
    }

    // ===================================================================================
    // =============================== PATTERN MATCHING ==================================
    // ===================================================================================

    @Override
    public void exitMatchRecognizeClause(@NotNull EPQLParser.MatchRecognizeClauseContext ctx) {
        LogicalOperator from = c.nodes.pop();
        LogicalOperator pattern = new PatternMatching(from, new PatternDefinition(c.patterns), c.measureAttributes);

        if (!c.nodes.isEmpty())
            c.nodes.peek().setInput(pattern, 0);
        else {
            c.root = pattern;
            root = c.root;
        }
        c.nodes.push(pattern);
        c.nodes.push(from);
        super.exitMatchRecognizeClause(ctx);
    }

    @Override
    public void exitSymbolDefinition(@NotNull EPQLParser.SymbolDefinitionContext ctx) {
        // Set the boolean expression for the symbol
        String symbol = ctx.symbol.getText();
        BooleanExpression be = c.booleanOp.pop();
        c.symbols.get(symbol).setCondition(be);
        // set possible pattern bindings
        for(int p = c.currentSymbolIndex; p >= 0; p--)
            for (int i = 0; i < c.symbolBindings.size(); i++)
                c.patterns.get(p).getSymbolDefinition().get(0).addBinding(c.symbolBindings.get(i));
        c.symbolBindings.clear();
        c.currentSymbolIndex++;
        super.exitSymbolDefinition(ctx);
    }

    @Override
    public void exitPatternSymbol(@NotNull EPQLParser.PatternSymbolContext ctx) {
        // handle symbol and kleene operator expressions
        Symbol s = new Symbol(ctx.name.getText().charAt(0));
        c.symbols.put(ctx.name.getText(), s); // Put the symbol to the map (required for boolean expression matching)
        Pattern result = s;
        if (ctx.asterisk != null)
            result = new KleeneStar(result);
        else if (ctx.plus != null)
            result = new KleenePlus(result);
        c.patterns.push(result); // Finally, push the pattern to the stack
        super.exitPatternSymbol(ctx);
    }

    @Override
    public void exitPatternDefinition(@NotNull EPQLParser.PatternDefinitionContext ctx) {
        if (ctx.subpattern != null) {
            // TODO: create subpattern !!
        }
        super.exitPatternDefinition(ctx);
    }

    /**
     * Generates a unique ID for pattern matching bindings
     *
     * @return a unique ID in this query
     */
    private String generateSymbolId() {
        return "var"+(idCounter++);
    }

    @Override
    public void exitLastMeasureAttribute(@NotNull EPQLParser.LastMeasureAttributeContext ctx) {
        MeasureAttribute ma = new MeasureAttribute(ctx.symbol.getText(), ctx.attr.getText(), ctx.rename.getText());
        ma.setFunction(MeasureAttribute.MeasureFunction.LAST);
        c.measureAttributes.add(ma);
        super.exitLastMeasureAttribute(ctx);
    }

    @Override
    public void exitFirstMeasureAttribute(@NotNull EPQLParser.FirstMeasureAttributeContext ctx) {
        MeasureAttribute ma = new MeasureAttribute(ctx.symbol.getText(), ctx.attr.getText(), ctx.rename.getText());
        ma.setFunction(MeasureAttribute.MeasureFunction.FIRST);
        c.measureAttributes.add(ma);
        super.exitFirstMeasureAttribute(ctx);
    }

    @Override
    public void exitPlainMeasureAttribute(@NotNull EPQLParser.PlainMeasureAttributeContext ctx) {
        MeasureAttribute ma = new MeasureAttribute(ctx.symbol.getText(), ctx.attr.getText(), ctx.rename.getText());
        c.measureAttributes.add(ma);
        super.exitPlainMeasureAttribute(ctx);
    }

    // ===================================================================================
    // ================================= CREATE STREAM ===================================
    // ===================================================================================

    @Override
    public void exitCreateStreamStatement(@NotNull EPQLParser.CreateStreamStatementContext ctx) {
        List<Attribute> attributes = new ArrayList<>(c.creationAttributes.size());
        attributes.addAll(c.creationAttributes);
        ddlOperators.add(new Create(ctx.name.getText(), attributes));
        c.creationAttributes.clear();
        super.exitCreateStreamStatement(ctx);
    }

    @Override
    public void exitAttributeDefinition(@NotNull EPQLParser.AttributeDefinitionContext ctx) {
        Attribute.DataType type = (Attribute.DataType) c.op.pop();
        Attribute att = new Attribute(ctx.identifier().getText(), type);
        if (type == Attribute.DataType.STRING || type == Attribute.DataType.GEOMETRY) {
            att.setProperty("MaxStringSize", String.valueOf(((ComparableNumber<?>) c.op.pop()).getNumber().intValue()));
        }
        c.creationAttributes.add(att);
        super.exitAttributeDefinition(ctx);
    }

    @Override
    public void exitStringDataType(@NotNull EPQLParser.StringDataTypeContext ctx) {
        c.op.push(parseNumber(ctx.num.getText()));
        c.op.push(Attribute.DataType.STRING);
        super.exitStringDataType(ctx);
    }

    @Override
    public void exitFloatDataType(@NotNull EPQLParser.FloatDataTypeContext ctx) {
        c.op.push(Attribute.DataType.FLOAT);
        super.exitFloatDataType(ctx);
    }

    @Override
    public void exitDoubleDataType(@NotNull EPQLParser.DoubleDataTypeContext ctx) {
        c.op.push(Attribute.DataType.DOUBLE);
        super.exitDoubleDataType(ctx);
    }

    @Override
    public void exitByteDataType(@NotNull EPQLParser.ByteDataTypeContext ctx) {
        c.op.push(Attribute.DataType.BYTE);
        super.exitByteDataType(ctx);
    }

    @Override
    public void exitShortDataType(@NotNull EPQLParser.ShortDataTypeContext ctx) {
        c.op.push(Attribute.DataType.SHORT);
        super.exitShortDataType(ctx);
    }

    @Override
    public void exitIntegerDataType(@NotNull EPQLParser.IntegerDataTypeContext ctx) {
        c.op.push(Attribute.DataType.INTEGER);
        super.exitIntegerDataType(ctx);
    }

    @Override
    public void exitLongDataType(@NotNull EPQLParser.LongDataTypeContext ctx) {
        c.op.push(Attribute.DataType.LONG);
        super.exitLongDataType(ctx);
    }

    // ===================================================================================
    // ================================== DROP STREAM ====================================
    // ===================================================================================

    @Override
    public void exitDropStreamStatement(@NotNull EPQLParser.DropStreamStatementContext ctx) {
        ddlOperators.add(new Drop(ctx.name.getText()));
        super.exitDropStreamStatement(ctx);
    }

    // ===================================================================================
    // ================================ INSERT STATEMENT =================================
    // ===================================================================================

    @Override
    public void exitInsertEventStatement(@NotNull EPQLParser.InsertEventStatementContext ctx) {
        String name = ctx.name.getText();
        List<Attribute> schema = schemaProvider.getSchema(name);
        Stack<Object> partStack = new Stack<>();
        while(!c.op.isEmpty())
            partStack.push(c.op.pop());
        List<Object> obj = new ArrayList<>();
        for(int i = 0; !partStack.isEmpty(); i++) {
            Object c = ((Constant)partStack.pop()).getConstantValue();
            if (c instanceof ComparableNumber) {
                Number n = ((ComparableNumber<?>) c).getNumber();
                switch(schema.get(i).getType()) {
                    case BYTE: c = n.byteValue(); break;
                    case SHORT: c = n.shortValue(); break;
                    case INTEGER: c = n.intValue(); break;
                    case LONG: c = n.longValue(); break;
                    case FLOAT: c = n.floatValue(); break;
                    case DOUBLE: c = n.doubleValue(); break;
                    default:
                    throw new RuntimeException("Argument at index "+i+" has to be of type string!");
                }
            }
            obj.add(c);
        }
        Object[] payload = obj.toArray();
        ddlOperators.add(new Insert(name, payload));
        super.exitInsertEventStatement(ctx);
    }


    // ===================================================================================
    // ===================================== Helper ======================================
    // ===================================================================================

    /**
     * Parses a number string and returns the corresponding number.
     *
     * @param s the number string
     * @return the corresponding number
     */
    private ComparableNumber<?> parseNumber(String s) {
        if (s.endsWith("f") || s.contains("."))
            return new ComparableNumber<Double>(Double.parseDouble(s.replace("f", "")));
        else
            return new ComparableNumber<Long>(Long.parseLong(s));
    }

    /**
     * Returns the parsed root node of the query plan.
     *
     * @return the root node of the query plan.
     */
    public LogicalOperator getNode() {
        return root;
    }

    /**
     * Returns the parsed root node of the query plan.
     *
     * @return the root node of the query plan.
     */
    public List<DDLOperator> getDDLNodes() {
        return ddlOperators;
    }

    /**
     * Sorts the given input and returns the corresponding
     * sort operator.
     *
     * @param input the input to be sorted
     * @return the sort operator sorting the input.
     */
    private LogicalOperator sort(LogicalOperator input) {
        List<Attribute> s = input.getSchema();

        SortAttribute[] sortedAttributes = new SortAttribute[s.size()];
        for(int i = 0; i < sortedAttributes.length; i++)
            sortedAttributes[i] = new SortAttribute(SchemaUtils.getQualifiedAttributeName(s.get(i)));
        return new Sort(input, sortedAttributes);
    }

    
    
}
