/*
 * Copyright (c) 2012, 2013, 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.jepc.v2.addons.epa;

import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.EPException;
import de.umr.ep.common.IncompatibleTypeException;
import de.umr.ep.event.Event;
import de.umr.ep.event.EventSchema;
import de.umr.ep.event.SchemaException;
import de.umr.jepc.v2.addons.epa.tpstream.Situation;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinitions;
import de.umr.jepc.v2.addons.epa.tpstream.aggregates.OutputMapping;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.Matcher.MatcherCallback;
import de.umr.jepc.v2.addons.epa.tpstream.partition.EventPartitioner;
import de.umr.jepc.v2.addons.epa.tpstream.partition.PartitionFactory;
import de.umr.jepc.v2.addons.epa.tpstream.processor.MultiPartitionProcessor;
import de.umr.jepc.v2.addons.epa.tpstream.processor.Processor;
import de.umr.jepc.v2.addons.epa.tpstream.processor.SinglePartitionProcessor;
import de.umr.jepc.v2.api.bridge.EventChannel;
import de.umr.jepc.v2.api.epa.EPA;
import de.umr.jepc.v2.api.epa.UserDefinedEPA;
import de.umr.jepc.v2.api.epa.pattern.Output;
import de.umr.jepc.v2.api.epa.pattern.Partition;
import de.umr.jepc.v2.api.epa.pattern.symbol.Binding;
import de.umr.jepc.v2.api.epa.pattern.symbol.Bindings;
import de.umr.jepc.v2.api.epa.pattern.symbol.NoSuchVariableException;
import de.umr.jepc.v2.bindings.NoSuchBindingException;

/**
 * Base class for situation processing operators
 * 
 * @author Michael Körber
 *
 */
public abstract class AbstractSituationMatcher extends UserDefinedEPA implements MatcherCallback {

	private static final long				serialVersionUID	= 1L;

	private static final Logger				log					= LoggerFactory
			.getLogger(AbstractSituationMatcher.class);

	//////////////////////////////////////////////////////////////////////
	//
	// DEFINITIONS
	//
	//////////////////////////////////////////////////////////////////////

	/** The time window */
	protected final long					window;

	/** The situation stream definitions */
	protected final SituationDefinitions	definitions;

	/** The output definition */
	protected final Output					output;

	/** The partition definition (optional) */
	protected final Partition				partitionBy;

	//////////////////////////////////////////////////////////////////////
	//
	// IMPL
	//
	//////////////////////////////////////////////////////////////////////

	/** Processing implementation, depending on partitioning schema */
	private Processor						processor;

	/** The mapping from temporal configurations to output events */
	private final OutputMapping				outputMapping;

	/**
	 * @param input
	 *            The input EPA
	 * @param mode
	 *            The mode of operation
	 * @param window
	 *            The time-window in which successful matches must occur
	 *            completely
	 * @param definitions
	 *            The situation definitions
	 * @param constraints
	 *            The temporal constraints
	 * @param output
	 *            the output definition
	 */
	public AbstractSituationMatcher(EPA input, long window, SituationDefinitions definitions, Output output) {
		this(input, window, definitions, output, new Partition());
	}

	/**
	 * @param input
	 *            The input EPA
	 * @param mode
	 *            The mode of operation
	 * @param window
	 *            The time-window in which successful matches must occur
	 *            completely
	 * @param definitions
	 *            The situation definitions
	 * @param constraints
	 *            The temporal constraints
	 * @param output
	 *            the output definition
	 * @param partitionBy
	 *            the partitioning schema
	 */
	public AbstractSituationMatcher(EPA input, long window, SituationDefinitions definitions, Output output,
			Partition partitionBy) {
		super(input);
		this.window = window;
		this.definitions = definitions;
		this.output = output;
		this.partitionBy = partitionBy;
		this.outputMapping = new OutputMapping(definitions, output);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void process(EventChannel input, Event event) {
		this.processor.process(event);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void destroy() throws EPException {
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void initialize(EventChannel... inputStreams) throws EPException {
		EventSchema schemaIn = inputStreams[0].getSchema();

		PartitionFactory pf = getPartitionFactory(schemaIn, this);
		if ( partitionBy.isEmpty() )
			this.processor = new SinglePartitionProcessor(pf);
		else
			this.processor = new MultiPartitionProcessor(createPartitioner(schemaIn, partitionBy), pf);
	}

	protected abstract PartitionFactory getPartitionFactory(EventSchema schemaIn, MatcherCallback callback)
			throws EPException;

	/**
	 * @param inputSchema
	 *            the schema of incoming events
	 * @param partitionBy
	 *            the partitioning schema
	 * @return the partitioner to use to assign incoming events to a partition
	 * @throws NoSuchVariableException
	 *             if a attribute reference by the given schema does not exist
	 *             in the schema of incoming events
	 */
	private EventPartitioner createPartitioner(EventSchema inputSchema, Partition partitionBy) throws SchemaException {
		// Create event partitioner
		List<Integer> idxList = new ArrayList<>();
		for (String field : partitionBy)
			idxList.add(inputSchema.getAttributeIndex(field));

		int[] indexes = new int[idxList.size()];
		for (int i = 0; i < indexes.length; i++)
			indexes[i] = idxList.get(i);

		return new EventPartitioner(indexes);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void onResult(Situation[] result, long currentTime) {
		try {
			publishResult(outputMapping.create(result, currentTime));
		}
		catch (NoSuchBindingException e) {
			log.error("Requested unbound variable in output", e);
		}
	}

	/**
	 * @return the defined time window
	 */
	public long getWindow() {
		return window;
	}

	/**
	 * @return The situation definitions
	 */
	public SituationDefinitions getDefinitions() {
		return definitions;
	}

	/**
	 * @return The output definition
	 */
	public Output getOutput() {
		return output;
	}

	/**
	 * @return The partitioning schema
	 */
	public Partition getPartitionBy() {
		return partitionBy;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws IncompatibleTypeException
	 */
	@Override
	public EventSchema computeOutputSchema(EventSchema... inputSchemas)
			throws SchemaException, IncompatibleTypeException {
		List<Attribute> attributes = new ArrayList<>();
		Bindings allBindings = definitions.getAllAggregates();
		for (String var : output) {
			Binding b = definitions.findAggregate(var);
			attributes.add(new Attribute(var, b.getValue().getDataType(inputSchemas[0], allBindings)));
		}
		return new EventSchema(attributes.toArray(new Attribute[attributes.size()]));
	}

}
