/*
 * Copyright (c) 2012, 2013, 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.jepc.v2.addons.epa;

import de.umr.ep.common.EPException;
import de.umr.ep.common.IncompatibleTypeException;
import de.umr.ep.event.EventSchema;
import de.umr.jepc.v2.addons.epa.tpstream.Constraints;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinitions;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.Matcher.MatcherCallback;
import de.umr.jepc.v2.addons.epa.tpstream.partition.ClassicAdaptivePartitionFactory;
import de.umr.jepc.v2.addons.epa.tpstream.partition.ClassicPartitionFactory;
import de.umr.jepc.v2.addons.epa.tpstream.partition.LowLatencyPartitionFactory;
import de.umr.jepc.v2.addons.epa.tpstream.partition.PartitionFactory;
import de.umr.jepc.v2.api.bridge.EventChannel;
import de.umr.jepc.v2.api.epa.EPA;
import de.umr.jepc.v2.api.epa.pattern.Output;
import de.umr.jepc.v2.api.epa.pattern.Partition;
import de.umr.jepc.v2.ql.TranslatorException;

/**
 * EPA for matching patterns among streams of situations. Situations are derived
 * from {@link EventChannel event streams} using {@link SituationDefinitions}. Those are
 * conditions and a set of aggregate functions, computing the maximum timespan for
 * which the condition holds true (longest event-sequence) and the aggregate values
 * for the events participating in the sequence.
 * Then, the resulting situation streams are passed to the matching-component, which 
 * matches them against a given pattern. The pattern is specified using {@link Constraints temporal constraints},
 * each of them defining valid temporal relationships between two situations.
 * The output of TPStream is a stream of events, each event assembled from a defined (see {@link Output}) 
 * subset of the computed aggregates.
 * The incoming event stream may be partitioned (see {@link Partition}) by an arbitrary number of attributes
 * and evaluated separately for each partition.
 * Furthermore, TPStream is able to run in two modes:
 * <ol>
 *  <li>{@link Mode#CLASSIC} for classical closed interval semantics. Meaning that matching of the pattern 
 *      is triggered only if the end of a situation is known</li>
 *  <li>{@link Mode#LOW_LATENCY} for low-latency matching, detecting patterns as early as possible, even if
 *      a situation's end is not known yet</li>
 * </ol>
 * 
 * 
 * @author Michael Körber
 *
 */
public class TPStream extends AbstractSituationMatcher {

    private static final long   serialVersionUID = 1L;

    /**
     * Configures the behavior of TPStream
     * @author Michael Körber
     */
    public static enum Mode {

        /**
         * Classical temporal pattern matching with closed interval semantics. 
         */
        CLASSIC,

        /**
         * Enables low-latency matching of temporal patterns.
         */
        LOW_LATENCY,
        
        /**
         * Classic mode with adaptive optimization
         */
        CLASSIC_ADAPTIVE;
    }

    /** The temporal constraints */
    private final Constraints               constraints;

    /** The mode of operation */
    private final Mode                      mode;

    /**
     * @param input The input EPA
     * @param mode The mode of operation
     * @param window The time-window in which successful matches must occur completely
     * @param definitions The situation definitions 
     * @param constraints The temporal constraints
     * @param output the output definition
     */
    public TPStream(EPA input, Mode mode, long window, SituationDefinitions definitions, Constraints constraints, Output output) {
        this(input, mode, window, definitions, constraints, output, new Partition());
    }

    /**
     * @param input The input EPA
     * @param mode The mode of operation
     * @param window The time-window in which successful matches must occur completely
     * @param definitions The situation definitions 
     * @param constraints The temporal constraints
     * @param output the output definition
     * @param partitionBy the partitioning schema
     */
    public TPStream(EPA input, Mode mode, long window, SituationDefinitions definitions, Constraints constraints, Output output, Partition partitionBy) {
        super(input,window,definitions,output,partitionBy);
        this.constraints = constraints;
        this.mode = mode;
    }


    /**
     * {@inheritDoc}
     * @throws IncompatibleTypeException 
     * @throws TranslatorException 
     */
    @Override
    protected PartitionFactory getPartitionFactory(EventSchema schemaIn, MatcherCallback callback) throws EPException {
        if ( mode == Mode.CLASSIC )
            return new ClassicPartitionFactory(schemaIn, definitions, constraints, window, this);
        else if ( mode == Mode.LOW_LATENCY )
            return new LowLatencyPartitionFactory(schemaIn, definitions, constraints, window, this);
        else if ( mode == Mode.CLASSIC_ADAPTIVE) 
            return new ClassicAdaptivePartitionFactory(schemaIn, definitions, constraints, window, this, 0.2);
        else
            throw new UnsupportedOperationException("Unknown Mode: " + mode);
    }
    


    /**
     * @return The temporal constraints
     */
    public Constraints getConstraints() {
        return constraints;
    }

    /**
     * @return The mode of operation
     */
    public Mode getMode() {
        return mode;
    }
}
