/*
 * Copyright (c) 2012, 2013
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.jepc.v2.addons.epa;

import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.PriorityQueue;

import de.umr.ep.common.EPException;
import de.umr.ep.event.Event;
import de.umr.ep.event.EventSchema;
import de.umr.ep.event.SchemaException;
import de.umr.jepc.v2.api.bridge.EventChannel;
import de.umr.jepc.v2.api.epa.EPA;
import de.umr.jepc.v2.api.epa.UserDefinedEPA;

/**
 * EPA that provides the union of multiple schema-equal event streams.
 *
 * @author Michael Körber
 * @author Bastian Hoßbach
 */
public class UnionAll extends UserDefinedEPA {

    /**
     * 
     */
    private static final long       serialVersionUID = 1L;

    /**
       * Holds for each input stream its current clock.
       */
    private Map<EventChannel, Long> clocks;

    /**
     * Buffers all input events that are not yet safe to forward to the output.
     */
    private PriorityQueue<Event>    results;

    /**
     * Constructs a new EPA for providing the union of event streams.
     *
     * @param inputEPAs arbitrary number of EPAs that provide input
     *          streams of the EPA
     */
    public UnionAll(EPA... inputEPAs) {
        // register input streams by invoking constructor of superclass
        super(inputEPAs);
        // init event buffer
        results = new PriorityQueue<>(inputEPAs.length, Comparator.comparingLong(Event::getT1));
        // init clocks
        clocks = new HashMap<>();
    }

    @Override
    public void process(EventChannel input, Event event) {
        // save event and update clock of stream
        results.add(event);
        clocks.put(input, event.getT1());
        // Report all buffered events that became safe due to clock update
        long minTimestamp = getMinTimestamp();
        while (!results.isEmpty() && results.peek().getT1() <= minTimestamp) {
            publishResult(results.poll());
        }
    }

    @Override
    public void initialize(EventChannel... inputStreams) throws EPException {
        EventSchema fst = inputStreams[0].getSchema();
        for (EventChannel rs : inputStreams) {
            if (!fst.equals(rs.getSchema()))
                throw new SchemaException("Union requires all input schemas to be identical.");
        }
        for (EventChannel input : inputStreams) {
            clocks.put(input, Long.MIN_VALUE);
        }
    }

    /**
     * Gets the minimum clock among all input streams.
     *
     * @return minimum clock among all input streams
     */
    private long getMinTimestamp() {
        long minTimestamp = Long.MAX_VALUE;
        for (long v : clocks.values())
            minTimestamp = Math.min(minTimestamp, v);
        return minTimestamp;
    }

    @Override
    public void destroy() {
        while(!results.isEmpty())
            publishResult(results.poll());
    }

	/**
	 * {@inheritDoc}
	 */
	@Override
	public EventSchema computeOutputSchema(EventSchema... inputSchemas) throws SchemaException {
		return inputSchemas[0];
	}
}
