package de.umr.jepc.v2.addons.epa.topk;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.PriorityQueue;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.EPException;
import de.umr.ep.event.Event;
import de.umr.ep.event.EventSchema;
import de.umr.ep.event.SchemaException;
import de.umr.jepc.v2.api.bridge.EventChannel;
import de.umr.jepc.v2.api.epa.EPA;
import de.umr.jepc.v2.api.epa.UserDefinedEPA;

public class GridMonitor extends UserDefinedEPA {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	protected final PriorityQueue<TopKEvent> valids;
	private final Cell[] grid;
	private final double cellExtent;	
	private final int numberCellsOnDim;
	private final int dimension;

	public GridMonitor(EPA input, int dimension, double cellExtent) {
		super(input);
		this.cellExtent = cellExtent;
		//Calculate cell on each dimension
		numberCellsOnDim = (int)Math.ceil(1.0/cellExtent);
		//Build the grid
		grid = new Cell[(int)Math.pow(numberCellsOnDim, dimension)];
		int[] initArr = new int[dimension];
		//Create cells and calculate coordinates
		for(int i = 0; i < grid.length; i++) {
			grid[i] = new Cell(initArr, i, cellExtent);
			initArr[0]++;
			for(int j = 0; j < initArr.length-1; j++) {
				if(initArr[j] == numberCellsOnDim) {
					initArr[j] = 0;
					initArr[j+1]++;
				} else {
					break;
				}
			}
		}
		
		this.dimension = dimension;
		int[] dimensions = new int[dimension];
		for(int i = 0; i < dimensions.length; i++) {
			dimensions[i] = (int) (1.0/cellExtent);
		}		
		//Min-Heap for valid events
		valids = new PriorityQueue<>(new Comparator<TopKEvent>(){
			@Override
			public int compare(TopKEvent arg0, TopKEvent arg1) {
				if(arg0.getEvent().getT2() > arg1.getEvent().getT2()) return 1;
				if(arg0.getEvent().getT2() < arg1.getEvent().getT2()) return -1;
				return 0;
			}}); 
	}
	
	public Cell getCellFromGrid(TopKEvent se) {
		Event event = se.getEvent();
		int sum = 0;
		for(int i = 0; i < dimension; i++) {
			sum += Math.floor(((Double)event.get(i))/cellExtent)*(int)Math.pow(numberCellsOnDim, i);
		}
		return grid[sum];
	}
	
	@Override
	public void process(EventChannel input, Event event) {
		TopKEvent se = new TopKEvent(event);
		//Add event to Grid
		getCellFromGrid(se).addEventToCell(se);
		//Add event to Valids
		valids.add(se);
		HashSet<TMAQuery> processList = new HashSet<>();
		//Check for expired events
		while(valids.peek().getEvent().getT2() < se.getEvent().getT1()) {
			TopKEvent toRemoveEvent = valids.remove();
			Cell c = getCellFromGrid(toRemoveEvent);
			c.removeEventFromCell(toRemoveEvent);
			//If there are influenced queries, check for updates
			for(TMAQuery tkq : c.getInfluenceList().values()) {
				processList.add(tkq);
				tkq.addEventToRemove(toRemoveEvent);
			}
		}		
		//If there influenced queries for the new event, check for updates
		for(TMAQuery tkq : getCellFromGrid(se).getInfluenceList().values()) {
			processList.add(tkq);
			tkq.addEventToAdd(se);
		}
		//For all queries that may be influenced need to process the events 
		for(TMAQuery tkq : processList) {
			tkq.processCycle();
		}
	}
	
	public PriorityQueue<TopKEvent> getValids() {
        return valids;
    }

    public int getDimension() {
        return dimension;
    }

    public Cell[] getGrid() {
        return grid;
    }

    public int getNumberCellsOnDim() {
        return numberCellsOnDim;
    }
	
	/**
     * {@inheritDoc}
     */
    @Override
    public EventSchema computeOutputSchema(EventSchema... inputSchemas) throws SchemaException {
        List<Attribute> out = new ArrayList<>();
        for (int i = 1; i <= inputSchemas[0].getNumAttributes(); i++) {
            for (Attribute a : inputSchemas[0]) {
                out.add(new Attribute(a.getName() + "_" + i, a.getType()));
            }
        }
        return new EventSchema(out.toArray(new Attribute[out.size()]));
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void initialize(EventChannel... inputStreams) throws EPException {
    }
	
	@Override
	public void destroy() throws EPException {
	}
}
