package de.umr.jepc.v2.addons.epa.topk;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.TreeSet;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.EPException;
import de.umr.ep.event.Event;
import de.umr.ep.event.EventSchema;
import de.umr.ep.event.SchemaException;
import de.umr.jepc.v2.api.bridge.EventChannel;
import de.umr.jepc.v2.api.epa.EPA;
import de.umr.jepc.v2.api.epa.UserDefinedEPA;
import de.umr.jepc.v2.api.epa.window.TimeWindow;

public class MinTopK extends UserDefinedEPA {

	private static final long serialVersionUID = 1L;

	private LinkedList<MinTopKWindow> windows = new LinkedList<>();
	private final TreeSet<TopKEvent> superTopK;
	private final long windowSize;
	private final long timeAccuracy;
	private TopKScoringFunction scoringFunction;
	private int k;

	public MinTopK(EPA input, int k, long timeAccuracy, TopKScoringFunction scoringFunction) {
		super(input);
		this.k = k;
		this.timeAccuracy = timeAccuracy;
		this.superTopK = new TreeSet<>();
		this.scoringFunction = scoringFunction;
		windowSize = ((TimeWindow) input).getSize();
		for (int i = 1; i <= Math.ceil((double) windowSize / timeAccuracy); i++) {
			windows.add(new MinTopKWindow(i * timeAccuracy));
		}
	}

	public TreeSet<TopKEvent> outputTopKResults() {
		return superTopK;
	}

	@Override
	public void process(EventChannel input, Event event) {
		if (!windows.isEmpty()) {
			while (event.getT1() >= windows.getFirst().getEndingTime())
				purgeExpiredWindow();
		}
		updateTopK(event);
	}

	private void updateTopK(Event event) {
		double score = scoringFunction.eval(event);
		// Check, if score is high enough
		if (!superTopK.isEmpty() && score < superTopK.last().getScore()) {
			if (windows.getLast().getTopKCounter() == k)
				return;
		}
		TopKEvent emd = new TopKEvent(event, -1, -1, score);
		superTopK.add(emd);

		Iterator<MinTopKWindow> iter = windows.iterator();
		while (iter.hasNext()) {
			MinTopKWindow pw = iter.next();
			if (pw.getEndingTime() > emd.getEvent().getT2()) {
				break;
			}
			// Window is not full yet
			if (pw.getTopKCounter() < k) {
				pw.incrementTopKCounter();
				// Set LBP one up
				if (pw.getLowerBoundPointer() == null || pw.getLowerBoundPointer().getScore() > emd.getScore()) {
					pw.setLowerBoundPointer(emd);
				}
				if (emd.getStartW() == -1) {
					emd.setStartW(pw.getId());
				}
				emd.setEndW(pw.getId());
			} else {
				if (pw.getLowerBoundPointer().getScore() < emd.getScore()) {
					// Move old LBP to the next window
					pw.getLowerBoundPointer().setStartW(pw.getLowerBoundPointer().getStartW() + 1);
					// Delete if it is in no window anymore
					if (pw.getLowerBoundPointer().getStartW() > pw.getLowerBoundPointer().getEndW()) {
						superTopK.remove(pw.getLowerBoundPointer());
					}
					// Set LBP one up
					pw.setLowerBoundPointer(superTopK.lower(pw.getLowerBoundPointer()));
					if (emd.getStartW() == -1) {
						emd.setStartW(pw.getId());
					}
					emd.setEndW(pw.getId());
				}
			}
		}
	}

	private void purgeExpiredWindow() {
		Iterator<TopKEvent> iter = superTopK.iterator();
		// Iterate through first k Element of superTopK
		for (int i = 0; i < k && iter.hasNext(); i++) {
			TopKEvent emd = iter.next();
			emd.setStartW(emd.getStartW() + 1);
			// Remove if StartWindow > EndWindow
			if (emd.getStartW() > emd.getEndW()) {
				iter.remove();
			}
		}
		// Remove window and append a new one
		windows.add(new MinTopKWindow(windows.getLast().getEndingTime() + timeAccuracy));
		windows.remove();
	}
	
	/**
     * {@inheritDoc}
     */
    @Override
    public EventSchema computeOutputSchema(EventSchema... inputSchemas) throws SchemaException {
        List<Attribute> out = new ArrayList<>();
        for (int i = 1; i <= inputSchemas[0].getNumAttributes(); i++) {
            for (Attribute a : inputSchemas[0]) {
                out.add(new Attribute(a.getName() + "_" + i, a.getType()));
            }
        }
        return new EventSchema(out.toArray(new Attribute[out.size()]));
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void initialize(EventChannel... inputStreams) throws EPException {
    }

	@Override
	public void destroy() throws EPException {
	}

    
}
