package de.umr.jepc.v2.addons.epa.topk;

import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.PriorityQueue;
import java.util.TreeSet;

import de.umr.ep.common.IncompatibleTypeException;
import de.umr.ep.event.EventSchema;
import de.umr.jepc.v2.ql.TranslatorException;

public class TMAQuery {

	/**
	 * 
	 */
	protected final int k;

	protected TopKScoringFunction scoringFunction;

	protected TreeSet<TopKEvent> topList;

	protected double topScore = Double.NEGATIVE_INFINITY;
	
	protected final GridMonitor gridMonitor;
	
	protected final Comparator<TopKEvent> eventComparator;
	
	private final List<Cell> checkStillInfluence;
	protected int recomputations = 0;
	
	protected final int id;
	protected static int ids = 0;
	protected List<TopKEvent> addEvents;
	protected List<TopKEvent> removeEvents;
	
	public TMAQuery(GridMonitor gridMonitor, int k, TopKScoringFunction scoringFunction, EventSchema schema) throws TranslatorException, IncompatibleTypeException {
		this.gridMonitor = gridMonitor;
		this.k = k;
		this.id = ids++;
		this.addEvents = new LinkedList<>();
		this.removeEvents = new LinkedList<>();
		this.checkStillInfluence = new LinkedList<>();
        this.scoringFunction = scoringFunction;
        this.eventComparator = new Comparator<TopKEvent>() {
    		@Override
    		public int compare(TopKEvent o1, TopKEvent o2) {
    			if(scoringFunction.eval(o1.getEvent()) < scoringFunction.eval(o2.getEvent())) return 1;
    			if(scoringFunction.eval(o1.getEvent()) > scoringFunction.eval(o2.getEvent())) return -1;
    			if(o1.getEvent().getT2() < o2.getEvent().getT2()) return 1;
    			if(o1.getEvent().getT2() > o2.getEvent().getT2()) return -1;
    			if(o1.getId() < o2.getId()) return 1;
    			if(o1.getId() > o2.getId()) return 1;
    			return 0;
    		}
    	};
		this.topList = new TreeSet<>(eventComparator);
	}
	
	public void processCycle() {
		processTMA();
		addEvents.clear();
		removeEvents.clear();
	}
	
	private void processTMA() {
		if(topList.size() < k) {
			computeTopKEvents();
			return;
		}
		//Insert new Events
		for(TopKEvent toInsertEvent : addEvents) {
			if(scoringFunction.eval(toInsertEvent.getEvent()) >= topScore) {
				topList.add(toInsertEvent);
				while(topList.size() > k) {
					topList.pollLast();
				}
				topScore = scoringFunction.eval(topList.last().getEvent());
			}
		}
		//Delete expired Events
		for(TopKEvent toRemoveEvent : removeEvents) {
			if(scoringFunction.eval(toRemoveEvent.getEvent()) >= topScore) {
				recomputations++;
				computeTopKEvents();
				cleanInfluenceList(checkStillInfluence);
				break;
			}
		}
	}
	
	private void cleanInfluenceList(List<Cell> events) {
		//Clean up the influence region
		while(!events.isEmpty()) {
			Cell c = events.remove(0);
			if(c.removeFromInfluenceList(this)) {
				for(int n : c.getNeighbours()) {
					if(n < 0) {
						continue;
					}
					events.add(gridMonitor.getGrid()[n]);
				}
			}
		}
	}
	
	public TreeSet<TopKEvent> getTopKEvents() {
		if(topList.size() < k) {
			computeTopKEvents();
		}
		return topList;
	}

	protected TreeSet<TopKEvent> computeTopKEvents() {
		topScore = Double.NEGATIVE_INFINITY;
		if(gridMonitor.valids.size() < k) {
			//valids < k, remove Query from influence lists
			LinkedList<Cell> toClean = new LinkedList<>();
			toClean.add(gridMonitor.getGrid()[gridMonitor.getGrid().length-1]);
			cleanInfluenceList(toClean);
			topList.clear();
			topList.addAll(gridMonitor.valids);
			return topList;
		}
		topList.clear();
		Cell[] grid = gridMonitor.getGrid();
		HashSet<Integer> alreadyQueued = new HashSet<>();
		//Order of the cells to check
		PriorityQueue<Cell> cellsToCheck = new PriorityQueue<>(new Comparator<Cell>() {
			@Override
			public int compare(Cell o1, Cell o2) {
				if(scoringFunction.eval(o1.getXyzMax()) < scoringFunction.eval(o2.getXyzMax())) return 1;
				if(scoringFunction.eval(o1.getXyzMax()) > scoringFunction.eval(o2.getXyzMax())) return -1;
				return 0;
			}
		});
		cellsToCheck.add((Cell) grid[grid.length-1]);
		alreadyQueued.add(grid.length-1);
		while(!cellsToCheck.isEmpty()) {
			Cell c = cellsToCheck.remove();
			if(scoringFunction.eval(c.getXyzMax()) < topScore) {
				continue;
			}
			//Add Event to toplist
			for(TopKEvent e : c.getEventList()) {
				double score = scoringFunction.eval(e.getEvent());
				if(topList.size() < k) {
					topList.add(e);
					if(topList.size() == k) {
						topScore = scoringFunction.eval(topList.last().getEvent());
					}
				}
				else if(score > topScore) {
					topList.add(e);
					topList.pollLast();
					topScore = scoringFunction.eval(topList.last().getEvent());
				}
			}
			//Add the neighbours
			for(int x : c.getNeighbours()) {
				if(x >= 0) {
					if(!alreadyQueued.contains(x)) {
						cellsToCheck.add(grid[x]);
						alreadyQueued.add(x);
					}
				}
			}
			c.addToInfluenceList(this);
		}
		checkStillInfluence.clear();
		checkStillInfluence.addAll(cellsToCheck);
		return topList;
	}
	

	public int getId() {
		return id;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + id;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		TMAQuery other = (TMAQuery) obj;
		if (id != other.id)
			return false;
		return true;
	}

	public int getRecomputations() {
		return recomputations;
	}

	public void addEventToRemove(TopKEvent toRemove) {
		removeEvents.add(toRemove);
	}

	public void addEventToAdd(TopKEvent addEvent) {
		addEvents.add(addEvent);
	}
}
