/*
 * Copyright (c) 2012, 2013
 * Database Research Group, University of Marburg.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.jepc.v2.addons.epa.tpstream;

import java.util.HashSet;
import java.util.Set;

/**
 * Defines a temporal constraint between two situation streams (a and b).
 * @author Michael Körber
 */
public final class Constraint {

    /** The id of the first situation stream */
    public final char               idA;

    /** The id of the second situation stream */
    public final char               idB;

    /** The TemporalRelations evaluated in a logical OR manner */
    public final Set<TemporalRelation> relations;

    /**
     * @param ida the id of the first stream
     * @param idB the id of the second stream
     * @param relations the relations, where at least one of them has to be
     *          fulfilled in order to fulfill this constraint
     */
    public Constraint(char ida, char idB, TemporalRelation... relations) {
        if (relations.length == 0)
            throw new IllegalArgumentException("Empty constraints not allowed");
        this.idA = ida;
        this.idB = idB;
        this.relations = new HashSet<>();
        for ( TemporalRelation ir : relations )
        	this.relations.add(ir);
    }

    /**
     * @{inheritDoc}
     */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + idA;
		result = prime * result + idB;
		return result;
	}

	/**
     * @{inheritDoc}
     */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Constraint other = (Constraint) obj;
		if (idA != other.idA)
			return false;
		if (idB != other.idB)
			return false;
		if (!relations.equals(other.relations))
			return false;
		return true;
	}

	/**
      * @{inheritDoc}
      */
    @Override
    public String toString() {
        return "Constraint [" + idA + ", " + idB + ", " + relations + "]";
    }

    /**
     * Describes one of the 13 possible relations between two time intervals.
     * @author Michael Körber
     */
    public static enum TemporalRelation {
        /**
         * <pre>
         * A: |----|
         * B:        |----|
         * </pre>
         */
        AABB,
        
        /**
         * <pre>
         * A: |----|
         * B:      |----|
         * </pre>
         */
        AA_BB,
        
        /**
         * <pre>
         * A: |----|
         * B:    |----|
         * </pre>
         */
        ABAB,
        
        /**
         * <pre>
         * A: |--------|
         * B:   |----|
         * </pre>
         */
        ABBA,
        
        /**
         * <pre>
         * A: |----|
         * B: |-------|
         * </pre>
         */
        A_BAB,
        
        /**
         * <pre>
         * A: |-------|
         * B:    |----|
         * </pre>
         */
        ABA_B,
        
        /**
         * <pre>
         * A: |----|
         * B: |----|
         * </pre>
         */
        A_BA_B,
        
        /**
         * <pre>
         * A:    |----|
         * B: |-------|
         * </pre>
         */
        BAA_B,
        
        /**
         * <pre>
         * A: |-------|
         * B: |----|
         * </pre>
         */
        A_BBA,
        
        /**
         * <pre>
         * A:   |----|
         * B: |--------|
         * </pre>
         */
        BAAB,
        
        /**
         * <pre>
         * A:    |----|
         * B: |----|
         * </pre>
         */
        BABA,
        
        /**
         * <pre>
         * A:      |----|
         * B: |----|
         * </pre>
         */
        BB_AA,

        /**
         * <pre>
         * A:        |----|
         * B: |----|
         * </pre>
         */
        BBAA
    }
}
