/*
 * Copyright (c) 2012, 2013
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.jepc.v2.addons.epa.tpstream;

import de.umr.jepc.v2.bindings.BoundVariables;

/**
 * Concrete instance of a situation
 * @author Michael Körber
 */
public class Situation {
    
    private static final BoundVariables NO_BINDINGS = new BoundVariables(0);

    /** The start of this interval */
    private final long           start;

    /** The end of this interval */
    private long                 end;

    /** The aggregates */
    private final BoundVariables aggregates;

    /** Closed indicator */
    private boolean              finished = false;

    /**
     * @param start the time-instant this situation starts
     * @param end the time-instant this situation ends
     */
    public Situation(long start, long end) {
        this(start, end, 0);
    }

    /**
     * @param start the time-instant this interval starts
     * @param end the time-instant this interval end
     * @param numBindings the number of aggregates
     */
    public Situation(long start, long end, int numBindings) {
        this.start = start;
        this.end = end;
        this.aggregates = numBindings > 0 ? new BoundVariables(numBindings) : NO_BINDINGS;
    }

    /**
     * Sets the aggregate defined at index idx to the given value
     * @param name the index of the aggregate
     * @param value the value to set
     */
    public void setAggregate(int idx, Object value) {
        aggregates.bind(idx, value);
    }

    /**
     * @return the computed aggreagte values
     */
    public BoundVariables getAggregates() {
        return aggregates;
    }

    /**
     * Finishes this situation
     * @param end the end timestamp
     */
    public void finish(long end) {
        this.finished = true;
        this.end = end;
    }

    /**
     * @return true if this interval was closed and won't be updated anymore
     */
    public boolean isFinished() {
        return finished;
    }
    /**
     * @return the duration of this interval
     */
    public long getDuration() {
      return end - start;
    }

    /**
     * @return the start instant
     */
    public long getStart() {
      return start;
    }

    /**
     * @return the end instant
     */
    public long getEnd() {
      return end;
    }

    /**
     * Updates the end to the given time instant
     * @param end the end instant to set
     */
    public void update(long end) {
      this.end = end;
    }

    /**
     * @{inheritDoc}
     */
    @Override
    public String toString() {
      return "[" + start + "," + end + "] - Aggregates: " + aggregates;
    }
}
