package de.umr.jepc.v2.addons.epa.tpstream;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import de.umr.jepc.v2.api.epa.pattern.symbol.Binding;
import de.umr.jepc.v2.api.epa.pattern.symbol.Bindings;
import de.umr.jepc.v2.api.epa.pattern.symbol.DuplicateBindingException;
import de.umr.jepc.v2.api.epa.pattern.symbol.NoSuchVariableException;

/**
 * Holds the definitions of situation streams
 * @author Michael Körber
 *
 */
public class SituationDefinitions implements Iterable<SituationDefinition> {
    
    /** The definitions */
    private final List<SituationDefinition> definitions = new ArrayList<>();

    /**
     * @param definitions The definitions
     */
    public SituationDefinitions(SituationDefinition... definitions) {
        this.definitions.addAll(Arrays.asList(definitions));
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Iterator<SituationDefinition> iterator() {
        return definitions.iterator();
    }
    
    /**
     * @return The number of definitions
     */
    public int getNumberOfDefinitions() {
        return definitions.size();
    }
    
    /**
     * Retrieves the definition at the given position
     * @param idx the index of the symbol to retrieve
     * @return the symbol at the given idx
     */
    public SituationDefinition getDefinition(Integer idx) {
        return definitions.get(idx);
    }
    
    /**
     * Retrieves the index of the given definition
     * @param s the id of the situation definition
     * @return the index
     */
    public int getDefinitionIndex(Character s) {
        int idx = 0;
        for ( SituationDefinition def : definitions ) {
            if ( def.getId() == s )
                return idx;
            idx++;
        }
        return -1;
    }

    
    /**
     * Retrieves the aggregate with the given name
     * @param name the name of the aggreagated value
     * @return the aggregate with the given name
     * @throws NoSuchVariableException if no aggregate with the given name is defined
     */
    public Binding findAggregate( String name ) throws NoSuchVariableException {
        for ( SituationDefinition s : definitions ) {
            for ( Binding b : s.getAggregates() ) {
                if ( b.getName().equals(name))
                    return b;
            }
        }
        throw new NoSuchVariableException("Aggregate: " + name + " not defined");
    }
    
    /**
     * @return A list of all aggregates defined by the situation-definitions
     * @throws DuplicateBindingException if a binding occurs in more than one situation definition
     */
    public Bindings getAllAggregates() throws DuplicateBindingException {
        Bindings result = new Bindings();
        for ( SituationDefinition s : definitions )
            result.add(s.getAggregates());
        return result;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        return definitions.toString();
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((definitions == null) ? 0 : definitions.hashCode());
        return result;
    }
    

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        SituationDefinitions other = (SituationDefinitions) obj;
        if (definitions == null) {
            if (other.definitions != null)
                return false;
        } else if (!definitions.equals(other.definitions))
            return false;
        return true;
    }
}
