/**
 * 
 */
package de.umr.jepc.v2.addons.epa.tpstream.buffer;

import java.util.Comparator;
import java.util.Iterator;

import de.umr.jepc.v2.addons.epa.tpstream.Situation;

/**
 * Buffers situations of a situation stream in an array-backed
 * {@link SortedRingBuffer ring buffer} structure and offers methods to 
 * search for situations based on their start and end timestamps.
 * @author Michael Körber
 *
 */
public class SituationBuffer {
	
    /**
     * To search by end-timestamp
     */
    private static final Comparator<Situation> CMP_END = new Comparator<Situation>() {
        @Override
        public int compare(Situation o1, Situation o2) {
            if ( o1 == o2 ) return 0;
            return Long.compare(o1.getEnd(), o2.getEnd());
        }
    };
    
    /**
     * To search by start-timestamp
     */
    private static final Comparator<Situation> CMP_START = new Comparator<Situation>() {
        @Override
        public int compare(Situation o1, Situation o2) {
            if ( o1 == o2 ) return 0;
            return Long.compare(o1.getStart(), o2.getStart());
        }
    };
    
    private final SortedRingBuffer<Situation> buffer = new SortedRingBuffer<>();
    
    /**
     * Adds the given situation to this buffer.
     * @param situation the situation to add
     */
    public void add( Situation situation ) {
        buffer.add(situation);
    }
    
    /**
     * Searches this buffer for a situation starting at the given point in time.
     * If such a situation is found, this method returns its index, if not the returned
     * index is the point of insertion.
     * @param time the point in time to search for 
     * @return the index of the situation starting at the given point in time or its point of insertion
     */
    public int searchStart( long time ) {
    	return wrapSearch( buffer.binarySearch( new Situation(time, Long.MAX_VALUE,0), CMP_START ) );
    }
    
    /**
     * Searches this buffer for a situation ending at the given point in time.
     * If such a situation is found, this method returns its index, if not the returned
     * index is the point of insertion.
     * @param time the point in time to search for 
     * @return the index of the situation ending at the given point in time or its point of insertion
     */
    public int searchEnd( long ref ) {
    	return wrapSearch( buffer.binarySearch( new Situation(Long.MIN_VALUE, ref,0), CMP_END ) );
    }
    
    /**
     * Wraps a negative value returned by the {@link SortedRingBuffer#binarySearch(Object, Comparator)} to a
     * valid index. This means transforms a negative value into the corresponding point of insertion. 
     * @param index the index to wrap
     * @return the wrapped index
     */
    private int wrapSearch( int index ) {
    	return index < 0 ? -(index+1) : index;
    }
    
    /**
     * @return an iterator returning all situations stored in this buffer
     */
    public Iterator<Situation> iterator() {
        return buffer.iterator(0, buffer.size());
    }
    
    
    /**
     * @param range the range to iterate
     * @return an iterator returning the situations falling into the given index range
     */
    public Iterator<Situation> iterator( IndexRange range ) {
    	return buffer.iterator(range._1, range._2);
    }
    
    /**
     * Removes all situations starting before or at the given point in time
     * @param time the point in time up to which situations are removed
     */
    public void cleanUp( long time ) {
//        int sres = buffer.binarySearch(new Situation(Long.MIN_VALUE, time), CMP_END);
        int sres = buffer.binarySearch(new Situation(time, Long.MAX_VALUE), CMP_START);
        if ( sres >= 0 )
            buffer.removeAllUntil(sres+1);
        else if ( sres < -1 )
            buffer.removeAllUntil(-(sres+1));
    }
    
    /**
     * @param index the index of the situation to retrieve
     * @return the situation stored at the given index position
     */
    public Situation get(int index) {
        return buffer.get(index);
    }

    /**
     * {@inheritDoc}
     */
    public int getSize() {
        return buffer.size();
    }
    

}
