package de.umr.jepc.v2.addons.epa.tpstream.matcher;

import java.util.List;

import de.umr.jepc.v2.addons.epa.TPStream.Mode;
import de.umr.jepc.v2.addons.epa.tpstream.Constraints;
import de.umr.jepc.v2.addons.epa.tpstream.Situation;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinitions;
import de.umr.jepc.v2.addons.epa.tpstream.buffer.SituationBuffers;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.Evaluator;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.EvaluatorUtil;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.Optimizer;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.RunningStats;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.TPStats;

/**
 * Adaptive implementation for {@link Mode#CLASSIC classic} temporal
 * pattern matching. This matcher only considers situation with a known
 * end and adapts the evaluation orders, if the tracked statistics change 
 * significantly.
 * @author Michael Körber
 *
 */
public final class AdaptiveFixedDFSMatcher implements Matcher {
    
    /** The number of update-invocations to ignore, until statistics are tracked */
    private static final int WARM_UP = 3_500;
    
    /** Relax factor to avoid thrashing */
    private static final int RELAX = 1_000;
    
    /** The threshold, when to trigger re-optimization */
    private final double reoptimizeThreshold;
    
    /** Update-invocation counter */
    private long counter = 0L;
	
    /** Flag telling whether warm-up is completed */
	private boolean warmedUp = false;
	
	/** The continuously updated statistics */
    private final RunningStats stats;
    
    /** The situation definitions */
    private final SituationDefinitions definitions;
    
    /** The constraints */
    private final Constraints constraints;

    /** The buffers */
    private final SituationBuffers buffers;

    /** The optimizer instance */
    private final Optimizer optimizer;
    
    /** The callback to invoke on new results */
    private final MatcherCallback callback;
    
    /** The length of the time-window */
    private final long window;
    
    /** The evaluators, one for each situation stream */
    private Evaluator[] evaluators;
    
    /** The working-set holding (partial) temporal configurations */
    private Situation[] work;

	/**
     * @param definitions The situation definitions
     * @param constraints The constraints
     * @param window The length of the time-window
     * @param optimizer The optimizer instance
     * @param callback The callback to invoke on new results
     */
	public AdaptiveFixedDFSMatcher(SituationDefinitions definitions, Constraints constrains, long window, Optimizer optimizer, MatcherCallback callback, double reoptimizeThreshold) {
		this.definitions = definitions;
		this.constraints = constrains;
		this.window = window;
		this.stats = new RunningStats(definitions, constrains);
        this.buffers = new SituationBuffers(definitions);
        this.optimizer = optimizer;
        this.callback = callback;
        this.evaluators = new Evaluator[definitions.getNumberOfDefinitions()];
        this.evaluators = optimize(stats.snapshot());
        this.work = new Situation[definitions.getNumberOfDefinitions()];
        this.reoptimizeThreshold = reoptimizeThreshold;
    }
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(Situation[] updates, long currentTime) {
		buffers.cleanup(currentTime-window);
		
		for (int idx = 0; idx < updates.length; idx++) {
			Situation interval = updates[idx];
			if ( interval == null || !interval.isFinished() )
		        continue;
		
			buffers.getBuffer(idx).add(interval);
	
			// Check if warmed up
			if ( !warmedUp ) {
				counter++;
				if ( counter > WARM_UP ) {
					counter = 0;
				    warmedUp = true;
				    // Set buffers to actual sizes and begin statistic tracking
				    stats.setBuffers(buffers);	
				}
			}
			
	
			Evaluator evalutaor = evaluators[idx];
			if ( evalutaor != null ) {
			    work[idx] = interval;
				evalutaor.evaluate(work, currentTime, stats);
				counter++;
				
				// If warmed up und RELAX has passed and stats differ above threshold... re-optimize
				if ( warmedUp && counter > RELAX && stats.maxDiff(evalutaor.getStatistics()) > reoptimizeThreshold ) {
					this.counter = 0;
				    this.evaluators = optimize(stats.snapshot());
				}
			}
			work[idx] = null;
		}
		// Update buffer size stats
		stats.updateBuffers(buffers);
	}

	/**
     * Creates new evaluators providing the most efficient evaluation strategy
     * under the given statistics.
     * @param statistics the current statistics
     * @return a new set of evaluators optimized for the current workload
     */
	private Evaluator[] optimize(TPStats statistics) {
		Evaluator[] result = new Evaluator[definitions.getNumberOfDefinitions()];
		for (Character c : optimizer.getTriggerSituations()) {
			List<Character> order = optimizer.calculateOrder(c, statistics);
			int idx = definitions.getDefinitionIndex(c);
			result[idx] = EvaluatorUtil.create(statistics, definitions, constraints, order, buffers, callback);
			System.out.println("  " + result[idx] );
		}
		System.out.println("============================================");
		return result;
	}
}
