package de.umr.jepc.v2.addons.epa.tpstream.matcher;

import java.util.List;

import de.umr.jepc.v2.addons.epa.TPStream.Mode;
import de.umr.jepc.v2.addons.epa.tpstream.Constraints;
import de.umr.jepc.v2.addons.epa.tpstream.Situation;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinitions;
import de.umr.jepc.v2.addons.epa.tpstream.buffer.SituationBuffers;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.Evaluator;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.EvaluatorUtil;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.Optimizer;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.TPStats;

/**
 * Matching implementation for {@link Mode#CLASSIC classic} temporal
 * pattern matching. This matcher only considers situation with a known
 * end.
 * @author Michael Körber
 *
 */
public final class FixedDFSMatcher implements Matcher {
    
	/** The situation definitions */
	private final SituationDefinitions definitions;
	
	/** The constraints */
	private final Constraints constraints;

	/** The buffers */
	private final SituationBuffers buffers;

	/** The optimizer instance */
	private final Optimizer optimizer;
	
	/** The callback to invoke on new results */
	private final MatcherCallback callback;
	
	/** The length of the time-window */
	private final long window;
	
	/** The evaluators, one for each situation stream */
	private Evaluator[] evaluators;
	
	/** The working-set holding (partial) temporal configurations */
	private Situation[] work;
	
	/**
	 * @param definitions The situation definitions
	 * @param constraints The constraints
	 * @param window The length of the time-window
	 * @param optimizer The optimizer instance
	 * @param callback The callback to invoke on new results
	 */
	public FixedDFSMatcher(SituationDefinitions definitions, Constraints constraints, long window, Optimizer optimizer, MatcherCallback callback) {
		this.definitions = definitions;
		this.constraints = constraints;
		this.optimizer = optimizer;
		this.window = window;
		this.callback = callback;
        this.buffers = new SituationBuffers(definitions);
        
        this.evaluators = optimize(new TPStats(definitions, constraints));
        this.work = new Situation[definitions.getNumberOfDefinitions()];
    }
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(Situation[] updates, long currentTime) {
		buffers.cleanup(currentTime - window);
		for (int i = 0; i < updates.length; i++) {
			Situation interval = updates[i];
			if ( interval == null || !interval.isFinished() )
		        continue;
			
			buffers.getBuffer(i).add(interval);
			Evaluator ev = evaluators[i];
			if ( ev != null ) {
			    work[i] = interval;
				ev.evaluate(work, currentTime);
			}
			work[i] = null;
			
        }
	}

	/**
	 * Creates new evaluators providing the most efficient evaluation strategy
	 * under the given statistics.
	 * @param statistics the current statistics
	 * @return a new set of evaluators optimized for the current workload
	 */
	private Evaluator[] optimize(TPStats statistics) {
		Evaluator[] result = new Evaluator[definitions.getNumberOfDefinitions()];
		for (Character c : optimizer.getTriggerSituations()) {
			List<Character> order = optimizer.calculateOrder(c, statistics);
			int idx = definitions.getDefinitionIndex(c);
			result[idx] = EvaluatorUtil.create(statistics, definitions, constraints, order, buffers, callback);
		}
		return result;
	}
}
