package de.umr.jepc.v2.addons.epa.tpstream.matcher;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;

import de.umr.jepc.v2.addons.epa.tpstream.Constraint;
import de.umr.jepc.v2.addons.epa.tpstream.Constraints;
import de.umr.jepc.v2.addons.epa.tpstream.Situation;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinition;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinitions;
import de.umr.jepc.v2.addons.epa.tpstream.buffer.SituationBuffers;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.Evaluator;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.EvaluatorUtil;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.Optimizer;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.TPStats;

public final class FloatingDFSMatcher implements Matcher {

    /** The situation definitions */
    private final SituationDefinitions definitions;
    
    /** The constraints */
    private final Constraints constraints;

    /** The buffers */
    private final SituationBuffers buffers;

    /** The optimizer instance */
    private final Optimizer optimizer;
    
    /** The callback to invoke on new results */
    private final MatcherCallback callback;
    
    /** The length of the time-window */
    private final long window;
    
    /** The evaluators, one for each situation stream */
    private Evaluator[] evaluators;
    
    /** The working-set holding (partial) temporal configurations */
    private Situation[] work;

    /** The shadow-buffer, holding started butn not finished situations */
    private final Situation[]          shadow;

    /** Holds for each situation stream the indexes of unrelated trigger situations */
    private final int[][]              unrelatedTrigger;

    /** Holds for each situation stream the indexes of related trigger situations */
    private final int[][]              relatedTrigger;

    /** Holds the indexes of all trigger situations */
    private final int[]                allTriggers;

    /** Indicator for each situation index if it is a trigger or not */
    private final boolean[]            triggers;

    /** Indicator for each situation index if it is not a trigger, but related to one */
    private final boolean[]            relatedNonTrigger;

    /**
     * @param definitions The situation definitions
     * @param constraints The constraints
     * @param window The length of the time-window
     * @param optimizer The optimizer instance
     * @param callback The callback to invoke on new results
     */
    public FloatingDFSMatcher(SituationDefinitions definitions, Constraints constraints, long window, Optimizer optimizer, MatcherCallback callback) {
        this.definitions = definitions;
        this.constraints = constraints;
        this.window = window;
        this.buffers = new SituationBuffers(definitions);
        this.optimizer = optimizer;
        this.callback = callback;
        this.evaluators = optimize(new TPStats(definitions, constraints));

        // Fancy low latency stuff!
        this.shadow = new Situation[definitions.getNumberOfDefinitions()];
        this.unrelatedTrigger = new int[definitions.getNumberOfDefinitions()][];
        this.relatedTrigger = new int[definitions.getNumberOfDefinitions()][];
        this.allTriggers = new int[optimizer.getTriggerSituations().size()];
        
        this.triggers = new boolean[definitions.getNumberOfDefinitions()];
        this.relatedNonTrigger = new boolean[definitions.getNumberOfDefinitions()];
        Arrays.fill(triggers, false);
        Arrays.fill(relatedNonTrigger, false);

        int i = 0;
        for (Character c : optimizer.getTriggerSituations()) {
            triggers[definitions.getDefinitionIndex(c)] = true;
            allTriggers[i++] = definitions.getDefinitionIndex(c);
        }

        populateRelations();
        popuplateRelatedNTs();

        this.work = new Situation[definitions.getNumberOfDefinitions()];
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void update(Situation[] updates, long currentTime) {
        buffers.cleanup(currentTime - window);

        for (int idx = 0; idx < updates.length; idx++) {
            Situation interval = updates[idx];
            if (interval == null || (!interval.isFinished() && !triggers[idx]))
                continue;

            if (triggers[idx] && !interval.isFinished()) {
                shadow[idx] = interval;
                // Regular match + unrelated active intervals
                Evaluator ev = evaluators[idx];
                work[idx] = interval;
                ev.evaluate(work, currentTime);
                if (unrelatedTrigger[idx] != null)
                    combine(ev, currentTime, unrelatedTrigger[idx], work, 0, 0);
            } else if (!triggers[idx] && interval.isFinished()) {
                buffers.getBuffer(idx).add(interval);
                // Perform all matches
                if (relatedNonTrigger[idx]) {
                    Evaluator ev = evaluators[idx];
                    work[idx] = interval;
                    combine(ev, currentTime, allTriggers, work, 0, 0);
                }

            } else if (triggers[idx] && interval.isFinished()) {
                shadow[idx] = null;
                buffers.getBuffer(idx).add(interval);
                if (relatedTrigger[idx] != null) {
                    Evaluator ev = evaluators[idx];
                    work[idx] = interval;
                    combine(ev, currentTime, relatedTrigger[idx], work, 0, 0);
                }
            }
            work[idx] = null;
        }
    }

    /**
     * Creates new evaluators providing the most efficient evaluation strategy
     * under the given statistics.
     * @param statistics the current statistics
     * @return a new set of evaluators optimized for the current workload
     */
    private Evaluator[] optimize(TPStats stats) {
        Evaluator[] result = new Evaluator[this.definitions.getNumberOfDefinitions()];
        for (SituationDefinition s : definitions) {
            List<Character> order = optimizer.calculateOrder(s.getId(), stats);
            int idx = definitions.getDefinitionIndex(s.getId());
            result[idx] = EvaluatorUtil.create(stats, definitions, constraints, order, buffers, callback);
        }
        return result;
    }

    private void combine(Evaluator ev, long currentTime, int[] shdwIdxs, Situation[] work, int step, int shadowCount) {
        if (step == shdwIdxs.length) {
            if (shadowCount > 0)
                ev.evaluate(work, currentTime);
            return;
        }

        int idx = shdwIdxs[step];
        combine(ev, currentTime, shdwIdxs, work, step + 1, shadowCount);
        if (shadow[idx] != null) {
            work[idx] = shadow[idx];
            combine(ev, currentTime, shdwIdxs, work, step + 1, shadowCount + 1);
            work[idx] = null;
        }
    }

    /**
     * Computes for each situation, if it is not a trigger, but related to one
     * a stores this information in the corresponding field.
     */
    private void popuplateRelatedNTs() {
        Set<Character> ts = optimizer.getTriggerSituations();
        for (int i = 0; i < this.relatedNonTrigger.length; i++) {
            char s = definitions.getDefinition(i).getId();
            for (Constraint evc : constraints.getSituationConstraints(s)) {
                if (evc.idA != s && ts.contains(evc.idA) || evc.idB != s && ts.contains(evc.idB))
                    relatedNonTrigger[i] = true;
            }
        }
    }

    /**
     * Computes for each trigger situation, the disjoint sets
     * of related and unrelated other triggers. This information
     * is stored in the corresponding fields.
     */
    private void populateRelations() {
        Set<Character> ts = optimizer.getTriggerSituations();
        for (Character c : ts) {
            List<Character> related = new ArrayList<>();
            List<Character> unrelated = new ArrayList<>(ts);
            unrelated.remove(c);

            for (Constraint evc : constraints.getSituationConstraints(c)) {
                if (unrelated.remove((Character) evc.idA))
                    related.add(evc.idA);
                else if (unrelated.remove((Character) evc.idB))
                    related.add(evc.idB);
            }

            int idx = definitions.getDefinitionIndex(c);
            if (!related.isEmpty()) {
                relatedTrigger[idx] = new int[related.size()];
                for (int i = 0; i < related.size(); i++) {
                    relatedTrigger[idx][i] = definitions.getDefinitionIndex(related.get(i));
                }
            }

            if (!unrelated.isEmpty()) {
                unrelatedTrigger[idx] = new int[unrelated.size()];
                for (int i = 0; i < unrelated.size(); i++) {
                    unrelatedTrigger[idx][i] = definitions.getDefinitionIndex(unrelated.get(i));
                }
            }
        }
    }
}
