/**
 * 
 */
package de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import de.umr.jepc.v2.addons.epa.tpstream.Constraint;
import de.umr.jepc.v2.addons.epa.tpstream.Situation;
import de.umr.jepc.v2.addons.epa.tpstream.Constraint.TemporalRelation;
import de.umr.jepc.v2.addons.epa.tpstream.buffer.IndexRange;
import de.umr.jepc.v2.addons.epa.tpstream.buffer.SituationBuffer;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.SelectivityEstimate;

/**
 * Evaluable implementation of a {@link Constraint temporal constraint}.
 * Offers methods to check this constraint on a (partial) temporal configuration,
 * as well as for searching a situation buffer for matches, using one situation as 
 * reference point-. 
 * @author Michael Körber
 *
 */
public final class EVConstraint {
    
    /** Value to signal, that a search may be skipped */
    private static final long UNSPECIFIED = Long.MIN_VALUE;
    
    /**
     * Represents a single temporal relation and
     * provides a method to check if it is fulfilled between
     * two given situations
     * @author Michael Körber
     *
     */
    private static interface TRChecker {
        /**
         * @param a the first situation
         * @param b the second situation
         * @return <code>true</code> if the represented relation is fulfilled by the given situations, 
         *  <code>false</code> otherwise
         */
        boolean check( Situation a, Situation b);
    }
    
    /**
     * A factory for creating two search ranges for a buffer:
     * One for searching intervals with valid start timestamps
     * and a second one for search interval with valid end timestamps.
     * The intersection of these ranges form an index-range of situations 
     * all fulfilling the represented temporal relation.
     * @author Michael Körber
     *
     */
    private static interface RangeGen {
        /**
         * @param reference the reference situation
         * @return a four element array [tsmin,tsmax,temin,temax] holding minimum and maximum 
         *  values for start and end points of situations fulfilling the represented temporal constraint 
         */
        long[] getRange( Situation reference );
    }
    
    /** Checks for all available temporal relations */
    private static final TRChecker[] ALL_CHECKS;
    
    /** Range generators for all available temporal relations */
    private static final RangeGen[] ALL_RANGES;
    
    /**
     * Populate checks and ranges
     */
    static {
        ALL_CHECKS = new TRChecker[TemporalRelation.values().length];
        ALL_RANGES = new RangeGen[TemporalRelation.values().length];
        
        
        ALL_CHECKS[TemporalRelation.AABB.ordinal()]   = (a,b) -> a.getEnd()   <  b.getStart();
        ALL_CHECKS[TemporalRelation.AA_BB.ordinal()]  = (a,b) -> a.getEnd()   == b.getStart();
        ALL_CHECKS[TemporalRelation.ABAB.ordinal()]   = (a,b) -> a.getStart() <  b.getStart() && a.getEnd() >  b.getStart() && a.getEnd() < b.getEnd();
        ALL_CHECKS[TemporalRelation.ABBA.ordinal()]   = (a,b) -> a.getStart() <  b.getStart() && a.getEnd() >  b.getEnd();
        ALL_CHECKS[TemporalRelation.A_BAB.ordinal()]  = (a,b) -> a.getStart() == b.getStart() && a.getEnd() <  b.getEnd();
        ALL_CHECKS[TemporalRelation.ABA_B.ordinal()]  = (a,b) -> a.getStart() <  b.getStart() && a.getEnd() == b.getEnd();
        
        ALL_CHECKS[TemporalRelation.A_BA_B.ordinal()] = (a,b) -> a.getStart() == b.getStart() && a.getEnd() == b.getEnd();
        
        ALL_CHECKS[TemporalRelation.BAA_B.ordinal()]  = (a,b) -> b.getStart() <  a.getStart() && b.getEnd() == a.getEnd();
        ALL_CHECKS[TemporalRelation.A_BBA.ordinal()]  = (a,b) -> b.getStart() == a.getStart() && b.getEnd() <  a.getEnd();
        ALL_CHECKS[TemporalRelation.BAAB.ordinal()]   = (a,b) -> b.getStart() <  a.getStart() && b.getEnd() >  a.getEnd();
        ALL_CHECKS[TemporalRelation.BABA.ordinal()]   = (a,b) -> b.getStart() <  a.getStart() && b.getEnd() >  a.getStart() && b.getEnd() < a.getEnd();
        ALL_CHECKS[TemporalRelation.BB_AA.ordinal()]  = (a,b) -> b.getEnd()   == a.getStart();
        ALL_CHECKS[TemporalRelation.BBAA.ordinal()]   = (a,b) -> b.getEnd()   <  a.getStart();
        
        
        ALL_RANGES[TemporalRelation.AABB.ordinal()]   = ref -> new long[] { ref.getEnd() + 1, UNSPECIFIED, UNSPECIFIED, UNSPECIFIED };
        ALL_RANGES[TemporalRelation.AA_BB.ordinal()]  = ref -> new long[] { ref.getEnd(), ref.getEnd()+1, UNSPECIFIED, UNSPECIFIED };
        ALL_RANGES[TemporalRelation.ABAB.ordinal()]   = ref -> new long[] { ref.getStart()+1, ref.getEnd(), ref.getEnd()+1, UNSPECIFIED };
        ALL_RANGES[TemporalRelation.ABBA.ordinal()]   = ref -> new long[] { ref.getStart()+1, UNSPECIFIED, UNSPECIFIED, ref.getEnd() };
        ALL_RANGES[TemporalRelation.A_BAB.ordinal()]  = ref -> new long[] { ref.getStart(), ref.getStart()+1, ref.getEnd()+1, UNSPECIFIED };
        ALL_RANGES[TemporalRelation.ABA_B.ordinal()]  = ref -> new long[] { ref.getStart()+1, UNSPECIFIED, ref.getEnd(), ref.getEnd()+1 };
        
        ALL_RANGES[TemporalRelation.A_BA_B.ordinal()] = ref -> new long[] { ref.getStart(), ref.getStart()+1, ref.getEnd(), ref.getEnd()+1 };
        
        ALL_RANGES[TemporalRelation.BAA_B.ordinal()]  = ref -> new long[] { UNSPECIFIED, ref.getStart(), ref.getEnd(), ref.getEnd()+1 };
        ALL_RANGES[TemporalRelation.A_BBA.ordinal()]  = ref -> new long[] { ref.getStart(), ref.getStart()+1, UNSPECIFIED, ref.getEnd() };
        ALL_RANGES[TemporalRelation.BAAB.ordinal()]   = ref -> new long[] { UNSPECIFIED, ref.getStart(), ref.getEnd()+1, UNSPECIFIED };
        ALL_RANGES[TemporalRelation.BABA.ordinal()]   = ref -> new long[] { UNSPECIFIED, ref.getStart(), ref.getStart()+1, ref.getEnd() };
        ALL_RANGES[TemporalRelation.BB_AA.ordinal()]  = ref -> new long[] { UNSPECIFIED, UNSPECIFIED, ref.getStart(), ref.getStart()+1 };
        ALL_RANGES[TemporalRelation.BBAA.ordinal()]   = ref -> new long[] { UNSPECIFIED, UNSPECIFIED, UNSPECIFIED, ref.getStart() };
    }
    
    /** The index of the "A"-situation in a partial temporal configuration */
    private final int idxA;
    
    /** The index of the "B"-situation in a partial temporal configuration */
    private final int idxB;
    
    /** The index of the refernce-situation when generating a search range */
    private final int refIdx;
    
    /** The constraint this evaluable is based on */
    private Constraint constraint;
    
    /** The range-generators to apply */
    private final RangeGen[] ranges;
    
    /** The checks to perform */
    private final TRChecker[] checker;
    
    /** The buffer to search with */
    private final SituationBuffer buffer;

    /**
     * @param constraint The constraint this evaluable is based on
     * @param idxA The index of the "A"-situation in a partial temporal configuration
     * @param idxB The index of the "B"-situation in a partial temporal configuration
     * @param searchB Indicator whether to search for "B"-situations or for "A"s.
     * @param buffer The buffer to search
     */
    public EVConstraint(Constraint constraint, int idxA, int idxB, boolean searchB, SituationBuffer buffer) {
        this.idxA = idxA;
        this.idxB = idxB;
        this.refIdx = searchB ? idxA : idxB;
        this.constraint = constraint;
        this.buffer  = buffer;
        this.ranges  = new RangeGen[constraint.relations.size()];
        this.checker = new TRChecker[constraint.relations.size()];
        
        int i = 0;
        for ( TemporalRelation tr : constraint.relations ) {
           checker[i] = ALL_CHECKS[tr.ordinal()];
           ranges[i++] = searchB ? ALL_RANGES[tr.ordinal()] : ALL_RANGES[TemporalRelation.values().length - 1 - tr.ordinal()];
        }
    }

    /**
     * Checks if the given temporal configuration satisfies this constraint
     * @param workingSet the temporal configuration to check 
     * @return <code>true</code> if this constraint is fulfilled by the given situations, 
         *  <code>false</code> otherwise
     */
    public boolean isSatisfied( Situation[] workingSet ) {
        return isSatisfied(workingSet[idxA], workingSet[idxB]);
    }
    
    /**
     * Checks if the given situations fulfill this constraint
     * @param a the "A"-situation
     * @param b the "B"-situation
     * @return <code>true</code> if this constraint is fulfilled by the given situations, 
     *  <code>false</code> otherwise
     */
    public boolean isSatisfied( Situation a, Situation b) {
        for ( TemporalRelation tr : constraint.relations )
            if ( ALL_CHECKS[tr.ordinal()].check(a, b) )
                return true;
        return false;
    }
    
    /**
     * Searches the buffer for situations fulfilling this constraint given a partial
     * temporal configuration
     * @param workingset the partial temporal configuration
     * @param estimate the selectivity estimate to update
     * @return the disjoint index ranges pointing to situation inside the buffer which fulfill this constraint.
     */
    public List<IndexRange> search( Situation[] workingset, SelectivityEstimate estimate ) {
        if ( buffer.getSize() > 0 ) {
            LinkedList<IndexRange> uRanges = new LinkedList<>();
            for ( RangeGen rg : ranges ) {
                IndexRange r = search( rg.getRange(workingset[refIdx]) );
                if ( r != IndexRange.EMPTY )
                    uRanges.push(r);
            }
            
            switch ( uRanges.size() ) {
            case 0:
                estimate.update(constraint, 0);
                return Collections.emptyList();
            case 1:
                estimate.update(constraint, (double) uRanges.size() / buffer.getSize() );
                return uRanges;
            default: return maxUnion(uRanges, estimate, buffer.getSize());
            }
        }
        return Collections.emptyList();
    }
    
    /**
     * Searches the buffer with the given query-bounds
     * and returns an IndexRange of situations fulfilling this constraint.
     * @param queryBounds the bounds to be fulfilled
     * @return a index range of situations fulfilling this constraint.
     */
    private IndexRange search( long[] queryBounds ) {
        int tsa = (queryBounds[0] == UNSPECIFIED) ? 0 : buffer.searchStart(queryBounds[0]);
        int tsb = (queryBounds[1] == UNSPECIFIED) ? buffer.getSize() : buffer.searchStart(queryBounds[1]);
        
        if ( tsa == tsb )
            return IndexRange.EMPTY;
        
        int tea = (queryBounds[2] == UNSPECIFIED) ? 0 : buffer.searchEnd(queryBounds[2]);
        int teb = (queryBounds[3] == UNSPECIFIED) ? buffer.getSize() : buffer.searchEnd(queryBounds[3]);
        
        
        // No results
        if ( tea == teb || tsb <= tea || teb <= tsa )
            return IndexRange.EMPTY;
        
        int lower = Math.max(tsa, tea);
        int upper = Math.min(tsb, teb);
        
        if ( lower == upper )
            return IndexRange.EMPTY;
        else
            return new IndexRange(lower, upper);
    }
    
    /**
     * Computes the max union of the given ranges. This means, all intervals
     * that may be unified are unified and disjoint intervals are kept as they are.
     * @param ranges the ranges to unify
     * @param estimate the selectivity estimate to update
     * @param bufferSize the actual buffer-size used to update the selectivity estimate
     * @return the max union of the given ranges.
     */
    private List<IndexRange> maxUnion( LinkedList<IndexRange> ranges, SelectivityEstimate estimate, double bufferSize ) {
        List<IndexRange> result = new ArrayList<>();
        int numResults = 0;
        while ( !ranges.isEmpty() ) {
            IndexRange current = ranges.pollFirst();
            Iterator<IndexRange> idxr = ranges.iterator();
            while ( idxr.hasNext() ) {
                IndexRange next = idxr.next();
                if ( current.intersects(next) ) {
                    current = current.union(next);
                    idxr.remove();
                    idxr = ranges.iterator();
                }
            }
            numResults += current.getSize();
            result.add(current);
        }
        estimate.update(constraint, numResults / bufferSize);
        return result;
    }
    
}
