package de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import de.umr.jepc.v2.addons.epa.tpstream.Constraint;
import de.umr.jepc.v2.addons.epa.tpstream.Situation;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinitions;
import de.umr.jepc.v2.addons.epa.tpstream.buffer.IndexRange;
import de.umr.jepc.v2.addons.epa.tpstream.buffer.SituationBuffer;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.EvaluationStep;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.SelectivityEstimate;
import de.umr.jepc.v2.addons.util.MultiIterator;

/**
 * Implements a single evalutation step. This step is backed by
 * one ore more {@link EVConstraint} handling searches and constraint
 * validation.
 * @author Michael Körber
 *
 */
public final class EVStepImpl implements EvaluationStep {
	
    /** The index of the situation to add inside the partial temporal configuration */
	private final int situationIdx;
	
	/** The situation buffer to consider in this step */ 
	private final SituationBuffer buffer;
	
	/** The applicable constraints in this step */
	private final List<EVConstraint> evconstraints;
	
	
	/**
	 * @param situationId the id of the situation stream to consider in this step 
	 * @param buffer the situation stream's buffer
	 * @param definitions the situation definitions
	 * @param applicableConstraints all applicable constraints to be fulfilled after adding a situation in this step
	 */
	public EVStepImpl(Character situationId, SituationBuffer buffer, SituationDefinitions definitions, List<Constraint> applicableConstraints ) {
		this.situationIdx = definitions.getDefinitionIndex(situationId);
		this.buffer = buffer;
		this.evconstraints = new ArrayList<>();
		for ( Constraint c : applicableConstraints ) 
			evconstraints.add( new EVConstraint(c, definitions.getDefinitionIndex(c.idA), definitions.getDefinitionIndex(c.idB), c.idB == situationId, buffer) );
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getSituationIndex() {
		return situationIdx;
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isSatisfied(Situation[] workingSet) {
		for ( EVConstraint evc : evconstraints )
			if ( !evc.isSatisfied(workingSet) )
				return false;
		return true;
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public Iterator<Situation> query(Situation[] workingSet, SelectivityEstimate estimate) {
		List<IndexRange> finalRanges;
		if ( evconstraints.isEmpty() ) 
			return buffer.iterator();
		if ( evconstraints.size() == 1 ) 
			finalRanges = evconstraints.get(0).search(workingSet, estimate);
		else 
			finalRanges = queryFull(workingSet, estimate);
		
		switch ( finalRanges.size() ) {
		case 0: return Collections.emptyIterator();
		case 1: return buffer.iterator(finalRanges.get(0));
		default: {
		    List<Iterator<Situation>> iters = new ArrayList<>();
	        for ( IndexRange r : finalRanges ) {
	            iters.add(buffer.iterator(r));
	        }
		    return new MultiIterator<Situation>(iters);
		}}
	}
	

	/**
	 * Query method invoked when there are more than one constraint to satisfy.
	 * The idea is to compute index-ranges for each constraint individually and
	 * intgersect them lateron. All remaining situations fulfill all defined constraints. 
	 * @param workingset the current partial temporal configuration
	 * @param estimate the selectivity estimate to update during search
	 * @return the list of index-ranges pointing to situations fulfilling all defined constraints
	 */
	private List<IndexRange> queryFull(Situation[] workingset, SelectivityEstimate estimate) {
		List<List<IndexRange>> iRanges = new ArrayList<>();
		for ( EVConstraint evc : evconstraints ) {
			List<IndexRange> uList = evc.search(workingset, estimate);
			if ( uList.isEmpty() )
				return Collections.emptyList();
			else
				iRanges.add(uList);
		}
		return intersectAll(iRanges);
	}
	
	/**
	 * Computes the maximum intersection of ranges obtained
	 * from queries on individual constraints.
	 * @param ranges the list of list of ranges to intersect.
	 * @return the max intersection of the given ranges.
	 */
	private List<IndexRange> intersectAll( List<List<IndexRange>> ranges ) {
		Iterator<List<IndexRange>> main = ranges.iterator();
		List<IndexRange> work = main.next();
		
		while ( main.hasNext() && !work.isEmpty() ) {
			List<IndexRange> tmp = main.next();
			List<IndexRange> newWork = new ArrayList<>();
			
			for ( IndexRange r1 : work ) {
				for ( IndexRange r2 : tmp ) {
					if ( r1.intersects(r2) )
						newWork.add(r1.intersect(r2));
				}
			}
			work = newWork;
		}
		return work;
	}

    /**
     * {@inheritDoc}
     */
    @Override
    public Iterator<Situation> iterateBuffer() {
        return buffer.iterator();
    }
}