/**
 * 
 */
package de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.impl;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.Iterator;
import java.util.List;

import de.umr.jepc.v2.addons.epa.tpstream.Constraints;
import de.umr.jepc.v2.addons.epa.tpstream.Situation;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.Matcher.MatcherCallback;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.EvaluationStep;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.Evaluator;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.RunningStats;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.SelectivityEstimate;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.TPStats;

/**
 * Iterative evaluator implementation, performing a depth first
 * search among the situation buffers.
 * Performs slower than the recursive version.
 * In each step executes a search on the corresponding buffer
 * and adds the results one by one to the working-set. After setting
 * a result, a recursive call is executed to process the next step.
 * @author Michael Körber
 *
 */
public final class IterativeEvaluator implements Evaluator {
	
	private static class SingletonIterator implements Iterator<Situation> {
		private Situation item;

		public SingletonIterator(Situation item) {
			this.item = item;
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public boolean hasNext() {
			return item != null;
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Situation next() {
			Situation res = item;
			item = null;
			return res;
		}
	}
	
	
    /** The statistics this evaluator is based on */
	private final TPStats statistics;
	
	/** The callback to invoke on new results */
	private final MatcherCallback callback;

	/** The steps to execute recursively */
	private final EvaluationStep[] steps;
	
	private final int[] idxs;
	
	private final SelectivityEstimate estimate;
	
	
	
	/**
	 * @param statistics The statistics this evaluator is based on
	 * @param constraints The defined temporal constraints
	 * @param steps The steps to execute recursively
	 * @param callback The callback to invoke on new results
	 */
	public IterativeEvaluator(TPStats statistics, Constraints constraints, List<EvaluationStep> steps, MatcherCallback callback) {
		super();
		this.statistics = statistics;
		this.callback = callback;
		this.steps = steps.toArray( new EvaluationStep[steps.size()] );
		this.estimate = new SelectivityEstimate(constraints);
		
		this.idxs = new int[this.steps.length];
		for ( int i = 0; i < this.steps.length; i++ )
			this.idxs[i] = this.steps[i].getSituationIndex();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public TPStats getStatistics() {
		return statistics;
	}
	
	/**
     * {@inheritDoc}
     */
    @Override
    public void evaluate(Situation[] workingSet, long currentTime, RunningStats statistics) {
        this.estimate.reset();
        evaluate(workingSet, currentTime);
        this.estimate.updateStats(statistics);
    }

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void evaluate(Situation[] workingSet, long currentTime) {
		Deque<Iterator<Situation>> stack = new ArrayDeque<>();
		
		// Initialize stack
		EvaluationStep zero = steps[0];
		if ( workingSet[zero.getSituationIndex()] != null )
			stack.push(new SingletonIterator(workingSet[zero.getSituationIndex()]));
		else
			stack.push(zero.query(workingSet, estimate));
		
		while (!stack.isEmpty()) {
			if ( stack.peek().hasNext() ) {
				workingSet[this.idxs[stack.size()-1]] = stack.peek().next();
				
				if ( stack.size() == this.idxs.length )
					callback.onResult(workingSet, currentTime);
				else {
					EvaluationStep step = steps[stack.size()];
					// Encountered a preset situation
					if ( workingSet[step.getSituationIndex()] != null ) {
						// If satisfied... move along
						if ( step.isSatisfied(workingSet) )
							stack.push( new SingletonIterator(workingSet[step.getSituationIndex()]) );
					}
					// Query
					else
						stack.push( step.query(workingSet, estimate) );
				}
			}
			// Reset working-set if situation was not preset
			else if ( stack.pop().getClass() != SingletonIterator.class ){
				workingSet[idxs[stack.size()]] = null;
			}
		}
	}
}
