/**
 * 
 */
package de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.impl;

import java.util.Iterator;
import java.util.List;

import de.umr.jepc.v2.addons.epa.tpstream.Constraints;
import de.umr.jepc.v2.addons.epa.tpstream.Situation;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.Matcher.MatcherCallback;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.EvaluationStep;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.eval.Evaluator;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.RunningStats;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.SelectivityEstimate;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.TPStats;

/**
 * Recursive evaluator implementation, performing a depth first
 * search among the situation buffers.
 * In each step executes a search on the corresponding buffer
 * and adds the results one by one to the working-set. After setting
 * a result, a recursive call is executed to process the next step.
 * @author Michael Körber
 *
 */
public final class RecursiveEvaluator implements Evaluator {
	
    /** The statistics this evaluator is based on */
	private final TPStats statistics;
	
	/** The callback to invoke on new results */
	private final MatcherCallback callback;

	/** The steps to execute recursively */
	private final EvaluationStep[] steps;
	
	private final SelectivityEstimate estimate;
	
	/**
	 * @param statistics The statistics this evaluator is based on
	 * @param constraints The defined temporal constraints
	 * @param steps The steps to execute recursively
	 * @param callback The callback to invoke on new results
	 */
	public RecursiveEvaluator(TPStats statistics, Constraints constraints, List<EvaluationStep> steps, MatcherCallback callback) {
		super();
		this.statistics = statistics;
		this.callback = callback;
		this.steps = steps.toArray( new EvaluationStep[steps.size()] );
		this.estimate = new SelectivityEstimate(constraints);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public TPStats getStatistics() {
		return statistics;
	}
	
	/**
     * {@inheritDoc}
     */
    @Override
    public void evaluate(Situation[] workingSet, long currentTime, RunningStats statistics) {
        this.estimate.reset();
        process(0, workingSet, currentTime);
        this.estimate.updateStats(statistics);
    }

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void evaluate(Situation[] workingSet, long currentTime) {
        process(0, workingSet, currentTime);
	}
	
	/**
	 * Processes the steps recursivly and passes successful matches to the callback.
	 * @param stepIndex the index of the step to execute
	 * @param workingset the current partial temporal configuration
	 * @param currentTime the current time
	 * @param estimate the selectivity estimate to update during execution
	 */
	private void process( int stepIndex, Situation[] workingset, long currentTime ) {
    	// Done! Publish result
    	if ( stepIndex == steps.length ) {
    		callback.onResult( workingset, currentTime );
    		return;
    	}
    	
    	EvaluationStep evi = steps[stepIndex];
    	// We reached a pre-set situation
    	if ( workingset[evi.getSituationIndex()] != null ) {
    		if ( evi.isSatisfied(workingset) )
    			process(stepIndex+1, workingset, currentTime);
    	}
    	// Shrink results by searching buffer using constraints
    	else {
    		Iterator<Situation> results = evi.query(workingset,estimate);
    		while ( results.hasNext() ) {
    			workingset[evi.getSituationIndex()] = results.next();
    			process(stepIndex+1, workingset, currentTime);
    		}
    		workingset[evi.getSituationIndex()] = null;
    	}
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for ( EvaluationStep es : steps ) {
            sb.append(',').append(es.getSituationIndex());
        }
        sb.deleteCharAt(0);
        return "EvaluationOrder: [" + sb.toString() + "]";
    }
	
	

}
