/**
 * 
 */
package de.umr.jepc.v2.addons.epa.tpstream.optimizer;

import java.util.List;
import java.util.Set;

import de.umr.jepc.v2.addons.epa.tpstream.Constraint;
import de.umr.jepc.v2.addons.epa.tpstream.Constraints;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.constraints.ConstraintGraph;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.TPStats;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats.TriggerWrapper;

/**
 * The optimizer estimates the costs of an evaluation order 
 * based on recent {@link TPStats statistics} retrieved during execution.
 * This way TPStream is able to adapt to changing workloads.
 * @author Michael Körber
 *
 */
public final class Optimizer {

    /** The temporal constraints */
	private final Constraints constraints;

	/** The constraint graph */
	private final ConstraintGraph graph;

	/**
	 * @param constraints the defined temporal constraints
	 */
	public Optimizer(Constraints constraints) {
		this.constraints = constraints;
		this.graph = new ConstraintGraph(constraints);
	}

	/**
	 * Returns all situation ids of trigger situations.
	 * A trigger situation is a situation that may lead to 
	 * a new match, because the defined pattern may end with
	 * this situation's end.
	 * @return the ids of all trigger situations
	 */
	public Set<Character> getTriggerSituations() {
		return graph.getTriggerSymbols();
	}

	/**
	 * Computes the optimal evaluation order for the given trigger
	 * situation and assuming the given statistics.
	 * @param trigger the trigger situation to compute the order for
	 * @param stats the most recent statistics to use
	 * @return the optimal evaluation order for the given trigger
	 */
	public List<Character> calculateOrder(Character trigger, TPStats stats) {
		// Fix card of trigger situation to 1
		TriggerWrapper tstats = new TriggerWrapper(trigger, stats);

		List<Character> current = null;
		double costs = Double.POSITIVE_INFINITY;

		// Loop valid orders and estimate the costs
		for (List<Character> order : graph.getValidEvaluationOrdersFrom(trigger)) {
			double tmp = estimateCosts(order, tstats);
			if (tmp < costs) {
				costs = tmp;
				current = order;
			}
		}
		return current;
	}

	/**
	 * Calculates the evaluation costs (sum of intermediate results) for
	 * the given evaluation order and assuming the given statistics.
	 * @param order the evaluation order to estimate the costs for
	 * @param stats the statistics to use during computation
	 * @return the estimated costs
	 */
	public double estimateCosts(List<Character> order, TPStats stats) {
		// The cardinality of the initial buffer
		double card = stats.getCardinality( order.get(0) );
		double costs = card;

		for (int i = 1; i < order.size(); i++) {
			// Selectivity for current situation
			double selectivity = 1.0;
			// The current situation type
			Character current = order.get(i);
			// The cardinality of the current intermediate result
			double currentCard = stats.getCardinality(current);
			
			// Estimate selectivity for all applicable constraints
			for (Constraint c : this.constraints.getSituationConstraints(current)) {
				if (   c.idA == current && order.indexOf(c.idB) < i
					|| c.idB == current && order.indexOf(c.idA) < i ) {
					selectivity *= stats.getSelectivity(c);
				}
			}

			// Calculate intermediate result size and add to costs
			card *= currentCard * selectivity;
			costs += card;
		}
		return costs;
	}
}
