package de.umr.jepc.v2.addons.epa.tpstream.optimizer.constraints;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.umr.jepc.v2.addons.epa.tpstream.Constraint;
import de.umr.jepc.v2.addons.epa.tpstream.Constraints;
import de.umr.jepc.v2.addons.epa.tpstream.Constraint.TemporalRelation;

/**
 * Models the constraint graph to be evaluated by TPStream. 
 * Provides informations about valid evaluation orders and 
 * trigger situations.
 * @author Michael Körber
 *
 */
public class ConstraintGraph {

	/** The graph as adjacency lists */
	private Map<Character,List<Character>> adjLists = new HashMap<>();
	
	/** A cache for already computed valid evaluation orders */
	private Map<Character, Set<List<Character>>> orderCache = new HashMap<>();
	
	/** Stores all ids of situations that may trigger result construction */
	private final Set<Character> triggerSymbols;
	
	/**
	 * Creates a constraint graph from the given constraint definitions 
	 * @param constraints the constraint definitions
	 */
	public ConstraintGraph( Constraints constraints ) {
		for ( Constraint c : constraints ) {
			adjLists.put(c.idA, new ArrayList<>());
			adjLists.put(c.idB, new ArrayList<>());
		}
		
		for ( Constraint c : constraints ) {
			adjLists.get(c.idA).add(c.idB);
			adjLists.get(c.idB).add(c.idA);
		}
		this.triggerSymbols = Collections.unmodifiableSet(computeTriggerSymbols(constraints));
	}
	
	/**
	 * @return the ids of situations that may trigger result construction
	 */
	public Set<Character> getTriggerSymbols() {
	    return triggerSymbols;
	}
	
	/**
	 * @return All id-list permutations resulting in a valid evaluation order
	 */
	public List<List<Character>> getValidEvaluationOrders() {
        List<List<Character>> result = new ArrayList<>();
        for ( Character k : adjLists.keySet() ) {
            result.addAll( getValidEvaluationOrdersFrom(k) );
        }
        return result;
    }
	
	/**
	 * @param start the id of the situation tostart with
	 * @return all valid evaluation orders starting with the situation
	 */
	public Set<List<Character>> getValidEvaluationOrdersFrom( Character start ) {
	    Set<List<Character>> result = orderCache.get(start);
		if ( result == null ) {
		    result = new HashSet<>();
    		Deque<Character> seq = new ArrayDeque<>();
    		seq.addLast(start);
    		validPathsRecursive( seq, result );
    		orderCache.put(start, result);
		}
		return result;
	}
	
	/**
	 * Recursivly computes the valid situation stream orderings assuming that the streams on the given stack
	 * already represent a valid order.
	 * @param current the current stack of situation ids representing a valid ordering of size |current|
	 * @param accum stores all valid orders
	 */
	private void validPathsRecursive( Deque<Character> current, Set<List<Character>> accum ) {
		if ( current.size() ==  adjLists.size() ) {
			accum.add( Collections.unmodifiableList(new ArrayList<>(current)) );
		}
		else {
			for ( Character c : current ) {
				for ( Character n : adjLists.get(c) ) {
					if ( !current.contains(n) ) {
						current.addLast(n);
						validPathsRecursive(current, accum);
						current.pollLast();
					}
				}
			}
		}
	}
	
	
	/**
	 * Computes all situations which may trigger a result construction.
	 * These are the situations a pattern may end with.
	 * @param constraints the constraints describing the pattern
	 * @return the situations possibly triggering a result construction
	 */
	private static Set<Character> computeTriggerSymbols(Constraints constraints) {
        Set<Character> remaining = new HashSet<>();
        // Add all situations as candidates
        for ( Constraint c : constraints ) {
            remaining.add(c.idA);
            remaining.add(c.idB);
        }
        
        boolean changed = false;
        do {
            changed = false;
            for ( Constraint c : constraints ) {
                
                // if a ends before b, remove a
                // if b ends before a, remove b
                if ( isAEndsBeforeB(c) && remaining.remove(c.idA) ||
                     isAEndsAfterB (c) && remaining.remove(c.idB) )
                    changed = true;
                // if a and b end simultaneously, and one of them was exluded already, exclude the second one, too
                else if ( isAEndsWithB(c) && (!remaining.contains(c.idA) ^ !remaining.contains(c.idB)) )
                    changed |= remaining.remove(c.idA) || remaining.remove(c.idB);
                
            }
        } while (changed);
        
//        // FIXME: For experiment only
//        // Keep only one from simultaneous endings
//        for ( Constraint c : constraints ) {
//            if ( isAEndsWithB(c) && remaining.contains(c.idA) && remaining.contains(c.idB) )
//                remaining.remove(c.idA);
//        }
        
        
        return remaining;
    }
    
    /**
     * Tells if the given constraint implies that A has to end before B
     * @param c the constraint to check
     * @return <code>true</code> if A ends before B, <code>false</code> otherwise
     */
    private static boolean isAEndsBeforeB(Constraint c) {
        for ( TemporalRelation ir : c.relations ) {
            switch ( ir ) {
            case AABB:
            case AA_BB:
            case A_BAB:
            case ABAB:
            case BAAB:
                break;
            default:
                return false;
            }
        }
        return true;
    }
    
    /**
     * Tells if the given constraint implies that A has to end with B
     * @param c the constraint to check
     * @return <code>true</code> if A ends with B, <code>false</code> otherwise
     */
    private static boolean isAEndsWithB(Constraint c) {
        for ( TemporalRelation ir : c.relations ) {
            switch ( ir ) {
            case A_BA_B:
            case ABA_B:
            case BAA_B:
                break;
            default:
                return false;
            }
        }
        return true;
    }
    
    /**
     * Tells if the given constraint implies that A has to end after B
     * @param c the constraint to check
     * @return <code>true</code> if A ends after B, <code>false</code> otherwise
     */
    private static boolean isAEndsAfterB(Constraint c) {
        for ( TemporalRelation ir : c.relations ) {
            switch ( ir ) {
            case BBAA:
            case BB_AA:
            case A_BBA:
            case BABA:
            case ABBA:
                break;
            default:
                return false;
            }
        }
        return true;
    }
	

}
 