package de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats;

import de.umr.jepc.v2.addons.epa.tpstream.Constraint;
import de.umr.jepc.v2.addons.epa.tpstream.Constraints;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinitions;
import de.umr.jepc.v2.addons.epa.tpstream.buffer.SituationBuffers;

/**
 * Extension of the static {@link TPStats statistics}, to continuously
 * update the values during execution of TPStream.
 * The values are update using an Exponential Moving Average (EMA).
 * @author Michael Körber
 *
 */
public class RunningStats extends TPStats {

    /** The smoothing factor for EMA-computation of buffer sizes */
	private static final double ALPHA_BUF = 0.01;
	
	/** 1 minus the smoothing factor for buffer sizes */
	private static final double ALPHA_BUF_N = 1.0-ALPHA_BUF;
	
	/** The smoothing factor for EMA-computation of selectivities */
	private static final double ALPHA_SEL = 0.01;
	
	/** 1 minus the smoothing factor for selectivities */
	private static final double ALPHA_SEL_N = 1.0-ALPHA_SEL;
	
	
	/**
	 * Creates statistics and sets all of them to the pre-defined
     * initial values.
	 * @param definitions the situation definitions
     * @param constraints the temporal constraints
	 */
	public RunningStats(SituationDefinitions definitions, Constraints constraints) {
		super(definitions, constraints);
	}
	
	/**
	 * Creates an unmodifiable snapshot of the current statistics
	 * @return a snapshot of the current values
	 */
	public TPStats snapshot() {
		return new TPStats(this);
	}
	
	/**
	 * Sets the buffer-sizes to the actual sizes
	 * @param buffers the buffers
	 */
	public void setBuffers(SituationBuffers buffers) {
		for ( int i = 0; i < buffers.getNumberOfBuffers(); i++ ) {
			cardinalities[i] = buffers.getBuffer(i).getSize();
			ratios[i] = cardinalities[0] / cardinalities[i];
		}
	}
	
	/**
     * Updates EMA of all buffer sizes with the actual values obtained
     * from the given instances.
     * @param buffers the buffers
     */
	public void updateBuffers( SituationBuffers buffers ) {
		for ( int i = 0; i < buffers.getNumberOfBuffers(); i++ ) {
			updateBuffer(i, (double) buffers.getBuffer(i).getSize());
			ratios[i] = cardinalities[0] / cardinalities[i];
		}
	}
	
	/**
	 * Updates EMA of the given buffer with the given size and the configured
	 * smoothing factor. 
	 * @param situationStreamIdx the index of the situation stream
	 * @param size the current size of the corresponding buffer
	 */
	private void updateBuffer( Integer situationStreamIdx, double size ) {
		double val = ALPHA_BUF * size + ALPHA_BUF_N * cardinalities[situationStreamIdx];
		cardinalities[situationStreamIdx] = val;
	}
	
	/**
	 * Updates the EMA of the selectivity for the given constraint with the given
	 * value and configured smoothing factor.
	 * @param constraint the constraint to update the selectivity for
	 * @param value the measured selectivity
	 */
	public void updateSelectivity( Constraint constraint, double value ) {
		double nval = ALPHA_SEL * value + ALPHA_SEL_N * getSelectivity(constraint);
		selectivities.put(constraint, nval);
	}
}
