package de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import de.umr.jepc.v2.addons.epa.tpstream.Constraint;
import de.umr.jepc.v2.addons.epa.tpstream.Constraints;

/**
 * Helper to track the selectivity of constraints
 * during a single invocation of the matcher. This helper is required,
 * since every constraint may be evaluated multiple times during one 
 * matcher invocation.
 * 
 * @author Michael Körber
 */
public final class SelectivityEstimate {

    /**
     * Tiny helper, to compute a per invocation average value
     * of the selectivities
     * @author Michael Körber
     *
     */
	private static class Avg {
	    /** The number of times a value was added */
		private int count = 0;
		
		/** The sum of selectivities */
		private double sel = 0;
		
		/** 
		 * @return the averaged selectivity
		 */
		double get() {
			return count > 0 ? sel / count : 0;
		}
		
		/**
		 * Adds the selectivity of a single invocation of the corresponding constraint
		 * @param value the selectivity € [0,1]
		 */
		void add( double value ) {
			count++;
			sel += value;
		}
		
		/**
		 * Resets this average
		 */
		void reset() {
		    count = 0;
		    sel = 0;
		}
		
		/**
		 * {@inheritDoc}
		 */
		public String toString() {
			return ""+get();
		}
	}
	
	/** Maps each constraint to the current average */
	private Map<Constraint,Avg> estimates = new HashMap<>();
	
	/**
	 * @param constraints all defined constraints
	 */
	public SelectivityEstimate( Constraints constraints ) {
		for ( Constraint c : constraints )
			estimates.put(c, new Avg());
	}
	
	/**
	 * Updates the selectivity of the given constraint with the given value
	 * @param constraint the constraint
	 * @param selectivity the selectivity
	 */
	public void update( Constraint constraint, double selectivity ) {
		estimates.get(constraint).add(selectivity);
	}
	
	/**
	 * Retrieves the current (since the last call to {@link #reset()} average selectivity
	 * of the given constraint.
	 * @param constraint the constraint
	 * @return the averaged selectivity of the given constraint 
	 */
	public double getSelectivity( Constraint constraint ) {
		return estimates.get(constraint).get();
	}
	
	/**
	 * Updates the given running statistics with the current 
	 * averages.
	 * @param runningStats the statistics to update
	 */
	public void updateStats( RunningStats runningStats ) {
		for ( Entry<Constraint, Avg> e : estimates.entrySet() ) {
			if ( e.getValue().count > 0 )
				runningStats.updateSelectivity(e.getKey(), e.getValue().get());
		}
	}
	
	/**
	 * Resets all averages
	 */
	public void reset() {
	    for ( Avg avg : estimates.values() )
	        avg.reset();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder("QueryEstimate[\n");
		for ( Entry<Constraint, Avg> e : estimates.entrySet() ) {
			sb.append("  ").append(e.getKey().idA).append("->");
			sb.append(e.getKey().idB).append(": ").append(e.getValue()).append("\n");
		}
		sb.append("]");
		return sb.toString();
	}
}
