package de.umr.jepc.v2.addons.epa.tpstream.optimizer.stats;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import de.umr.jepc.v2.addons.epa.tpstream.Constraint;
import de.umr.jepc.v2.addons.epa.tpstream.Constraint.TemporalRelation;
import de.umr.jepc.v2.addons.epa.tpstream.Constraints;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinition;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinitions;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.Optimizer;

/**
 * Holds the statistics required by the {@link Optimizer} in order
 * to compute the optimal evaluation order for TPStream.
 * Those statistics are:
 * <ul>
 *  <li>The cardinalities of the situation buffers</li>
 *  <li>The selectivities of the individual temporal constraints</li>
 * </ul>
 * 
 * @author Michael Körber
 */
public class TPStats {

    /** Initially assume a cardinality of 1000 for each buffer */
    private static final double             INITIAL_CARDINALITY = 1000;

    /** The situation definitions */
    private final SituationDefinitions      definitions;

    /** The actual cardinalities */
    protected final double[]                cardinalities;

    /** The cardinality ratios, relative to the first buffer (cardinalities[0]) */
    protected final double[]                ratios;

    /** The selectivities per constraint */
    protected final Map<Constraint, Double> selectivities;

    /**
     * Clones the given statistics
     * @param template the statistics to clone
     */
    protected TPStats(TPStats template) {
        this.definitions = template.definitions;
        this.selectivities = new HashMap<>(template.selectivities);
        this.cardinalities = new double[template.cardinalities.length];
        this.ratios = new double[template.ratios.length];
        System.arraycopy(template.cardinalities, 0, cardinalities, 0, cardinalities.length);
        System.arraycopy(template.ratios, 0, ratios, 0, ratios.length);
    }

    /**
     * Creates statistics and sets all of them to the pre-defined
     * initial values.
     * @param definitions the situation definitions
     * @param constraints the temporal constraints
     */
    public TPStats(SituationDefinitions definitions, Constraints constraints) {
        this.definitions = definitions;
        this.selectivities = new HashMap<>();
        for (Constraint c : constraints) {
            selectivities.put(c, initialSelectivity(c));
        }
        this.cardinalities = new double[definitions.getNumberOfDefinitions()];
        this.ratios = new double[definitions.getNumberOfDefinitions()];

        for (SituationDefinition s : definitions) {
            cardinalities[definitions.getDefinitionIndex(s.getId())] = INITIAL_CARDINALITY;
        }
        Arrays.fill(ratios, 1.0);
    }

    /**
     * Computes the maximum value difference between this and the given statistics
     * in percent. This is either the variation of the selectivity of a constraint
     * or a change in the ratios of the situation buffer sizes. 
     * @param other the statistics to compare to
     * @return the maximum percentage difference
     */
    public double maxDiff(TPStats other) {
        double d = 0.0;

        for (int i = 0; i < cardinalities.length; i++) {
            d = Math.max(d, Math.abs(ratios[i] - other.ratios[i]));
        }

        for (Entry<Constraint, Double> se : selectivities.entrySet()) {
            d = Math.max(d, Math.abs(se.getValue() - other.selectivities.get(se.getKey())));
        }
        return d;
    }

    /**
     * Retrieves the cardinality for the situation buffer belonging to the given id.
     * @return the cardinality
     */
    public double getCardinality(Character situationStreamId) {
        return cardinalities[definitions.getDefinitionIndex(situationStreamId)];
    }

    /**
     * Retrieves the selectivity of the given constraint.
     * @param constraint the constraint
     * @return the selectivity of the given constraint
     */
    public double getSelectivity(Constraint constraint) {
        return selectivities.get(constraint);
    }

    /**
     * Retrieves the initial estimate of the selectivity for the given constraint.
     * @param constraint the constraints
     * @return the initial estimate of the selectivity for the given constraint
     */
    private static double initialSelectivity(Constraint constraint) {
        double s = 0.0;
        for (TemporalRelation ir : constraint.relations) {
            s += initialSelectivity(ir);
        }
        return s;
    }

    /**
     * Retrieves the initial estimate of the selectivity for the given temporal relation.
     * @param temporalRelation the temporal relation
     * @return the initial estimate of the selectivity for the given temporal relation
     */
    private static double initialSelectivity(TemporalRelation temporalRelation) {
        switch (temporalRelation) {
        case AABB:
        case BBAA:
            return 0.445;
        case ABBA:
        case BAAB:
            return 0.03;
        case ABAB:
        case BABA:
            return 0.01;
        case A_BA_B:
            return 0.0006;
        default:
            return 0.0049;
        }
    }

}
