/**
 * 
 */
package de.umr.jepc.v2.addons.epa.tpstream.partition;

import java.util.ArrayList;
import java.util.List;

import de.umr.ep.common.IncompatibleTypeException;
import de.umr.ep.event.EventSchema;
import de.umr.jepc.v2.addons.epa.tpstream.Constraints;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinition;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinitions;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.Matcher;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.Matcher.MatcherCallback;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.Optimizer;
import de.umr.jepc.v2.addons.epa.tpstream.partition.deriver.AbstractDeriver;
import de.umr.jepc.v2.api.epa.pattern.symbol.Binding;
import de.umr.jepc.v2.api.epa.pattern.symbol.Bindings;
import de.umr.jepc.v2.bindings.EVBinding;
import de.umr.jepc.v2.bindings.EVBindings;
import de.umr.jepc.v2.expressions.bool.EVBooleanExpression;
import de.umr.jepc.v2.ql.ExpressionTranslator;
import de.umr.jepc.v2.ql.TranslatorException;

/**
 * Abstract implementation of a {@link PartitionFactory}.
 * Pre-compiles all predicates and constraints and makes them available for
 * concrete implementors.
 * @author Michael Körber
 *
 */
public abstract class AbstractPartitionFactory implements PartitionFactory {
    
    /** The optimizer for the given operator configuration */
    private final Optimizer optimizer;
    
    /** The situation definitions */
    private final SituationDefinitions definitions;
    
    /** The constraint definitions */
    private final Constraints constraints;
    
    /** The window constraint */
    private final long window;
    
    /** The callback to invoke on new results */
    private final MatcherCallback callback;
    
    /** The set of compiled conditions */
    private final EVBooleanExpression[] conditions;
    
    /** The set of compiled aggregates */
    private final EVBindings[] aggregates;
    
    
    /**
     * @param schema the schema of incoming events
     * @param definitions the situation stream definitions
     * @param constraints the defined constraints
     * @param window the defined time-window
     * @param callback the callback to invoke on new results
     * @throws TranslatorException if conditions could not be translated
     * @throws IncompatibleTypeException if conditions could not be translated
     */
    protected AbstractPartitionFactory( EventSchema schema, SituationDefinitions definitions, Constraints constraints, long window, MatcherCallback callback ) throws TranslatorException, IncompatibleTypeException {
        this.optimizer = new Optimizer(constraints);
        this.definitions = definitions;
        this.constraints = constraints;
        this.window = window;
        this.callback = callback;
        this.conditions = new EVBooleanExpression[definitions.getNumberOfDefinitions()];
        this.aggregates = new EVBindings[definitions.getNumberOfDefinitions()];
        
        int i = 0;
        for ( SituationDefinition s : definitions ) {
            // Do not allow bindings to be accessed!
            conditions[i] = ExpressionTranslator.INSTANCE.translateBooleanExpression(s.getCondition(), schema, new Bindings());
            List<EVBinding> tmp = new ArrayList<>();
            for (Binding b : s.getAggregates()) {
                tmp.add(ExpressionTranslator.INSTANCE.translateVariableBinding(b, schema, new Bindings()));
            }
            this.aggregates[i] = new EVBindings( tmp.toArray(new EVBinding[tmp.size()]) );
            i++;
        }
    }
    
    /**
     * Creates a new Deriver instance for deriving situations from event streams
     * @param condition the condition to evaluate on incoming events
     * @param aggreagtes the aggregates to compute
     * @return a new Deriver instance
     */
    protected abstract AbstractDeriver createDeriver( EVBooleanExpression condition, EVBindings aggreagtes );
    
    /**
     * Creates a new matcher instance according to the given configuration
     * @param definitions the situation stream definitions
     * @param constraints the defined constraints
     * @param window the defined time-window
     * @param optimizer the optimizer used to adapt evaluation order
     * @param callback the callback to invoke on new results
     * @return a new matcher-instance
     */
    protected abstract Matcher createMatcher(SituationDefinitions definitions, Constraints constraints, long window, Optimizer optimizer, MatcherCallback callback);

    /**
     * {@inheritDoc}
     */
    @Override
    public TPSPartition create() {
        AbstractDeriver[] derivers = new AbstractDeriver[definitions.getNumberOfDefinitions()];
        for ( int i = 0; i < derivers.length; i++ ) {
            derivers[i] = createDeriver(conditions[i], aggregates[i]);
        }
        return new TPSPartition(derivers, createMatcher(definitions,constraints,window,optimizer,callback));
    }

}
