/*
 * Copyright (c) 2012, 2013
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.jepc.v2.addons.epa.tpstream.partition.deriver;

import de.umr.ep.event.Event;
import de.umr.jepc.v2.addons.epa.TPStream.Mode;
import de.umr.jepc.v2.addons.epa.tpstream.Situation;
import de.umr.jepc.v2.api.epa.pattern.symbol.Binding.BindingTime;
import de.umr.jepc.v2.bindings.EVBinding;
import de.umr.jepc.v2.bindings.EVBindings;
import de.umr.jepc.v2.expressions.bool.EVBooleanExpression;

/**
 * Abstract implementation of the deriver component. Provides
 * methods for updating the defined aggregate values to concrete
 * implementations. 
 *
 * @author Michael Körber
 */
public abstract class AbstractDeriver {

    /** The condition to evaluate */
    protected final EVBooleanExpression condition;

    /** The aggreagates to compute */
    protected final EVBindings          aggregates;

    /**
     * @param condition the condition to evaluate on incoming events
     * @param aggregates the aggregates to compute for each situation
     */
    protected AbstractDeriver(EVBooleanExpression condition, EVBindings aggregates) {
        this.condition = condition;
        this.aggregates = aggregates;
    }

    /**
     * Updates the state of this deriver with the current event.
     * @param event the event to update the state with
     * @return a situation instance, if required by the configured {@link Mode mode} of operation, null otherwise.
     */
    public abstract Situation update(Event event);

    /**
     * Updates all aggregate values of the currently constructed situation using the given event
     * @param event the event to update aggregates with
     * @param current the situation to update
     * @param isFirst indicator if this event initiated the given situation
     */
    protected final void updateAggregates(Event event, Situation current, boolean isFirst) {
        int i = 0;
        for (EVBinding b : aggregates) {
            if (isFirst || b.getBindingTime() == BindingTime.INCREMENTAL)
                current.setAggregate(i++, b.getExpression().eval(event, current.getAggregates()));
        }
    }

    /**
     * @{inheritDoc}
     */
    @Override
    public final String toString() {
        return "State [condition=" + condition + "]";
    }
}
