/*
 * Copyright (c) 2012, 2013
 * Database Research Group, University of Marburg.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.jepc.v2.addons.epa.tpstream;

import java.util.Arrays;
import java.util.List;

import org.junit.Assert;
import org.junit.Test;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.Attribute.DataType;
import de.umr.ep.common.EPException;
import de.umr.ep.event.Event;
import de.umr.ep.event.EventSchema;
import de.umr.ep.event.impl.SimpleEvent;
import de.umr.jepc.v2.addons.epa.TPStream;
import de.umr.jepc.v2.addons.epa.TPStream.Mode;
import de.umr.jepc.v2.addons.epa.tpstream.Constraint.TemporalRelation;
import de.umr.jepc.v2.addons.util.UserDefinedEPATest;
import de.umr.jepc.v2.api.epa.Stream;
import de.umr.jepc.v2.api.epa.pattern.Output;
import de.umr.jepc.v2.api.epa.pattern.Partition;
import de.umr.jepc.v2.api.epa.pattern.symbol.Binding;
import de.umr.jepc.v2.api.epa.pattern.symbol.Binding.BindingTime;
import de.umr.jepc.v2.api.epa.pattern.symbol.Bindings;
import de.umr.jepc.v2.api.epa.window.TimeWindow;
import de.umr.jepc.v2.api.expression.BooleanExpression;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.DoubleConstant;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.IntConstant;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.Variable;
import de.umr.jepc.v2.api.expression.arithmetic.compund.numeric.Abs;
import de.umr.jepc.v2.api.expression.arithmetic.compund.numeric.Subtract;
import de.umr.jepc.v2.api.expression.predicate.Between;
import de.umr.jepc.v2.api.expression.predicate.Equal;
import de.umr.jepc.v2.api.expression.predicate.Greater;
import de.umr.jepc.v2.api.expression.predicate.LessEq;
import de.umr.jepc.v2.api.provider.AlreadyRegisteredException;
import de.umr.jepc.v2.api.provider.NoSuchStreamException;
import de.umr.jepc.v2.api.provider.QueryHandle;
import de.umr.jepc.v2.api.provider.StreamHandle;
import de.umr.jepc.v2.api.util.CollectingSink;

/**
 * The MinimalityTest
 * DESCRIPTION Here
 * 
 * @author Michael Körber
 */
public class LivenessSemanticTest extends UserDefinedEPATest {

    private static final String   STREAM   = "test";

    private static final String   QUERY    = "tpstream";

    private static final Bindings BINDINGS = new Bindings(new Binding("n_id", new Variable("id"), BindingTime.INITIAL));

    private CollectingSink        sink;

    private StreamHandle          sh;
    private QueryHandle           qh;

    private void setup(long window, SituationDefinitions symbols, Constraints constraints, DataType valueType)
            throws AlreadyRegisteredException, NoSuchStreamException, EPException {
        EventSchema schema = new EventSchema(new Attribute("id", DataType.LONG), new Attribute("value", valueType));
        Output output = new Output("n_id");
        Partition partitionBy = new Partition("id");
        TPStream def = new TPStream(new TimeWindow(new Stream(STREAM), 1), Mode.LOW_LATENCY, window, symbols, constraints, output, partitionBy);

        sink = new CollectingSink();

        sh = provider.registerStream(STREAM, schema);
        qh = provider.createQuery(QUERY, def);
        provider.registerSink(qh, sink);

    }

//    @Test
    public void testDoubleDetectionAvoidance() throws AlreadyRegisteredException, NoSuchStreamException, EPException {

        Variable field = new Variable("value");
        IntConstant origin = new IntConstant(0);
        IntConstant three = new IntConstant(3);
        IntConstant dist = new IntConstant(2);

        BooleanExpression exp1 = new LessEq(new Abs(new Subtract(origin, field)), dist);
        BooleanExpression exp2 = new LessEq(new Abs(new Subtract(three, field)), dist);

        SituationDefinitions symbols = new SituationDefinitions(new SituationDefinition('a', exp1, BINDINGS), new SituationDefinition('b', exp2));

        Constraints constraints = new Constraints(new Constraint('a', 'b', TemporalRelation.ABAB));

        setup(100, symbols, constraints, DataType.INTEGER);

        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { 1L }, 5L, 6L));

        sh.pushEvent(new Object[] { 1L, 0 }, 1L);
        sh.pushEvent(new Object[] { 1L, 1 }, 2L);
        sh.pushEvent(new Object[] { 1L, 2 }, 3L);
        sh.pushEvent(new Object[] { 1L, 3 }, 4L);
        sh.pushEvent(new Object[] { 1L, 3 }, 5L);
        sh.pushEvent(new Object[] { 1L, 3 }, 6L);
        sh.pushEvent(new Object[] { 1L, 3 }, 7L);
        sh.pushEvent(new Object[] { 1L, 3 }, 8L);
        sh.pushEvent(new Object[] { 1L, 7 }, 9L);
        sh.pushEvent(new Object[] { 1L, 7 }, 10L);
        sh.pushEvent(new Object[] { 1L, 7 }, 11L);

        compareResults(expected, results);
    }

//    @Test
    public void testDoubleDetectionAvoidance2() throws AlreadyRegisteredException, NoSuchStreamException, EPException {
        Variable field = new Variable("value");
        IntConstant origin = new IntConstant(0);
        IntConstant three = new IntConstant(3);
        IntConstant dist = new IntConstant(5);

        SituationDefinitions symbols = new SituationDefinitions(
                new SituationDefinition('a', new LessEq(new Abs(new Subtract(origin, field)), dist), BINDINGS),
                new SituationDefinition('b', new Equal(field, three)));

        Constraints constraints = new Constraints(new Constraint('a', 'b', TemporalRelation.ABBA));

        setup(100, symbols, constraints, DataType.INTEGER);

        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { 1L }, 7L, 8L), new SimpleEvent(new Object[] { 1L }, 14L, 15L));

        long timestamp = 1;
        sh.pushEvent(new Object[] { 1L, 0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0 }, timestamp++);

        compareResults(expected, results);
    }

    /**
     * <pre>
     * Constraint : AA.B.B
     * Window-Size: 5
     * Cases:
     * 1. Detect if start of a is beyond window-start
     * W:        |----|
     * A: |-----------|
     * B:             .
     * 
     * 2. Do not yield if b grows to a real interval
     * W:          |----|
     * A: |-------------|
     * B:             |-|
     * </pre>
     * @throws EPException 
     * @throws NoSuchStreamException 
     * @throws AlreadyRegisteredException 
     */
    //    @Test
    //    public void testAExceedingWindowBInstant() throws AlreadyRegisteredException, NoSuchStreamException, EPException {
    //        
    //        Variable field = new Variable("value");
    //        DoubleConstant origin = new DoubleConstant(0.0);
    //        DoubleConstant lower = new DoubleConstant(-0.1);
    //        DoubleConstant upper = new DoubleConstant(10.1);
    //        
    //        Between pred1 = new Between(field, lower, upper);
    //        Equal   pred2 = new Equal(origin,field);
    //        
    //        
    //        Symbols symbols = new Symbols(
    //          new Symbol('a', pred1, BINDINGS ),
    //          new Symbol('b', pred2 )
    //        );
    //        
    //        Constraints cs = new Constraints(
    //            new Constraint('a', 'b', IntervalRelation.AA_B_B)
    //        );
    //        
    //        setup(5, symbols, cs, DataType.DOUBLE);
    //        
    //        List<Event> results = sink.getEvents();
    //
    //        List<Event> expected = Arrays.asList(
    //            new SimpleEvent(new Object[] { 1L }, 12L, 13L)
    //        );
    //        
    //        
    //        long timestamp = 1;
    //        sh.pushEvent( new Object[] { 1L, 10.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  9.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  8.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  7.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  6.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  5.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  4.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  3.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  2.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  1.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  0.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L, -1.0 }, timestamp++ );
    //        
    //        compareResults(expected, results);
    //        
    //        sh.pushEvent( new Object[] { 1L, -1.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  6.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  5.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  4.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  3.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  2.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  1.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  0.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L,  0.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L, -1.0 }, timestamp++ );
    //        sh.pushEvent( new Object[] { 1L, -1.0 }, timestamp++ );
    //        
    //        compareResults(expected, results);
    //    }

    /**
     * <pre>
     * Constraint : ABA.B
     * Window-Size: 5
     * Cases:
     * 1. Detect if start of a is beyond window-start
     * W:        |----|
     * A: |-----------|
     * B:           |-|
     * 
     * 2. Do not yield if a exceeds b
     * W:           |----|
     * A: |--------------|
     * B:             |-|
     * </pre>
     * @throws EPException 
     * @throws NoSuchStreamException 
     * @throws AlreadyRegisteredException 
     */
    @Test
    public void testAExceedingWindowBInterval() throws AlreadyRegisteredException, NoSuchStreamException, EPException {
        Variable field = new Variable("value");
        DoubleConstant origin = new DoubleConstant(0.0);
        DoubleConstant lower = new DoubleConstant(-0.1);
        DoubleConstant upper = new DoubleConstant(10.1);

        Between pred1 = new Between(field, lower, upper);
        Equal pred2 = new Equal(origin, field);

        SituationDefinitions symbols = new SituationDefinitions(new SituationDefinition('a', pred1, BINDINGS), new SituationDefinition('b', pred2));

        Constraints cs = new Constraints(new Constraint('a', 'b', TemporalRelation.ABA_B));

        setup(5, symbols, cs, DataType.DOUBLE);

        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { 1L }, 15L, 16L));

        long timestamp = 1;
        sh.pushEvent(new Object[] { 1L, 10.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 9.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 8.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 7.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 6.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 5.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 4.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, -1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, -1.0 }, timestamp++);

        compareResults(expected, results);

        sh.pushEvent(new Object[] { 1L, 6.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 5.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 4.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, -1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, -1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, -1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, -1.0 }, timestamp++);

        compareResults(expected, results);
    }

    /**
     * <pre>
     * Constraint : ABBA
     * Window-Size: 5
     * Cases:
     * 1. Do not detect if they close at the same time
     * W:        |----|
     * A: |-----------|
     * B:          |--|
     * 
     * 2. Do detect if a exceeds b and b.start is within window
     * W:           |----|
     * A: |---------------
     * B:            |--|
     * 
     * 2. Do NOTdetect if a exceeds b and b.start is NOT within window
     * W:           |----|
     * A: |---------------
     * B:         |-----|
     * </pre>
     * @throws EPException 
     * @throws NoSuchStreamException 
     * @throws AlreadyRegisteredException 
     */
//    @Test
    public void testWithin() throws AlreadyRegisteredException, NoSuchStreamException, EPException {
        Variable field = new Variable("value");
        DoubleConstant origin = new DoubleConstant(0.0);
        DoubleConstant lower = new DoubleConstant(-0.1);
        DoubleConstant upper = new DoubleConstant(10.1);

        Between pred1 = new Between(field, lower, upper);
        Equal pred2 = new Equal(origin, field);

        SituationDefinitions symbols = new SituationDefinitions(new SituationDefinition('a', pred1, BINDINGS), new SituationDefinition('b', pred2));

        Constraints cs = new Constraints(new Constraint('a', 'b', TemporalRelation.ABBA));

        setup(5, symbols, cs, DataType.DOUBLE);

        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { 1L }, 24L, 25L));

        List<Event> expected2 = Arrays.asList(new SimpleEvent(new Object[] { 1L }, 24L, 25L), new SimpleEvent(new Object[] { 1L }, 41L, 42L));

        long timestamp = 1;
        sh.pushEvent(new Object[] { 1L, 6.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 5.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 4.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, -1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, -1.0 }, timestamp++);

        Assert.assertTrue(results.isEmpty());

        sh.pushEvent(new Object[] { 1L, 6.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 5.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 4.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1.0 }, timestamp++);

        compareResults(expected, results);
        sh.pushEvent(new Object[] { 1L, 1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1.0 }, timestamp++);

        compareResults(expected, results);

        sh.pushEvent(new Object[] { 1L, 6.0 }, timestamp++);
        compareResults(expected, results);
        sh.pushEvent(new Object[] { 1L, 5.0 }, timestamp++);
        compareResults(expected, results);
        sh.pushEvent(new Object[] { 1L, 4.0 }, timestamp++);
        compareResults(expected, results);
        sh.pushEvent(new Object[] { 1L, 3.0 }, timestamp++);
        compareResults(expected, results);
        sh.pushEvent(new Object[] { 1L, 2.0 }, timestamp++);
        compareResults(expected, results);
        sh.pushEvent(new Object[] { 1L, 1.0 }, timestamp++);
        compareResults(expected, results);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        compareResults(expected, results);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        compareResults(expected, results);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        compareResults(expected, results);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1.0 }, timestamp++);
        compareResults(expected2, results);
        sh.pushEvent(new Object[] { 1L, 1.0 }, timestamp++);
        compareResults(expected2, results);
        sh.pushEvent(new Object[] { 1L, -1.0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, -1.0 }, timestamp++);

        compareResults(expected2, results);
    }

    /**
     * <pre>
     * Constraint : ABBA, ABAB
     * Window-Size: 5
     * Cases:
     * 1. Detect two matches:
     * W:    |----| |----|  
     * A: |-----------|
     * B:     |--|   |----
     * @throws EPException 
     * @throws NoSuchStreamException 
     * @throws AlreadyRegisteredException 
     */
    @Test
    public void testADoubleInvocation() throws AlreadyRegisteredException, NoSuchStreamException, EPException {
        //    Coordinate[] coords = new Coordinate[] {
        //        new Coordinate(-0.1, -0.1),
        //        new Coordinate(10.1, -0.1),
        //        new Coordinate(10.1, 10.1),
        //        new Coordinate(-0.1, 10.1),
        //        new Coordinate(-0.1, -0.1)
        //    };
        //    
        //    String   field  = "geom";
        //    Geometry origin = gf.createPoint( new Coordinate(0, 0));
        //    Geometry area  = gf.createPolygon(coords);
        //    Geometry outer  = gf.createPoint( new Coordinate(11, 0));
        //    
        //    Symbol[] symbols = new Symbol[] {
        //      new Symbol('a', new Within( field, area), new Variable("id","id",BindingTime.INITIAL) ),
        //      new Symbol('b', new Or(new Equal( field, origin), new IsWithinDistance( 2.0, field, outer) ) )
        //    };
        //    
        //    Constraint[] cs = new Constraint[] {
        //        new Constraint('a', 'b', IntervalRelation.ABBA, IntervalRelation.ABAB )
        //    };
        //    
        //    long     window = 5;
        //    String[] output = new String[] {"id"};
        //    
        //    TSOp op = new TSOp();
        //    TPStream spm = createSPM(Mode.FLOATING, symbols, cs, window, output);
        //    addOP(spm, op);
        //    
        //    long timestamp = 1;
        //    push( timestamp++, 1L, "Point(  6  6 )");
        //    push( timestamp++, 1L, "Point(  5  5 )");
        //    push( timestamp++, 1L, "Point(  4  4 )");
        //    push( timestamp++, 1L, "Point(  3  3 )");
        //    push( timestamp++, 1L, "Point(  2  2 )");
        //    push( timestamp++, 1L, "Point(  1  1 )");
        //    push( timestamp++, 1L, "Point(  0  0 )");
        //    push( timestamp++, 1L, "Point(  0  0 )");
        //    push( timestamp++, 1L, "Point(  1  0 )");
        //    push( timestamp++, 1L, "Point(  3  0 )");
        //    push( timestamp++, 1L, "Point(  8  0 )");
        //    push( timestamp++, 1L, "Point(  9  0 )");
        //    push( timestamp++, 1L, "Point(  10  0 )");
        //    push( timestamp++, 1L, "Point(  11  0 )");
        //    push( timestamp++, 1L, "Point(  11  11 )");
        //    push( timestamp++, 1L, "Point(  11  11 )");
        //    
        //    
        //    
        //    Assert.assertEquals(op.getResultCount(),2);
        //    Assert.assertEquals(op.getTimestamp(0),10);
        //    Assert.assertEquals(op.getTimestamp(1),15);
    }

    /**
     * <pre>
     * Constraint : AABB
     * Window-Size: 5
     * Cases:
     * 1. Detect if a exceeds window-starz:
     * W:    |----|  
     * A: |----|
     * B:        |----
     * 
     * 1. Do not detect if a left the window completely
     * W:       |----|  
     * A: |----|
     * B:           |----
     * @throws EPException 
     * @throws NoSuchStreamException 
     * @throws AlreadyRegisteredException 
     */
//    @Test
    public void testAABB() throws AlreadyRegisteredException, NoSuchStreamException, EPException {
        Variable field = new Variable("value");
        DoubleConstant origin = new DoubleConstant(0.0);
        DoubleConstant dest = new DoubleConstant(3.0);

        Equal pred1 = new Equal(field, origin);
        Equal pred2 = new Equal(field, dest);

        SituationDefinitions symbols = new SituationDefinitions(new SituationDefinition('a', pred1, BINDINGS), new SituationDefinition('b', pred2));

        Constraints cs = new Constraints(new Constraint('a', 'b', TemporalRelation.AABB));

        setup(5, symbols, cs, DataType.INTEGER);

        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { 1L }, 6L, 7L));

        long timestamp = 1;
        sh.pushEvent(new Object[] { 1L, 0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);

        compareResults(expected, results);

        sh.pushEvent(new Object[] { 1L, 0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);

        compareResults(expected, results);

        //    String   field  = "geom";
        //    Geometry origin = gf.createPoint( new Coordinate(0, 0));
        //    Geometry dest   = gf.createPoint( new Coordinate(3, 0));
        //    
        //    Symbol[] symbols = new Symbol[] {
        //      new Symbol('a', new Equal( field, origin), new Variable("id","id",BindingTime.INITIAL) ),
        //      new Symbol('b', new Equal( field, dest) )
        //    };
        //    
        //    Constraint[] cs = new Constraint[] {
        //        new Constraint('a', 'b', IntervalRelation.AABB)
        //    };
        //    
        //    long     window = 5;
        //    String[] output = new String[] {"id"};
        //    
        //    TSOp op = new TSOp();
        //    TPStream spm = createSPM(Mode.FLOATING, symbols, cs, window, output);
        //    addOP(spm, op);
        //    
        //    long timestamp = 1;
        //    push( timestamp++, 1L, "Point(  0 0 )");
        //    push( timestamp++, 1L, "Point(  0 0 )");
        //    push( timestamp++, 1L, "Point(  1 0 )");
        //    push( timestamp++, 1L, "Point(  2 0 )");
        //    push( timestamp++, 1L, "Point(  3 0 )");
        //    push( timestamp++, 1L, "Point(  3 0 )");
        //    push( timestamp++, 1L, "Point(  3 0 )");
        //    push( timestamp++, 1L, "Point(  3 0 )");
        //    push( timestamp++, 1L, "Point(  3 0 )");
        //    Assert.assertEquals(op.getResultCount(),1);
        //    Assert.assertEquals(op.getTimestamp(0),7);
        //    
        //    op.reset();
        //    
        //    push( timestamp++, 1L, "Point(  0 0 )");
        //    push( timestamp++, 1L, "Point(  0 0 )");
        //    push( timestamp++, 1L, "Point(  1 0 )");
        //    push( timestamp++, 1L, "Point(  2 0 )");
        //    push( timestamp++, 1L, "Point(  2 0 )");
        //    push( timestamp++, 1L, "Point(  2 0 )");
        //    push( timestamp++, 1L, "Point(  2 0 )");
        //    push( timestamp++, 1L, "Point(  3 0 )");
        //    push( timestamp++, 1L, "Point(  3 0 )");
        //    push( timestamp++, 1L, "Point(  3 0 )");
        //    push( timestamp++, 1L, "Point(  3 0 )");
        //    push( timestamp++, 1L, "Point(  3 0 )");
        //    Assert.assertEquals(op.getResultCount(),0);
    }

    @Test
    public void testSimultaneousEnd() throws AlreadyRegisteredException, NoSuchStreamException, EPException {
        Variable field = new Variable("value");

        Greater pred1 = new Greater(field, new IntConstant(1));
        Greater pred2 = new Greater(field, new IntConstant(3));

        SituationDefinitions symbols = new SituationDefinitions(new SituationDefinition('a', pred1, BINDINGS), new SituationDefinition('b', pred2));

        Constraints cs = new Constraints(new Constraint('a', 'b', TemporalRelation.ABA_B));

        setup(50, symbols, cs, DataType.INTEGER);

        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { 1L }, 9L, 10L));

        long timestamp = 1;
        sh.pushEvent(new Object[] { 1L, 0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 1 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 2 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 3 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 4 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 5 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 6 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 7 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0 }, timestamp++);
        compareResults(expected, results);
        sh.pushEvent(new Object[] { 1L, 0 }, timestamp++);
        sh.pushEvent(new Object[] { 1L, 0 }, timestamp++);
    }

}
