/**
 * 
 */
package de.umr.jepc.v2.addons.epa.tpstream;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.junit.Test;

import de.umr.jepc.v2.addons.epa.tpstream.Constraint.TemporalRelation;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.FixedDFSMatcher;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.FloatingDFSMatcher;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.Matcher;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.Matcher.MatcherCallback;
import de.umr.jepc.v2.addons.epa.tpstream.optimizer.Optimizer;
import de.umr.jepc.v2.api.expression.logical.True;

/**
 * @author Michael Körber
 *
 */
public class MatcherEvaluation {

    private static class CountingCallback implements MatcherCallback {
        private long count = 0;

        /**
         * {@inheritDoc}
         */
        @Override
        public void onResult(Situation[] result, long currentTime) {
            count++;
        }
    }

    private static interface MatcherFactory {
        Matcher create(SituationDefinitions ss, Constraints cs, long window, MatcherCallback callback);
    }

    private static final MatcherFactory FDFS_MF = new MatcherFactory() {
                                                    @Override
                                                    public Matcher create(SituationDefinitions ss, Constraints cs, long window, MatcherCallback callback) {
                                                        Optimizer opt = new Optimizer(cs);
                                                        return new FloatingDFSMatcher(ss, cs, window, opt, callback);
                                                    }

                                                    @Override
                                                    public String toString() {
                                                        return "FDFS Matcher";
                                                    }
                                                };

    private static final MatcherFactory DFS_MF  = new MatcherFactory() {
                                                    @Override
                                                    public Matcher create(SituationDefinitions ss, Constraints cs, long window, MatcherCallback callback) {
                                                        Optimizer opt = new Optimizer(cs);
                                                        return new FixedDFSMatcher(ss, cs, window, opt, callback);
                                                    }

                                                    @Override
                                                    public String toString() {
                                                        return "DFS Matcher ";
                                                    }
                                                };

    @Test
    public void run() {
        int numSymbols = 3;
        int interval = 10;
        long window = 20_000;

        runTest(DFS_MF, numSymbols, interval, window);
        runTest(FDFS_MF, numSymbols, interval, window);

    }

    private static void runTest(MatcherFactory mf1, int numSymbols, int intervalLength, long window) {
        char sym = 'a';
        List<SituationDefinition> ss = new ArrayList<>();
        List<Constraint> cs = new ArrayList<>();

        ss.add(new SituationDefinition(sym, new True()));

        for (int i = 1; i < numSymbols; i++) {
            cs.add(new Constraint(sym++, sym, TemporalRelation.AABB));
            ss.add(new SituationDefinition(sym, new True()));
        }

        SituationDefinitions symbols = new SituationDefinitions(ss.toArray(new SituationDefinition[ss.size()]));
        Constraints constraints = new Constraints(cs.toArray(new Constraint[cs.size()]));

        CountingCallback cb1 = new CountingCallback();

        Matcher m1 = mf1.create(symbols, constraints, window, cb1);

        long start = System.currentTimeMillis();
        long ts = 1;
        
        Situation[] work = new Situation[numSymbols];
        for (int i = 1; i <= 5_000; i++, ts += (intervalLength)) {
            Arrays.fill(work, null);
            if ( i % numSymbols == 0 ) {
                work[numSymbols-1] = new Situation(ts, ts + 1);
                m1.update(work, ts-window);
                work[numSymbols-1] = null;
            }
            for (int j = 1; j <= numSymbols; j++) {
                if (i % j == 0) {
                        work[j-1] = new Situation(ts, ts+1);
                        work[j-1].finish(ts + intervalLength);
                }
            }
            m1.update(work, ts - window + intervalLength);
        }
        long end = System.currentTimeMillis();

        System.out.println("Matcher: " + mf1 + ", #symbols: " + numSymbols + ", interval-length: " + intervalLength + ", window: " + window
                + ", time: " + (end - start) + "ms, #results: " + cb1.count);
    }

}
