package de.umr.jepc.v2.api;

import de.umr.ep.common.EPException;
import de.umr.ep.event.EventSchema;
import de.umr.jepc.v2.api.epa.EPA;
import de.umr.jepc.v2.api.provider.AlreadyRegisteredException;
import de.umr.jepc.v2.api.provider.DataSource;
import de.umr.jepc.v2.api.provider.EventSink;
import de.umr.jepc.v2.api.provider.ExternalStreamHandle;
import de.umr.jepc.v2.api.provider.NoSuchQueryException;
import de.umr.jepc.v2.api.provider.NoSuchStreamException;
import de.umr.jepc.v2.api.provider.QueryHandle;
import de.umr.jepc.v2.api.provider.StreamHandle;
import de.umr.jepc.v2.api.schema.StreamInfo;

/**
 * An EPProvider offers all methods for creating, updating and removing
 * continuous queries.
 * @author Michael Körber
 *
 */
public interface EPProvider {
    
    /**
     * Retrieves metadata of the given stream
     * @param name the stream's name
     * @return the associated meta information
     */
    StreamInfo getStreamInfo(String name) throws NoSuchStreamException;

	/**
	 * Registers a new raw event-stream
	 * @param name a unique name for the new stream
	 * @param schema the schema of the new stream
	 * @return a handle to access the newly registered stream
	 * @throws AlreadyRegisteredException if a stream with the given name is already registered
	 * @throws EPException on any error registering the stream
	 */
	StreamHandle registerStream(String name, EventSchema schema) throws AlreadyRegisteredException, EPException;
	
	/**
     * Registers a new external event-stream
     * @param name a unique name for the new stream
     * @param schema the schema of the new stream
     * @return a handle to access the newly registered stream
     * @throws AlreadyRegisteredException if a stream with the given name is already registered
     * @throws EPException on any error registering the stream
     */
    ExternalStreamHandle registerExternalStream(String name, EventSchema schema) throws AlreadyRegisteredException, EPException;

	/**
	 * Creates a new continuous query
	 * @param name the unique name of this query
	 * @param epn the EPA-graph representing this query
	 * @return a handle to access the newly created query
	 * @throws AlreadyRegisteredException if a query with the given name is already registered
	 * @throws NoSuchStreamException if this query consumes a raw stream which is not yet registered
	 * @throws EPException on any error creating the query
	 */
	QueryHandle createQuery(String name, EPA epn) throws AlreadyRegisteredException, NoSuchStreamException, EPException;
	
	/**
	 * Attaches the given sink to the given query
	 * @param query the handle referencing the target query
	 * @param sink the sink to attach
	 * @throws NoSuchQueryException if the referenced query is not defined 
	 */
	void registerSink( QueryHandle query, EventSink sink ) throws NoSuchQueryException;
	
	/**
	 * Removes the given stream. All subsequent queries will also be removed!
	 * @param stream the stream to remove
	 * @throws NoSuchStreamException if the referenced stream is not defined
	 */
	void removeStream( DataSource stream ) throws NoSuchStreamException;
	
	/**
	 * Removes the given query
	 * @param query the query to remove
	 * @throws NoSuchQueryException if the referenced query is not defined 
	 */
	void removeQuery( QueryHandle query ) throws NoSuchQueryException;
	
	/**
	 * Removes the given sink from the given query
	 * @param query the query to remove the sink from
	 * @param sink the sink to remove
	 * @throws NoSuchQueryException if the referenced query is not defined 
	 */
	void removeSink( QueryHandle query, EventSink sink ) throws NoSuchQueryException;
	
	/**
	 * Removes the given sink from all queries it is attached to
	 * @param sink the sink to remove
	 */
	void removeSink( EventSink sink );
	
	/**
     * Shuts down this provider. Any subsequent calls to any of this provider's methods will result
     * in undefined behaviour.
     */
	void destroy();

}
