/*
 * Copyright (c) 2012, 2013, 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.jepc.v2.test.bridge;

import java.util.Arrays;
import java.util.List;

import org.junit.Assert;
import org.junit.Test;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.Attribute.DataType;
import de.umr.ep.common.EPException;
import de.umr.ep.event.Event;
import de.umr.ep.event.EventSchema;
import de.umr.ep.event.SchemaException;
import de.umr.ep.event.impl.SimpleEvent;
import de.umr.jepc.v2.api.bridge.EPBridgeFactory;
import de.umr.jepc.v2.api.epa.Aggregator;
import de.umr.jepc.v2.api.epa.EPA;
import de.umr.jepc.v2.api.epa.Stream;
import de.umr.jepc.v2.api.epa.aggregates.Aggregate;
import de.umr.jepc.v2.api.epa.aggregates.Average;
import de.umr.jepc.v2.api.epa.aggregates.Count;
import de.umr.jepc.v2.api.epa.aggregates.Group;
import de.umr.jepc.v2.api.epa.aggregates.Maximum;
import de.umr.jepc.v2.api.epa.aggregates.Minimum;
import de.umr.jepc.v2.api.epa.aggregates.Stddev;
import de.umr.jepc.v2.api.epa.aggregates.Sum;
import de.umr.jepc.v2.api.epa.window.CountWindow;
import de.umr.jepc.v2.api.epa.window.TimeWindow;
import de.umr.jepc.v2.api.provider.AlreadyRegisteredException;
import de.umr.jepc.v2.api.provider.NoSuchStreamException;
import de.umr.jepc.v2.api.provider.QueryHandle;
import de.umr.jepc.v2.api.provider.StreamHandle;
import de.umr.jepc.v2.api.util.CollectingSink;

/**
 * @author Michael Körber
 *
 */
public class AggregatorTest extends AbstractSemanticTest {

    // private static final Logger log =
    // LoggerFactory.getLogger(AggregatorTest.class);

    private static final Stream STREAM = new Stream("test");
    private static final String QUERY  = "agg";

    public AggregatorTest(EPBridgeFactory bf) {
        super(bf);
    }

    /**
     * Tests aggregates (min, max and count) of a counting window with size 4 and jump of 2
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void jumpingCountWindowTest() throws EPException {
        Aggregator def = new Aggregator(
                new CountWindow(STREAM, 4, 2),
                //new Stream("myStream1",new TimeWindow(10, 2)),
                new Minimum("y","myMin"),new Maximum("y","myMax"), new Count("y", "countY"));
        
        CollectingSink sink = new CollectingSink();
        
        StreamHandle sh = provider.registerStream(STREAM.getName(), new EventSchema(
                new Attribute("y", DataType.INTEGER)));
        
        QueryHandle qh = provider.createQuery(QUERY, def);
        
        provider.registerSink(qh, sink);
        List<Event> results = sink.getEvents();
    
        Integer number = 10;
        Long ts = 1L;
    
        for(int i = 0; i < 21; i++){
            sh.pushEvent(new Object[]{number}, ts++);
            number += 10;
        }

        List<Event> expected = Arrays.asList(
                new SimpleEvent(new Object[] { 10, 20, 2L} , 2L, 4L ),
                new SimpleEvent(new Object[] { 10, 40, 4L} , 4L, 6L ),
                new SimpleEvent(new Object[] { 30, 60, 4L} , 6L, 8L ),
                new SimpleEvent(new Object[] { 50, 80, 4L} , 8L, 10L ),
                new SimpleEvent(new Object[] { 70, 100, 4L} , 10L, 12L ),
                new SimpleEvent(new Object[] { 90, 120, 4L} , 12L, 14L ),
                new SimpleEvent(new Object[] { 110, 140, 4L} , 14L, 16L )
        );

        compareResults(expected, results);
    }

    /**
     * Tests aggregates (avg, count, max, min, sum, stddev) by an attribute on a TimeWindow with size 6
     * @throws Exception
     */
    @Test
    public void simpleTest() throws Exception {

        Aggregate[] aggregates = { new Group("id"), new Average("value", "avg"), new Count("value", "count"), new Maximum("value", "max"),
                new Minimum("value", "min"), new Sum("value", "sum"), new Stddev("value", "stddev") };

        Aggregator def = new Aggregator(new TimeWindow(STREAM, 6), aggregates);

        CollectingSink sink = new CollectingSink();

        StreamHandle sh = provider.registerStream(STREAM.getName(), new EventSchema(
                new Attribute("id", DataType.LONG), 
                new Attribute("value", DataType.DOUBLE)));
        
        QueryHandle qh = provider.createQuery(QUERY, def);
        
        provider.registerSink(qh, sink);
        List<Event> results = sink.getEvents();
        
        List<Event> expected = Arrays.asList( 
            new SimpleEvent(new Object[] { 1L, 1.0, 1L, 1.0, 1.0, 1.0, 0.0 }, 1L, 2L),
            new SimpleEvent(new Object[] { 1L, 1.5, 2L, 2.0, 1.0, 3.0, 0.7071067811865476 }, 2L, 3L),
            new SimpleEvent(new Object[] { 1L, 2.0, 3L, 3.0, 1.0, 6.0, 1.0 }, 3L, 4L),
            new SimpleEvent(new Object[] { 1L, 2.5, 4L, 4.0, 1.0, 10.0, 1.2909944487358056 }, 4L, 5L),
            new SimpleEvent(new Object[] { 1L, 3.0, 5L, 5.0, 1.0, 15.0, 1.5811388300841898 }, 5L, 6L)
        );
        

        sh.pushEvent(new Object[] { 1L, 1.0 }, 1L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { 1L, 2.0 }, 2L);
        Assert.assertEquals(1,results.size());
        sh.pushEvent(new Object[] { 1L, 3.0 }, 3L);
        Assert.assertEquals(2,results.size());
        sh.pushEvent(new Object[] { 1L, 4.0 }, 4L);
        Assert.assertEquals(3,results.size());
        sh.pushEvent(new Object[] { 1L, 5.0 }, 5L);
        Assert.assertEquals(4,results.size());
        
        // Flush
        sh.pushEvent(new Object[] { 1L, 0.0 }, 6L);
        Assert.assertEquals(5,results.size());

        compareResults(expected, results);
    }

    /**
     * Tests an aggregate (sum) with type conversion from byte to long and float to double
     * Aggregates with Timewindow (size: 6)
     * @throws Exception
     */
    @Test
    public void typeConversionTest() throws Exception {
        EventSchema schema = new EventSchema(new Attribute("id", DataType.LONG), new Attribute("value_i", DataType.BYTE),
                new Attribute("value_f", DataType.FLOAT));

        Aggregate[] aggregates = { new Group("id"), new Sum("value_i", "sum_i"), new Sum("value_f", "sum_f") };

        Aggregator def = new Aggregator(new TimeWindow(STREAM, 6), aggregates);

        CollectingSink sink = new CollectingSink();

        StreamHandle sh = provider.registerStream(STREAM.getName(), schema);
        QueryHandle qh = provider.createQuery(QUERY, def);
        provider.registerSink(qh, sink);

        for (int i = 1; i <= 5; i++) {
            sh.pushEvent(new Object[] { 1L, (byte) i, (float) i }, i);
        }
        // Flush
        sh.pushEvent(new Object[] { 1L, (byte) 0, (float) 0 }, 6L);

        Event[] expected = { new SimpleEvent(new Object[] { 1L, 1L, 1.0 }, 1L, 2L), new SimpleEvent(new Object[] { 1L, 3L, 3.0 }, 2L, 3L),
                new SimpleEvent(new Object[] { 1L, 6L, 6.0 }, 3L, 4L), new SimpleEvent(new Object[] { 1L, 10L, 10.0 }, 4L, 5L),
                new SimpleEvent(new Object[] { 1L, 15L, 15.0 }, 5L, 6L)

        };
        compareResults(Arrays.asList(expected), sink.getEvents());
    }

    /**
     * Short test, weather the new aggregated (sum) events still ordered after their timestamps
     * Aggregate with Timewindow (size: 6)
     * @throws AlreadyRegisteredException
     * @throws EPException
     */
    @Test
    public void timeorderTest() throws EPException {
        EventSchema schema = new EventSchema(new Attribute("id", DataType.LONG), new Attribute("value", DataType.DOUBLE));
        Aggregate[] aggregates = { new Group("id"), new Sum("value", "sum"), };

        Aggregator def = new Aggregator(new TimeWindow(STREAM, 6), aggregates);

        CollectingSink sink = new CollectingSink();

        StreamHandle sh = provider.registerStream(STREAM.getName(), schema);
        QueryHandle qh = provider.createQuery("agg", def);
        provider.registerSink(qh, sink);

        sh.pushEvent(new Object[] { 1L, 1.0 }, 1L);
        sh.pushEvent(new Object[] { 2L, 1.0 }, 2L);
        sh.pushEvent(new Object[] { 2L, 1.0 }, 3L);

        // Flush
        sh.pushEvent(new Object[] { 3L, 1.0 }, 12l);

        Assert.assertEquals(4, sink.getEvents().size());
        for (int i = 1; i < sink.getEvents().size(); i++) {
            Assert.assertTrue(sink.getEvents().get(i - 1).getT1() <= sink.getEvents().get(i).getT1());
        }
    }

    /**
     * Tests aggregates (avg, max, min) with a timewindow (size: 3) on two groups
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void aggregator1() throws EPException {
        EPA epaA1 = new Aggregator(new TimeWindow(STREAM, 3),
                new Group("x"), new Average("y", "myavg"),
                new Maximum("y", "mymax"), new Minimum("y", "mymin"));

        CollectingSink sink = new CollectingSink();

        StreamHandle sh = provider.registerStream(STREAM.getName(), new EventSchema(
                new Attribute("x", Attribute.DataType.STRING), 
                new Attribute("y", Attribute.DataType.INTEGER)));
        
        QueryHandle qh = provider.createQuery(QUERY, epaA1);
        
        provider.registerSink(qh, sink);
        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(
                new SimpleEvent(new Object[] { "A", 40.0, 40, 40 }, 800001, 800003),
                new SimpleEvent(new Object[] { "B", 80.0, 80, 80 }, 800001, 800004),
                new SimpleEvent(new Object[] { "A", 40.5, 41, 40 }, 800003, 800004),
                new SimpleEvent(new Object[] { "A", 41.0, 41, 41 }, 800004, 800005),
                new SimpleEvent(new Object[] { "A", 41.5, 42, 41 }, 800005, 800006),
                new SimpleEvent(new Object[] { "A", 42.0, 42, 42 }, 800006, 800008),
                new SimpleEvent(new Object[] { "B", 80.0, 80, 80 }, 800007, 800010),
                new SimpleEvent(new Object[] { "B", 80.0, 80, 80 }, 800012, 800015));

        sh.pushEvent(new Object[] { "A", 40 }, 800001L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "B", 80 }, 800001L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "A", 41 }, 800003L);
        Assert.assertEquals(1,results.size());
        sh.pushEvent(new Object[] { "A", 42 }, 800005L);
        Assert.assertEquals(4,results.size());
        sh.pushEvent(new Object[] { "B", 80 }, 800007L);
        Assert.assertEquals(5,results.size());
        sh.pushEvent(new Object[] { "B", 80 }, 800012L);
        Assert.assertEquals(7,results.size());
        sh.pushEvent(new Object[] { "B", 85 }, 800015L);
        Assert.assertEquals(8,results.size());
        compareResults(expected, results);
    }

    /**
     * Tests aggregates (avg of floats becomes double, count on shorts becomes long)
     * with a timewindow (size: 2) on three groups and grouptype double
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void aggregator2() throws EPException {
        EPA epaAggregator = new Aggregator(new TimeWindow(STREAM, 2),
                new Average("z", "newZ"), new Group("y"), new Count("x", "newX"));

        CollectingSink sink = new CollectingSink();

        StreamHandle sh = provider.registerStream(STREAM.getName(), new EventSchema(
                new Attribute("x", Attribute.DataType.SHORT),
                new Attribute("y", Attribute.DataType.DOUBLE), 
                new Attribute("z", Attribute.DataType.FLOAT)));
        
        QueryHandle qh = provider.createQuery(QUERY, epaAggregator);

        provider.registerSink(qh, sink);
        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { 0.050999999046325684, 3.0, 1L }, 100025, 100027),
                new SimpleEvent(new Object[] { 0.052000001072883606, 0.0, 1L }, 100026, 100028),
                new SimpleEvent(new Object[] { 0.05299999937415123, 1.0, 1L }, 100026, 100028),

                new SimpleEvent(new Object[] { 0.05400000140070915, 2.0, 1L }, 100027, 100029),
                new SimpleEvent(new Object[] { 0.054999999701976776, 3.0, 1L }, 100027, 100029),

                new SimpleEvent(new Object[] { 0.0560000017285347, 0.0, 1L }, 100028, 100030),
                new SimpleEvent(new Object[] { 0.05700000002980232, 1.0, 1L }, 100028, 100030),

                new SimpleEvent(new Object[] { 0.057999998331069946, 2.0, 1L }, 100029, 100031),
                new SimpleEvent(new Object[] { 0.05900000035762787, 3.0, 1L }, 100029, 100031),

                new SimpleEvent(new Object[] { 0.05999999865889549, 0.0, 1L }, 100030, 100032),
                new SimpleEvent(new Object[] { 0.061000000685453415, 1.0, 1L }, 100030, 100032),

                new SimpleEvent(new Object[] { 0.06199999898672104, 2.0, 1L }, 100031, 100033),
                new SimpleEvent(new Object[] { 0.06300000101327896, 3.0, 1L }, 100031, 100033),

                new SimpleEvent(new Object[] { 0.06400000303983688, 0.0, 1L }, 100032, 100034),
                new SimpleEvent(new Object[] { 0.06499999761581421, 1.0, 1L }, 100032, 100034),

                new SimpleEvent(new Object[] { 0.06599999964237213, 2.0, 1L }, 100033, 100035),
                new SimpleEvent(new Object[] { 0.06700000166893005, 3.0, 1L }, 100033, 100035),

                new SimpleEvent(new Object[] { 0.06800000369548798, 0.0, 1L }, 100034, 100036),
                new SimpleEvent(new Object[] { 0.0689999982714653, 1.0, 1L }, 100034, 100036),

                new SimpleEvent(new Object[] { 0.07000000029802322, 2.0, 1L }, 100035, 100037),
                new SimpleEvent(new Object[] { 0.07100000232458115, 3.0, 1L }, 100035, 100037),

                new SimpleEvent(new Object[] { 0.07199999690055847, 0.0, 1L }, 100036, 100038),
                new SimpleEvent(new Object[] { 0.0729999989271164, 1.0, 1L }, 100036, 100038),

                new SimpleEvent(new Object[] { 0.07400000095367432, 2.0, 1L }, 100037, 100039),
                new SimpleEvent(new Object[] { 0.07500000298023224, 3.0, 1L }, 100037, 100039),

                new SimpleEvent(new Object[] { 0.07599999755620956, 0.0, 1L }, 100038, 100040),
                new SimpleEvent(new Object[] { 0.07699999958276749, 1.0, 1L }, 100038, 100040),

                new SimpleEvent(new Object[] { 0.07800000160932541, 2.0, 1L }, 100039, 100041),
                new SimpleEvent(new Object[] { 0.07900000363588333, 3.0, 1L }, 100039, 100041),

                new SimpleEvent(new Object[] { 0.07999999821186066, 0.0, 1L }, 100040, 100042),
                new SimpleEvent(new Object[] { 0.08100000023841858, 1.0, 1L }, 100040, 100042),

                new SimpleEvent(new Object[] { 0.0820000022649765, 2.0, 1L }, 100041, 100043),
                new SimpleEvent(new Object[] { 0.08299999684095383, 3.0, 1L }, 100041, 100043),

                new SimpleEvent(new Object[] { 0.08399999886751175, 0.0, 1L }, 100042, 100044),
                new SimpleEvent(new Object[] { 0.08500000089406967, 1.0, 1L }, 100042, 100044));

        sh.pushEvent(new Object[] { (short) (51), 3.0, (float) (0.051) }, 100025L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { (short) (52), 0.0, (float) (0.052) }, 100026L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { (short) (53), 1.0, (float) (0.053) }, 100026L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { (short) (54), 2.0, (float) (0.054) }, 100027L);
        Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { (short) (55), 3.0, (float) (0.055) }, 100027L);
        Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { (short) (56), 0.0, (float) (0.056) }, 100028L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { (short) (57), 1.0, (float) (0.057) }, 100028L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { (short) (58), 2.0, (float) (0.058) }, 100029L);
        Assert.assertEquals(5, results.size());
        sh.pushEvent(new Object[] { (short) (59), 3.0, (float) (0.059) }, 100029L);
        Assert.assertEquals(5, results.size());
        sh.pushEvent(new Object[] { (short) (60), 0.0, (float) (0.06) }, 100030L);
        Assert.assertEquals(7, results.size());
        sh.pushEvent(new Object[] { (short) (61), 1.0, (float) (0.061) }, 100030L);
        Assert.assertEquals(7, results.size());
        sh.pushEvent(new Object[] { (short) (62), 2.0, (float) (0.062) }, 100031L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { (short) (63), 3.0, (float) (0.063) }, 100031L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { (short) (64), 0.0, (float) (0.064) }, 100032L);
        Assert.assertEquals(11, results.size());
        sh.pushEvent(new Object[] { (short) (65), 1.0, (float) (0.065) }, 100032L);
        Assert.assertEquals(11, results.size());
        sh.pushEvent(new Object[] { (short) (66), 2.0, (float) (0.066) }, 100033L);
        Assert.assertEquals(13, results.size());
        sh.pushEvent(new Object[] { (short) (67), 3.0, (float) (0.067) }, 100033L);
        Assert.assertEquals(13, results.size());
        sh.pushEvent(new Object[] { (short) (68), 0.0, (float) (0.068) }, 100034L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { (short) (69), 1.0, (float) (0.069) }, 100034L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { (short) (70), 2.0, (float) (0.07) }, 100035L);
        Assert.assertEquals(17, results.size());
        sh.pushEvent(new Object[] { (short) (71), 3.0, (float) (0.071) }, 100035L);
        Assert.assertEquals(17, results.size());
        sh.pushEvent(new Object[] { (short) (72), 0.0, (float) (0.072) }, 100036L);
        Assert.assertEquals(19, results.size());
        sh.pushEvent(new Object[] { (short) (73), 1.0, (float) (0.073) }, 100036L);
        Assert.assertEquals(19, results.size());
        sh.pushEvent(new Object[] { (short) (74), 2.0, (float) (0.074) }, 100037L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { (short) (75), 3.0, (float) (0.075) }, 100037L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { (short) (76), 0.0, (float) (0.076) }, 100038L);
        Assert.assertEquals(23, results.size());
        sh.pushEvent(new Object[] { (short) (77), 1.0, (float) (0.077) }, 100038L);
        Assert.assertEquals(23, results.size());
        sh.pushEvent(new Object[] { (short) (78), 2.0, (float) (0.078) }, 100039L);
        Assert.assertEquals(25, results.size());
        sh.pushEvent(new Object[] { (short) (79), 3.0, (float) (0.079) }, 100039L);
        Assert.assertEquals(25, results.size());
        sh.pushEvent(new Object[] { (short) (80), 0.0, (float) (0.08) }, 100040L);
        Assert.assertEquals(27, results.size());
        sh.pushEvent(new Object[] { (short) (81), 1.0, (float) (0.081) }, 100040L);
        Assert.assertEquals(27, results.size());
        sh.pushEvent(new Object[] { (short) (82), 2.0, (float) (0.082) }, 100041L);
        Assert.assertEquals(29, results.size());
        sh.pushEvent(new Object[] { (short) (83), 3.0, (float) (0.083) }, 100041L);
        Assert.assertEquals(29, results.size());
        sh.pushEvent(new Object[] { (short) (84), 0.0, (float) (0.084) }, 100042L);
        Assert.assertEquals(31, results.size());
        sh.pushEvent(new Object[] { (short) (85), 1.0, (float) (0.085) }, 100042L);
        Assert.assertEquals(31, results.size());
        sh.pushEvent(new Object[] { (short) (86), 2.0, (float) (0.086) }, 100043L);
        Assert.assertEquals(33, results.size());
        sh.pushEvent(new Object[] { (short) (87), 3.0, (float) (0.087) }, 100043L);
        Assert.assertEquals(33, results.size());
        sh.pushEvent(new Object[] { (short) (88), 0.0, (float) (0.088) }, 100044L);
        Assert.assertEquals(35, results.size());
        sh.pushEvent(new Object[] { (short) (89), 1.0, (float) (0.089) }, 100044L);
        Assert.assertEquals(35, results.size());

        compareResults(expected, results);
    }

    /**
     * Tests aggregates (avg, count, max, min and sum of integers)
     * with a timewindow (size: 5) on three groups and grouptype String
     * Each millisecond one Event per group
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void aggregator3() throws EPException {
        EPA epaA1 = new Aggregator(new TimeWindow(STREAM, 5), new Group("x"), new Average("y", "myavg"), new Count("x", "mycount"),
                new Maximum("y", "mymax"), new Minimum("y", "mymin"), new Sum("y", "mysum"));

        CollectingSink sink = new CollectingSink();
        List<Event> results = sink.getEvents();

        StreamHandle sh = provider.registerStream(STREAM.getName(),
                new EventSchema(new Attribute("x", Attribute.DataType.STRING), new Attribute("y", Attribute.DataType.INTEGER)));

        QueryHandle qh = provider.createQuery(QUERY, epaA1);
        provider.registerSink(qh, sink);

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { "A", 0.0, 1L, 0, 0, 0L }, 800000, 800001),
                new SimpleEvent(new Object[] { "B", 0.0, 1L, 0, 0, 0L }, 800000, 800001),
                new SimpleEvent(new Object[] { "C", 0.0, 1L, 0, 0, 0L }, 800000, 800001),
                new SimpleEvent(new Object[] { "A", 0.5, 2L, 1, 0, 1L }, 800001, 800002),
                new SimpleEvent(new Object[] { "B", 0.5, 2L, 1, 0, 1L }, 800001, 800002),
                new SimpleEvent(new Object[] { "C", 0.5, 2L, 1, 0, 1L }, 800001, 800002),
                new SimpleEvent(new Object[] { "A", 1.0, 3L, 2, 0, 3L }, 800002, 800003),
                new SimpleEvent(new Object[] { "B", 1.0, 3L, 2, 0, 3L }, 800002, 800003),
                new SimpleEvent(new Object[] { "C", 1.0, 3L, 2, 0, 3L }, 800002, 800003),
                new SimpleEvent(new Object[] { "A", 1.5, 4L, 3, 0, 6L }, 800003, 800004),
                new SimpleEvent(new Object[] { "B", 1.5, 4L, 3, 0, 6L }, 800003, 800004),
                new SimpleEvent(new Object[] { "C", 1.5, 4L, 3, 0, 6L }, 800003, 800004),
                new SimpleEvent(new Object[] { "A", 2.0, 5L, 4, 0, 10L }, 800004, 800005),
                new SimpleEvent(new Object[] { "B", 2.0, 5L, 4, 0, 10L }, 800004, 800005),
                new SimpleEvent(new Object[] { "C", 2.0, 5L, 4, 0, 10L }, 800004, 800005),
                new SimpleEvent(new Object[] { "A", 3.0, 5L, 5, 1, 15L }, 800005, 800006),
                new SimpleEvent(new Object[] { "B", 3.0, 5L, 5, 1, 15L }, 800005, 800006),
                new SimpleEvent(new Object[] { "C", 3.0, 5L, 5, 1, 15L }, 800005, 800006),
                new SimpleEvent(new Object[] { "A", 4.0, 5L, 6, 2, 20L }, 800006, 800007),
                new SimpleEvent(new Object[] { "B", 4.0, 5L, 6, 2, 20L }, 800006, 800007),
                new SimpleEvent(new Object[] { "C", 4.0, 5L, 6, 2, 20L }, 800006, 800007),
                new SimpleEvent(new Object[] { "A", 5.0, 5L, 7, 3, 25L }, 800007, 800008),
                new SimpleEvent(new Object[] { "B", 5.0, 5L, 7, 3, 25L }, 800007, 800008),
                new SimpleEvent(new Object[] { "C", 5.0, 5L, 7, 3, 25L }, 800007, 800008),
                new SimpleEvent(new Object[] { "A", 6.0, 5L, 8, 4, 30L }, 800008, 800009),
                new SimpleEvent(new Object[] { "B", 6.0, 5L, 8, 4, 30L }, 800008, 800009),
                new SimpleEvent(new Object[] { "C", 6.0, 5L, 8, 4, 30L }, 800008, 800009));

        sh.pushEvent(new Object[] { "A", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "B", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "C", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "A", 1 }, 800001L);
        Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { "B", 1 }, 800001L);
        Assert.assertEquals(2, results.size());
        sh.pushEvent(new Object[] { "C", 1 }, 800001L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "A", 2 }, 800002L);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { "B", 2 }, 800002L);
        Assert.assertEquals(5, results.size());
        sh.pushEvent(new Object[] { "C", 2 }, 800002L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "A", 3 }, 800003L);
        Assert.assertEquals(7, results.size());
        sh.pushEvent(new Object[] { "B", 3 }, 800003L);
        Assert.assertEquals(8, results.size());
        sh.pushEvent(new Object[] { "C", 3 }, 800003L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { "A", 4 }, 800004L);
        Assert.assertEquals(10, results.size());
        sh.pushEvent(new Object[] { "B", 4 }, 800004L);
        Assert.assertEquals(11, results.size());
        sh.pushEvent(new Object[] { "C", 4 }, 800004L);
        Assert.assertEquals(12, results.size());
        sh.pushEvent(new Object[] { "A", 5 }, 800005L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { "B", 5 }, 800005L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { "C", 5 }, 800005L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { "A", 6 }, 800006L);
        Assert.assertEquals(18, results.size());
        sh.pushEvent(new Object[] { "B", 6 }, 800006L);
        Assert.assertEquals(18, results.size());
        sh.pushEvent(new Object[] { "C", 6 }, 800006L);
        Assert.assertEquals(18, results.size());
        sh.pushEvent(new Object[] { "A", 7 }, 800007L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "B", 7 }, 800007L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "C", 7 }, 800007L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "A", 8 }, 800008L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "B", 8 }, 800008L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "C", 8 }, 800008L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "A", 9 }, 800009L);
        Assert.assertEquals(27, results.size());
        sh.pushEvent(new Object[] { "B", 9 }, 800009L);
        Assert.assertEquals(27, results.size());
        sh.pushEvent(new Object[] { "C", 9 }, 800009L);
        Assert.assertEquals(27, results.size());

        compareResults(expected, results);
    }

    /**
     * Tests aggregates (avg, count, max, min and sum of integers)
     * with a timewindow (size: 5) on three groups and grouptype String
     * Each two milliseconds one Event per group
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void aggregator4() throws EPException {
        EPA epaA1 = new Aggregator(new TimeWindow(STREAM, 5), new Group("x"), new Average("y", "myavg"), new Count("x", "mycount"),
                new Maximum("y", "mymax"), new Minimum("y", "mymin"), new Sum("y", "mysum"));

        CollectingSink sink = new CollectingSink();
        List<Event> results = sink.getEvents();

        StreamHandle sh = provider.registerStream(STREAM.getName(),
                new EventSchema(new Attribute("x", Attribute.DataType.STRING), new Attribute("y", Attribute.DataType.INTEGER)));

        QueryHandle qh = provider.createQuery(QUERY, epaA1);
        provider.registerSink(qh, sink);

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { "A", 0.0, 1L, 0, 0, 0L }, 800000L, 800002L),
                new SimpleEvent(new Object[] { "B", 0.0, 1L, 0, 0, 0L }, 800000L, 800002L),
                new SimpleEvent(new Object[] { "C", 0.0, 1L, 0, 0, 0L }, 800000L, 800002L),
                new SimpleEvent(new Object[] { "A", 1.0, 2L, 2, 0, 2L }, 800002L, 800004L),
                new SimpleEvent(new Object[] { "B", 1.0, 2L, 2, 0, 2L }, 800002L, 800004L),
                new SimpleEvent(new Object[] { "C", 1.0, 2L, 2, 0, 2L }, 800002L, 800004L),
                new SimpleEvent(new Object[] { "A", 2.0, 3L, 4, 0, 6L }, 800004L, 800005L),
                new SimpleEvent(new Object[] { "B", 2.0, 3L, 4, 0, 6L }, 800004L, 800005L),
                new SimpleEvent(new Object[] { "C", 2.0, 3L, 4, 0, 6L }, 800004L, 800005L),
                new SimpleEvent(new Object[] { "A", 3.0, 2L, 4, 2, 6L }, 800005L, 800006L),
                new SimpleEvent(new Object[] { "B", 3.0, 2L, 4, 2, 6L }, 800005L, 800006L),
                new SimpleEvent(new Object[] { "C", 3.0, 2L, 4, 2, 6L }, 800005L, 800006L),
                new SimpleEvent(new Object[] { "A", 4.0, 3L, 6, 2, 12L }, 800006L, 800007L),
                new SimpleEvent(new Object[] { "B", 4.0, 3L, 6, 2, 12L }, 800006L, 800007L),
                new SimpleEvent(new Object[] { "C", 4.0, 3L, 6, 2, 12L }, 800006L, 800007L),
                new SimpleEvent(new Object[] { "A", 5.0, 2L, 6, 4, 10L }, 800007L, 800008L),
                new SimpleEvent(new Object[] { "B", 5.0, 2L, 6, 4, 10L }, 800007L, 800008L),
                new SimpleEvent(new Object[] { "C", 5.0, 2L, 6, 4, 10L }, 800007L, 800008L),
                new SimpleEvent(new Object[] { "A", 6.0, 3L, 8, 4, 18L }, 800008L, 800009L),
                new SimpleEvent(new Object[] { "B", 6.0, 3L, 8, 4, 18L }, 800008L, 800009L),
                new SimpleEvent(new Object[] { "C", 6.0, 3L, 8, 4, 18L }, 800008L, 800009L),
                new SimpleEvent(new Object[] { "A", 7.0, 2L, 8, 6, 14L }, 800009L, 800010L),
                new SimpleEvent(new Object[] { "B", 7.0, 2L, 8, 6, 14L }, 800009L, 800010L),
                new SimpleEvent(new Object[] { "C", 7.0, 2L, 8, 6, 14L }, 800009L, 800010L),
                new SimpleEvent(new Object[] { "A", 8.0, 3L, 10, 6, 24L }, 800010L, 800011L),
                new SimpleEvent(new Object[] { "B", 8.0, 3L, 10, 6, 24L }, 800010L, 800011L),
                new SimpleEvent(new Object[] { "C", 8.0, 3L, 10, 6, 24L }, 800010L, 800011L),
                new SimpleEvent(new Object[] { "A", 9.0, 2L, 10, 8, 18L }, 800011L, 800012L),
                new SimpleEvent(new Object[] { "B", 9.0, 2L, 10, 8, 18L }, 800011L, 800012L),
                new SimpleEvent(new Object[] { "C", 9.0, 2L, 10, 8, 18L }, 800011L, 800012L),
                new SimpleEvent(new Object[] { "A", 10.0, 3L, 12, 8, 30L }, 800012L, 800013L),
                new SimpleEvent(new Object[] { "B", 10.0, 3L, 12, 8, 30L }, 800012L, 800013L),
                new SimpleEvent(new Object[] { "C", 10.0, 3L, 12, 8, 30L }, 800012L, 800013L),
                new SimpleEvent(new Object[] { "A", 11.0, 2L, 12, 10, 22L }, 800013L, 800014L),
                new SimpleEvent(new Object[] { "B", 11.0, 2L, 12, 10, 22L }, 800013L, 800014L),
                new SimpleEvent(new Object[] { "C", 11.0, 2L, 12, 10, 22L }, 800013L, 800014L),
                new SimpleEvent(new Object[] { "A", 12.0, 3L, 14, 10, 36L }, 800014L, 800015L),
                new SimpleEvent(new Object[] { "B", 12.0, 3L, 14, 10, 36L }, 800014L, 800015L),
                new SimpleEvent(new Object[] { "C", 12.0, 3L, 14, 10, 36L }, 800014L, 800015L),
                new SimpleEvent(new Object[] { "A", 13.0, 2L, 14, 12, 26L }, 800015L, 800016L),
                new SimpleEvent(new Object[] { "B", 13.0, 2L, 14, 12, 26L }, 800015L, 800016L),
                new SimpleEvent(new Object[] { "C", 13.0, 2L, 14, 12, 26L }, 800015L, 800016L),
                new SimpleEvent(new Object[] { "A", 14.0, 3L, 16, 12, 42L }, 800016L, 800017L),
                new SimpleEvent(new Object[] { "B", 14.0, 3L, 16, 12, 42L }, 800016L, 800017L),
                new SimpleEvent(new Object[] { "C", 14.0, 3L, 16, 12, 42L }, 800016L, 800017L),
                new SimpleEvent(new Object[] { "A", 15.0, 2L, 16, 14, 30L }, 800017L, 800018L),
                new SimpleEvent(new Object[] { "B", 15.0, 2L, 16, 14, 30L }, 800017L, 800018L),
                new SimpleEvent(new Object[] { "C", 15.0, 2L, 16, 14, 30L }, 800017L, 800018L));

        sh.pushEvent(new Object[] { "A", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "B", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "C", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "A", 2 }, 800002L);
        Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { "B", 2 }, 800002L);
        Assert.assertEquals(2, results.size());
        sh.pushEvent(new Object[] { "C", 2 }, 800002L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "A", 4 }, 800004L);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { "B", 4 }, 800004L);
        Assert.assertEquals(5, results.size());
        sh.pushEvent(new Object[] { "C", 4 }, 800004L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "A", 6 }, 800006L);
        Assert.assertEquals(10, results.size());
        sh.pushEvent(new Object[] { "B", 6 }, 800006L);
        Assert.assertEquals(11, results.size());
        sh.pushEvent(new Object[] { "C", 6 }, 800006L);
        Assert.assertEquals(12, results.size());
        sh.pushEvent(new Object[] { "A", 8 }, 800008L);
        Assert.assertEquals(16, results.size());
        sh.pushEvent(new Object[] { "B", 8 }, 800008L);
        Assert.assertEquals(17, results.size());
        sh.pushEvent(new Object[] { "C", 8 }, 800008L);
        Assert.assertEquals(18, results.size());
        sh.pushEvent(new Object[] { "A", 10 }, 800010L);
        Assert.assertEquals(22, results.size());
        sh.pushEvent(new Object[] { "B", 10 }, 800010L);
        Assert.assertEquals(23, results.size());
        sh.pushEvent(new Object[] { "C", 10 }, 800010L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "A", 12 }, 800012L);
        Assert.assertEquals(28, results.size());
        sh.pushEvent(new Object[] { "B", 12 }, 800012L);
        Assert.assertEquals(29, results.size());
        sh.pushEvent(new Object[] { "C", 12 }, 800012L);
        Assert.assertEquals(30, results.size());
        sh.pushEvent(new Object[] { "A", 14 }, 800014L);
        Assert.assertEquals(34, results.size());
        sh.pushEvent(new Object[] { "B", 14 }, 800014L);
        Assert.assertEquals(35, results.size());
        sh.pushEvent(new Object[] { "C", 14 }, 800014L);
        Assert.assertEquals(36, results.size());
        sh.pushEvent(new Object[] { "A", 16 }, 800016L);
        Assert.assertEquals(40, results.size());
        sh.pushEvent(new Object[] { "B", 16 }, 800016L);
        Assert.assertEquals(41, results.size());
        sh.pushEvent(new Object[] { "C", 16 }, 800016L);
        Assert.assertEquals(42, results.size());
        sh.pushEvent(new Object[] { "A", 18 }, 800018L);
        Assert.assertEquals(46, results.size());
        sh.pushEvent(new Object[] { "B", 18 }, 800018L);
        Assert.assertEquals(47, results.size());
        sh.pushEvent(new Object[] { "C", 18 }, 800018L);
        Assert.assertEquals(48, results.size());

        compareResults(expected, results);
    }

    /**
     * Tests aggregates (avg, count, max, min and sum of integers)
     * with a timewindow (size: 5) on three groups and grouptype String
     * Each five milliseconds one Event per group
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void aggregator5() throws EPException {
        EPA epaA1 = new Aggregator(new TimeWindow(STREAM, 5), new Group("x"), new Average("y", "myavg"), new Count("x", "mycount"),
                new Maximum("y", "mymax"), new Minimum("y", "mymin"), new Sum("y", "mysum"));

        CollectingSink sink = new CollectingSink();
        List<Event> results = sink.getEvents();

        StreamHandle sh = provider.registerStream(STREAM.getName(),
                new EventSchema(new Attribute("x", Attribute.DataType.STRING), new Attribute("y", Attribute.DataType.INTEGER)));

        QueryHandle qh = provider.createQuery(QUERY, epaA1);
        provider.registerSink(qh, sink);

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { "A", 0.0, 1L, 0, 0, 0L }, 800000L, 800005L),
                new SimpleEvent(new Object[] { "B", 0.0, 1L, 0, 0, 0L }, 800000L, 800005L),
                new SimpleEvent(new Object[] { "C", 0.0, 1L, 0, 0, 0L }, 800000L, 800005L),
                new SimpleEvent(new Object[] { "A", 5.0, 1L, 5, 5, 5L }, 800005L, 800010L),
                new SimpleEvent(new Object[] { "B", 5.0, 1L, 5, 5, 5L }, 800005L, 800010L),
                new SimpleEvent(new Object[] { "C", 5.0, 1L, 5, 5, 5L }, 800005L, 800010L),
                new SimpleEvent(new Object[] { "A", 10.0, 1L, 10, 10, 10L }, 800010L, 800015L),
                new SimpleEvent(new Object[] { "B", 10.0, 1L, 10, 10, 10L }, 800010L, 800015L),
                new SimpleEvent(new Object[] { "C", 10.0, 1L, 10, 10, 10L }, 800010L, 800015L),
                new SimpleEvent(new Object[] { "A", 15.0, 1L, 15, 15, 15L }, 800015L, 800020L),
                new SimpleEvent(new Object[] { "B", 15.0, 1L, 15, 15, 15L }, 800015L, 800020L),
                new SimpleEvent(new Object[] { "C", 15.0, 1L, 15, 15, 15L }, 800015L, 800020L),
                new SimpleEvent(new Object[] { "A", 20.0, 1L, 20, 20, 20L }, 800020L, 800025L),
                new SimpleEvent(new Object[] { "B", 20.0, 1L, 20, 20, 20L }, 800020L, 800025L),
                new SimpleEvent(new Object[] { "C", 20.0, 1L, 20, 20, 20L }, 800020L, 800025L),
                new SimpleEvent(new Object[] { "A", 25.0, 1L, 25, 25, 25L }, 800025L, 800030L),
                new SimpleEvent(new Object[] { "B", 25.0, 1L, 25, 25, 25L }, 800025L, 800030L),
                new SimpleEvent(new Object[] { "C", 25.0, 1L, 25, 25, 25L }, 800025L, 800030L),
                new SimpleEvent(new Object[] { "A", 30.0, 1L, 30, 30, 30L }, 800030L, 800035L),
                new SimpleEvent(new Object[] { "B", 30.0, 1L, 30, 30, 30L }, 800030L, 800035L),
                new SimpleEvent(new Object[] { "C", 30.0, 1L, 30, 30, 30L }, 800030L, 800035L),
                new SimpleEvent(new Object[] { "A", 35.0, 1L, 35, 35, 35L }, 800035L, 800040L),
                new SimpleEvent(new Object[] { "B", 35.0, 1L, 35, 35, 35L }, 800035L, 800040L),
                new SimpleEvent(new Object[] { "C", 35.0, 1L, 35, 35, 35L }, 800035L, 800040L),
                new SimpleEvent(new Object[] { "A", 40.0, 1L, 40, 40, 40L }, 800040L, 800045L),
                new SimpleEvent(new Object[] { "B", 40.0, 1L, 40, 40, 40L }, 800040L, 800045L),
                new SimpleEvent(new Object[] { "C", 40.0, 1L, 40, 40, 40L }, 800040L, 800045L));

        sh.pushEvent(new Object[] { "A", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "B", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "C", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "A", 5 }, 800005L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "B", 5 }, 800005L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "C", 5 }, 800005L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "A", 10 }, 800010L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "B", 10 }, 800010L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "C", 10 }, 800010L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "A", 15 }, 800015L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { "B", 15 }, 800015L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { "C", 15 }, 800015L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { "A", 20 }, 800020L);
        Assert.assertEquals(12, results.size());
        sh.pushEvent(new Object[] { "B", 20 }, 800020L);
        Assert.assertEquals(12, results.size());
        sh.pushEvent(new Object[] { "C", 20 }, 800020L);
        Assert.assertEquals(12, results.size());
        sh.pushEvent(new Object[] { "A", 25 }, 800025L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { "B", 25 }, 800025L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { "C", 25 }, 800025L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { "A", 30 }, 800030L);
        Assert.assertEquals(18, results.size());
        sh.pushEvent(new Object[] { "B", 30 }, 800030L);
        Assert.assertEquals(18, results.size());
        sh.pushEvent(new Object[] { "C", 30 }, 800030L);
        Assert.assertEquals(18, results.size());
        sh.pushEvent(new Object[] { "A", 35 }, 800035L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "B", 35 }, 800035L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "C", 35 }, 800035L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "A", 40 }, 800040L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "B", 40 }, 800040L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "C", 40 }, 800040L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "A", 45 }, 800045L);
        Assert.assertEquals(27, results.size());
        sh.pushEvent(new Object[] { "B", 45 }, 800045L);
        Assert.assertEquals(27, results.size());
        sh.pushEvent(new Object[] { "C", 45 }, 800045L);
        Assert.assertEquals(27, results.size());

        compareResults(expected, results);
    }

    /**
     * Tests aggregates (avg, count, max, min and sum of integers)
     * with a timewindow (size: 5) on three groups and grouptype String
     * Each six milliseconds one Event per group
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void aggregator6() throws EPException {
        EPA epaA1 = new Aggregator(new TimeWindow(STREAM, 5), new Group("x"), new Average("y", "myavg"), new Count("x", "mycount"),
                new Maximum("y", "mymax"), new Minimum("y", "mymin"), new Sum("y", "mysum"));

        CollectingSink sink = new CollectingSink();
        List<Event> results = sink.getEvents();

        StreamHandle sh = provider.registerStream(STREAM.getName(),
                new EventSchema(new Attribute("x", Attribute.DataType.STRING), new Attribute("y", Attribute.DataType.INTEGER)));

        QueryHandle qh = provider.createQuery(QUERY, epaA1);
        provider.registerSink(qh, sink);

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { "A", 0.0, 1L, 0, 0, 0L }, 800000L, 800005L),
                new SimpleEvent(new Object[] { "B", 0.0, 1L, 0, 0, 0L }, 800000L, 800005L),
                new SimpleEvent(new Object[] { "C", 0.0, 1L, 0, 0, 0L }, 800000L, 800005L),
                new SimpleEvent(new Object[] { "A", 6.0, 1L, 6, 6, 6L }, 800006L, 800011L),
                new SimpleEvent(new Object[] { "B", 6.0, 1L, 6, 6, 6L }, 800006L, 800011L),
                new SimpleEvent(new Object[] { "C", 6.0, 1L, 6, 6, 6L }, 800006L, 800011L),
                new SimpleEvent(new Object[] { "A", 12.0, 1L, 12, 12, 12L }, 800012L, 800017L),
                new SimpleEvent(new Object[] { "B", 12.0, 1L, 12, 12, 12L }, 800012L, 800017L),
                new SimpleEvent(new Object[] { "C", 12.0, 1L, 12, 12, 12L }, 800012L, 800017L),
                new SimpleEvent(new Object[] { "A", 18.0, 1L, 18, 18, 18L }, 800018L, 800023L),
                new SimpleEvent(new Object[] { "B", 18.0, 1L, 18, 18, 18L }, 800018L, 800023L),
                new SimpleEvent(new Object[] { "C", 18.0, 1L, 18, 18, 18L }, 800018L, 800023L),
                new SimpleEvent(new Object[] { "A", 24.0, 1L, 24, 24, 24L }, 800024L, 800029L),
                new SimpleEvent(new Object[] { "B", 24.0, 1L, 24, 24, 24L }, 800024L, 800029L),
                new SimpleEvent(new Object[] { "C", 24.0, 1L, 24, 24, 24L }, 800024L, 800029L),
                new SimpleEvent(new Object[] { "A", 30.0, 1L, 30, 30, 30L }, 800030L, 800035L),
                new SimpleEvent(new Object[] { "B", 30.0, 1L, 30, 30, 30L }, 800030L, 800035L),
                new SimpleEvent(new Object[] { "C", 30.0, 1L, 30, 30, 30L }, 800030L, 800035L),
                new SimpleEvent(new Object[] { "A", 36.0, 1L, 36, 36, 36L }, 800036L, 800041L),
                new SimpleEvent(new Object[] { "B", 36.0, 1L, 36, 36, 36L }, 800036L, 800041L),
                new SimpleEvent(new Object[] { "C", 36.0, 1L, 36, 36, 36L }, 800036L, 800041L),
                new SimpleEvent(new Object[] { "A", 42.0, 1L, 42, 42, 42L }, 800042L, 800047L),
                new SimpleEvent(new Object[] { "B", 42.0, 1L, 42, 42, 42L }, 800042L, 800047L),
                new SimpleEvent(new Object[] { "C", 42.0, 1L, 42, 42, 42L }, 800042L, 800047L),
                new SimpleEvent(new Object[] { "A", 48.0, 1L, 48, 48, 48L }, 800048L, 800053L),
                new SimpleEvent(new Object[] { "B", 48.0, 1L, 48, 48, 48L }, 800048L, 800053L),
                new SimpleEvent(new Object[] { "C", 48.0, 1L, 48, 48, 48L }, 800048L, 800053L));

        sh.pushEvent(new Object[] { "A", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "B", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "C", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "A", 6 }, 800006L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "B", 6 }, 800006L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "C", 6 }, 800006L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "A", 12 }, 800012L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "B", 12 }, 800012L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "C", 12 }, 800012L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "A", 18 }, 800018L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { "B", 18 }, 800018L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { "C", 18 }, 800018L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { "A", 24 }, 800024L);
        Assert.assertEquals(12, results.size());
        sh.pushEvent(new Object[] { "B", 24 }, 800024L);
        Assert.assertEquals(12, results.size());
        sh.pushEvent(new Object[] { "C", 24 }, 800024L);
        Assert.assertEquals(12, results.size());
        sh.pushEvent(new Object[] { "A", 30 }, 800030L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { "B", 30 }, 800030L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { "C", 30 }, 800030L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { "A", 36 }, 800036L);
        Assert.assertEquals(18, results.size());
        sh.pushEvent(new Object[] { "B", 36 }, 800036L);
        Assert.assertEquals(18, results.size());
        sh.pushEvent(new Object[] { "C", 36 }, 800036L);
        Assert.assertEquals(18, results.size());
        sh.pushEvent(new Object[] { "A", 42 }, 800042L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "B", 42 }, 800042L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "C", 42 }, 800042L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "A", 48 }, 800048L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "B", 48 }, 800048L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "C", 48 }, 800048L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "A", 54 }, 800054L);
        Assert.assertEquals(27, results.size());
        sh.pushEvent(new Object[] { "B", 54 }, 800054L);
        Assert.assertEquals(27, results.size());
        sh.pushEvent(new Object[] { "C", 54 }, 800054L);
        Assert.assertEquals(27, results.size());

        compareResults(expected, results);
    }

    /**
     * Tests aggregates (avg, count, max, min and sum of integers)
     * with a timewindow (size: 5) on three groups and grouptype String
     * Each seven milliseconds one Event per group
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void aggregator7() throws EPException {
        EPA epaA1 = new Aggregator(new TimeWindow(STREAM, 5), new Group("x"), new Average("y", "myavg"), new Count("x", "mycount"),
                new Maximum("y", "mymax"), new Minimum("y", "mymin"), new Sum("y", "mysum"));

        CollectingSink sink = new CollectingSink();
        List<Event> results = sink.getEvents();

        StreamHandle sh = provider.registerStream(STREAM.getName(),
                new EventSchema(new Attribute("x", Attribute.DataType.STRING), new Attribute("y", Attribute.DataType.INTEGER)));

        QueryHandle qh = provider.createQuery(QUERY, epaA1);
        provider.registerSink(qh, sink);

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { "A", 0.0, 1L, 0, 0, 0L }, 800000L, 800005L),
                new SimpleEvent(new Object[] { "B", 0.0, 1L, 0, 0, 0L }, 800000L, 800005L),
                new SimpleEvent(new Object[] { "C", 0.0, 1L, 0, 0, 0L }, 800000L, 800005L),
                new SimpleEvent(new Object[] { "A", 7.0, 1L, 7, 7, 7L }, 800007L, 800012L),
                new SimpleEvent(new Object[] { "B", 7.0, 1L, 7, 7, 7L }, 800007L, 800012L),
                new SimpleEvent(new Object[] { "C", 7.0, 1L, 7, 7, 7L }, 800007L, 800012L),
                new SimpleEvent(new Object[] { "A", 14.0, 1L, 14, 14, 14L }, 800014L, 800019L),
                new SimpleEvent(new Object[] { "B", 14.0, 1L, 14, 14, 14L }, 800014L, 800019L),
                new SimpleEvent(new Object[] { "C", 14.0, 1L, 14, 14, 14L }, 800014L, 800019L),
                new SimpleEvent(new Object[] { "A", 21.0, 1L, 21, 21, 21L }, 800021L, 800026L),
                new SimpleEvent(new Object[] { "B", 21.0, 1L, 21, 21, 21L }, 800021L, 800026L),
                new SimpleEvent(new Object[] { "C", 21.0, 1L, 21, 21, 21L }, 800021L, 800026L),
                new SimpleEvent(new Object[] { "A", 28.0, 1L, 28, 28, 28L }, 800028L, 800033L),
                new SimpleEvent(new Object[] { "B", 28.0, 1L, 28, 28, 28L }, 800028L, 800033L),
                new SimpleEvent(new Object[] { "C", 28.0, 1L, 28, 28, 28L }, 800028L, 800033L),
                new SimpleEvent(new Object[] { "A", 35.0, 1L, 35, 35, 35L }, 800035L, 800040L),
                new SimpleEvent(new Object[] { "B", 35.0, 1L, 35, 35, 35L }, 800035L, 800040L),
                new SimpleEvent(new Object[] { "C", 35.0, 1L, 35, 35, 35L }, 800035L, 800040L),
                new SimpleEvent(new Object[] { "A", 42.0, 1L, 42, 42, 42L }, 800042L, 800047L),
                new SimpleEvent(new Object[] { "B", 42.0, 1L, 42, 42, 42L }, 800042L, 800047L),
                new SimpleEvent(new Object[] { "C", 42.0, 1L, 42, 42, 42L }, 800042L, 800047L),
                new SimpleEvent(new Object[] { "A", 49.0, 1L, 49, 49, 49L }, 800049L, 800054L),
                new SimpleEvent(new Object[] { "B", 49.0, 1L, 49, 49, 49L }, 800049L, 800054L),
                new SimpleEvent(new Object[] { "C", 49.0, 1L, 49, 49, 49L }, 800049L, 800054L),
                new SimpleEvent(new Object[] { "A", 56.0, 1L, 56, 56, 56L }, 800056L, 800061L),
                new SimpleEvent(new Object[] { "B", 56.0, 1L, 56, 56, 56L }, 800056L, 800061L),
                new SimpleEvent(new Object[] { "C", 56.0, 1L, 56, 56, 56L }, 800056L, 800061L));

        sh.pushEvent(new Object[] { "A", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "B", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "C", 0 }, 800000L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "A", 7 }, 800007L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "B", 7 }, 800007L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "C", 7 }, 800007L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "A", 14 }, 800014L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "B", 14 }, 800014L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "C", 14 }, 800014L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "A", 21 }, 800021L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { "B", 21 }, 800021L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { "C", 21 }, 800021L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { "A", 28 }, 800028L);
        Assert.assertEquals(12, results.size());
        sh.pushEvent(new Object[] { "B", 28 }, 800028L);
        Assert.assertEquals(12, results.size());
        sh.pushEvent(new Object[] { "C", 28 }, 800028L);
        Assert.assertEquals(12, results.size());
        sh.pushEvent(new Object[] { "A", 35 }, 800035L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { "B", 35 }, 800035L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { "C", 35 }, 800035L);
        Assert.assertEquals(15, results.size());
        sh.pushEvent(new Object[] { "A", 42 }, 800042L);
        Assert.assertEquals(18, results.size());
        sh.pushEvent(new Object[] { "B", 42 }, 800042L);
        Assert.assertEquals(18, results.size());
        sh.pushEvent(new Object[] { "C", 42 }, 800042L);
        Assert.assertEquals(18, results.size());
        sh.pushEvent(new Object[] { "A", 49 }, 800049L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "B", 49 }, 800049L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "C", 49 }, 800049L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "A", 56 }, 800056L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "B", 56 }, 800056L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "C", 56 }, 800056L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "A", 63 }, 800063L);
        Assert.assertEquals(27, results.size());
        sh.pushEvent(new Object[] { "B", 63 }, 800063L);
        Assert.assertEquals(27, results.size());
        sh.pushEvent(new Object[] { "C", 63 }, 800063L);
        Assert.assertEquals(27, results.size());

        compareResults(expected, results);
    }

    /**
     * Tests aggregate (avg of integers)
     * with a timewindow (size: 20) and different time distances between events
     *
     * @throws AlreadyRegisteredException
     * @throws SchemaException
     * @throws EPException
     */
    @Test
    public void aggregator8() throws EPException {
        EPA epaC2 = new Aggregator(new TimeWindow(STREAM, 20), new Average("m", "avg_m"));

        CollectingSink sink = new CollectingSink();
        List<Event> results = sink.getEvents();

        StreamHandle sh = provider.registerStream(STREAM.getName(),
                new EventSchema(new Attribute("n", Attribute.DataType.STRING), new Attribute("m", Attribute.DataType.INTEGER)));

        QueryHandle qh = provider.createQuery(QUERY, epaC2);
        provider.registerSink(qh, sink);

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { 1.0 }, 600001L, 600002L),
                new SimpleEvent(new Object[] { 2.5 }, 600002L, 600003L), new SimpleEvent(new Object[] { 3.3333333333333335 }, 600003L, 600004L),
                new SimpleEvent(new Object[] { 5.2 }, 600004L, 600005L), new SimpleEvent(new Object[] { 7.142857142857143 }, 600005L, 600006L),
                new SimpleEvent(new Object[] { 8.125 }, 600006L, 600011L), new SimpleEvent(new Object[] { 11.88888888888889 }, 600011L, 600012L),
                new SimpleEvent(new Object[] { 14.9 }, 600012L, 600013L), new SimpleEvent(new Object[] { 13.545454545454545 }, 600013L, 600017L),
                new SimpleEvent(new Object[] { 12.5 }, 600017L, 600021L), new SimpleEvent(new Object[] { 12.583333333333334 }, 600021L, 600022L),
                new SimpleEvent(new Object[] { 13.363636363636363 }, 600022L, 600023L), new SimpleEvent(new Object[] { 14.2 }, 600023L, 600024L),
                new SimpleEvent(new Object[] { 15.75 }, 600024L, 600025L), new SimpleEvent(new Object[] { 15.0 }, 600025L, 600026L),
                new SimpleEvent(new Object[] { 15.0 }, 600026L, 600029L), new SimpleEvent(new Object[] { 13.428571428571429 }, 600029L, 600031L),
                new SimpleEvent(new Object[] { 8.666666666666666 }, 600031L, 600032L), new SimpleEvent(new Object[] { 2.0 }, 600032L, 600033L),
                new SimpleEvent(new Object[] { 3.0 }, 600033L, 600037L), new SimpleEvent(new Object[] { 4.0 }, 600037L, 600041L),
                new SimpleEvent(new Object[] { 5.0 }, 600041L, 600045L), new SimpleEvent(new Object[] { 6.0 }, 600045L, 600049L),
                new SimpleEvent(new Object[] { 7.0 }, 600049L, 600053L), new SimpleEvent(new Object[] { 8.0 }, 600053L, 600057L),
                new SimpleEvent(new Object[] { 9.0 }, 600057L, 600061L), new SimpleEvent(new Object[] { 10.0 }, 600061L, 600065L),
                new SimpleEvent(new Object[] { 11.0 }, 600065L, 600069L), new SimpleEvent(new Object[] { 12.0 }, 600069L, 600073L),
                new SimpleEvent(new Object[] { 13.0 }, 600073L, 600077L), new SimpleEvent(new Object[] { 14.0 }, 600077L, 600081L),
                new SimpleEvent(new Object[] { 15.0 }, 600081L, 600085L), new SimpleEvent(new Object[] { 16.0 }, 600085L, 600089L));

        sh.pushEvent(new Object[] { "Event", 1 }, 600001L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Event", 4 }, 600002L);
        Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { "Event", 5 }, 600003L);
        Assert.assertEquals(2, results.size());
        sh.pushEvent(new Object[] { "Event", 7 }, 600004L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "Event", 9 }, 600004L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "Processing", 11 }, 600005L);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { "Processing", 13 }, 600005L);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { "Event", 15 }, 600006L);
        Assert.assertEquals(5, results.size());
        sh.pushEvent(new Object[] { "Event", 42 }, 600011L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "Bla", 42 }, 600012L);
        Assert.assertEquals(7, results.size());
        sh.pushEvent(new Object[] { "Bla", 0 }, 600013L);
        Assert.assertEquals(8, results.size());
        sh.pushEvent(new Object[] { "Bla", 1 }, 600017L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { "Bla", 2 }, 600021L);
        Assert.assertEquals(10, results.size());
        sh.pushEvent(new Object[] { "Bla", 3 }, 600025L);
        Assert.assertEquals(14, results.size());
        sh.pushEvent(new Object[] { "Bla", 4 }, 600029L);
        Assert.assertEquals(16, results.size());
        sh.pushEvent(new Object[] { "Bla", 5 }, 600033L);
        Assert.assertEquals(19, results.size());
        sh.pushEvent(new Object[] { "Bla", 6 }, 600037L);
        Assert.assertEquals(20, results.size());
        sh.pushEvent(new Object[] { "Bla", 7 }, 600041L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "Bla", 8 }, 600045L);
        Assert.assertEquals(22, results.size());
        sh.pushEvent(new Object[] { "Bla", 9 }, 600049L);
        Assert.assertEquals(23, results.size());
        sh.pushEvent(new Object[] { "Bla", 10 }, 600053L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "Bla", 11 }, 600057L);
        Assert.assertEquals(25, results.size());
        sh.pushEvent(new Object[] { "Bla", 12 }, 600061L);
        Assert.assertEquals(26, results.size());
        sh.pushEvent(new Object[] { "Bla", 13 }, 600065L);
        Assert.assertEquals(27, results.size());
        sh.pushEvent(new Object[] { "Bla", 14 }, 600069L);
        Assert.assertEquals(28, results.size());
        sh.pushEvent(new Object[] { "Bla", 15 }, 600073L);
        Assert.assertEquals(29, results.size());
        sh.pushEvent(new Object[] { "Bla", 16 }, 600077L);
        Assert.assertEquals(30, results.size());
        sh.pushEvent(new Object[] { "Bla", 17 }, 600081L);
        Assert.assertEquals(31, results.size());
        sh.pushEvent(new Object[] { "Bla", 18 }, 600085L);
        Assert.assertEquals(32, results.size());
        sh.pushEvent(new Object[] { "Bla", 19 }, 600089L);
        Assert.assertEquals(33, results.size());

        compareResults(expected, results);
    }

    /**
     * Tests aggregate (avg of integers)
     * with a timewindow (size: 20) and different time distances between events
     * equal to aggregator8() ???
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void aggregator9() throws EPException {
        // if(provider instanceof WebMethodsprovider)
        // return;
        EPA epaC9 = new Aggregator(new TimeWindow(STREAM, 20), new Average("m", "avg_m"));

        CollectingSink sink = new CollectingSink();
        List<Event> results = sink.getEvents();

        StreamHandle sh = provider.registerStream(STREAM.getName(),
                new EventSchema(new Attribute("n", Attribute.DataType.STRING), new Attribute("m", Attribute.DataType.INTEGER)));

        QueryHandle qh = provider.createQuery(QUERY, epaC9);
        provider.registerSink(qh, sink);

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { 1.0 }, 600001L, 600002L),
                new SimpleEvent(new Object[] { 2.5 }, 600002L, 600003L), new SimpleEvent(new Object[] { 3.3333333333333335 }, 600003L, 600004L),
                new SimpleEvent(new Object[] { 5.2 }, 600004L, 600005L), new SimpleEvent(new Object[] { 7.142857142857143 }, 600005L, 600006L),
                new SimpleEvent(new Object[] { 8.125 }, 600006L, 600011L), new SimpleEvent(new Object[] { 11.88888888888889 }, 600011L, 600012L),
                new SimpleEvent(new Object[] { 14.9 }, 600012L, 600013L), new SimpleEvent(new Object[] { 13.545454545454545 }, 600013L, 600017L),
                new SimpleEvent(new Object[] { 12.5 }, 600017L, 600021L), new SimpleEvent(new Object[] { 12.583333333333334 }, 600021L, 600022L),
                new SimpleEvent(new Object[] { 13.363636363636363 }, 600022L, 600023L), new SimpleEvent(new Object[] { 14.2 }, 600023L, 600024L),
                new SimpleEvent(new Object[] { 15.75 }, 600024L, 600025L), new SimpleEvent(new Object[] { 15.0 }, 600025L, 600026L),
                new SimpleEvent(new Object[] { 15.0 }, 600026L, 600029L), new SimpleEvent(new Object[] { 13.428571428571429 }, 600029L, 600031L),
                new SimpleEvent(new Object[] { 8.666666666666666 }, 600031L, 600032L), new SimpleEvent(new Object[] { 2.0 }, 600032L, 600033L),
                new SimpleEvent(new Object[] { 3.0 }, 600033L, 600037L), new SimpleEvent(new Object[] { 4.0 }, 600037L, 600041L),
                new SimpleEvent(new Object[] { 5.0 }, 600041L, 600045L), new SimpleEvent(new Object[] { 6.0 }, 600045L, 600049L),
                new SimpleEvent(new Object[] { 7.0 }, 600049L, 600053L), new SimpleEvent(new Object[] { 8.0 }, 600053L, 600057L),
                new SimpleEvent(new Object[] { 9.0 }, 600057L, 600061L), new SimpleEvent(new Object[] { 10.0 }, 600061L, 600065L),
                new SimpleEvent(new Object[] { 11.0 }, 600065L, 600069L), new SimpleEvent(new Object[] { 12.0 }, 600069L, 600073L),
                new SimpleEvent(new Object[] { 13.0 }, 600073L, 600077L), new SimpleEvent(new Object[] { 14.0 }, 600077L, 600081L),
                new SimpleEvent(new Object[] { 15.0 }, 600081L, 600085L), new SimpleEvent(new Object[] { 16.0 }, 600085L, 600089L));

        sh.pushEvent(new Object[] { "Event", 1 }, 600001L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Event", 4 }, 600002L);
        Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { "Event", 5 }, 600003L);
        Assert.assertEquals(2, results.size());
        sh.pushEvent(new Object[] { "Event", 7 }, 600004L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "Event", 9 }, 600004L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { "Processing", 11 }, 600005L);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { "Processing", 13 }, 600005L);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { "Event", 15 }, 600006L);
        Assert.assertEquals(5, results.size());
        sh.pushEvent(new Object[] { "Event", 42 }, 600011L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "Bla", 42 }, 600012L);
        Assert.assertEquals(7, results.size());
        sh.pushEvent(new Object[] { "Bla", 0 }, 600013L);
        Assert.assertEquals(8, results.size());
        sh.pushEvent(new Object[] { "Bla", 1 }, 600017L);
        Assert.assertEquals(9, results.size());
        sh.pushEvent(new Object[] { "Bla", 2 }, 600021L);
        Assert.assertEquals(10, results.size());
        sh.pushEvent(new Object[] { "Bla", 3 }, 600025L);
        Assert.assertEquals(14, results.size());
        sh.pushEvent(new Object[] { "Bla", 4 }, 600029L);
        Assert.assertEquals(16, results.size());
        sh.pushEvent(new Object[] { "Bla", 5 }, 600033L);
        Assert.assertEquals(19, results.size());
        sh.pushEvent(new Object[] { "Bla", 6 }, 600037L);
        Assert.assertEquals(20, results.size());
        sh.pushEvent(new Object[] { "Bla", 7 }, 600041L);
        Assert.assertEquals(21, results.size());
        sh.pushEvent(new Object[] { "Bla", 8 }, 600045L);
        Assert.assertEquals(22, results.size());
        sh.pushEvent(new Object[] { "Bla", 9 }, 600049L);
        Assert.assertEquals(23, results.size());
        sh.pushEvent(new Object[] { "Bla", 10 }, 600053L);
        Assert.assertEquals(24, results.size());
        sh.pushEvent(new Object[] { "Bla", 11 }, 600057L);
        Assert.assertEquals(25, results.size());
        sh.pushEvent(new Object[] { "Bla", 12 }, 600061L);
        Assert.assertEquals(26, results.size());
        sh.pushEvent(new Object[] { "Bla", 13 }, 600065L);
        Assert.assertEquals(27, results.size());
        sh.pushEvent(new Object[] { "Bla", 14 }, 600069L);
        Assert.assertEquals(28, results.size());
        sh.pushEvent(new Object[] { "Bla", 15 }, 600073L);
        Assert.assertEquals(29, results.size());
        sh.pushEvent(new Object[] { "Bla", 16 }, 600077L);
        Assert.assertEquals(30, results.size());
        sh.pushEvent(new Object[] { "Bla", 17 }, 600081L);
        Assert.assertEquals(31, results.size());
        sh.pushEvent(new Object[] { "Bla", 18 }, 600085L);
        Assert.assertEquals(32, results.size());
        sh.pushEvent(new Object[] { "Bla", 19 }, 600089L);
        Assert.assertEquals(33, results.size());

        compareResults(expected, results);
    }

    /**
     * Tests aggregate (avg of byte)
     * with a countwindow (size: 3)
     * each millisecond one event
     *
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void aggregator10() throws EPException { // count window without grouping
        EPA epaA1 = new Aggregator(new CountWindow(STREAM, 3), new Average("y", "myavg"));

        CollectingSink sink = new CollectingSink();
        List<Event> results = sink.getEvents();

        StreamHandle sh = provider.registerStream(STREAM.getName(), new EventSchema(new Attribute("y", Attribute.DataType.BYTE)));

        QueryHandle qh = provider.createQuery(QUERY, epaA1);
        provider.registerSink(qh, sink);

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { 1.0 }, 801L, 802L), new SimpleEvent(new Object[] { 1.5 }, 802L, 803L),
                new SimpleEvent(new Object[] { 2.0 }, 803L, 804L), new SimpleEvent(new Object[] { 3.0 }, 804L, 805L),
                new SimpleEvent(new Object[] { 4.0 }, 805L, 806L), new SimpleEvent(new Object[] { 5.0 }, 806L, 807L),
                new SimpleEvent(new Object[] { 6.0 }, 807L, 808L), new SimpleEvent(new Object[] { 7.0 }, 808L, 809L));

        sh.pushEvent(new Object[] { (byte) 1 }, 801L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { (byte) 2 }, 802L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { (byte) 3 }, 803L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { (byte) 4 }, 804L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { (byte) 5 }, 805L);
        Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { (byte) 6 }, 806L);
        Assert.assertEquals(2, results.size());
        sh.pushEvent(new Object[] { (byte) 7 }, 807L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { (byte) 8 }, 808L);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { (byte) 9 }, 809L);
        Assert.assertEquals(5, results.size());
        sh.pushEvent(new Object[] { (byte) 10 }, 810L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { (byte) 11 }, 811L);
        Assert.assertEquals(7, results.size());
        sh.pushEvent(new Object[] { (byte) 12 }, 812L);
        Assert.assertEquals(8, results.size());

        compareResults(expected, results);
    }

    /**
     * Tests aggregates (avg, max and of longs) on two groups (type of byte)
     * with a countwindow (size: 2)
     *
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void aggregator11() throws EPException { // count window with grouping
        EPA epaA1 = new Aggregator(new CountWindow(STREAM, 2),
                new Group("x"), new Average("y", "myavg"), new Maximum("y", "mymax"), new Minimum("y", "mymin"));

        CollectingSink sink = new CollectingSink();
        List<Event> results = sink.getEvents();

        StreamHandle sh = provider.registerStream(STREAM.getName(),
                new EventSchema(new Attribute("x", Attribute.DataType.BYTE), new Attribute("y", Attribute.DataType.LONG)));

        QueryHandle qh = provider.createQuery(QUERY, epaA1);
        provider.registerSink(qh, sink);

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { (byte) 1, 40.0, 40L, 40L }, 800001L, 800003L),
                new SimpleEvent(new Object[] { (byte) 2, 80.0, 80L, 80L }, 800001L, 800005L),
                new SimpleEvent(new Object[] { (byte) 1, 41.0, 41L, 41L }, 800003L, 800005L),
                new SimpleEvent(new Object[] { (byte) 1, 41.5, 42L, 41L }, 800005L, 800007L),
                new SimpleEvent(new Object[] { (byte) 1, 42.0, 42L, 42L }, 800007L, 800012L),
                new SimpleEvent(new Object[] { (byte) 2, 80.0, 80L, 80L }, 800007L, 800012L));

        sh.pushEvent(new Object[] { (byte) 1, 40L }, 800001L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { (byte) 2, 80L }, 800001L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { (byte) 1, 41L }, 800003L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { (byte) 1, 42L }, 800005L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { (byte) 2, 80L }, 800007L);
        Assert.assertEquals(1, results.size());

        sh.pushEvent(new Object[] { (byte) 2, 80L }, 800012L);
        Assert.assertEquals(3, results.size());

        sh.pushEvent(new Object[] { (byte) 2, 85L }, 800015L);
        Assert.assertEquals(4, results.size());

        sh.pushEvent(new Object[] { (byte) 2, 85L }, 800020L);
        Assert.assertEquals(6, results.size());

        compareResults(expected, results);

    }

    /**
     * Tests aggregates (avg, max, min, count) on two groups (type of String)
     * with a countwindow (size: 3)
     *
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void aggregator12() throws EPException { // count window with grouping and simultanous  events

        EPA epaA1 = new Aggregator(new CountWindow(STREAM, 3), new Group("x"), new Average("z", "agg1"), new Maximum("z", "agg2"),
                new Minimum("z", "agg3"), new Count("x", "agg4"));

        CollectingSink sink = new CollectingSink();
        List<Event> results = sink.getEvents();

        StreamHandle sh = provider.registerStream(STREAM.getName(), new EventSchema(new Attribute("x", Attribute.DataType.STRING),
                new Attribute("y", Attribute.DataType.INTEGER), new Attribute("z", Attribute.DataType.INTEGER)));

        QueryHandle qh = provider.createQuery(QUERY, epaA1);
        provider.registerSink(qh, sink);

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { "Processing", 2.0, 2, 2, 1L }, 800000L, 800001L),
                new SimpleEvent(new Object[] { "Event", 2.5, 3, 2, 2L }, 800000L, 800001L),
                new SimpleEvent(new Object[] { "Processing", 6.0, 6, 6, 1L }, 800001L, 800002L),
                new SimpleEvent(new Object[] { "Event", 6.5, 7, 6, 2L }, 800001L, 800002L),
                new SimpleEvent(new Object[] { "Processing", 10.0, 10, 10, 1L }, 800002L, 800003L),
                new SimpleEvent(new Object[] { "Event", 10.5, 11, 10, 2L }, 800002L, 800003L),
                new SimpleEvent(new Object[] { "Processing", 14.0, 14, 14, 1L }, 800003L, 800004L),
                new SimpleEvent(new Object[] { "Event", 14.5, 15, 14, 2L }, 800003L, 800004L),
                new SimpleEvent(new Object[] { "Processing", 18.0, 18, 18, 1L }, 800004L, 800005L),
                new SimpleEvent(new Object[] { "Event", 18.5, 19, 18, 2L }, 800004L, 800005L));

        sh.pushEvent(new Object[] { "Event", 0, 0 }, 800000L + 0);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Processing", 0, 0 }, 800000L + 0);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Event", 1, 1 }, 800000L + 0);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Event", 2, 2 }, 800000L + 0);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Processing", 2, 2 }, 800000L + 0);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Event", 3, 3 }, 800000L + 0);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Event", 4, 4 }, 800000L + 1);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Processing", 4, 4 }, 800000L + 1);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Event", 5, 5 }, 800000L + 1);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Event", 6, 6 }, 800000L + 1);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Processing", 6, 6 }, 800000L + 1);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Event", 7, 7 }, 800000L + 1);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Event", 8, 8 }, 800000L + 2);
        Assert.assertEquals(2, results.size());
        sh.pushEvent(new Object[] { "Processing", 8, 8 }, 800000L + 2);
        Assert.assertEquals(2, results.size());
        sh.pushEvent(new Object[] { "Event", 9, 9 }, 800000L + 2);
        Assert.assertEquals(2, results.size());
        sh.pushEvent(new Object[] { "Event", 10, 10 }, 800000L + 2);
        Assert.assertEquals(2, results.size());
        sh.pushEvent(new Object[] { "Processing", 10, 10 }, 800000L + 2);
        Assert.assertEquals(2, results.size());
        sh.pushEvent(new Object[] { "Event", 11, 11 }, 800000L + 2);
        Assert.assertEquals(2, results.size());
        sh.pushEvent(new Object[] { "Event", 12, 12 }, 800000L + 3);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { "Processing", 12, 12 }, 800000L + 3);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { "Event", 13, 13 }, 800000L + 3);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { "Event", 14, 14 }, 800000L + 3);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { "Processing", 14, 14 }, 800000L + 3);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { "Event", 15, 15 }, 800000L + 3);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { "Event", 16, 16 }, 800000L + 4);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "Processing", 16, 16 }, 800000L + 4);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "Event", 17, 17 }, 800000L + 4);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "Event", 18, 18 }, 800000L + 4);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "Processing", 18, 18 }, 800000L + 4);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "Event", 19, 19 }, 800000L + 4);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { "Event", 20, 20 }, 800000L + 5);
        Assert.assertEquals(8, results.size());
        sh.pushEvent(new Object[] { "Processing", 20, 20 }, 800000L + 5);
        Assert.assertEquals(8, results.size());
        sh.pushEvent(new Object[] { "Event", 21, 21 }, 800000L + 5);
        Assert.assertEquals(8, results.size());
        sh.pushEvent(new Object[] { "Event", 22, 22 }, 800000L + 5);
        Assert.assertEquals(8, results.size());
        sh.pushEvent(new Object[] { "Processing", 22, 22 }, 800000L + 5);
        Assert.assertEquals(8, results.size());
        sh.pushEvent(new Object[] { "Event", 23, 23 }, 800000L + 5);
        Assert.assertEquals(8, results.size());
        sh.pushEvent(new Object[] { "Event", 24, 24 }, 800000L + 6);
        Assert.assertEquals(10, results.size());
        sh.pushEvent(new Object[] { "Processing", 24, 24 }, 800000L + 6);
        Assert.assertEquals(10, results.size());
        sh.pushEvent(new Object[] { "Event", 25, 25 }, 800000L + 6);
        Assert.assertEquals(10, results.size());
        sh.pushEvent(new Object[] { "Event", 26, 26 }, 800000L + 6);
        Assert.assertEquals(10, results.size());
        sh.pushEvent(new Object[] { "Processing", 26, 26 }, 800000L + 6);
        Assert.assertEquals(10, results.size());
        sh.pushEvent(new Object[] { "Event", 27, 27 }, 800000L + 6);
        Assert.assertEquals(10, results.size());

        compareResults(expected, results);
    }

    /**
     * Tests an aggregate (avg of integers)
     * with a timewindow (size: 5) and 'holes' between event timestamps,
     * so that some timewindows empty
     *
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void aggregator13() throws EPException { // aggregator
                                                                                                             // on
                                                                                                         // stream
                                                                                                         // with
                                                                                                         // 'holes'
        EPA myAggregator = new Aggregator(new TimeWindow(STREAM, 5), new Average("a", "avgA"));

        CollectingSink sink = new CollectingSink();
        List<Event> results = sink.getEvents();

        StreamHandle sh = provider.registerStream(STREAM.getName(), new EventSchema(new Attribute("a", Attribute.DataType.INTEGER),
                new Attribute("b", Attribute.DataType.INTEGER), new Attribute("c", Attribute.DataType.INTEGER)));

        QueryHandle qh = provider.createQuery(QUERY, myAggregator);

        provider.registerSink(qh, sink);

        List<Event> expected = Arrays.asList(new SimpleEvent(new Object[] { 42.0 }, 100, 103), new SimpleEvent(new Object[] { 42.5 }, 103, 105),
                new SimpleEvent(new Object[] { 43.0 }, 105, 108), new SimpleEvent(new Object[] { 44.0 }, 113, 118),
                new SimpleEvent(new Object[] { 45.0 }, 123, 128), new SimpleEvent(new Object[] { 46.0 }, 128, 130),
                new SimpleEvent(new Object[] { 46.5 }, 130, 133), new SimpleEvent(new Object[] { 47.0 }, 133, 135));

        sh.pushEvent(new Object[] { 42, 998, 997 }, 100L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { 43, 998, 997 }, 103L);
        Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { 44, 998, 997 }, 113L);
        Assert.assertEquals(3, results.size());
        sh.pushEvent(new Object[] { 45, 998, 997 }, 123L);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { 46, 998, 997 }, 128L);
        Assert.assertEquals(5, results.size());
        sh.pushEvent(new Object[] { 47, 998, 997 }, 130L);
        Assert.assertEquals(6, results.size());
        sh.pushEvent(new Object[] { 48, 998, 997 }, 139L);
        Assert.assertEquals(8, results.size());

        compareResults(expected, results);
    }

}
