/*
 * Copyright (c) 2012, 2013, 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.jepc.v2.test.bridge;

import java.util.Arrays;
import java.util.List;

import org.junit.Assert;
import org.junit.Test;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.EPException;
import de.umr.ep.event.Event;
import de.umr.ep.event.EventSchema;
import de.umr.ep.event.impl.SimpleEvent;
import de.umr.jepc.v2.api.bridge.EPBridgeFactory;
import de.umr.jepc.v2.api.epa.Correlator;
import de.umr.jepc.v2.api.epa.EPA;
import de.umr.jepc.v2.api.epa.Stream;
import de.umr.jepc.v2.api.epa.window.CountWindow;
import de.umr.jepc.v2.api.epa.window.TimeWindow;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.DoubleConstant;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.IntConstant;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.LongConstant;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.Variable;
import de.umr.jepc.v2.api.expression.arithmetic.compund.numeric.Add;
import de.umr.jepc.v2.api.expression.arithmetic.compund.numeric.Pow;
import de.umr.jepc.v2.api.expression.arithmetic.compund.numeric.Sqrt;
import de.umr.jepc.v2.api.expression.arithmetic.compund.numeric.Subtract;
import de.umr.jepc.v2.api.expression.predicate.Equal;
import de.umr.jepc.v2.api.expression.predicate.Greater;
import de.umr.jepc.v2.api.expression.predicate.LessEq;
import de.umr.jepc.v2.api.provider.AlreadyRegisteredException;
import de.umr.jepc.v2.api.provider.NoSuchStreamException;
import de.umr.jepc.v2.api.provider.QueryHandle;
import de.umr.jepc.v2.api.provider.StreamHandle;
import de.umr.jepc.v2.api.util.CollectingSink;

/**
 * @author Michael Körber
 *
 */
public class CorrelatorTest extends AbstractSemanticTest {

//    private static final Logger log          = LoggerFactory.getLogger(CorrelatorTest.class);

    private static final Stream STREAM_LEFT  = new Stream("left");
    private static final Stream STREAM_RIGHT = new Stream("right");
    private static final String QUERY        = "join";

    public CorrelatorTest(EPBridgeFactory bf) {
        super(bf);
    }

    /**
     * Simple correlation of two event streams each with a timewindow (size: 7)
     * The correlation condition tests, weather the first payload of events is equal
     * @throws AlreadyRegisteredException
     * @throws EPException
     */
    @Test
    public void simpleTest() throws EPException {
        EventSchema sl = new EventSchema(new Attribute("a", Attribute.DataType.STRING), new Attribute("b", Attribute.DataType.INTEGER));

        EventSchema sr = new EventSchema(new Attribute("c", Attribute.DataType.STRING), new Attribute("d", Attribute.DataType.INTEGER));

        CollectingSink sink = new CollectingSink();

        Correlator def = new Correlator(new TimeWindow(STREAM_LEFT, 7L), new TimeWindow(STREAM_RIGHT, 7L),
                new Equal(new Variable("a"), new Variable("c")));

        StreamHandle hl = provider.registerStream(STREAM_LEFT.getName(), sl);
        StreamHandle hr = provider.registerStream(STREAM_RIGHT.getName(), sr);
        QueryHandle qh = provider.createQuery(QUERY, def);

        provider.registerSink(qh, sink);
        List<Event> results = sink.getEvents();
        
        List<Event> expected = Arrays.asList( 
            new SimpleEvent(new Object[] { "a", 1, "a", 3 }, 5, 11), 
            new SimpleEvent(new Object[] { "d", 2, "d", 2 }, 6, 10),
            new SimpleEvent(new Object[] { "a", 2, "a", 3 }, 9, 11), 
            new SimpleEvent(new Object[] { "b", 3, "b", 4 }, 12, 14)
        );

        hl.pushEvent(new Object[] { "c", 1 }, 1);
        Assert.assertTrue(results.isEmpty());
        hr.pushEvent(new Object[] { "b", 1 }, 1L);
        Assert.assertTrue(results.isEmpty());
        hr.pushEvent(new Object[] { "d", 2 }, 3L);
        Assert.assertTrue(results.isEmpty());
        hr.pushEvent(new Object[] { "a", 3 }, 4L);
        Assert.assertTrue(results.isEmpty());
        hl.pushEvent(new Object[] { "a", 1 }, 5L);
        Assert.assertTrue(results.isEmpty());
        hl.pushEvent(new Object[] { "d", 2 }, 6L);
        Assert.assertTrue(results.isEmpty());
        hr.pushEvent(new Object[] { "b", 4 }, 7L);
        Assert.assertEquals(2, results.size());
        hl.pushEvent(new Object[] { "a", 2 }, 9L);
        Assert.assertEquals(2, results.size());
        hr.pushEvent(new Object[] { "e", 5 }, 10L);
        Assert.assertEquals(3, results.size());
        hl.pushEvent(new Object[] { "b", 3 }, 12L);
        Assert.assertEquals(3, results.size());

        // Flush
        hl.pushEvent(new Object[] { "x", 6 }, 13L);
        Assert.assertEquals(3, results.size());
        hr.pushEvent(new Object[] { "y", 6 }, 13L);
        Assert.assertEquals(4, results.size());

        compareResults(expected, sink.getEvents());
    }

    /**
     * Tests a correlation of an event stream with itself, but different timewindows; first one has a timewindow with size of 10 and the second of 5.
     * The correlation condition tests, weather the second payload of an event is greater than the other.
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void selfCorrelation1() throws EPException { // testing self-joins
        EPA selfJoin = new Correlator(
                new TimeWindow(STREAM_LEFT, 10), "s1",
                new TimeWindow(STREAM_LEFT, 5), "s2",
                new Greater(new Variable("s1.y"), new Variable("s2.y")));

        CollectingSink sink = new CollectingSink();
        List<Event> results = sink.getEvents();

        StreamHandle sh = provider.registerStream(STREAM_LEFT.getName(),
                new EventSchema(new Attribute("x", Attribute.DataType.STRING), new Attribute("y", Attribute.DataType.FLOAT)));

        QueryHandle qh = provider.createQuery(QUERY, selfJoin);
        provider.registerSink(qh, sink);

        List<Event> expected = Arrays.asList(
                new SimpleEvent(new Object[] { "Stock_A", 60.0f, "Stock_B", 50.0f }, 300L, 305L),
                new SimpleEvent(new Object[] { "Stock_A", 60.0f, "Stock_C", 50.0f }, 301L, 306L),
                new SimpleEvent(new Object[] { "Stock_D", 60.0f, "Stock_B", 50.0f }, 301L, 305L),
                new SimpleEvent(new Object[] { "Stock_D", 60.0f, "Stock_C", 50.0f }, 301L, 306L),
                new SimpleEvent(new Object[] { "Stock_E", 70.0f, "Stock_A", 60.0f }, 302L, 305L),
                new SimpleEvent(new Object[] { "Stock_E", 70.0f, "Stock_B", 50.0f }, 302L, 305L),
                new SimpleEvent(new Object[] { "Stock_E", 70.0f, "Stock_D", 60.0f }, 302L, 306L),
                new SimpleEvent(new Object[] { "Stock_E", 70.0f, "Stock_C", 50.0f }, 302L, 306L),
                new SimpleEvent(new Object[] { "Stock_A", 60.0f, "Stock_F", 50.0f }, 302L, 307L),
                new SimpleEvent(new Object[] { "Stock_D", 60.0f, "Stock_F", 50.0f }, 302L, 307L),
                new SimpleEvent(new Object[] { "Stock_E", 70.0f, "Stock_F", 50.0f }, 302L, 307L),
                new SimpleEvent(new Object[] { "Stock_G", 80.0f, "Stock_A", 60.0f }, 303L, 305L),
                new SimpleEvent(new Object[] { "Stock_G", 80.0f, "Stock_B", 50.0f }, 303L, 305L),
                new SimpleEvent(new Object[] { "Stock_G", 80.0f, "Stock_D", 60.0f }, 303L, 306L),
                new SimpleEvent(new Object[] { "Stock_G", 80.0f, "Stock_F", 50.0f }, 303L, 307L),
                new SimpleEvent(new Object[] { "Stock_G", 80.0f, "Stock_E", 70.0f }, 303L, 307L),
                new SimpleEvent(new Object[] { "Stock_G", 80.0f, "Stock_C", 50.0f }, 303L, 306L),
                new SimpleEvent(new Object[] { "Stock_H", 60.0f, "Stock_B", 50.0f }, 303L, 305L),
                new SimpleEvent(new Object[] { "Stock_H", 60.0f, "Stock_F", 50.0f }, 303L, 307L),
                new SimpleEvent(new Object[] { "Stock_H", 60.0f, "Stock_C", 50.0f }, 303L, 306L),
                new SimpleEvent(new Object[] { "Stock_G", 80.0f, "Stock_H", 60.0f }, 303L, 308L),
                new SimpleEvent(new Object[] { "Stock_E", 70.0f, "Stock_H", 60.0f }, 303L, 308L),
                new SimpleEvent(new Object[] { "Stock_H", 60.0f, "Stock_I", 50.0f }, 304L, 309L),
                new SimpleEvent(new Object[] { "Stock_G", 80.0f, "Stock_I", 50.0f }, 304L, 309L),
                new SimpleEvent(new Object[] { "Stock_E", 70.0f, "Stock_I", 50.0f }, 304L, 309L),
                new SimpleEvent(new Object[] { "Stock_D", 60.0f, "Stock_I", 50.0f }, 304L, 309L),
                new SimpleEvent(new Object[] { "Stock_A", 60.0f, "Stock_I", 50.0f }, 304L, 309L),
                new SimpleEvent(new Object[] { "Stock_J", 70.0f, "Stock_A", 60.0f }, 304L, 305L),
                new SimpleEvent(new Object[] { "Stock_J", 70.0f, "Stock_B", 50.0f }, 304L, 305L),
                new SimpleEvent(new Object[] { "Stock_J", 70.0f, "Stock_D", 60.0f }, 304L, 306L),
                new SimpleEvent(new Object[] { "Stock_J", 70.0f, "Stock_H", 60.0f }, 304L, 308L),
                new SimpleEvent(new Object[] { "Stock_J", 70.0f, "Stock_I", 50.0f }, 304L, 309L),
                new SimpleEvent(new Object[] { "Stock_J", 70.0f, "Stock_F", 50.0f }, 304L, 307L),
                new SimpleEvent(new Object[] { "Stock_J", 70.0f, "Stock_C", 50.0f }, 304L, 306L),
                new SimpleEvent(new Object[] { "Stock_G", 80.0f, "Stock_J", 70.0f }, 304L, 309L),
                new SimpleEvent(new Object[] { "Stock_K", 60.0f, "Stock_C", 50.0f }, 305L, 306L),
                new SimpleEvent(new Object[] { "Stock_K", 60.0f, "Stock_F", 50.0f }, 305L, 307L),
                new SimpleEvent(new Object[] { "Stock_J", 70.0f, "Stock_K", 60.0f }, 305L, 310L),
                new SimpleEvent(new Object[] { "Stock_G", 80.0f, "Stock_K", 60.0f }, 305L, 310L),
                new SimpleEvent(new Object[] { "Stock_E", 70.0f, "Stock_K", 60.0f }, 305L, 310L),
                new SimpleEvent(new Object[] { "Stock_K", 60.0f, "Stock_I", 50.0f }, 305L, 309L),
                new SimpleEvent(new Object[] { "Stock_L", 99.0f, "Stock_C", 50.0f }, 305L, 306L),
                new SimpleEvent(new Object[] { "Stock_L", 99.0f, "Stock_D", 60.0f }, 305L, 306L),
                new SimpleEvent(new Object[] { "Stock_L", 99.0f, "Stock_F", 50.0f }, 305L, 307L),
                new SimpleEvent(new Object[] { "Stock_L", 99.0f, "Stock_J", 70.0f }, 305L, 309L),
                new SimpleEvent(new Object[] { "Stock_L", 99.0f, "Stock_K", 60.0f }, 305L, 310L),
                new SimpleEvent(new Object[] { "Stock_L", 99.0f, "Stock_I", 50.0f }, 305L, 309L),
                new SimpleEvent(new Object[] { "Stock_L", 99.0f, "Stock_H", 60.0f }, 305L, 308L),
                new SimpleEvent(new Object[] { "Stock_L", 99.0f, "Stock_G", 80.0f }, 305L, 308L),
                new SimpleEvent(new Object[] { "Stock_L", 99.0f, "Stock_E", 70.0f }, 305L, 307L),
                new SimpleEvent(new Object[] { "Stock_M", 55.0f, "Stock_F", 50.0f }, 306L, 307L),
                new SimpleEvent(new Object[] { "Stock_H", 60.0f, "Stock_M", 55.0f }, 306L, 311L),
                new SimpleEvent(new Object[] { "Stock_L", 99.0f, "Stock_M", 55.0f }, 306L, 311L),
                new SimpleEvent(new Object[] { "Stock_K", 60.0f, "Stock_M", 55.0f }, 306L, 311L),
                new SimpleEvent(new Object[] { "Stock_J", 70.0f, "Stock_M", 55.0f }, 306L, 311L),
                new SimpleEvent(new Object[] { "Stock_G", 80.0f, "Stock_M", 55.0f }, 306L, 311L),
                new SimpleEvent(new Object[] { "Stock_E", 70.0f, "Stock_M", 55.0f }, 306L, 311L),
                new SimpleEvent(new Object[] { "Stock_D", 60.0f, "Stock_M", 55.0f }, 306L, 311L),
                new SimpleEvent(new Object[] { "Stock_A", 60.0f, "Stock_M", 55.0f }, 306L, 310L),
                new SimpleEvent(new Object[] { "Stock_M", 55.0f, "Stock_I", 50.0f }, 306L, 309L),
                new SimpleEvent(new Object[] { "Stock_N", 150.0f, "Stock_E", 70.0f }, 306L, 307L),
                new SimpleEvent(new Object[] { "Stock_N", 150.0f, "Stock_F", 50.0f }, 306L, 307L),
                new SimpleEvent(new Object[] { "Stock_N", 150.0f, "Stock_H", 60.0f }, 306L, 308L),
                new SimpleEvent(new Object[] { "Stock_N", 150.0f, "Stock_L", 99.0f }, 306L, 310L),
                new SimpleEvent(new Object[] { "Stock_N", 150.0f, "Stock_M", 55.0f }, 306L, 311L),
                new SimpleEvent(new Object[] { "Stock_N", 150.0f, "Stock_K", 60.0f }, 306L, 310L),
                new SimpleEvent(new Object[] { "Stock_N", 150.0f, "Stock_J", 70.0f }, 306L, 309L),
                new SimpleEvent(new Object[] { "Stock_N", 150.0f, "Stock_I", 50.0f }, 306L, 309L),
                new SimpleEvent(new Object[] { "Stock_N", 150.0f, "Stock_G", 80.0f }, 306L, 308L));

        sh.pushEvent(new Object[] { "Stock_A", 60.0f }, 300L);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { "Stock_B", 50.0f }, 300L);
        Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { "Stock_C", 50.0f }, 301L);
        Assert.assertEquals(2, results.size());
        sh.pushEvent(new Object[] { "Stock_D", 60.0f }, 301L);
        Assert.assertEquals(4, results.size());
        sh.pushEvent(new Object[] { "Stock_E", 70.0f }, 302L);
        Assert.assertEquals(8, results.size());
        sh.pushEvent(new Object[] { "Stock_F", 50.0f }, 302L);
        Assert.assertEquals(11, results.size());
        sh.pushEvent(new Object[] { "Stock_G", 80.0f }, 303L);
        Assert.assertEquals(17, results.size());
        sh.pushEvent(new Object[] { "Stock_H", 60.0f }, 303L);
        Assert.assertEquals(22, results.size());
        sh.pushEvent(new Object[] { "Stock_I", 50.0f }, 304L);
        Assert.assertEquals(27, results.size());
        sh.pushEvent(new Object[] { "Stock_J", 70.0f }, 304L);
        Assert.assertEquals(35, results.size());
        sh.pushEvent(new Object[] { "Stock_K", 60.0f }, 305L);
        Assert.assertEquals(41, results.size());
        sh.pushEvent(new Object[] { "Stock_L", 99.0f }, 305L);
        Assert.assertEquals(50, results.size());
        sh.pushEvent(new Object[] { "Stock_M", 55.0f }, 306L);
        Assert.assertEquals(60, results.size());
        sh.pushEvent(new Object[] { "Stock_N", 150.0f }, 306L);
        Assert.assertEquals(69, results.size());

        compareResults(expected, results);
    }

    /**
     * Tests a correlation of two event streams; each with a timewindow of 7.
     * The correlation condition tests, weather the geometrical distance is equal or smaller 2.2
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void distanceCorrelation1() throws EPException {
        EventSchema s1 = new EventSchema(new Attribute("s1_id", Attribute.DataType.STRING), new Attribute("s1_x",
                Attribute.DataType.FLOAT), new Attribute("s1_y", Attribute.DataType.FLOAT));
        EventSchema s2 = new EventSchema(new Attribute("s2_id", Attribute.DataType.STRING), new Attribute("s2_x",
                Attribute.DataType.FLOAT), new Attribute("s2_y", Attribute.DataType.FLOAT));

        StreamHandle hl = provider.registerStream(STREAM_LEFT.getName(), s1);
        StreamHandle hr = provider.registerStream(STREAM_RIGHT.getName(), s2);

        Correlator corr = new Correlator(
                new TimeWindow(STREAM_LEFT, 7), STREAM_LEFT.getName(),
                new TimeWindow(STREAM_RIGHT, 7), STREAM_RIGHT.getName(),
                new LessEq(
                        new Sqrt(
                                new Add(
                                        new Pow(    new Subtract(new Variable(STREAM_LEFT.getName()+".s1_x"), new Variable(STREAM_RIGHT.getName()+".s2_x")),
                                                    new IntConstant(2)),
                                        new Pow(    new Subtract(new Variable(STREAM_LEFT.getName()+".s1_y"), new Variable(STREAM_RIGHT.getName()+".s2_y")),
                                                    new LongConstant(2L)))),
                        new DoubleConstant(2.2)
                )
        );
        QueryHandle qh = provider.createQuery(QUERY, corr);
        CollectingSink sink = new CollectingSink();
        provider.registerSink(qh, sink);

        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(
                new SimpleEvent(new Object[] { "s1_1", 60.0f, 60.0f, "s2_1", 60.0f, 61.0f}, 300L, 307L),
                new SimpleEvent(new Object[] { "s1_2", 60.5f, 69.0f, "s2_2", 60.0f, 70.0f}, 301L, 307L),
                new SimpleEvent(new Object[] { "s1_4", 58.5f, 60.0f, "s2_4", 60.0f, 61.0f}, 318L, 324L),
                new SimpleEvent(new Object[] { "s1_5", -1.0f, 0.0f, "s2_5", 1.0f, 0.2f}, 326L, 332L),
                new SimpleEvent(new Object[] { "s1_6", 0.0f, 0.0f, "s2_6", 1.0f, 1.0f}, 334L, 341L),
                new SimpleEvent(new Object[] { "s1_8", 1.0f, 3.19f, "s2_6", 1.0f, 1.0f}, 338L, 341L)
        );

        hl.pushEvent(new Object[] { "s1_1", 60.0f, 60.0f }, 300L);
        hr.pushEvent(new Object[] { "s2_1", 60.0f, 61.0f }, 300L);

        //first no match, but with "s1_2"
        hr.pushEvent(new Object[] { "s2_2", 60.0f, 70.0f }, 300L);
        hl.pushEvent(new Object[] { "s1_2", 60.5f, 69.0f }, 301L);

        //no match
        hl.pushEvent(new Object[] { "s1_3", 57.5f, 60.0f }, 309L);
        hr.pushEvent(new Object[] { "s2_3", 60.0f, 61.0f }, 310L);

        hl.pushEvent(new Object[] { "s1_4", 58.5f, 60.0f }, 317L);
        hr.pushEvent(new Object[] { "s2_4", 60.0f, 61.0f }, 318L);

        hl.pushEvent(new Object[] { "s1_5", -1.0f, 0.0f }, 325L);
        hr.pushEvent(new Object[] { "s2_5", 1.0f, 0.2f }, 326L);

        hr.pushEvent(new Object[] { "s2_6", 1.0f, 1.0f }, 334L);
        hl.pushEvent(new Object[] { "s1_6", 0.0f, 0.0f }, 334L);

        //second is a match with "s2_6"
        hl.pushEvent(new Object[] { "s1_7", 1.0f, 3.3f }, 337L);
        hl.pushEvent(new Object[] { "s1_8", 1.0f, 3.19f }, 338L);

        hr.pushEvent(new Object[] { "flush", 110f, 110f }, 350L);

//        results.forEach(System.out::println);

        compareResults(expected, results);
    }

    /**
     * Tests a correlation of two event streams; each with a c of 7, which jumps 7 milliseconds.
     * The correlation condition tests, weather the geometrical distance is equal or smaller 2.2
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void distanceCorrelation2() throws EPException {
        EventSchema s1 = new EventSchema(new Attribute("id_1", Attribute.DataType.STRING), new Attribute("x",
                Attribute.DataType.FLOAT), new Attribute("y", Attribute.DataType.FLOAT));
        EventSchema s2 = new EventSchema(new Attribute("id_2", Attribute.DataType.STRING), new Attribute("x",
                Attribute.DataType.FLOAT), new Attribute("y", Attribute.DataType.FLOAT));

        StreamHandle hl = provider.registerStream(STREAM_LEFT.getName(), s1);
        StreamHandle hr = provider.registerStream(STREAM_RIGHT.getName(), s2);

        Correlator corr = new Correlator(
                new TimeWindow(STREAM_LEFT, 7,7), STREAM_LEFT.getName(),
                new TimeWindow(STREAM_RIGHT, 7,7), STREAM_RIGHT.getName(),
                new LessEq(
                        new Sqrt(
                                new Add(
                                        new Pow(    new Subtract(new Variable(STREAM_LEFT.getName()+".x"), new Variable(STREAM_RIGHT.getName()+".x")),
                                                new IntConstant(2)),
                                        new Pow(    new Subtract(new Variable(STREAM_LEFT.getName()+".y"), new Variable(STREAM_RIGHT.getName()+".y")),
                                                new LongConstant(2L)))),
                        new DoubleConstant(2.2)
                )
        );
        QueryHandle qh = provider.createQuery(QUERY, corr);
        CollectingSink sink = new CollectingSink();
        provider.registerSink(qh, sink);

        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(
                new SimpleEvent(new Object[] { "s1_1", 60.0f, 60.0f, "s2_1", 60.0f, 61.0f}, 294L, 301L),
                new SimpleEvent(new Object[] { "s1_4", 58.5f, 60.0f, "s2_4", 60.0f, 61.0f}, 315L, 322L),
                new SimpleEvent(new Object[] { "s1_5", -1.0f, 0.0f, "s2_5", 1.0f, 0.2f}, 322L, 329L),
                new SimpleEvent(new Object[] { "s1_6", 0.0f, 0.0f, "s2_6", 1.0f, 1.0f}, 329L, 336L)
        );

        hl.pushEvent(new Object[] { "s1_1", 60.0f, 60.0f }, 300L);
        hr.pushEvent(new Object[] { "s2_1", 60.0f, 61.0f }, 300L);

        hr.pushEvent(new Object[] { "s2_2", 60.0f, 70.0f }, 300L);
        hl.pushEvent(new Object[] { "s1_2", 60.5f, 69.0f }, 301L);

        hl.pushEvent(new Object[] { "s1_3", 57.5f, 60.0f }, 309L);
        hr.pushEvent(new Object[] { "s2_3", 60.0f, 61.0f }, 310L);

        hl.pushEvent(new Object[] { "s1_4", 58.5f, 60.0f }, 317L);
        hr.pushEvent(new Object[] { "s2_4", 60.0f, 61.0f }, 318L);

        hl.pushEvent(new Object[] { "s1_5", -1.0f, 0.0f }, 325L);
        hr.pushEvent(new Object[] { "s2_5", 1.0f, 0.2f }, 326L);

        hr.pushEvent(new Object[] { "s2_6", 1.0f, 1.0f }, 334L);
        hl.pushEvent(new Object[] { "s1_6", 0.0f, 0.0f }, 334L);

        hl.pushEvent(new Object[] { "s1_7", 1.0f, 3.3f }, 337L);
        hl.pushEvent(new Object[] { "s1_8", 1.0f, 3.19f }, 338L);

        hr.pushEvent(new Object[] { "flush", 110f, 110f }, 350L);

        compareResults(expected, results);
    }

    /**
     * Tests a correlation of two event streams; first with a countWindow (size 1) and second has a countWindow with size 3
     * The correlation condition tests, weather the geometrical distance is equal or smaller 2.2
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void distanceCorrelation3() throws EPException {
        EventSchema s1 = new EventSchema(new Attribute("id_1", Attribute.DataType.STRING), new Attribute("x",
                Attribute.DataType.FLOAT), new Attribute("y", Attribute.DataType.FLOAT));
        EventSchema s2 = new EventSchema(new Attribute("id_2", Attribute.DataType.STRING), new Attribute("x",
                Attribute.DataType.FLOAT), new Attribute("y", Attribute.DataType.FLOAT));

        StreamHandle hl = provider.registerStream(STREAM_LEFT.getName(), s1);
        StreamHandle hr = provider.registerStream(STREAM_RIGHT.getName(), s2);

        CountWindow wl = new CountWindow(STREAM_LEFT, 1);
//        QueryHandle qh_wl = provider.createQuery(QUERY+"wl", wl);
//        CollectingSink sink_wl = new CollectingSink();
//        provider.registerSink(qh_wl, sink_wl);
        CountWindow wr = new CountWindow(STREAM_RIGHT, 3);
//        QueryHandle qh_wr = provider.createQuery(QUERY+"wr", wr);
//        CollectingSink sink_wr = new CollectingSink();
//        provider.registerSink(qh_wr, sink_wr);

        Correlator corr = new Correlator(
                wl, STREAM_LEFT.getName(),
                wr, STREAM_RIGHT.getName(),
                new LessEq(
                        new Sqrt(
                                new Add(
                                        new Pow(    new Subtract(new Variable(STREAM_LEFT.getName()+".x"), new Variable(STREAM_RIGHT.getName()+".x")),
                                                new IntConstant(2)),
                                        new Pow(    new Subtract(new Variable(STREAM_LEFT.getName()+".y"), new Variable(STREAM_RIGHT.getName()+".y")),
                                                new LongConstant(2L)))),
                        new DoubleConstant(2.2)
                )
        );
        QueryHandle qh = provider.createQuery(QUERY, corr);
        CollectingSink sink = new CollectingSink();
        provider.registerSink(qh, sink);

//        List<Event> results_wl = sink_wl.getEvents();
//        List<Event> results_wr = sink_wr.getEvents();

        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(
                new SimpleEvent(new Object[] { "s1_1", 60.0f, 60.0f, "s2_1", 60.0f, 61.0f }, 300L, 301L),
                new SimpleEvent(new Object[] { "s1_2", 60.5f, 69.0f, "s2_2", 60.0f, 70.0f }, 301L, 309L),
                new SimpleEvent(new Object[] { "s1_4", 58.5f, 60.0f, "s2_3", 60.0f, 61.0f }, 317L, 325L),
                new SimpleEvent(new Object[] { "s1_4", 58.5f, 60.0f, "s2_1", 60.0f, 61.0f }, 317L, 318L),
                new SimpleEvent(new Object[] { "s1_4", 58.5f, 60.0f, "s2_4", 60.0f, 61.0f }, 318L, 325L),
                new SimpleEvent(new Object[] { "s1_5", -1.0f, 0.0f, "s2_5", 1.0f, 0.2f }, 326L, 334L),
                new SimpleEvent(new Object[] { "s1_6", 0.0f, 0.0f, "s2_5", 1.0f, 0.2f }, 334L, 337L),
                new SimpleEvent(new Object[] { "s1_6", 0.0f, 0.0f, "s2_6", 1.0f, 1.0f }, 334L, 337L),
                new SimpleEvent(new Object[] { "s1_8", 1.0f, 3.19f, "s2_6", 1.0f, 1.0f }, 338L, 352L)
        );

        hl.pushEvent(new Object[] { "s1_1", 60.0f, 60.0f }, 300L);
        hr.pushEvent(new Object[] { "s2_1", 60.0f, 61.0f }, 300L);

        hr.pushEvent(new Object[] { "s2_2", 60.0f, 70.0f }, 300L);
        hl.pushEvent(new Object[] { "s1_2", 60.5f, 69.0f }, 301L);

        hl.pushEvent(new Object[] { "s1_3", 57.5f, 60.0f }, 309L);
        hr.pushEvent(new Object[] { "s2_3", 60.0f, 61.0f }, 310L);

        hl.pushEvent(new Object[] { "s1_4", 58.5f, 60.0f }, 317L);
        hr.pushEvent(new Object[] { "s2_4", 60.0f, 61.0f }, 318L);

        hl.pushEvent(new Object[] { "s1_5", -1.0f, 0.0f }, 325L);
        hr.pushEvent(new Object[] { "s2_5", 1.0f, 0.2f }, 326L);

        hr.pushEvent(new Object[] { "s2_6", 1.0f, 1.0f }, 334L);
        hl.pushEvent(new Object[] { "s1_6", 0.0f, 0.0f }, 334L);

        hl.pushEvent(new Object[] { "s1_7", 1.0f, 3.3f }, 337L);
        hl.pushEvent(new Object[] { "s1_8", 1.0f, 3.19f }, 338L);

        hr.pushEvent(new Object[] { "flush1", 110f, 110f }, 350L);
        hr.pushEvent(new Object[] { "flush1", 110f, 110f }, 351L);
        hr.pushEvent(new Object[] { "flush1", 110f, 110f }, 352L);
        hl.pushEvent(new Object[] { "flush2", 110f, 110f }, 352L);

        hr.pushEvent(new Object[] { "flush3", 110f, 110f }, 353L);

        results.forEach(System.out::println);

//        System.out.println();
//        System.out.println("---wl---");
//        results_wl.forEach(System.out::println);
//        System.out.println();
//        System.out.println("---wr---");
//        results_wr.forEach(System.out::println);

        compareResults(expected, results);
    }

    /**
     * Tests a correlation of two event streams; first with a countWindow (size 1) and second has a countWindow with size 5 and jump 3
     * The correlation condition tests, weather the geometrical distance is equal or smaller 2.2
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void distanceCorrelation4() throws EPException {
        EventSchema s1 = new EventSchema(new Attribute("id_1", Attribute.DataType.STRING), new Attribute("x",
                Attribute.DataType.FLOAT), new Attribute("y", Attribute.DataType.FLOAT));
        EventSchema s2 = new EventSchema(new Attribute("id_2", Attribute.DataType.STRING), new Attribute("x",
                Attribute.DataType.FLOAT), new Attribute("y", Attribute.DataType.FLOAT));

        StreamHandle hl = provider.registerStream(STREAM_LEFT.getName(), s1);
        StreamHandle hr = provider.registerStream(STREAM_RIGHT.getName(), s2);

        Correlator corr = new Correlator(
                new CountWindow(STREAM_LEFT, 1), STREAM_LEFT.getName(),
                new CountWindow(STREAM_RIGHT, 3, 3), STREAM_RIGHT.getName(),
                new LessEq(
                        new Sqrt(
                                new Add(
                                        new Pow(    new Subtract(new Variable(STREAM_LEFT.getName()+".x"), new Variable(STREAM_RIGHT.getName()+".x")),
                                                new IntConstant(2)),
                                        new Pow(    new Subtract(new Variable(STREAM_LEFT.getName()+".y"), new Variable(STREAM_RIGHT.getName()+".y")),
                                                new LongConstant(2L)))),
                        new DoubleConstant(2.2)
                )
        );
        QueryHandle qh = provider.createQuery(QUERY, corr);
        CollectingSink sink = new CollectingSink();
        provider.registerSink(qh, sink);

        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(
                new SimpleEvent(new Object[] { "s1_4", 58.5f, 60.0f, "s2_1", 60.0f, 61.0f }, 317L, 325L),
                new SimpleEvent(new Object[] { "s1_4", 58.5f, 60.0f, "s2_3", 60.0f, 61.0f }, 317L, 325L),
                new SimpleEvent(new Object[] { "s1_6", 0.0f, 0.0f, "s2_5", 1.0f, 0.2f }, 334L, 337L),
                new SimpleEvent(new Object[] { "s1_6", 0.0f, 0.0f, "s2_6", 1.0f, 1.0f }, 334L, 337L)
        );

        hl.pushEvent(new Object[] { "s1_1", 60.0f, 60.0f }, 300L);
        hr.pushEvent(new Object[] { "s2_1", 60.0f, 61.0f }, 300L);

        hr.pushEvent(new Object[] { "s2_2", 60.0f, 70.0f }, 300L);
        hl.pushEvent(new Object[] { "s1_2", 60.5f, 69.0f }, 301L);

        hl.pushEvent(new Object[] { "s1_3", 57.5f, 60.0f }, 309L);
        hr.pushEvent(new Object[] { "s2_3", 60.0f, 61.0f }, 310L);

        hl.pushEvent(new Object[] { "s1_4", 58.5f, 60.0f }, 317L);
        hr.pushEvent(new Object[] { "s2_4", 60.0f, 61.0f }, 318L);

        hl.pushEvent(new Object[] { "s1_5", -1.0f, 0.0f }, 325L);
        hr.pushEvent(new Object[] { "s2_5", 1.0f, 0.2f }, 326L);

        hr.pushEvent(new Object[] { "s2_6", 1.0f, 1.0f }, 334L);
        hl.pushEvent(new Object[] { "s1_6", 0.0f, 0.0f }, 334L);

        hl.pushEvent(new Object[] { "s1_7", 1.0f, 3.3f }, 337L);
        hl.pushEvent(new Object[] { "s1_8", 1.0f, 3.19f }, 338L);

        hr.pushEvent(new Object[] { "flush1", 110f, 110f }, 350L);
        hr.pushEvent(new Object[] { "flush1", 110f, 110f }, 351L);
        hr.pushEvent(new Object[] { "flush1", 110f, 110f }, 352L);
        hl.pushEvent(new Object[] { "flush2", 110f, 110f }, 352L);

        hr.pushEvent(new Object[] { "flush3", 110f, 110f }, 353L);


        compareResults(expected, results);
    }



    /**
     * Tests a correlation of two event streams; first with a timeWindow (size 3, jump 2) and second has a countWindow with size 3 and jump 3
     * The correlation condition tests, weather the geometrical distance is equal or smaller 2.2
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void distanceCorrelation5() throws EPException {
        EventSchema s1 = new EventSchema(new Attribute("id_1", Attribute.DataType.STRING), new Attribute("x",
                Attribute.DataType.FLOAT), new Attribute("y", Attribute.DataType.FLOAT));
        EventSchema s2 = new EventSchema(new Attribute("id_2", Attribute.DataType.STRING), new Attribute("x",
                Attribute.DataType.FLOAT), new Attribute("y", Attribute.DataType.FLOAT));

        StreamHandle hl = provider.registerStream(STREAM_LEFT.getName(), s1);
        StreamHandle hr = provider.registerStream(STREAM_RIGHT.getName(), s2);

        Correlator corr = new Correlator(
                new TimeWindow(STREAM_LEFT, 3, 2), STREAM_LEFT.getName(),
                new CountWindow(STREAM_RIGHT, 3, 3), STREAM_RIGHT.getName(),
                new LessEq(
                        new Sqrt(
                                new Add(
                                        new Pow(    new Subtract(new Variable(STREAM_LEFT.getName()+".x"), new Variable(STREAM_RIGHT.getName()+".x")),
                                                new IntConstant(2)),
                                        new Pow(    new Subtract(new Variable(STREAM_LEFT.getName()+".y"), new Variable(STREAM_RIGHT.getName()+".y")),
                                                new LongConstant(2L)))),
                        new DoubleConstant(2.2)
                )
        );
        QueryHandle qh = provider.createQuery(QUERY, corr);
        CollectingSink sink = new CollectingSink();
        provider.registerSink(qh, sink);

        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(
                new SimpleEvent(new Object[] { "s1_4", 58.5f, 60.0f, "s2_1", 60.0f, 61.0f }, 316L, 320L),
                new SimpleEvent(new Object[] { "s1_4", 58.5f, 60.0f, "s2_3", 60.0f, 61.0f }, 316L, 320L),
                new SimpleEvent(new Object[] { "s1_6", 0.0f, 0.0f, "s2_5", 1.0f, 0.2f }, 334L, 336L),
                new SimpleEvent(new Object[] { "s1_6", 0.0f, 0.0f, "s2_6", 1.0f, 1.0f }, 334L, 336L)
        );

        hl.pushEvent(new Object[] { "s1_1", 60.0f, 60.0f }, 300L);
        hr.pushEvent(new Object[] { "s2_1", 60.0f, 61.0f }, 300L);

        hr.pushEvent(new Object[] { "s2_2", 60.0f, 70.0f }, 300L);
        hl.pushEvent(new Object[] { "s1_2", 60.5f, 69.0f }, 301L);

        hl.pushEvent(new Object[] { "s1_3", 57.5f, 60.0f }, 309L);
        hr.pushEvent(new Object[] { "s2_3", 60.0f, 61.0f }, 310L);

        hl.pushEvent(new Object[] { "s1_4", 58.5f, 60.0f }, 317L);
        hr.pushEvent(new Object[] { "s2_4", 60.0f, 61.0f }, 318L);

        hl.pushEvent(new Object[] { "s1_5", -1.0f, 0.0f }, 325L);
        hr.pushEvent(new Object[] { "s2_5", 1.0f, 0.2f }, 326L);

        hr.pushEvent(new Object[] { "s2_6", 1.0f, 1.0f }, 334L);
        hl.pushEvent(new Object[] { "s1_6", 0.0f, 0.0f }, 334L);

        hl.pushEvent(new Object[] { "s1_7", 1.0f, 3.3f }, 337L);
        hl.pushEvent(new Object[] { "s1_8", 1.0f, 3.19f }, 338L);

        hr.pushEvent(new Object[] { "flush1", 110f, 110f }, 350L);
        hr.pushEvent(new Object[] { "flush1", 110f, 110f }, 351L);
        hr.pushEvent(new Object[] { "flush1", 110f, 110f }, 352L);
        hl.pushEvent(new Object[] { "flush2", 110f, 110f }, 352L);

        hr.pushEvent(new Object[] { "flush3", 110f, 110f }, 353L);

        compareResults(expected, results);
    }

    /**
     * Tests a correlation of two event streams; first with a timewindow of 6 and second with 4.
     * The correlation condition tests, weather the geometrical distance is equal or smaller 2.2
     * The first stream receives 15000 events and after that the second receives some matching events to test the streambuffers
     * @throws AlreadyRegisteredException
     * @throws NoSuchStreamException
     * @throws EPException
     */
    @Test
    public void distanceCorrelation6() throws EPException {
        EventSchema s1 = new EventSchema(new Attribute("s1_id", Attribute.DataType.STRING), new Attribute("s1_x",
                Attribute.DataType.FLOAT), new Attribute("s1_y", Attribute.DataType.FLOAT));
        EventSchema s2 = new EventSchema(new Attribute("s2_id", Attribute.DataType.STRING), new Attribute("s2_x",
                Attribute.DataType.FLOAT), new Attribute("s2_y", Attribute.DataType.FLOAT));

        StreamHandle hl = provider.registerStream(STREAM_LEFT.getName(), s1);
        StreamHandle hr = provider.registerStream(STREAM_RIGHT.getName(), s2);

        Correlator corr = new Correlator(
                new TimeWindow(STREAM_LEFT, 6), STREAM_LEFT.getName(),
                new TimeWindow(STREAM_RIGHT, 4), STREAM_RIGHT.getName(),
                new LessEq(
                        new Sqrt(
                                new Add(
                                        new Pow(    new Subtract(new Variable(STREAM_LEFT.getName()+".s1_x"), new Variable(STREAM_RIGHT.getName()+".s2_x")),
                                                new IntConstant(2)),
                                        new Pow(    new Subtract(new Variable(STREAM_LEFT.getName()+".s1_y"), new Variable(STREAM_RIGHT.getName()+".s2_y")),
                                                new LongConstant(2L)))),
                        new DoubleConstant(2.2)
                )
        );
        QueryHandle qh = provider.createQuery(QUERY, corr);
        CollectingSink sink = new CollectingSink();
        provider.registerSink(qh, sink);

        List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(
                new SimpleEvent(new Object[] { "s1_15000", 15060.0f, 15060.0f, "s2_1", 15060.0f, 15061.0f}, 15000L, 15004L),
                new SimpleEvent(new Object[] { "s1_14999", 15059.0f, 15059.0f, "s2_3", 15060.0f, 15060.0f}, 15002L, 15005L),
                new SimpleEvent(new Object[] { "s1_15000", 15060.0f, 15060.0f, "s2_3", 15060.0f, 15060.0f}, 15002L, 15006L)
        );

        int k = 15000;
        for (int i = 0; i < k; i++) {
            hl.pushEvent(new Object[] { "s1_"+i , 60.0f + i, 60.0f + i }, 0L + i);
        }

        hl.pushEvent(new Object[] { "s1_"+k, 60.0f+k, 60.0f+k }, 0L + k);
        hr.pushEvent(new Object[] { "s2_1", 60.0f+k, 61.0f+k }, 0L + k);

        //no match
        hr.pushEvent(new Object[] { "s2_2", 60.0f+k, 65.0f+k }, 1L + k);

        hr.pushEvent(new Object[] { "s2_3", 60.0f+k, 60.0f+k }, 2L + k);

        hl.pushEvent(new Object[] { "flush", 0f, 0.0f }, 20L + k);
        hr.pushEvent(new Object[] { "flush", 5f, 0.0f }, 20L + k);

        compareResults(expected, results);
    }


//     @Test
//     public void spatialTest() {
//     GeometryFactory gf = new GeometryFactory();
//
//     Attribute[] schema = {
//     new Attribute("id", Attribute.DataType.LONG),
//     new Attribute("pos", DataType.GEOMETRY),
//     new Attribute(Constants.T_START, Attribute.DataType.LONG),
//     new Attribute(Constants.T_END, Attribute.DataType.LONG )
//     };
//
//     Attribute[] schemaOut = {
//     new Attribute("l_id", DataType.LONG),
//     new Attribute("l_pos", DataType.GEOMETRY),
//     new Attribute("r_id", DataType.LONG),
//     new Attribute("r_pos", DataType.GEOMETRY),
//     new Attribute(Constants.T_START, Attribute.DataType.LONG),
//     new Attribute(Constants.T_END, Attribute.DataType.LONG )
//     };
//
//     CollectingOP op = new CollectingOP();
//
//     Stream src = new Stream("dummy");
//
//     Correlator def = new Correlator("SpatialCorrelator",
//     src, "l",
//     src, "r",
//     new And( new Less( "l.id", "r.id" ),
//     new IsWithinDistance(3.0, "l.pos", "r.pos") ) );
//
//     NativeCorrelator nc = new NativeCorrelator(def,
//     schema, schema, op );
//     Assert.assertEquals(nc.getOutputSchema(), schemaOut);
//
//     for ( int i = 0; i < 10; i++ ) {
//     Point p1 = gf.createPoint( new Coordinate(1.0*i, 1.0*i) );
//     Point p2 = gf.createPoint( new Coordinate(10 - 1.0*i, 10 - 1.0*i) );
//
//     Object[] e1 = { 1L, p1, (long) i, (long) i+1 };
//     Object[] e2 = { 2L, p2, (long) i, (long) i+1 };
//     nc.process(src, e1);
//     nc.process(src, e2);
//     }
//
//     log.debug("Results: ");
//     op.getResults("SpatialCorrelator").forEach( event ->
//     log.debug(Arrays.toString(event)));
//
//     Assert.assertEquals(op.getResults("SpatialCorrelator").size(), 3);
//     for ( Object[] e : op.getResults("SpatialCorrelator") ) {
//     Assert.assertTrue( e[1] instanceof Point );
//     Assert.assertTrue( e[3] instanceof Point );
//     }
//     Assert.assertEquals( op.getResults("SpatialCorrelator").get(0)[1],
//     gf.createPoint( new Coordinate(4.0,4.0) ) );
//     Assert.assertEquals( op.getResults("SpatialCorrelator").get(0)[3],
//     gf.createPoint( new Coordinate(6.0,6.0) ) );
//
//     Assert.assertEquals( op.getResults("SpatialCorrelator").get(1)[1],
//     gf.createPoint( new Coordinate(5.0,5.0) ) );
//     Assert.assertEquals( op.getResults("SpatialCorrelator").get(1)[3],
//     gf.createPoint( new Coordinate(5.0,5.0) ) );
//
//     Assert.assertEquals( op.getResults("SpatialCorrelator").get(2)[1],
//     gf.createPoint( new Coordinate(6.0,6.0) ) );
//     Assert.assertEquals( op.getResults("SpatialCorrelator").get(2)[3],
//     gf.createPoint( new Coordinate(4.0,4.0) ) );
//     }
}
