/*
 * Copyright (c) 2012, 2013, 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.jepc.v2.test.bridge;

import java.util.Arrays;
import java.util.List;

import org.junit.Assert;
import org.junit.Test;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.Attribute.DataType;
import de.umr.ep.common.EPException;
import de.umr.ep.event.Event;
import de.umr.ep.event.EventSchema;
import de.umr.ep.event.impl.SimpleEvent;
import de.umr.jepc.v2.api.bridge.EPBridgeFactory;
import de.umr.jepc.v2.api.epa.Filter;
import de.umr.jepc.v2.api.epa.Stream;
import de.umr.jepc.v2.api.epa.window.TimeWindow;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.DoubleConstant;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.Variable;
import de.umr.jepc.v2.api.expression.logical.And;
import de.umr.jepc.v2.api.expression.logical.Not;
import de.umr.jepc.v2.api.expression.logical.Or;
import de.umr.jepc.v2.api.expression.predicate.Between;
import de.umr.jepc.v2.api.expression.predicate.Equal;
import de.umr.jepc.v2.api.expression.predicate.Greater;
import de.umr.jepc.v2.api.expression.predicate.GreaterEq;
import de.umr.jepc.v2.api.expression.predicate.Less;
import de.umr.jepc.v2.api.expression.predicate.LessEq;
import de.umr.jepc.v2.api.expression.predicate.Unequal;
import de.umr.jepc.v2.api.provider.AlreadyRegisteredException;
import de.umr.jepc.v2.api.provider.QueryHandle;
import de.umr.jepc.v2.api.provider.StreamHandle;
import de.umr.jepc.v2.api.util.CollectingSink;

/**
 * @author Michael Körber
 *
 */
public class FilterTest extends AbstractSemanticTest {

//	private static final Logger log = LoggerFactory.getLogger(FilterTest.class);
    
    private static final Stream STREAM = new Stream("test");
    private static final String QUERY  = "filter";

	public FilterTest(EPBridgeFactory bf) {
		super(bf);
	}

	/**
	 * Tests >= condition
	 * @throws AlreadyRegisteredException
	 * @throws EPException
	 */
	@Test
	public void filter1() throws EPException {

		Filter f = new Filter(new TimeWindow(STREAM, 1),
				new GreaterEq(new Variable("value"), new DoubleConstant(0.5)));

		CollectingSink sink = new CollectingSink();

		StreamHandle sh = provider.registerStream(STREAM.getName(), new EventSchema(
		        new Attribute("id", DataType.LONG), 
		        new Attribute("value", DataType.DOUBLE)));
		
		QueryHandle qh = provider.createQuery(QUERY, f);
		
		provider.registerSink(qh, sink);
		List<Event> results = sink.getEvents();

		List<Event> expected = Arrays.asList(
	        new SimpleEvent(new Object[] { 1L, 0.7 }, 1L, 2L), 
	        new SimpleEvent(new Object[] { 1L, 0.5 }, 1L, 2L)
        );

		sh.pushEvent(new Object[] { 1L, 0.7 }, 1);
		Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { 1L, 0.3 }, 1);
        Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { 1L, 0.5 }, 1);
        Assert.assertEquals(2, results.size());
        sh.pushEvent(new Object[] { 1L, 0.1 }, 1);
        Assert.assertEquals(2, results.size());
		
        compareResults(expected, results);
	}

	/**
	 * Tests (>)&&(<=) condition
	 * @throws AlreadyRegisteredException
	 * @throws EPException
	 */
	@Test
	public void filter2() throws EPException {

		Filter f = new Filter(new TimeWindow(STREAM, 1), new And(
				new Greater(new Variable("value"), new DoubleConstant(0.2)),
				new LessEq(new Variable("value"), new DoubleConstant(0.5))) );

		CollectingSink sink = new CollectingSink();

		StreamHandle sh = provider.registerStream(STREAM.getName(), new EventSchema(
				new Attribute("id", DataType.LONG),
				new Attribute("value", DataType.DOUBLE)));

		QueryHandle qh = provider.createQuery(QUERY, f);

		provider.registerSink(qh, sink);
		List<Event> results = sink.getEvents();

		List<Event> expected = Arrays.asList(
				new SimpleEvent(new Object[] { 1L, 0.3 }, 1, 2),
				new SimpleEvent(new Object[] { 1L, 0.5 }, 1, 2)
		);

		sh.pushEvent(new Object[] { 1L, 0.7 }, 1);
		Assert.assertTrue(results.isEmpty());
		sh.pushEvent(new Object[] { 1L, 0.3 }, 1);
		Assert.assertEquals(1, results.size());
		sh.pushEvent(new Object[] { 1L, 0.5 }, 1);
		Assert.assertEquals(2, results.size());
		sh.pushEvent(new Object[] { 1L, 0.1 }, 1);
		Assert.assertEquals(2, results.size());

		compareResults(expected, results);
	}

	/**
	 * Tests (<)||(between) condition
	 * @throws AlreadyRegisteredException
	 * @throws EPException
	 */
	@Test
	public void filter3() throws EPException {

		Filter f = new Filter(new TimeWindow(STREAM, 1), new Or(
				new Less(new Variable("value"), new DoubleConstant(0.0)),
				new Between(new Variable("value"), new DoubleConstant(100.0), new DoubleConstant(Double.MAX_VALUE))) );

		CollectingSink sink = new CollectingSink();

		StreamHandle sh = provider.registerStream(STREAM.getName(), new EventSchema(
				new Attribute("id", DataType.LONG),
				new Attribute("value", DataType.DOUBLE)));

		QueryHandle qh = provider.createQuery(QUERY, f);

		provider.registerSink(qh, sink);
		List<Event> results = sink.getEvents();

		List<Event> expected = Arrays.asList(
                new SimpleEvent(new Object[] { 3L, -0.1}, 1L, 2L),
                new SimpleEvent(new Object[] { 2L, 100.0}, 3L, 4L),
                new SimpleEvent(new Object[] { 3L, 100.1}, 4L, 5L),
                new SimpleEvent(new Object[] { 4L, 110.0}, 4L, 5L),
                new SimpleEvent(new Object[] { 5L, 1.7976931348623157E308}, 5L, 6L)
		);

		sh.pushEvent(new Object[] { 1L, 0.0 }, 1);
		Assert.assertTrue(results.isEmpty());
		sh.pushEvent(new Object[] { 2L, 0.1 }, 1);
        Assert.assertTrue(results.isEmpty());
		sh.pushEvent(new Object[] { 3L, -0.1 }, 1);
		Assert.assertEquals(1, results.size());
		sh.pushEvent(new Object[] { 1L, 99.9 }, 2);
		Assert.assertEquals(1, results.size());
		sh.pushEvent(new Object[] { 2L, 100.0 }, 3);
		Assert.assertEquals(2, results.size());
		sh.pushEvent(new Object[] { 3L, 100.1 }, 4);
		Assert.assertEquals(3, results.size());
		sh.pushEvent(new Object[] { 4L, 110.0 }, 4);
		Assert.assertEquals(4, results.size());
		sh.pushEvent(new Object[] { 5L, Double.MAX_VALUE }, 5);
		Assert.assertEquals(5, results.size());
		sh.pushEvent(new Object[] { 5L, 12.2 }, 6);
		Assert.assertEquals(5, results.size());

		compareResults(expected, results);
	}

	/**
	 * Tests not unequal condition
	 * @throws AlreadyRegisteredException
	 * @throws EPException
	 */
	@Test
	public void filter4() throws EPException {

		Filter f = new Filter(	new TimeWindow(STREAM, 1),
								new Not(
									new Unequal(new Variable("value"), new DoubleConstant(0.2))
		) );

		CollectingSink sink = new CollectingSink();

		StreamHandle sh = provider.registerStream(STREAM.getName(), new EventSchema(
				new Attribute("id", DataType.LONG),
				new Attribute("value", DataType.DOUBLE)));

		QueryHandle qh = provider.createQuery(QUERY, f);

		provider.registerSink(qh, sink);
		List<Event> results = sink.getEvents();

		List<Event> expected = Arrays.asList(
                new SimpleEvent(new Object[] { 3L, 0.2}, 1L, 2L)
		);

		sh.pushEvent(new Object[] { 1L, 0.1 }, 1);
		Assert.assertTrue(results.isEmpty());
		sh.pushEvent(new Object[] { 2L, 0.1999 }, 1);
        Assert.assertTrue(results.isEmpty());
		sh.pushEvent(new Object[] { 3L, 0.2 }, 1);
		Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { 4L, 0.2001 }, 1);
        Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { 4L, -0.2 }, 1);
        Assert.assertEquals(1, results.size());

		compareResults(expected, results);
	}

	/**
	 * Tests  equal condition
	 * @throws AlreadyRegisteredException
	 * @throws EPException
	 */
	@Test
	public void filter5() throws EPException {

		Filter f = new Filter(new TimeWindow(STREAM, 1), new Equal(
				new Variable("value"), new DoubleConstant(0.2)) );

		CollectingSink sink = new CollectingSink();

		StreamHandle sh = provider.registerStream(STREAM.getName(), new EventSchema(
				new Attribute("id", DataType.LONG),
				new Attribute("value", DataType.DOUBLE)));

		QueryHandle qh = provider.createQuery(QUERY, f);

		provider.registerSink(qh, sink);
		List<Event> results = sink.getEvents();

        List<Event> expected = Arrays.asList(
                new SimpleEvent(new Object[] { 3L, 0.2}, 1L, 2L)
        );

        sh.pushEvent(new Object[] { 1L, 0.1 }, 1);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { 2L, 0.1999 }, 1);
        Assert.assertTrue(results.isEmpty());
        sh.pushEvent(new Object[] { 3L, 0.2 }, 1);
        Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { 4L, 0.2001 }, 1);
        Assert.assertEquals(1, results.size());
        sh.pushEvent(new Object[] { 4L, -0.2 }, 1);
        Assert.assertEquals(1, results.size());

        compareResults(expected, results);
	}
	
	//
	// @Test
	// public void spatialTest() {
	//
	// GeometryFactory gf = new GeometryFactory();
	//
	// Attribute[] schema = {
	// new Attribute("id", DataType.LONG),
	// new Attribute("pos", DataType.GEOMETRY),
	// new Attribute(Constants.T_START, Attribute.DataType.LONG),
	// new Attribute(Constants.T_END, Attribute.DataType.LONG)
	// };
	//
	// FilterIndex index = new FilterIndexFactory(new
	// FilterIndexConfiguration()).createFilterIndex(schema);
	//
	// Coordinate[] rect = {
	// new Coordinate(0, 0),
	// new Coordinate(1, 0),
	// new Coordinate(1, 1),
	// new Coordinate(0, 1),
	// new Coordinate(0, 0)
	// };
	//
	// Polygon area = gf.createPolygon(rect);
	//
	// Filter f1 = new Filter("F1", new Stream("stream"), new Within("pos",
	// area));
	// index.insertFilter(f1);
	//
	// Object[] e1 = { 1L, gf.createPoint( new Coordinate(0, 0) ), 1L, 2L };
	// Object[] e2 = { 1L, gf.createPoint( new Coordinate(0.5, 0) ), 2L, 3L };
	// Object[] e3 = { 1L, gf.createPoint( new Coordinate(0, 0.5) ), 3L, 4L };
	// Object[] e4 = { 1L, gf.createPoint( new Coordinate(0.5, 0.5) ), 4L, 5L };
	// Object[] e5 = { 1L, gf.createPoint( new Coordinate(2.0, 2.0) ), 5L, 6L };
	//
	// Collection<Filter> results = null;
	// results = index.checkEvent(e1);
	// assertTrue(results.isEmpty());
	//
	// results = index.checkEvent(e2);
	// assertTrue(results.isEmpty());
	//
	// results = index.checkEvent(e3);
	// assertTrue(results.isEmpty());
	//
	// results = index.checkEvent(e4);
	// assertFalse(results.isEmpty());
	//
	// results = index.checkEvent(e5);
	// assertTrue(results.isEmpty());
	//
	// }

}
