package de.umr.jepc.v2.ql;

import java.util.ArrayList;
import java.util.List;
import java.util.ServiceLoader;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.umr.ep.common.IncompatibleTypeException;
import de.umr.ep.event.EventSchema;
import de.umr.ep.event.SchemaException;
import de.umr.jepc.v2.api.epa.pattern.symbol.Binding;
import de.umr.jepc.v2.api.epa.pattern.symbol.Bindings;
import de.umr.jepc.v2.api.epa.pattern.symbol.NoSuchVariableException;
import de.umr.jepc.v2.api.expression.ArithmeticExpression;
import de.umr.jepc.v2.api.expression.BooleanExpression;
import de.umr.jepc.v2.api.expression.Expression;
import de.umr.jepc.v2.bindings.EVBinding;
import de.umr.jepc.v2.expressions.arithmetic.EVNumericExpression;
import de.umr.jepc.v2.expressions.arithmetic.EVStringExpression;
import de.umr.jepc.v2.expressions.bool.EVBooleanExpression;

public class ExpressionTranslator {
    
    private static final Logger log = LoggerFactory.getLogger(ExpressionTranslator.class);
    
    public static final ExpressionTranslator INSTANCE = new ExpressionTranslator();
	
	private static class TranslatorMapping {
		
		private List<Translator<?,?>> translators = new ArrayList<>();
		
		
		public void add( Translator<?, ?> translator ) {
			translators.add(translator);
		}
		
		@SuppressWarnings("unchecked")
		public <TIN extends Expression,TOUT> Translator<TIN,TOUT> get( Class<TIN> inType, Class<TOUT> outType ) throws TranslatorException {
			for ( Translator<?,?> t : translators ) {
				if ( t.getInputType().isAssignableFrom(inType) && outType.isAssignableFrom(t.getOutputType()) )
					return (Translator<TIN, TOUT>) t;
			}
			throw new TranslatorException("No Translator found: " + inType + " --> " + outType );
		}
	}
	
	
	private final TranslatorMapping translators = new TranslatorMapping();
	
	@SuppressWarnings("rawtypes")
	public ExpressionTranslator() {
		log.debug("Loading registered translators.");
		
		ServiceLoader<Translator> tsl = ServiceLoader.load(Translator.class);
		for ( Translator<?,?> t : tsl ) {
			log.debug("Registering translator: {}", t );
			translators.add(t);
		}
		log.debug("Finished loading registered translators.");
	}
	
	public EVBinding translateVariableBinding( Binding in, EventSchema schema, Bindings bindings )  throws TranslatorException, IncompatibleTypeException {
	    try {
            return new EVBinding(in, schema, bindings);
        } catch (NoSuchVariableException e) {
            throw new TranslatorException("Variable not found!",e);
        }
	    catch (SchemaException e) {
			throw new TranslatorException("Error translating binding.",e);
		}
	}
	
	public EVBooleanExpression translateBooleanExpression( BooleanExpression in, EventSchema schema, Bindings bindings ) throws TranslatorException, IncompatibleTypeException {
		return translate(in, EVBooleanExpression.class, schema, bindings);
	}
	
	public EVStringExpression translateStringExpression( ArithmeticExpression in, EventSchema schema, Bindings bindings ) throws TranslatorException, IncompatibleTypeException {
		return translate(in, EVStringExpression.class, schema, bindings);
	}
	
	public EVNumericExpression translateNumericExpression( ArithmeticExpression in, EventSchema schema, Bindings bindings ) throws TranslatorException, IncompatibleTypeException {
		return translate(in, EVNumericExpression.class, schema, bindings);
	}
	
	private <T> T translate(Expression in, Class<T> desiredType, EventSchema schema, Bindings bindings) throws TranslatorException, IncompatibleTypeException {
		try {
            return desiredType.cast( translators.get(in.getClass(), desiredType).translate(in, this, schema, bindings));
        } catch (NoSuchVariableException e) {
            throw new TranslatorException("Variable not found!",e);
        }
		catch (SchemaException e) {
			throw new TranslatorException("Error translating expression.",e);
		}
	}
}
