package de.umr.jepc.v2.ql;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import org.junit.Before;
import org.junit.Test;

import de.umr.ep.common.Attribute;
import de.umr.ep.common.Attribute.DataType;
import de.umr.ep.common.IncompatibleTypeException;
import de.umr.ep.event.Event;
import de.umr.ep.event.EventSchema;
import de.umr.ep.event.SchemaException;
import de.umr.ep.event.impl.SimpleEvent;
import de.umr.jepc.v2.api.epa.pattern.symbol.Bindings;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.DoubleConstant;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.FloatConstant;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.IntConstant;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.LongConstant;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.StringConstant;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.Variable;
import de.umr.jepc.v2.api.expression.arithmetic.compund.numeric.Add;
import de.umr.jepc.v2.api.expression.logical.And;
import de.umr.jepc.v2.api.expression.logical.False;
import de.umr.jepc.v2.api.expression.logical.Not;
import de.umr.jepc.v2.api.expression.logical.Or;
import de.umr.jepc.v2.api.expression.logical.True;
import de.umr.jepc.v2.api.expression.predicate.Equal;
import de.umr.jepc.v2.api.expression.predicate.Greater;
import de.umr.jepc.v2.api.expression.predicate.GreaterEq;
import de.umr.jepc.v2.api.expression.predicate.Less;
import de.umr.jepc.v2.api.expression.predicate.LessEq;
import de.umr.jepc.v2.api.expression.predicate.Unequal;
import de.umr.jepc.v2.bindings.BoundVariables;
import de.umr.jepc.v2.expressions.arithmetic.EVNumericExpression;
import de.umr.jepc.v2.expressions.arithmetic.EVStringExpression;
import de.umr.jepc.v2.expressions.arithmetic.constant.EVDoubleConstant;
import de.umr.jepc.v2.expressions.arithmetic.constant.EVFloatConstant;
import de.umr.jepc.v2.expressions.arithmetic.constant.EVIntConstant;
import de.umr.jepc.v2.expressions.arithmetic.constant.EVLongConstant;
import de.umr.jepc.v2.expressions.arithmetic.constant.EVStringConstant;
import de.umr.jepc.v2.expressions.arithmetic.field.EVByteField;
import de.umr.jepc.v2.expressions.arithmetic.field.EVDoubleField;
import de.umr.jepc.v2.expressions.arithmetic.field.EVFloatField;
import de.umr.jepc.v2.expressions.arithmetic.field.EVIntField;
import de.umr.jepc.v2.expressions.arithmetic.field.EVLongField;
import de.umr.jepc.v2.expressions.arithmetic.field.EVShortField;
import de.umr.jepc.v2.expressions.arithmetic.field.EVStringField;
import de.umr.jepc.v2.expressions.arithmetic.numeric.EVAdd;
import de.umr.jepc.v2.expressions.bool.EVAnd;
import de.umr.jepc.v2.expressions.bool.EVBooleanExpression;
import de.umr.jepc.v2.expressions.bool.EVNot;
import de.umr.jepc.v2.expressions.bool.EVOr;
import de.umr.jepc.v2.expressions.predicate.numeric.EVNumericEqual;
import de.umr.jepc.v2.expressions.predicate.numeric.EVNumericGreater;
import de.umr.jepc.v2.expressions.predicate.numeric.EVNumericGreaterEq;
import de.umr.jepc.v2.expressions.predicate.numeric.EVNumericLess;
import de.umr.jepc.v2.expressions.predicate.numeric.EVNumericLessEq;
import de.umr.jepc.v2.expressions.predicate.numeric.EVNumericUnequal;
import de.umr.jepc.v2.expressions.predicate.string.EVStringEqual;
import de.umr.jepc.v2.expressions.predicate.string.EVStringGreater;
import de.umr.jepc.v2.expressions.predicate.string.EVStringGreaterEq;
import de.umr.jepc.v2.expressions.predicate.string.EVStringLess;
import de.umr.jepc.v2.expressions.predicate.string.EVStringLessEq;
import de.umr.jepc.v2.expressions.predicate.string.EVStringUnequal;

public class ExpressionTranslatorTest {

	ExpressionTranslator t;
	EventSchema schema;
	Event event;
	BoundVariables vars;

	@Before
	public void setUp() throws Exception {
		t = new ExpressionTranslator();
		schema = new EventSchema(new Attribute("A", DataType.BYTE), new Attribute("B", DataType.SHORT),
				new Attribute("C", DataType.INTEGER), new Attribute("D", DataType.LONG),
				new Attribute("E", DataType.FLOAT), new Attribute("F", DataType.DOUBLE),
				new Attribute("G", DataType.STRING));

		Object[] payload = { (byte) 42, (short) 42, 42, 42L, 42f, 42d, "42" };
		event = new SimpleEvent(payload, 42, 43);
		vars = null;
	}

	@Test
	public void testTranslateIntConstant() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVNumericExpression result = t.translateNumericExpression(new IntConstant(42), schema, new Bindings());
		assertTrue(result instanceof EVIntConstant);
		Number ev = result.eval(event, null);
		assertTrue(ev instanceof Integer);
		assertEquals(42, ev);
	}

	@Test
	public void testTranslateLongConstant() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVNumericExpression result = t.translateNumericExpression(new LongConstant(42L), schema, new Bindings());
		assertTrue(result instanceof EVLongConstant);
		Number ev = result.eval(event, null);
		assertTrue(ev instanceof Long);
		assertEquals(42L, ev);
	}

	@Test
	public void testTranslateFloatConstant() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVNumericExpression result = t.translateNumericExpression(new FloatConstant(42f), schema, new Bindings());
		assertTrue(result instanceof EVFloatConstant);
		Number ev = result.eval(event, null);
		assertTrue(ev instanceof Float);
		assertEquals(42f, ev);
	}

	@Test
	public void testTranslateDoubleConstant() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVNumericExpression result = t.translateNumericExpression(new DoubleConstant(42d), schema, new Bindings());
		assertTrue(result instanceof EVDoubleConstant);
		Number ev = result.eval(event, null);
		assertTrue(ev instanceof Double);
		assertEquals(42d, ev);
	}

	@Test
	public void testTranslateStringConstant() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVStringExpression result = t.translateStringExpression(new StringConstant("42"), schema, new Bindings());
		assertTrue(result instanceof EVStringConstant);
		String ev = result.eval(event, null);
		assertEquals("42", ev);
	}

	/*
	 * FIELDS
	 */
	@Test
	public void testTranslateByteField() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVNumericExpression result = t.translateNumericExpression(new Variable("A"), schema, new Bindings());
		assertTrue(result instanceof EVByteField);
		Number ev = result.eval(event, null);
		assertTrue(ev instanceof Byte);
		assertEquals(42, ev.byteValue());
	}

	@Test
	public void testTranslateShortField() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVNumericExpression result = t.translateNumericExpression(new Variable("B"), schema, new Bindings());
		assertTrue(result instanceof EVShortField);
		Number ev = result.eval(event, null);
		assertTrue(ev instanceof Short);
		assertEquals(42, ev.shortValue());
	}

	@Test
	public void testTranslateIntField() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVNumericExpression result = t.translateNumericExpression(new Variable("C"), schema, new Bindings());
		assertTrue(result instanceof EVIntField);
		Number ev = result.eval(event, null);
		assertTrue(ev instanceof Integer);
		assertEquals(42, ev.intValue());
	}

	@Test
	public void testTranslateLongField() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVNumericExpression result = t.translateNumericExpression(new Variable("D"), schema, new Bindings());
		assertTrue(result instanceof EVLongField);
		Number ev = result.eval(event, null);
		assertTrue(ev instanceof Long);
		assertEquals(42L, ev.longValue());
	}

	@Test
	public void testTranslateFloatField() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVNumericExpression result = t.translateNumericExpression(new Variable("E"), schema, new Bindings());
		assertTrue(result instanceof EVFloatField);
		Number ev = result.eval(event, null);
		assertTrue(ev instanceof Float);
		assertEquals(42f, ev.floatValue(), 0.01f);
	}

	@Test
	public void testTranslateDoubleField() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVNumericExpression result = t.translateNumericExpression(new Variable("F"), schema, new Bindings());
		assertTrue(result instanceof EVDoubleField);
		Number ev = result.eval(event, null);
		assertTrue(ev instanceof Double);
		assertEquals(42d, ev.doubleValue(), 0.01d);
	}

	@Test
	public void testTranslateStringField() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVStringExpression result = t.translateStringExpression(new Variable("G"), schema, new Bindings());
		assertTrue(result instanceof EVStringField);
		assertEquals("42", result.eval(event, null));
	}

	/*
	 * Predicates
	 */
	@Test
	public void testTranslateNumericLess() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t.translateBooleanExpression(new Less(new IntConstant(42), new IntConstant(42)),
				schema, new Bindings());
		assertTrue(result instanceof EVNumericLess);
		assertFalse(result.eval(event, vars));
	}

	@Test
	public void testTranslateNumericLessEq() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t.translateBooleanExpression(new LessEq(new IntConstant(42), new IntConstant(42)),
				schema, new Bindings());
		assertTrue(result instanceof EVNumericLessEq);
		assertTrue(result.eval(event, vars));
	}

	@Test
	public void testTranslateNumericGreater() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t.translateBooleanExpression(new Greater(new IntConstant(42), new IntConstant(42)),
				schema, new Bindings());
		assertTrue(result instanceof EVNumericGreater);
		assertFalse(result.eval(event, vars));
	}

	@Test
	public void testTranslateNumericGreaterEq() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t
				.translateBooleanExpression(new GreaterEq(new IntConstant(42), new IntConstant(42)), schema, new Bindings());
		assertTrue(result instanceof EVNumericGreaterEq);
		assertTrue(result.eval(event, vars));
	}

	@Test
	public void testTranslateNumericUnqueal() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t.translateBooleanExpression(new Unequal(new IntConstant(42), new IntConstant(42)),
				schema, new Bindings());
		assertTrue(result instanceof EVNumericUnequal);
		assertFalse(result.eval(event, vars));
	}

	@Test
	public void testTranslateNumericEqual() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t.translateBooleanExpression(new Equal(new IntConstant(42), new IntConstant(42)),
				schema, new Bindings());
		assertTrue(result instanceof EVNumericEqual);
		assertTrue(result.eval(event, vars));
	}

	// String

	@Test
	public void testTranslateStringLess() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t
				.translateBooleanExpression(new Less(new StringConstant("42"), new StringConstant("42")), schema, new Bindings());
		assertTrue(result instanceof EVStringLess);
		assertFalse(result.eval(event, vars));
	}

	@Test
	public void testTranslateStringLessEq() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t
				.translateBooleanExpression(new LessEq(new StringConstant("42"), new StringConstant("42")), schema, new Bindings());
		assertTrue(result instanceof EVStringLessEq);
		assertTrue(result.eval(event, vars));
	}

	@Test
	public void testTranslateStringGreater() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t
				.translateBooleanExpression(new Greater(new StringConstant("42"), new StringConstant("42")), schema, new Bindings());
		assertTrue(result instanceof EVStringGreater);
		assertFalse(result.eval(event, vars));
	}

	@Test
	public void testTranslateStringGreaterEq() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t
				.translateBooleanExpression(new GreaterEq(new StringConstant("42"), new StringConstant("42")), schema, new Bindings());
		assertTrue(result instanceof EVStringGreaterEq);
		assertTrue(result.eval(event, vars));
	}

	@Test
	public void testTranslateStringUnqueal() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t
				.translateBooleanExpression(new Unequal(new StringConstant("42"), new StringConstant("42")), schema, new Bindings());
		assertTrue(result instanceof EVStringUnequal);
		assertFalse(result.eval(event, vars));
	}

	@Test
	public void testTranslateStringEqual() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t
				.translateBooleanExpression(new Equal(new StringConstant("42"), new StringConstant("42")), schema, new Bindings());
		assertTrue(result instanceof EVStringEqual);
		assertTrue(result.eval(event, vars));
	}

	// INCOMPATIBLE TYPES

	@Test
	public void testTranslateIncLess() throws IncompatibleTypeException, TranslatorException, SchemaException {
		try {
			t.translateBooleanExpression(new Less(new StringConstant("42"), new IntConstant(42)), schema, new Bindings());
			fail("Not allowed to compare strings with numbers");
		} catch (IncompatibleTypeException ite) {

		}
	}

	@Test
	public void testTranslateIncLessEq() throws IncompatibleTypeException, TranslatorException, SchemaException {
		try {
			t.translateBooleanExpression(new LessEq(new StringConstant("42"), new IntConstant(42)), schema, new Bindings());
			fail("Not allowed to compare strings with numbers");
		} catch (IncompatibleTypeException ite) {

		}
	}

	@Test
	public void testTranslateIncGreater() throws IncompatibleTypeException, TranslatorException, SchemaException {
		try {
			t.translateBooleanExpression(new Greater(new StringConstant("42"), new IntConstant(42)), schema, new Bindings());
			fail("Not allowed to compare strings with numbers");
		} catch (IncompatibleTypeException ite) {

		}
	}

	@Test
	public void testTranslateIncGreaterEq() throws IncompatibleTypeException, TranslatorException, SchemaException {
		try {
			t.translateBooleanExpression(new GreaterEq(new StringConstant("42"), new IntConstant(42)), schema, new Bindings());
			fail("Not allowed to compare strings with numbers");
		} catch (IncompatibleTypeException ite) {

		}
	}

	@Test
	public void testTranslateIncUnqueal() throws IncompatibleTypeException, TranslatorException, SchemaException {
		try {
			t.translateBooleanExpression(new Unequal(new StringConstant("42"), new IntConstant(42)), schema, new Bindings());
			fail("Not allowed to compare strings with numbers");
		} catch (IncompatibleTypeException ite) {

		}
	}

	@Test
	public void testTranslateIncEqual() throws IncompatibleTypeException, TranslatorException, SchemaException {
		try {
			t.translateBooleanExpression(new Equal(new StringConstant("42"), new IntConstant(42)), schema, new Bindings());
			fail("Not allowed to compare strings with numbers");
		} catch (IncompatibleTypeException ite) {

		}
	}

	/*
	 * Bool
	 */
	@Test
	public void testTranslateAnd() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t.translateBooleanExpression(new And(new True(), new False()), schema, new Bindings());
		assertTrue(result instanceof EVAnd);
		assertFalse(result.eval(event, vars));
	}
	
	@Test
	public void testTranslateOr() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t.translateBooleanExpression(new Or(new True(), new False()), schema, new Bindings());
		assertTrue(result instanceof EVOr);
		assertTrue(result.eval(event, vars));
	}
	
	@Test
	public void testTranslateNot() throws IncompatibleTypeException, TranslatorException, SchemaException {
		EVBooleanExpression result = t.translateBooleanExpression(new Not(new False()), schema, new Bindings());
		assertTrue(result instanceof EVNot);
		assertTrue(result.eval(event, vars));
	}

	/*
	 * Arithmetic
	 */

	@Test
	public void testTranslateAdd() throws IncompatibleTypeException, TranslatorException, SchemaException {
		Add p = new Add(new IntConstant(21), new IntConstant(21));
		EVNumericExpression result = t.translateNumericExpression(p, schema, new Bindings());
		assertTrue(result instanceof EVAdd);
		Number ev = result.eval(event, vars);
		assertTrue(ev instanceof Integer);
		assertEquals(42, ev);
	}

}
