/**
 * 
 */
package de.umr.jepc.v2.federated.api;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.umr.ep.common.EPException;
import de.umr.ep.event.EventSchema;
import de.umr.jepc.v2.api.EPProvider;
import de.umr.jepc.v2.api.epa.EPA;
import de.umr.jepc.v2.api.epa.Stream;
import de.umr.jepc.v2.api.provider.AlreadyRegisteredException;
import de.umr.jepc.v2.api.provider.DataSource;
import de.umr.jepc.v2.api.provider.EventSink;
import de.umr.jepc.v2.api.provider.ExternalStreamHandle;
import de.umr.jepc.v2.api.provider.NoSuchQueryException;
import de.umr.jepc.v2.api.provider.NoSuchStreamException;
import de.umr.jepc.v2.api.provider.QueryHandle;
import de.umr.jepc.v2.api.provider.StreamHandle;
import de.umr.jepc.v2.api.schema.StreamInfo;
import de.umr.jepc.v2.federated.api.epa.ForeignInput;
import de.umr.jepc.v2.federated.api.epa.QualifiedStream;
import de.umr.jepc.v2.federated.api.epa.ResourceBoundary;
import de.umr.jepc.v2.federated.api.provider.Resource;
import de.umr.jepc.v2.federated.api.provider.ResourceManager;
import de.umr.jepc.v2.federated.api.provider.Resources;
import de.umr.jepc.v2.federated.api.provider.SplitStrategy;
import de.umr.jepc.v2.federated.api.provider.query.FederatedQuery;
import de.umr.jepc.v2.federated.api.provider.query.SubQuery;
import de.umr.jepc.v2.federated.api.util.ResourceException;

/**
 * Provider implementing a federated EP system. All available processing {@link Resource resources}
 * are registered and queries are distributed using the given {@link SplitStrategy strategy}.
 * 
 * @param <T> the concrete resource type to manage
 * 
 * @author Michael Körber
 *
 */
public class FederatedEPProvider<T extends Resource> implements EPProvider{
	
    /** The logger */
	private static final Logger log = LoggerFactory.getLogger(FederatedEPProvider.class);

	/** The registered resources */ 
    private final Resources<T> resources = new Resources<>();
    
    /** The strategy for distributing queries */
    private final SplitStrategy<? super T> strategy;
    
    /** The master-resource receiving the results of all queries */
    private final T masterResource;
    
    /** A manager implementation, taking care of query deployment */
    private final ResourceManager<T> manager;
    
    /** All currently active queries */
    private final Map<String, FederatedQuery> queries = new HashMap<>();

    /**
     * @param strategy the split strategy to use
     * @param masterResource the master resource receiving all query results
     * @param mgr the manager handling query deployment
     */
    public FederatedEPProvider(SplitStrategy<? super T> strategy, T masterResource, ResourceManager<T> mgr) {
    	this.strategy = strategy;
    	this.masterResource = masterResource;
    	try {
    	    resources.addResource(masterResource);
    	} catch ( ResourceException re ) {
    	    throw new IllegalStateException("Could not add master-resource",re);
    	}
    	this.manager = mgr;
    }

    /**
     * Adds a new resource
     * @param resource the resource to add
     * @throws ResourceException when adding a resource with a duplicate identifier.
     */
    public void addResource(T resource) throws ResourceException {
        log.debug("New resource added: " + resource);
    	resources.addResource(resource);
    }
    
    /**
     * {@inheritDoc}
     */
    public QueryHandle createQuery( String name, EPA query ) throws AlreadyRegisteredException, EPException {
        if ( queries.containsKey(name) )
            throw new AlreadyRegisteredException("A query with name \"" + name + "\" already exists");
        
    	log.debug("Distributing query \"" + name + "\": " + query);
    	log.trace("Invoking split strategy");
    	EPA splitQuery = strategy.split(query, resources);
    	log.trace("Finished invoking split strategy: " + splitQuery);
    	log.trace("Building federated query");
    	FederatedQuery fq = new FederatedQuery(name);
    	
    	Deque<String> toResources = new ArrayDeque<>();
    	toResources.push(masterResource.getIdentifier());
    	processGraphRek(splitQuery, fq, toResources);
    	
    	log.trace("Finished building federated query: " + fq);
    	log.trace("Deploying query");
    	
    	SubQuery end = fq.iterator().next();
    	
    	ForeignInput sinkIn = new ForeignInput(end.getResourceId(), name, end.getOutputSchema());
    	
    	manager.deploySink(masterResource, name, sinkIn, fq.getSink());
    	for ( SubQuery qp : fq ) {
    	    manager.deployQuery(resources.getResourceById(qp.getResourceId()), 
    	                        resources.getResourceById(qp.getToResourceId()), 
    	                        fq.getName(), qp.getQueryPart());
    	}
    	
    	queries.put(name, fq);
    	log.trace("Finished deploying query");
    	return fq;
    }
    
    /**
     * {@inheritDoc}
     */
    public void removeQuery( QueryHandle handle ) throws NoSuchQueryException {
        FederatedQuery fq = queries.remove(handle.getName());
        if ( fq == null )
            throw new NoSuchQueryException("No query found for name: " + handle.getName());
        
        Iterator<SubQuery> qpi = fq.reverseIterator();
        while ( qpi.hasNext() ) {
            SubQuery qp = qpi.next();
            try {
                manager.removeQuery(resources.getResourceById(qp.getResourceId()), fq.getName());
            } catch (ResourceException e) {
                log.error("Query: " + handle.getName() + " could not be removed from resource: " + qp.getResourceId(), e);
            }
        }
        try {
            manager.removeQuery(masterResource, fq.getName());
        } catch (ResourceException e) {
            log.error("Could not remove sink from master resource.", e);
        }
        
    }
    
    /**
     * Transforms a {@link SplitStrategy#split(EPA, Resources) splitted query} into the internal
     * representation. This means creating sub-queries and for each of them computing foreign inputs
     * and target resrouces. Resource-specific stream-names are converted into their local representation.
     * @param query the split query containing {@link ResourceBoundary boundaries}
     * @param accum the accumulator storing constructed sub-query
     * @param toResources yet another accumulator to retrieve the target resource for the currently processed sub-query
     * @return the output-schema of the given query.
     * @throws EPException on any error when assembling the sub-queries
     */
    private EventSchema processGraphRek( EPA query, FederatedQuery accum, Deque<String> toResources ) throws EPException {
    	EventSchema[] inputSchemas = new EventSchema[query.getInputEPAs().length];
    	
    	// Remeber resource to push results to
    	if ( query instanceof ResourceBoundary ) {
    	    toResources.push(((ResourceBoundary) query).getResourceId());
    	}
    	
    	// Recursively process inputs
    	for ( int i = 0; i < inputSchemas.length; i++ ) {
    		EPA src = query.getInputEPAs()[i];
    		inputSchemas[i] = processGraphRek(src, accum, toResources);
    		// Physically split the query by replacing boundaries with foreign input references
    		if ( src instanceof ResourceBoundary ) {
    			ResourceBoundary fence = (ResourceBoundary) src;
    			ForeignInput foreignInput = new ForeignInput(fence.getResourceId(), accum.getName(), inputSchemas[i]);
    			query.setInput(i, foreignInput);
    		}
    		// Strip qualified windows from final graph
    		else if ( src instanceof QualifiedStream ) {
    		    query.setInput(i, src.getInputEPAs()[0] );
    		}
    	}
    	
    	// Src -- replace compound stream name with local stream name (if specified)
    	if ( query instanceof Stream ) {
    	    Stream stream = (Stream) query;
    		String resource = toResources.peek();
    		StreamInfo si = this.resources.getStreamInfo(resource, stream.getName() );
    		return stream.computeOutputSchema(si.getSchema());
    	}
    	// ResourceBoundary handling --> Create new sub-query
    	else if ( query instanceof ResourceBoundary ) {
    	    toResources.pop();
    		ResourceBoundary fence = (ResourceBoundary) query;
    		EventSchema schemaOut = fence.computeOutputSchema(inputSchemas);
    		accum.addSubQuery( new SubQuery(fence.getResourceId(), toResources.peek(), fence.getInputEPAs()[0], schemaOut) );
    		return schemaOut;
    	}
    	// Anything else
    	else {
    		return query.computeOutputSchema(inputSchemas);
    	}
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public StreamInfo getStreamInfo(String name) throws NoSuchStreamException {
        try {
            List<T> tgts = resources.getResourcesForStream(name);
            if ( !tgts.isEmpty() )
                return tgts.get(0).getStreamByName(name);
        } catch ( ResourceException re ) {
            throw new NoSuchStreamException("No stream found for name: " + name, re );
        }
        throw new NoSuchStreamException("No stream found for name: " + name );
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public StreamHandle registerStream(String name, EventSchema schema) throws AlreadyRegisteredException, EPException {
        throw new UnsupportedOperationException("Streams are provided by federated resources");
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public ExternalStreamHandle registerExternalStream(String name, EventSchema schema) throws AlreadyRegisteredException, EPException {
        throw new UnsupportedOperationException("Streams are provided by federated resources");
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void removeStream(DataSource stream) throws NoSuchStreamException {
        throw new UnsupportedOperationException("Streams are provided by federated resources");
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public void registerSink(QueryHandle handle, EventSink sink) throws NoSuchQueryException {
        FederatedQuery fq = queries.get(handle.getName());
        if ( fq == null )
            throw new NoSuchQueryException("No query found for name: " + handle.getName());
        fq.getSink().addSink(sink);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void removeSink(QueryHandle handle, EventSink sink) throws NoSuchQueryException {
        FederatedQuery fq = queries.get(handle.getName());
        if ( fq == null )
            throw new NoSuchQueryException("No query found for name: " + handle.getName());
        fq.getSink().removeSink(sink);
        
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void removeSink(EventSink sink) {
        for ( FederatedQuery fq : queries.values() )
            fq.getSink().removeSink(sink);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void destroy() {
        log.debug("Shutting down.");
        log.trace("Removing all active queries");
        for ( FederatedQuery fq : queries.values() ) {
            try {
                removeQuery(fq);
                log.trace("Removed query: " + fq);
            } catch (NoSuchQueryException e) {
                log.error("Could not remove query: " + fq);
            }
        }
        log.debug("Finished shutting down.");
    }
}
