/**
 * 
 */
package de.umr.jepc.v2.federated.api.epa;

import de.umr.ep.event.EventSchema;
import de.umr.ep.event.SchemaException;
import de.umr.jepc.v2.api.epa.EPA;
import de.umr.jepc.v2.federated.api.provider.Resource;

/**
 * EPA for feeding external streams into the engine.
 * Holds information about the originating {@link Resource resource},
 * the query it belongs to and the {@link EventSchema schema} of incoming events.
 * @author Michael Körber
 *
 */
public class ForeignInput implements EPA {
	
	private static final long serialVersionUID = 1L;

	/** No input EPAs on foreign inputs */
	private static final EPA[] INPUTS = {};

	/** The resource this input originates at */
	private final String originResourceId;
	
	/** The name of the federated query */
	private final String query;
	
	/** The schema of incoming events */
	private final EventSchema schema;
	
	/**
	 * Creates a new instance
	 * @param originResourceId the id of the resource, this input is received from
	 * @param query the name of the federated query
	 * @param schema the schema of incoming events
	 */
	public ForeignInput(String originResourceId, String query, EventSchema schema) {
		this.originResourceId = originResourceId;
		this.query = query;
		this.schema = schema;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public EPA[] getInputEPAs() {
		return INPUTS;
	}
	
	/**
	 * @return the id of the resource, this input is received from
	 */
	public String getOriginResourceId() {
		return originResourceId;
	}
	
	/**
	 * @return the name of the federated query
	 */
	public String getQuery() {
        return query;
    }

    /**
	 * @return the schema of incoming events
	 */
	public EventSchema getSchema() {
		return schema;
	}
	
	/**
	 * @return a unique name composed of the federated query name and the origin resource id
	 */
	public String getUniqueStreamName() {
	    return query + "_" + originResourceId;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public EventSchema computeOutputSchema(EventSchema... inputSchemas) throws SchemaException {
		return schema;
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public void setInput(int idx, EPA input) {
		throw new UnsupportedOperationException("Rebase of ForeignInput not permitted");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString() {
		return "ForeignInput [resourceId=" + originResourceId + ", schema=" + schema + "]";
	}
	
	
}
