/**
 * 
 */
package de.umr.jepc.v2.federated.api.provider;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.umr.jepc.v2.api.provider.NoSuchStreamException;
import de.umr.jepc.v2.api.schema.StreamInfo;
import de.umr.jepc.v2.federated.api.util.ResourceException;

/**
 * Manages all available {@link Resource resources} and allows for retrieving
 * information about provided event streams or locating the resource for a given stream.
 * @author Michael Körber
 *
 */
public class Resources<T extends Resource> {

    /** The managed resources */
	private final Map<String, T> resources = new HashMap<>();
	
	/** 
	 * Adds the given resource 
	 * @param resource the resource to add
	 * @throws ResourceException if this resource was registered already (duplicate identifier). 
	 */
	public void addResource(T resource) throws ResourceException {
        if ( resources.containsKey( resource.getIdentifier()) ) 
            throw new ResourceException("Resource with id: " + resource.getIdentifier() + " already registered");
        resources.put( resource.getIdentifier(), resource );
    }
	
	/**
	 * Retrieves the resource for the given id.
	 * @param id the id of the resource to get.
	 * @return the resource for the given ID.
	 * @throws ResourceException if no resource is registered with the given id.
	 */
	public T getResourceById( String id ) throws ResourceException {
		T result = resources.get(id);
		if ( result == null ) 
			throw new ResourceException("No such resource: " + id);
		return result;
	}
	
	/**
	 * Retrieves the resrouces hosting the event stream with the given name.
	 * The returned list will contain all resources hosting a stream with the 
	 * given name.
	 * @param streamName the stream name
	 * @return the resources hosting the given stream
	 * @throws ResourceException if no resource with the given id is registered.
	 */
	public List<T> getResourcesForStream( String streamName ) throws ResourceException {
		List<T> result = new ArrayList<>();
		for ( T res : resources.values() ) {
		    try {
		        res.getStreamByName(streamName);
		        result.add(res);
		    } catch ( NoSuchStreamException nse ) {
		    }
		}
		if ( result.isEmpty() )
		    throw new ResourceException("No resource is providing stream: " + streamName);
		
		return result;
	}
	
	/**
	 * @param resource
	 * @param stream
	 * @return
	 * @throws NoSuchStreamException
	 * @throws ResourceException
	 */
	public StreamInfo getStreamInfo( String resource, String stream ) throws NoSuchStreamException, ResourceException {
	    return getResourceById(resource).getStreamByName(stream);
	}
}
