/**
 * 
 */
package de.umr.jepc.v2.federated.api.provider.strategy;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Deque;
import java.util.HashSet;
import java.util.List;

import de.umr.jepc.v2.api.epa.EPA;
import de.umr.jepc.v2.api.epa.Stream;
import de.umr.jepc.v2.federated.api.epa.QualifiedStream;
import de.umr.jepc.v2.federated.api.epa.ResourceBoundary;
import de.umr.jepc.v2.federated.api.provider.Resource;
import de.umr.jepc.v2.federated.api.provider.Resources;
import de.umr.jepc.v2.federated.api.provider.SplitStrategy;
import de.umr.jepc.v2.federated.api.util.ResourceException;

/**
 * A straight forward implementation of a {@link SplitStrategy split-strategy} not relying
 * on context-specific resource information. Sub-queries are kept on the same resource as long
 * as possible and are only forwarded, if further computation on the current resource is impossible.
 * This class may be used as blue-print for more sophisticated strategies.
 * @author Michael Körber
 *
 */
public class SimpleSplitStrategy implements SplitStrategy<Resource> {
	
	/**
	 * {@inheritDoc}
	 * @throws ResourceException 
	 */
	@Override
	public EPA split(EPA query, Resources<? extends Resource> resources) throws ResourceException {
		Deque<Resource> context = new ArrayDeque<>();
		EPA newQuery = splitRec(query, resources, context);
		return new ResourceBoundary(context.peek().getIdentifier(),newQuery);
	}
	
	private EPA splitRec( EPA query, Resources<? extends Resource> resources, Deque<Resource> context ) throws ResourceException {
		if ( query instanceof Stream ) {
		    Stream s = (Stream) query;
			context.push( resources.getResourcesForStream(s.getName()).get(0) );
		}
		else if ( query instanceof QualifiedStream ) {
		    QualifiedStream qw = (QualifiedStream) query;
		    context.push( resources.getResourceById(qw.getQualifier()) );
		}
		else {
			Resource[] inputResources = new Resource[query.getInputEPAs().length];
			int [] inputDepths = new int[query.getInputEPAs().length];
			
			for ( int i = 0; i < inputResources.length; i++ ) {
				query.setInput(i, splitRec(query.getInputEPAs()[i], resources, context));
				inputDepths[i] = query.getInputEPAs()[i].getDepth();
				inputResources[i] = context.peek();
			}
			if ( distinctValues(inputResources) > 1 ) {
				// Assign resource with max inputs
				Resource resource = getBestResource(inputResources, inputDepths);
				// Insert fence for parts not computed on current resource
				for ( int i = 0; i < inputResources.length; i++ ) {
					if ( inputResources[i] != resource ) {
						ResourceBoundary fence = new ResourceBoundary( inputResources[i].getIdentifier(), query.getInputEPAs()[i]);
						query.setInput(i, fence);
					}
				}
				context.push(resource);
			}
		}
		return query;
	}
	
	private static <T> int distinctValues( T[] arr ) {
		return new HashSet<>( Arrays.asList(arr) ).size();
	}
	
	private static Resource getBestResource( Resource[] rs, int[] depths ) {
		List<Resource> currentMaxs = new ArrayList<>();
		int count = 0;
		
		
		for ( Resource r : rs ) {
			int rc = count(r, rs);
			if ( rc > count ) {
				count = rc;
				currentMaxs.clear();
			}
			if ( rc == count )
				currentMaxs.add(r);
		}
		
		Resource result = currentMaxs.iterator().next();
		int minDepth = Integer.MAX_VALUE;
		for ( Resource r : currentMaxs ) {
			int md = maxDepth(r, rs, depths);
			if ( md < minDepth ) {
				result = r;
				minDepth = md;
			}
		}
		return result;
	}
	
	private static int count( Resource r, Resource[] rs ) {
		int count = 0; 
		for ( Resource chk : rs )
			if ( chk == r )
				count++;
		return count;
	}
	
	private static int maxDepth( Resource r, Resource[] rs, int[] depths ) {
		int md = 0;
		for ( int i = 0; i < rs.length; i++ ) {
			if ( r == rs[i] )
				md = Math.max(md, depths[i]);
		}
		return md;
	}
	
	
	

}
