/**
 * 
 */
package de.umr.jepc.v2.federated.api.resource;

import java.util.HashMap;
import java.util.Map;

import de.umr.ep.common.EPException;
import de.umr.ep.event.Event;
import de.umr.jepc.v2.api.EPProvider;
import de.umr.jepc.v2.api.epa.EPA;
import de.umr.jepc.v2.api.epa.Stream;
import de.umr.jepc.v2.api.provider.AlreadyRegisteredException;
import de.umr.jepc.v2.api.provider.EventSink;
import de.umr.jepc.v2.api.provider.ExternalStreamHandle;
import de.umr.jepc.v2.api.provider.NoSuchQueryException;
import de.umr.jepc.v2.api.provider.NoSuchStreamException;
import de.umr.jepc.v2.api.provider.QueryHandle;
import de.umr.jepc.v2.federated.api.epa.ForeignInput;
import de.umr.jepc.v2.federated.api.resource.query.LocalQuery;

/**
 * Wraps an {@link FederationAwareEPProvider EP provider} and provides basic functionality
 * for processing distributed queries. Especially takes care of passing sub-query results
 * via a given {@link RemoteEventChannel} to the successor and inject input events received on an
 * {@link RemoteEventChannel} into the processing engine. 
 * This wrapper is not specific to an implementation (e.g. centralized or distributed), but
 * offers functionality required in all distributed setups.
 * 
 * @author Michael Körber
 *
 */
public class ResourceInstance {
    
    /** The underlying engine */
    private final EPProvider engine;
    
    /** The currently deployed queries */
    private final Map<String, LocalQuery> localQueries = new HashMap<>();
    
    /**
     * @param engine the processing engine
     */
    public ResourceInstance(EPProvider engine) {
        this.engine = engine;
    }

    /**
     * Deploys the given (sub-) query and wires the result stream to the given channel.
     * @param name the name of the (sub-) query
     * @param epn the operator graph
     * @param resultChannel the channel to send results to
     * @throws EPException on any error while deploying the query
     */
    public void deployQuery( String name, EPA epn, RemoteEventChannel resultChannel ) throws EPException {
        LocalQuery lq = localQueries.get(name);
        if ( lq != null )
            throw new AlreadyRegisteredException("Query with name: " + name + " already registered.");
            
        lq = new LocalQuery(name); 
        EPA localEPN = registerForeignStreams(epn, lq);
        QueryHandle localHandle = engine.createQuery(name, localEPN);
        lq.setLocalHandle(localHandle);
        
        engine.registerSink(localHandle, new EventSink() {
            @Override
            public void receiveEvent(Event event) {
                resultChannel.pushEvent(event);
            }
        });
        localQueries.put(name, lq);
    }
    
    /**
     * Pushes an event received from the given resource and query into the processing engine.
     * @param fromResourceId the id of the originating resource
     * @param query the name of the query this event belongs to
     * @param event the event to push
     * @throws NoSuchQueryException if no query could be found for the given name
     */
    public void pushForeignEvent( String fromResourceId, String query,  Event event) throws NoSuchQueryException {
        LocalQuery lq = localQueries.get(query); 
        if ( lq == null )
            throw new NoSuchQueryException("Query: " + lq + " not registered");
        
        lq.getForeignStream(fromResourceId).pushEvent(event);
    }
    
    
    /**
     * Removes the given (sub-) query and unregisters all associated foreign streams
     * @param name the name of the query
     * @throws NoSuchQueryException if no query is registered with the given name 
     */
    public void removeQuery( String name ) throws NoSuchQueryException {
        LocalQuery lq = localQueries.get(name);
        if ( lq == null )
            throw new NoSuchQueryException("Query: " + name + " not registered");
        
        engine.removeSink(lq.getSink());
        engine.removeQuery(lq.getLocalHandle());
        for ( ExternalStreamHandle fsh : lq.getForeignStreams() )
            try {
                engine.removeStream(fsh);
            } catch (NoSuchStreamException e) {
            }
    }
    
    /**
     * Registers all foreign inputs with the underlying engine and stores
     * a reference to their handlers. This way incoming events from foreign resources
     * can directly be injected into the processing engine. 
     * @param epn the operator graph
     * @param lc the local query instance to register the foreign inputs with
     * @throws AlreadyRegisteredException if a stream with the given name is already registered
     * @throws EPException on any error registering the stream with the underlying engine
     */
    private EPA registerForeignStreams( EPA epn, LocalQuery lc ) throws AlreadyRegisteredException, EPException {
        if ( epn instanceof ForeignInput ) {
            ForeignInput fi = (ForeignInput) epn;
            lc.addForeignStream( fi, engine.registerExternalStream(fi.getUniqueStreamName(), fi.getSchema()) );
            return new Stream(fi.getUniqueStreamName());
        }
        else {
            EPA[] inputs = epn.getInputEPAs();
            for ( int i = 0; i < inputs.length; i++ ) {
                epn.setInput(i, registerForeignStreams(inputs[i], lc));
            }
            return epn;
        }
    }
    
    /**
     * Shuts down the underlying engine
     */
    public void destroy() {
        this.engine.destroy();
    }
}
