/*
 * Copyright (c) 2012, 2013, 2014
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates;

import de.umr.ep.common.Attribute;
import de.umr.ep.event.EventSchema;
import de.umr.ep.event.SchemaException;
import de.umr.jepc.v2.api.epa.aggregates.Aggregate;
import de.umr.jepc.v2.api.epa.aggregates.Average;
import de.umr.jepc.v2.api.epa.aggregates.Count;
import de.umr.jepc.v2.api.epa.aggregates.Maximum;
import de.umr.jepc.v2.api.epa.aggregates.Minimum;
import de.umr.jepc.v2.api.epa.aggregates.Stddev;
import de.umr.jepc.v2.api.epa.aggregates.Sum;
import de.umr.jepc.v2.api.epa.pattern.symbol.NoSuchVariableException;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.max.ByteMaxFunction;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.max.DoubleMaxFunction;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.max.FloatMaxFunction;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.max.IntMaxFunction;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.max.LongMaxFunction;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.max.ShortMaxFunction;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.min.ByteMinFunction;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.min.DoubleMinFunction;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.min.FloatMinFunction;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.min.IntMinFunction;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.min.LongMinFunction;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.min.ShortMinFunction;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.sum.DoubleSumFunction;
import de.umr.jepc.v2.bridges.nat.epa.impl.aggregator.aggregates.sum.LongSumFunction;

/**
 * Translates Mock-Up aggregates into {@link AggregateFunction functions}.
 * 
 * @author Michael Körber
 *
 */
public final class AggregateTranslator {

  private AggregateTranslator() {
  }

  /**
   * Creates a partial aggregate function from the given aggregate
   * 
   * @param agg the aggregate mock-up
   * @param inputSchema the schema of incoming events
   * @return partial aggregate function for the given mock-up
 * @throws NoSuchVariableException 
   */
  public static AggregateFunction<?, ?, ?> translateAggregate(Aggregate agg, EventSchema inputSchema) throws SchemaException {
    if (agg instanceof Average)
      return new AverageFunction();
    else if (agg instanceof Count)
      return new CountFunction();
    else if (agg instanceof Stddev)
      return new StdDevFunction();
    else if (agg instanceof Sum)
      return createSum((Sum) agg, inputSchema.byName(agg.getAttributeIn()));
    else if (agg instanceof Minimum)
      return createMin((Minimum) agg, inputSchema.byName(agg.getAttributeIn()));
    else if (agg instanceof Maximum)
      return createMax((Maximum) agg, inputSchema.byName(agg.getAttributeIn()));
    else
      throw new IllegalArgumentException("Unknown aggregate: " + agg.getClass().getSimpleName());
  }

  /**
   * Creates a sum aggregate
   * 
   * @param s the mock-up
   * @param a the attribute the sum is calculated on
   * @return the sum aggregate function
   */
  private static AggregateFunction<?, ?, ?> createSum(Sum s, Attribute a) {
    switch (a.getType()) {
    case BYTE:
    case SHORT:
    case INTEGER:
    case LONG:
      return new LongSumFunction();
    case FLOAT:
    case DOUBLE:
      return new DoubleSumFunction();
    default:
      throw new IllegalArgumentException("Cannot sum attribute: " + s.getAttributeIn() + ":" + a.getType());
    }
  }

  /**
   * Creates a minimum aggregate
   * 
   * @param s the mock-up
   * @param a the attribute the minimum is calculated on
   * @return the minimum aggregate function
   */
  private static AggregateFunction<?, ?, ?> createMin(Minimum s, Attribute a) {
    switch (a.getType()) {
    case BYTE:
      return new ByteMinFunction();
    case SHORT:
      return new ShortMinFunction();
    case INTEGER:
      return new IntMinFunction();
    case LONG:
      return new LongMinFunction();
    case FLOAT:
      return new FloatMinFunction();
    case DOUBLE:
      return new DoubleMinFunction();
    default:
      throw new IllegalArgumentException("Cannot calc minimum on attribute: " + s.getAttributeIn() + ":"
          + a.getType());
    }
  }

  /**
   * Creates a maximum aggregate
   * 
   * @param s the mock-up
   * @param a the attribute the maximum is calculated on
   * @return the maximum aggregate function
   */
  private static AggregateFunction<?, ?, ?> createMax(Maximum s, Attribute a) {
    switch (a.getType()) {
    case BYTE:
      return new ByteMaxFunction();
    case SHORT:
      return new ShortMaxFunction();
    case INTEGER:
      return new IntMaxFunction();
    case LONG:
      return new LongMaxFunction();
    case FLOAT:
      return new FloatMaxFunction();
    case DOUBLE:
      return new DoubleMaxFunction();
    default:
      throw new IllegalArgumentException("Cannot calc maximum on attribute: " + s.getAttributeIn() + ":"
          + a.getType());
    }
  }

}
