/*
 * Copyright (c) 2012, 2013, 2014, 2015
 * Database Research Group, University of Marburg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.umr.jepc.v2.bridges.nat.epa.impl.correlator;

import java.util.Iterator;
import java.util.PriorityQueue;

import de.umr.ep.common.IncompatibleTypeException;
import de.umr.ep.event.Event;
import de.umr.ep.event.EventSchema;
import de.umr.ep.event.SchemaException;
import de.umr.ep.event.impl.CompoundEvent;
import de.umr.jepc.v2.api.bridge.EPACallback;
import de.umr.jepc.v2.api.bridge.EventChannel;
import de.umr.jepc.v2.api.epa.Correlator;
import de.umr.jepc.v2.api.epa.pattern.symbol.Bindings;
import de.umr.jepc.v2.bridges.nat.epa.NativeOperator;
import de.umr.jepc.v2.bridges.nat.epa.util.AscTimestampComparator;
import de.umr.jepc.v2.expressions.bool.EVBooleanExpression;
import de.umr.jepc.v2.ql.ExpressionTranslator;
import de.umr.jepc.v2.ql.TranslatorException;

/**
 * Implementation of a native binary join over data streams for JEPC. For
 * conceptual issues see Jürgen Krämer. Continuous Queries over Data Streams –
 * Semantics and Implementation. <br>
 * 
 * @author Michael Körber
 */
public class NativeCorrelator implements NativeOperator {
	
	private final EventChannel left;
	
	private final EventChannel right;

	/**
	 * Predicate which determines if a pair of given events should be joined.
	 * This predicate is user defined and given by the constructor.
	 */
	private final EVBooleanExpression joinPredicate;

	private final EventSchema outputSchema;

	private final PriorityQueue<Event> out;
	private final IntervalSweepArea statusLeft;
	private final IntervalSweepArea statusRight;
	
	// Clocks of the input-streams
	private long minTStartL = Long.MIN_VALUE;
	private long minTStartR = Long.MIN_VALUE;
	
	private EPACallback callback;

	public NativeCorrelator(Correlator def, EventChannel left, EventChannel right) throws TranslatorException, IncompatibleTypeException, SchemaException {
		this.left = left;
		this.right = right;
		this.outputSchema = def.computeOutputSchema(left.getSchema(),right.getSchema());
		this.joinPredicate = new ExpressionTranslator().translateBooleanExpression(def.getCondition(), this.outputSchema, new Bindings());
		this.out = new PriorityQueue<>(new AscTimestampComparator());	
		this.statusLeft = new IntervalSweepArea();
		this.statusRight = new IntervalSweepArea();
	}
	
	@Override
	public EventSchema getSchema() {
		return outputSchema;
	}
	
	@Override
	public void setCallback(EPACallback callback) {
		this.callback = callback;
	}

	@Override
	public void process(EventChannel input, Event event) {
		if ( input == left )
			pushLeft(event);
		if ( input == right )
			pushRight(event);
	}

	/**
	 * Called for incoming events on the left side
	 * 
	 * @param event
	 *            the incoming event
	 */
	private void pushLeft(Event event) {
		statusLeft.add(event);

		// Fix this event in the access object
		CompoundEvent access = new CompoundEvent();
		access.setE1(event);
		
		// Remove all expired events from the right sweep-area
		statusRight.removeWithEndTimeStampLQ(event.getT1());

		// Iterate all potential right partners
		Iterator<Event> iterator = statusRight.queryStartTimestampLess(event.getT2());
		while (iterator.hasNext()) {
			Event potentialPartner = iterator.next();
			access.setE2(potentialPartner);
			if (joinPredicate.eval(access))
				out.add(access.materialize());
		}

		if (event.getT1() >= minTStartL) {
			minTStartL = event.getT1();
			// Flush only if necessary
			// if (tstart <= minTStartR || rightRelation)
			flush();
		}
	}

	/**
	 * Called for incoming events on the right side
	 * 
	 * @param event
	 *            the incoming event
	 */
	private void pushRight(Event event) {
		statusRight.add(event);

		// Fix this event in the access object
		CompoundEvent access = new CompoundEvent();
		access.setE2(event);

		// Remove all expired events from the left sweep-area
		statusLeft.removeWithEndTimeStampLQ(event.getT1());

		// Iterate all potential right partners
		Iterator<Event> iterator = statusLeft.queryStartTimestampLess(event.getT2());
		while (iterator.hasNext()) {
			Event potentialPartner = iterator.next();
			access.setE1(potentialPartner);
			if (joinPredicate.eval(access))
				out.add(access.materialize());
		}

		if (event.getT1() >= minTStartR) {
			minTStartR = event.getT1();
			// Flush only if necessary
			// if (tstart <= minTStartL || leftRelation )
			flush();
		}
	}

	/**
	 * Flushes the output-heap to the minimum of both input clocks
	 */
	protected void flush() {
		long minTS = Math.min(minTStartL, minTStartR);

		while (!out.isEmpty() && out.peek().getT1() <= minTS)
			callback.receive(out.poll());
	}
}
