package de.umr.jepc.v2.bridges.nat.epa.impl.window;

import java.util.HashMap;
import java.util.Map;
import java.util.PriorityQueue;

import de.umr.ep.event.Event;
import de.umr.ep.event.EventSchema;
import de.umr.ep.event.SchemaException;
import de.umr.jepc.v2.api.bridge.EPACallback;
import de.umr.jepc.v2.api.bridge.EventChannel;
import de.umr.jepc.v2.api.epa.window.PartitionedCountWindow;
import de.umr.jepc.v2.bridges.nat.epa.util.AscTimestampComparator;

/**
 * Implements a partitioned count-based window
 * @author Michael Körber
 */
public class NativePartitionedCountWindow extends NativeWindow implements EPACallback {

    /**
     * Separator for building a unique partition identifier
     */
    private static final char                    HASH_SEP = ':';

    /**
     * The window definition 
     */
    private final PartitionedCountWindow         def;

    /**
     * Stores the positions of all attributes that are used to partition the input
     * event stream.
     */
    private final int[]                          partitionAttributes;

    /**
     * Stores all active partitions.
     */
    private final Map<String, NativeCountWindow> partitions;

    /**
     * Buffers the output of partitioned count-based windows. Only events with a
     * timestamp that all active partitions have reached are safe to be reported!
     */
    private final PriorityQueue<Event>           partitionedCountBuffer;

    /**
     * @param input the input channel
     * @param def the window definition
     * @throws SchemaException if a defined paritioning attribute is not present in the input stream
     */
    public NativePartitionedCountWindow(EventChannel input, PartitionedCountWindow def) throws SchemaException {
        super(input, def);
        this.def = def;
        this.partitionAttributes = computPartitionAttributes(input.getSchema(), def.getPartitionAttributes());
        partitions = new HashMap<>();
        partitionedCountBuffer = new PriorityQueue<>(128, new AscTimestampComparator());
    }

    /**
     * Computes the positions of the given partition attributes
     * @param schema the schema of incoming events
     * @param partitionBy the partition attributes
     * @return the positions of the given partition attributes on incoming events
     * @throws SchemaException 
     */
    private static int[] computPartitionAttributes(EventSchema schema, String[] partitionBy) throws SchemaException {
        int[] partitionAttributes = new int[partitionBy.length];
        for (int i = 0; i < partitionAttributes.length; i++) {
            partitionAttributes[i] = schema.getAttributeIndex(partitionBy[i]);
        }
        return partitionAttributes;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void process(EventChannel input, Event event) {
        String hash;
        if (partitionAttributes.length == 1) {
            hash = event.get(partitionAttributes[0]).toString();
        } else {
            StringBuilder sb = new StringBuilder();
            for (int idx : partitionAttributes) {
                sb.append(event.get(idx)).append(HASH_SEP);
            }
            hash = sb.toString();
        }

        NativeCountWindow partition = partitions.get(hash);
        if (partition == null) {
            partition = new NativeCountWindow(input,def);
            partition.setCallback(this);
            partitions.put(hash, partition);
        }
        partition.process(input, event);
        flush();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void receive(Event event) {
        partitionedCountBuffer.add(event);
    }

    /**
     * Flushes the output-heap to the actual stream-time
     */
    private void flush() {
        long minTimestamp = getMinTimestamp(); // TODO Splits wenn eine Partition hängt!!
        while (!partitionedCountBuffer.isEmpty() && partitionedCountBuffer.peek().getT1() <= minTimestamp) {
            callback.receive(partitionedCountBuffer.poll());
        }
    }

    /**
     * Gets the minimum clock among all input streams.
     * @return minimum clock among all input streams
     */
    private long getMinTimestamp() {
        long minTimestamp = Long.MAX_VALUE;
        for (NativeCountWindow w : partitions.values())
            minTimestamp = Math.min(minTimestamp, w.tstart);
        return minTimestamp;
    }
}
