/**
 * 
 */
package de.umr.jepc.v2.addons.epa.tpstream.evaluation.latency;

import java.util.Arrays;

import de.umr.ep.common.IncompatibleTypeException;
import de.umr.ep.event.Event;
import de.umr.jepc.v2.addons.epa.tpstream.Constraint;
import de.umr.jepc.v2.addons.epa.tpstream.Constraint.TemporalRelation;
import de.umr.jepc.v2.addons.epa.tpstream.Constraints;
import de.umr.jepc.v2.addons.epa.tpstream.Situation;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinition;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinitions;
import de.umr.jepc.v2.addons.epa.tpstream.evaluation.util.EVUtil;
import de.umr.jepc.v2.addons.epa.tpstream.evaluation.util.data.EventGenerator;
import de.umr.jepc.v2.addons.epa.tpstream.evaluation.util.data.generator.BaseEventGenerator;
import de.umr.jepc.v2.addons.epa.tpstream.evaluation.util.data.situations.SituationDef;
import de.umr.jepc.v2.addons.epa.tpstream.evaluation.util.data.situations.duration.UniformDurationGenerator;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.Matcher.MatcherCallback;
import de.umr.jepc.v2.addons.epa.tpstream.partition.ClassicPartitionFactory;
import de.umr.jepc.v2.addons.epa.tpstream.partition.LowLatencyPartitionFactory;
import de.umr.jepc.v2.addons.epa.tpstream.partition.TPSPartition;
import de.umr.jepc.v2.api.epa.pattern.symbol.Binding;
import de.umr.jepc.v2.api.epa.pattern.symbol.Bindings;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.IntConstant;
import de.umr.jepc.v2.api.expression.arithmetic.atomic.Variable;
import de.umr.jepc.v2.api.expression.predicate.Equal;
import de.umr.jepc.v2.ql.TranslatorException;

/**
 * @author Michael Körber
 *
 */
public class ApplicationTimeLatency {
	
	private static final int NUM_RUNS = 100;
    
    private static class LatencyCallback implements MatcherCallback {
        private long count = 0;
        private double duration = 0;
        
        public double getAvgDuration() {
            return (duration / count);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void onResult(Situation[] result, long currentTime) {
            long[] ts = new long[result.length*2];
            int i = 0;
            for ( Situation si : result ) {
                ts[i++] = si.getStart();
                ts[i++] = si.getEnd();
            }
            Arrays.sort(ts);
            count++;
            duration += (currentTime - ts[ts.length-2]);
        }
    }
    
    private static final long WINDOW = 1000;
    
    private static final SituationDefinitions SYMBOLS = new SituationDefinitions(
            new SituationDefinition('a', new Equal( new Variable("A1"), new IntConstant(1)), new Bindings( new Binding("V1", new Variable("A1")))),
            new SituationDefinition('b', new Equal( new Variable("A2"), new IntConstant(1)))
    );
    
    private static EventGenerator getGen( double skew ) {
        return new BaseEventGenerator(EVUtil.STREAM, EVUtil.NUM_EVENTS, 0, 1,
                new SituationDef( new UniformDurationGenerator(10, 100), new UniformDurationGenerator(10, 50)),
                new SituationDef( new UniformDurationGenerator(10, (int) (100 * skew)), new UniformDurationGenerator(10, 50))
        );
    }
    
    
    private static void execute(Constraints pattern, double skew) throws TranslatorException, IncompatibleTypeException, InterruptedException {
        long[] latencies1 = new long[NUM_RUNS];
        long[] latencies2 = new long[NUM_RUNS];
        
        
        for ( int i = -1; i < latencies1.length; i++ ) {
            EventGenerator gen1 = getGen(skew);
            LatencyCallback cb1 = new LatencyCallback();
            LatencyCallback cb2 = new LatencyCallback();
            
            
            TPSPartition p1 = new ClassicPartitionFactory(gen1.getSchema(), SYMBOLS, pattern, WINDOW, cb1).create();
            TPSPartition p2 = new LowLatencyPartitionFactory(gen1.getSchema(), SYMBOLS, pattern, WINDOW, cb2).create();
            
            while ( gen1.hasNext() ) {
                Event e = gen1.next();
                p1.update(e);
                p2.update(e);
            }
            
            if ( i >= 0 ) {
                latencies1[i] = (long) cb1.getAvgDuration();
                latencies2[i] = (long) cb2.getAvgDuration();
            }
        }
        
        long avg1 = 0;
        long avg2 = 0;
        for (int i = 0; i < latencies1.length; i++) {
            avg1 += latencies1[i];
            avg2 += latencies2[i];
        }
        avg1 /= (latencies1.length);
        avg2 /= (latencies2.length);
        
        System.out.println("ISEQ,     skew=" + skew + ". Avg: "+ avg1 + "ms, details: " + Arrays.toString(latencies1));
        System.out.println("TPStream, skew=" + skew + ". Avg: "+ avg2 + "ms, details: " + Arrays.toString(latencies2));
    }
    
    
    public static void main(String[] args) throws TranslatorException, IncompatibleTypeException, InterruptedException {
        System.out.println("****************************************");
        System.out.println("         Experiment: Low Latency");
        
        TemporalRelation[] trs = { TemporalRelation.AABB, TemporalRelation.AA_BB, TemporalRelation.ABAB, TemporalRelation.A_BAB, TemporalRelation.BAAB };
        
        for ( TemporalRelation tr : trs ) {
            
            Constraints c = new Constraints( new Constraint('a', 'b', tr) );
            
            System.out.println("Pattern: " + c);
            execute( c, 0.45 );
            execute( c, 0.725 );
            execute( c, 1.0 );
            execute( c, 1.55 );
            execute( c, 2.1 );
        }
        
        System.out.println("****************************************");
    }
    
    


}
