package de.umr.jepc.v2.addons.epa.tpstream.evaluation.util.iseq.aggressive;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import de.umr.jepc.v2.addons.epa.TPStream.Mode;
import de.umr.jepc.v2.addons.epa.tpstream.Situation;
import de.umr.jepc.v2.addons.epa.tpstream.SituationDefinitions;
import de.umr.jepc.v2.addons.epa.tpstream.evaluation.util.iseq.util.ArrivalAction;
import de.umr.jepc.v2.addons.epa.tpstream.evaluation.util.iseq.util.TComp;
import de.umr.jepc.v2.addons.epa.tpstream.evaluation.util.iseq.util.Evaluator.Check;
import de.umr.jepc.v2.addons.epa.tpstream.matcher.Matcher;

/**
 * Matching implementation for {@link Mode#CLASSIC classic} temporal
 * pattern matching. This matcher only considers situation with a known
 * end.
 * @author Michael Körber
 *
 */
public final class AggressiveISEQLightMatcher implements Matcher {
    
	/** The situation definitions */
	private final SituationDefinitions definitions;
	
	private final ArrivalAction[] actions;
	
	/** The buffers */
	private final LinkedList<Situation>[] buffers;

	/** The callback to invoke on new results */
	private final MatcherCallback callback;
	
	/** The length of the time-window */
	private final long window;
	
	/** The working-set holding (partial) temporal configurations */
    private Situation[] work;
    
    private long lastClock = Long.MIN_VALUE;
    
	
	/**
	 * @param definitions The situation definitions
	 * @param constraints The constraints
	 * @param window The length of the time-window
	 * @param optimizer The optimizer instance
	 * @param callback The callback to invoke on new results
	 */
	@SuppressWarnings("unchecked")
	public AggressiveISEQLightMatcher(SituationDefinitions definitions, long window, MatcherCallback callback) {
		this.definitions = definitions;
		this.window = window;
		this.callback = callback;
        this.work = new Situation[this.definitions.getNumberOfDefinitions()];
        this.buffers = new LinkedList[this.definitions.getNumberOfDefinitions()];
        for ( int i = 0; i < buffers.length; i++ )
        	buffers[i] = new LinkedList<>();
        
        this.actions = new ArrivalAction[] {
        		new ArrivalAction(false),
        		new ArrivalAction(false,0),
        		new ArrivalAction(true,0,1)
        };
        
    }
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public void update(Situation[] updates, long currentTime) {
		windowPurge(currentTime);
		for (int i = 0; i < updates.length; i++) {
			Situation interval = updates[i];
			
			if ( interval == null || !interval.isFinished() || interval.getDuration() > window )
				continue;
			
			if ( interval == null || !interval.isFinished() || interval.getDuration() > window )
				continue;
			
			if ( actions[i].doBuffer(buffers) )
				buffers[i].add(interval);
			
			if ( actions[i].isTrigger(buffers) )
				eval( interval );
			
        }
	}
	
	final Check breakB = new Check(1,2,TComp.TE_LT_TS);
	final Check breakA = new Check(0,1,TComp.TE_LT_TS);
	
	final Check[] checkB = {
			new Check(1, 2, TComp.TS_LT_TS),
			new Check(1, 2, TComp.TE_GE_TS),
			new Check(1, 2, TComp.TE_LT_TE)
			
	};
	
	final Check[] checkA = {
			new Check(0,1, TComp.TS_LT_TS ),
			new Check(0,1, TComp.TE_GE_TS ),
			new Check(0,2, TComp.TE_LT_TS ),
			new Check(0,1, TComp.TE_LT_TE ),
	};
	
	private void eval( Situation s ) {
		work[2] = s;
		
		// Join Bs
		Iterator<Situation> bs = buffers[1].descendingIterator();
		
		while ( bs.hasNext() ) {
			work[1] = bs.next();
			// End reached
			if ( breakB != null && breakB.eval(work) )
				break;
			
			boolean passed = true;
			for ( int i = 0; i < checkB.length && passed; i++ )
				passed &= checkB[i].eval(work);
			
			if ( passed ) {
				// Join As
				Iterator<Situation>  as = buffers[0].descendingIterator();
				while ( as.hasNext() ) {
					work[0] = as.next();
					
					if ( breakA != null && breakA.eval(work) )
						break;
					
					boolean passedA = true;
					for ( int i = 0; i < checkA.length && passedA; i++ )
						passedA &= checkA[i].eval(work);
					
					if ( passedA )
						callback.onResult(work, lastClock);
				}
				work[0] = null;
			}
		}
		work[1] = null;
		work[2] = null;
	}

	private void windowPurge(long now) {
		if ( now <= lastClock )
			return;
		lastClock = now;
		long minTS = now - window;
		
		for ( List<Situation> buf : buffers ) {
			ListIterator<Situation> items = buf.listIterator();
			while ( items.hasNext() ) {
				if ( items.next().getStart() < minTS )
					items.remove();
			}
		}
	}
}
